/*
$Revision: 1.2 $

Last Modification Was on $Date: 2003/05/07 23:02:23 $

By $Author: Lann $

$Log: facedata.c,v $
Revision 1.2  2003/05/07 23:02:23  Lann
Revision V1_00_02

Revision 1.1.1.1  2003/05/07 21:50:38  Lann
First Release Version


*/

/*1 $Header: /usr/local/cvsrep/Genesis/tools/facedata.c,v 1.2 2003/05/07 23:02:23 Lann Exp $

Module: FACEDATA.C

Copyright 1994, 1995 A.J. Lincoln & Co.,Inc., Concord, Mass.


1*/

/*<f><s>Description of the Module*/

/**************************************************************************

This module is used to generate a file which is required to bind LJ and UFST.
Note that UFST requires a directory file called if.fnt which is used by
UFST among other things to locate the individual Intellifont files. (Note that
no such directory is used for the TrueType faces). This directory includes the
full path/name of all the IF faces. Two symbol set directory files are also
required - both by UFST and IF. The if.fnt file is created by a modified
ixmak.exe which reads the lj.ini file to get directory data. This file is
placed in the same directory as the IF faces. The symbol set files do not
contain data having to do with	locations and thus do not need to be
reconstructed except possibly for different CPU's.

The binding file produced by functions in this module contains all of the
information required to build the face masters and their associated aliases.
The input data required for building this data is the following:
(1) the Intellifont face (.CDI) files
(2) the Intellifont directory file, if.fnt
(3) the TrueType face files
(4) a TrueType directory file, tt.txt (similar to ixmak.txt used in making if.fnt).
(4) lj.ini which gives directories for finding these files
(5) the include file, symassoc.h, which contains data for associating symbol
    sets with faces
(6) a symbol set file called char_set.cco containing information about the on-board
    symbol sets.

For IF faces the required data is obtained using UFST functions. For TT faces 
the data is obtained by special routines which reads some of the tag sections 
in each TT file. If the PCLT is missing from the TT file, missing information can
be supplied on the tt.txt file.

The binding file is actually a set of MASTER_CG structures (see scaly_ff,h) for 
all of the faces.

The following is a rough outline of the program.

- read lj.ini to get the appropriate directories for the various files
- using the if.fnt file for the list of IF faces obtain the required data
  using CGIFsegments() in UFST
- using the text file directory of TT files interpret the appropriate tag
  section of each file.


***************************************************************************/

#include    <stdlib.h>
#include    <stdio.h>
#include    <math.h>
#include    <string.h>

#include    "lpstd.h" 
#include    "cgconfig.h"  

/* The GLOBAL name has been defined in lpstd.h, and must be undefined to
   eliminate a compiler warning. It will be redefined in port.h, which is
   included in ljport.h. (KMR 11/15/96) */

#ifdef GLOBAL
#undef GLOBAL
#endif

#ifdef SUCCESS
#undef SUCCESS
#endif

#include    "ljport.h"
#include	 "ifmem.h"
#include	 "cgif.h"
#include  	"scaly_ff.h"
INSTANCE_CG	*p_inst_cg;

#include    "cache.h"
#include	 "ix.h"
#include	 "bitmap.h"
#include	 "sclcrt.h"
#include	 "if_type.h"
#include  "ttdef.h"
#define ORIGINAL
#include	"symassoc.h"
#undef ORIGINAL

#include "pcl5e.h"
UFST_BUFS ufst_bufs;

extern int get_if_stuff(SL32 tfnum, HP_FACE * p_hp_face);
extern int get_tt_info(FILE * in_file, MASTER_CG * p_master);
int f_PCLT_present;	 /* passed back by ttgetinfo */

#ifndef _POSIX_PATH_MAX
#define	_POSIX_PATH_MAX 255
#endif

#define     F_ONE     0x10000L /* fixed point 1.0000 (16dot16 format) */
#define MAX_OPEN_LIBRARIES   5       /* max # of open library files */
#define BITMAP_WIDTH         2       /* must be 1, 2, 4 or 8 */
#if BYTEORDER == LOHI
#define     MY_SYMSET     (((UW16)'W') | ((UW16)'1' << 8)) /* Windows 3.1 */
#else
#define     MY_SYMSET     (((UW16)'1') | ((UW16)'W' << 8)) /* Windows 3.1 */
#endif
#define FIBUF (MEMptr(fontindex.hfnt_index) + ENTRY_CT_SIZE)
#define ENTRY_CT_SIZE sizeof(SL32)
EXTERN LPSB8        FIpath(PINDEX_ENTRY);
EXTERN FONTINDEX fontindex;


static int PIreadini(void);
static int install_symbits(MASTER_CG * p_master);
static BOOLEAN init_intellifont();
static int if_headers();
static int tt_headers();
static void	write_info(char * p_f, MASTER_CG * p_master);
char field1[50];   /* to pass possible data to ttgetinf.c */
char field2[50];
char tfn[_POSIX_PATH_MAX];
static int f_info = FALSE;

/*< The following are defined to satisfy the linker when using some UFST modules. 
     Note that the followng UFST modules are required to run if_headers:

	  bucket.c
	  cgif.c
	  dll.c
	  extmem.c
	  fc.c
	  fm.c
	  if_init.c
	  ix.c
	  path.c
	  symmap.c					>*/


CHUNK bit[16];
FONTCONTEXT fontcontext;      /* our local FONTCONTEXT */
IF_STATE if_state;				/* declared elsewhere usually */
CHR_DEF_HDR chr_def_hdr;		/* usually declared in chr_def.c */
FILE * out_file, * info_file;

typedef struct
	{
	char s[3];
	}COMM;

#define Max_tbl 8

char d[1000];
char * p_fontext[8];
char path[_POSIX_PATH_MAX];
char path1[_POSIX_PATH_MAX];


/*2 PIreadini()

Return values:

0 -> read lj.ini
1 -> could not open lj.ini

Description:

Reads the two character options from lj.ini. Note that except for "fl" that
these give the directory path for various parts of the face libraries.

2*/  /*<f>*/



static int PIreadini(void)
	{ 	
	FILE * ini_file;
	int i, j = 0;
	COMM pcl_table[Max_tbl] = {	
	"bm", "sc", "st", "pt", "ss", "if", "tt", "fl"};
	char * p_value;
	char string[500];

	if ( (ini_file = fopen("lj.ini", "ra")) == NULL )
		 {
         if ( (ini_file = fopen("/lj.ini", "ra")) == NULL )
           if ( (ini_file = fopen("/lj/lj.ini", "ra")) == NULL )
			return 1;
        }

	while ( fscanf(ini_file, "%s", string) != EOF )
		{
		char * p_equal = strchr(string, '=');
		if ( p_equal != NULL )
			{
			* p_equal = '\0';
			p_value = p_equal + 1;
			if ( strlen(string) == 2 )	 /* can be bm, sc, st, pt, ss, if, tt, fl */
				{
				for ( i = 0; i < Max_tbl; ++ i )
					{
					if ( strcmp(string, pcl_table[i].s) == 0 )
						break;
					}
				if ( (p_value = ( i < Max_tbl ) ? p_equal + 1 : NULL) != NULL )
					{
					if ( i < Max_tbl - 1 )
						{
						strcpy(&d[j], p_value);
						p_fontext[i] = &d[j];
						j += strlen(p_value) + 1;
						}
					}
				}
			}
		}
	return 0;
	}

/*2 install_symbits()

Return values:

0 -> success
1 -> error	(send message to stderr)

Description:

Given a pointer to the current	MASTER_CG structure, this function fills in the
ss_associate field which has a bit set at approprate location for each symbol
set supported by the face. An array of NAME_BITS structures (see symassoc.h) is
used to associate the first 12 characters of the face name of each face with
a 32 bit bit string. For most names it is only necessary to match the number of 
characters in the complete string in NAME_LIST with the corresponding number of
characters in the face name obtained from the alias segment of the face library.
Note that this name is 16 bytes long is not necessarily terminated with '\0'.
The alias names include characters which indicate bold, italic, etc. Thes are
ignored. An exceptional case is Univers which comes in 2 flavors: plain and 
condensed (indicated by a cap C somewhere in the second part of the complete
name. See below for how we handle this.

It is assumed that the name of the face is already in the face name field of the
MASTER_CG.

NEW NOTES -

The association of symbol sets with faces is according to PCL5 Comparison Guide (1992) 
from HP. Symbol sets are indexed by R8, E1, etc. not by 10U, ... It is possible to add
faces which are not in the list of standard faces ( not to mention the fact that some
faces with the same name mayh not carry a full complement of glyphs. For now we assume
that any face not in the list is a TrueType face (since this is the way the world is going).
For TrueType we use the info in the PCLT - tt.txt - default in that order. Note in
particular that a TrueType bound font must have a symbol set associated with it.

2*/  /*<f>*/

static int install_symbits(MASTER_CG * p_master)
	{
	int num = sizeof(name_bits) / sizeof(NAME_BITS);
	int i, j, l;
	char name[13], * p_str;

	/*< Can not guarantee that face_name is null terminated >*/

	for ( i = 0 ; i < num ; ++ i )
		{
		p_str = name_bits[i].name;
		l = strlen(p_str);
		for ( j = 0; j < l ; ++ j )
			{
			if ( *(p_str + j) != p_master->face_name[j] )
			break;
			}

/*< If the character where equality fails is '$' then see if there is a 'C'
    in the rest of the name. If yes then it's Univers Cond; else Univers. >*/

		if ( j == l )
			{
			if ( p_master->face_name[0] == 'U' )
				{
				for ( j = l ; j < 16 ; ++ j )
					{
					if ( p_master->face_name[j] == 'C' )
						++ i;
					}	
				}
			break;
			}
		}
	if ( i == num )
		{
		fprintf(stderr, "\n Could not match name_bits %s", name);
		return(1);
		}
	p_master->ss_associate = name_bits[i].bits;
	if ( i < 12 )
		{
		int j;
		for ( j = 0; j < 8; ++ j )
			p_master->char_comp[j] = char_comp[i].cc[j];
		}
	return(0);
	}

/*2 init_intellifont()

Return values:

0 -> SUCCESS
1 -> error (send error message to STDERR)

Description:

Does the required initialization of UFST for reading IF data.

2*/  /*<f>*/

static BOOLEAN init_intellifont()
	{
	IFCONFIG   config_block;
	UW16      status;
/*   LPUB8  ss_addr[3];*/	/* Variable unused and removed (KMR 11/25/96) */

  /*
   *   Set up default library path and key Intellifont/UFST file locations.
   *   'ufstPath' is for if.ss, if.fnt, psname.cfg files.
   *   'typePath' is for library files.
   */

    strcpy (config_block.ufstPath, "\\td\\type");
    strcpy (config_block.typePath, "\\td\\type");
	config_block.max_char_size = 0xFFFFFFFF;
    config_block.num_files = 5;
    config_block.bit_map_width = 4;
	config_block.cc_buf_size = (128 * 1024);
	config_block.cc_buf_ptr = (unsigned char*)malloc(config_block.cc_buf_size);

    status = CGIFinit ();
    if (status)
	    {
        fprintf (stderr, "CGIFinit() error: %d\n", status);
        return 1;
		}
    status = CGIFconfig (&config_block);
    if (status)
	    {
        fprintf (stderr, "CGIFconfig() error: %d\n", status);
        return 1;
	    }
    status = CGIFenter ();
    if (status)
	    {
        fprintf (stderr, "CGIFenter() error: %u\n",status);
        return 1;
	    }
    return 0;
	}


/*2 if_headers()

Return values:

0 -> SUCCESS
1 -> error (with message)

Description:

Using lj.ini as the source for the IF directory and if.fnt (located in this directory)
and if.fnt as a directory of all the IF faces, this function fills in the MASTER_CG
structure for each IF fase (using UFST).

2*/  /*<f>*/

static int if_headers()
	{
	SW16           i, rc;
	PINDEX_ENTRY   p;
	HP_FACE 		hp_face, * p_face;
	MASTER_CG      master_cg, * p_master;

	p_face = & hp_face;
	p_master = & master_cg;
	if ( init_intellifont() )
		return(1);

    fontcontext.xspot        = F_ONE;       /* 1.0 */
    fontcontext.yspot        = F_ONE;       /* 1.0 */
    fontcontext.xbold        = 0;           /* disabled */
    fontcontext.ybold        = 0;           /*   "      */
    fontcontext.point_size   = 12 * 8;      /* 12 point * 8ths of pixels */
    fontcontext.set_size     = 12 * 8;      /*    "                      */
    fontcontext.shear.x      = 0;           /* no shear */
    fontcontext.shear.y      = F_ONE;       /*   "      */
    fontcontext.rotate.x     = F_ONE;       /* no rotation */
    fontcontext.rotate.y     = 0;           /*      "      */
    fontcontext.xres         = 11811L;      /*  300dpi = 1181 pixels/meter  */
    fontcontext.yres         = 11811L;      /*                              */
    fontcontext.ssnum        = MY_SYMSET;   /* desktop */
    fontcontext.format       = 0;           /* bitmap, Intellifont */

    for(i = 0, p = (PINDEX_ENTRY)FIBUF; i < fontindex.entry_ct; i++, p++)
		{

/*< Do not include the universal sets. >*/

		if ( p->tfnum < 6000 )
			continue;
		fontcontext.font_id = p->tfnum;
		if (rc = CGIFfont (&fontcontext))
			{
			fprintf(stderr, "\n Error %d from CGIFfont with face %d", rc, p->tfnum); 
			return (1);
			}     
		if ( ! get_if_stuff(p->tfnum, p_face) )
		  {
			p_master->face_id = p->tfnum;
			p_master->face_type = IF_TYPE | UNBOUND_BIT; /* assume all IF fonts are unbound */
			p_master->u_depth = p_face->u_depth;
			p_master->space = p_face->thin_space;
			p_master->scale_factor = p_face->scale;
			p_master->x_res = p_face->x_res;
			p_master->y_res = p_face->y_res;
			p_master->spacing = p_face->spacing;
			p_master->style_SWP = p_face->style;
			p_master->stroke_wt = p_face->weight;
			p_master->typeface = p_face->vendor;
			memmove(p_master->face_name, p_face->name, sizeof(p_face->name));

		  if ( ! install_symbits( p_master))
		    {
				if ( fwrite(&master_cg, sizeof(MASTER_CG), 1, out_file) != 1 )
				  {
				  fprintf(stderr, "\nError in writing to output");
				  return(1);
				  }
			  }
			}
		}
	return(0);
	
	}

/*2 tt_headers()

Return values:

0 -> SUCCESS
1 -> error (with message)

Description:

The header information for TrueType faces is obtained from various items from the tag
sections which are part of the TrueType face file. The faces are assumed to reside 
in the directory indicated by the tt options in lj.ini. Since UFST does not use a
directory for TrueType faces a text list is used to list all the TrueType faces to be
processed. This list is used by both tt_headers and the interpreter when it is logging
in the TrueType faces. The face_id field created in tt_headers is just the index 
into this text list of file names. It is up to the interpreter to change this index
to a pointer to the file name of the face. (Directory path is added at the time the
face is accessed).

NEW NOTE:

This text file may also be used to obtain important data if the PCLT section is missing.
If this data is present and the PCLT data is present the text data replaces the PCLT data.

2*/  /*<f>*/

static int tt_headers()
  {
	FILE * in_text;		/* file with list of face file names */
	int i = 0, j, l;
	int nfld;	/* number of scanned fields */
	char fn[_POSIX_PATH_MAX];
	MASTER_CG master_cg, * p_master;
	FILE * in_file;
	char string[1000], * p_s;
	strcpy(fn, path);
	strcat(fn, "tt.txt");
	strcpy(tfn, path);
	strcat(tfn,"charcode.txt");
	if ( f_info )
		info_file = fopen(tfn, "wa");
	else
		info_file = NULL;

	if ( (in_text = fopen(fn, "ra"))== NULL )
	  {
	  fprintf(stderr, "\nUnable to open TT directory %s", path);
	  return(1);
	  }
	p_master = & master_cg;
	memset(field1, 0, sizeof(field1));
	memset(field2, 0, sizeof(field2));
	nfld = fread(string, 1, sizeof(string), in_text);

	for ( j = 0, p_s = string; j < nfld; ++ j, ++ p_s )
		if ( * p_s == 0x0D )
			* p_s = ' ';
		else if ( * p_s == 0x0A )
			* p_s = 0;
	* p_s = 0;
	p_s = string;
	while ( (l = strlen(p_s)) > 2 )
		{
		nfld = sscanf(p_s, "%s%s%s", tfn, field1, field2);
		p_s += l + 1;

/*< We use the usual convention that if the string starts with '#' disregard it. >*/

	  if ( tfn[0] == '#' )
		continue;
	  //strcpy(fn, path);
	  strcpy(fn, tfn);
	  if ( ( in_file = fopen(fn, "rb"))== NULL)
	    {
	    fprintf(stderr, "\nUnable to open TT file %s", tfn);
			return(1);
	    }
		if (f_info)
			fprintf(info_file, "\n\n\n CHARACTER MAPPING FOR %s\n\n", fn); 
		if ( get_tt_info(in_file, p_master) )
		  {
		  fprintf(stderr, "\nError in get_tt_info");
			return(1);
			}
		fclose(in_file);
		p_master->face_id = (UNS32) i;	 /* index into tt_txt */
		p_master->face_type = TT_TYPE;
/*< Here we put in some known facts about Wingding and Symbol. Some TrueType versions do
    not contain PCLT data. The versions of these fonts from Agfa contain incorrect information,
	at least they do not agree with HP. >*/

		if ( stricmp(tfn, "SYMBOL.TTF") == 0 )
			{
			p_master->sym_set = 621;
			p_master->typeface = 16686;
			memset(p_master->char_comp, 0, 8);
			}
		if ( stricmp(tfn, "WINGDING.TTF") == 0 )
			{
			p_master->sym_set = 18540;
			p_master->typeface = 31402;
			memset(p_master->char_comp, 0, 8);
			}
		if ( p_master->sym_set == 0 )
			p_master->face_type |= UNBOUND_BIT;
		if ( f_info )
			write_info(fn, p_master);
		memset(field1, 0, sizeof(field1));
		memset(field2, 0, sizeof(field2));
		install_symbits( p_master);		/* do not intend to use the symbits */
	    {
			if ( fwrite(&master_cg, sizeof(MASTER_CG), 1, out_file) != 1 )
			  {
			  fprintf(stderr, "\nError in writing to output");
		  	return(1);
			  }
			}
		++ i;
	  }
	return(0);
	
	}




/*2 main()

Return values:

0 -> success
1 -> error (message will be sent to stderr)

Description:

This is the main program which does the following:

(1) reads the lj.ini file to locate the IF face directory and the  IF, TT faces
(2) opens the output file to be created
(3) call if_headers to spew out the IF data
(4) call tt_headers to do the same thing for TT data
(5) close files.

If the /p or -p switch is set then print out an info file for the TT fonts.

2*/  /*<f>*/

int main (int argc, char *argv[])
	{
	char ifnt[5];
	char fn[_POSIX_PATH_MAX];
	char * p_s;

	if (argc ==2)
		{
		p_s = argv[1];
		if (* p_s == '/' || * p_s == '-')
			{
			++ p_s;
			if ( * p_s == 'p' || *p_s == 'P' )
				f_info = TRUE;
			}
		}
		
	strcpy(ifnt, "IFNT");
	if ( PIreadini())
		path[0] = path1[0] = '\0';
	else
	  {
		if ( p_fontext[1] == NULL )
			path[0] = '\0';
		else
			strcpy(path, p_fontext[1]);	 /* IF directory  */
		if ( p_fontext[5] == NULL )
			path1[0] = '\0';
		else
			strcpy(path1, p_fontext[5]);	 /* IF directory  */
	  }

	strcpy(fn, path);
	strcat(fn, "scalstuf.lj3");
	if ( (out_file = fopen(fn, "wb")) == NULL  )
	  {
	  fprintf(stderr, "\nUnable to open output file.");
	  return(1);
	  }

/*< Here we identify the file with "IFNT", just like scalable.lj3.>*/

	fwrite(ifnt, 1, 4, out_file);
	 
	if ( if_headers () )
	  {
	  fprintf(stderr, "\nError in getting IF header data.");
	  return(1);
	  }
	strcpy(path, "\\ttfonts\\");		/* TT faces should be here */
	if ( tt_headers() )
	  {
	  fprintf(stderr, "\nError in getting TT header data. ");
	  return(1);
	  }
	fclose(out_file);
	}

/*< The following are added just to keep the linker happy. >*/

UW16 ifrender(PBUCKET pb, INTR xt, INTR yt)
{
	return(0);
	
}

UW16 ifset_trans(PBUCKET pb, SL32 l1, SL32 l2,	SL32 l3, SL32 l4)
{
	return(0);
	
}

UW16 ifmake_gaso_and_stats(PBUCKET pb, PCHAR_STATS pcs)
{
	return(0);
	
}

UW16 makechar(PFONTCONTEXT fc, PHIFBITMAP phbm)
{
	return(0);
	
}

UW16 MAKfontSize(PFONTCONTEXT fc, SW16 sw1, LPSL32 sl)
  {
  return(0);
  }

UW16 PCLeo(LPSB8 psb8, LPUL32 pul32, SL32 sl32)
  {
  return(0);
  }

	UW16 PCLeo2IF_ssnum(PFONTCONTEXT fc, UW16 u1, LPUW16 p_u)
	  {
	  return(0);
	  
	  }

	UW16 PCLload_font(LPUB8 p_u, PBUCKET p_b)
	  {
	  return(0);
	  
	  }

void PCLchId2ptr() {}
void PCLglyphID2Ptr() {}
unsigned short PCLEOmake_cd(unsigned short a, int b, int c) 
{
	return 0;
}
short special_case = 0;

/* This function is defined in tt_if.c. (KMR 11/25/96) */
/*	UW16 ttload_font(PFONT_DES p_des, PBUCKET p_b)
	  {
	  return(0);
	  
	  }*/

/* This function is defined in tt_if.c. (KMR 11/25/96) */
/*	UW16 ttunload_font( PBUCKET p_b)
	  {
	  return(0);
	  
	  }*/

/*2 write_info()

Return values:

None

Description:

Writes the contents of p_master to the info_file.

2*/  /*<f>*/

static void print_bits(char * p_a)
	{
	int i;
	unsigned char mask;
	for ( i=0; i < 8; ++ i, ++ p_a)
		{
		mask = (unsigned char)'x80';
		while (mask)
			{
			if ( * p_a & mask )
				fprintf(info_file, "1");
			else
				fprintf(info_file, "0");
			mask >>= 1;
			}
		fprintf(info_file, " ");
		}
	}

static void write_info(char * p_f, MASTER_CG * p_master)
	{
	fprintf(info_file, "\n\n Font master data for %s", p_f); 
	if ( f_PCLT_present )
		fprintf(info_file, ", a PCLT table is in the file\n");
	else
		fprintf(info_file, ", a PCLT table is not in the file\n");
	fprintf(info_file, "\n under score depth %d", p_master->u_depth);
	fprintf(info_file, "\n design units escapement %d", p_master->space);
	fprintf(info_file, "\n scale_factor %d", p_master->scale_factor);
	fprintf(info_file, "\n character complement ");
	print_bits(p_master->char_comp);
	fprintf(info_file, "\n symbol set %d", p_master->sym_set);
	fprintf(info_file, "\n spacing (0 => fixed, 1 => proportional) %d", p_master->spacing);
	fprintf(info_file, "\n stroke weight %d", p_master->stroke_wt);
	fprintf(info_file, "\n style word %d", p_master->style_SWP);
	fprintf(info_file, "\n typeface ID %d", p_master->typeface);
	fprintf(info_file, "\n name %s ", p_master->face_name);
	fprintf(info_file, "\n----------------------------------------------------");
	}









