/*
$Revision: 1.2 $

Last Modification Was on $Date: 2003/05/07 23:02:24 $

By $Author: Lann $

$Log: ttgetinf.c,v $
Revision 1.2  2003/05/07 23:02:24  Lann
Revision V1_00_02

Revision 1.1.1.1  2003/05/07 21:50:38  Lann
First Release Version

 * 
 *    Rev 1.7   21 Nov 1997  9:56:40   tbb
 * changed ifndef _sparc to ifndef HI2LOW.
 * 
 *    Rev 1.6   19 Sep 1997 16:39:10   tbb
 * For TrueType fonts the OS2 ave. char width is valid only for fixed pitch
 * fonts but we were using it for all fonts. If the PCLT exists the ppitch value
 * can be used for fixed pitch and prop. For prop. fonts if the PCLT does not
 * exist the value of space in the Master structure is now set to zero and
 * UFST is used to obtain the "pitch" within read_scalstuf.
 * 
 *    Rev 1.5   05 Jul 1997 14:53:22   tbb
 * Changes for GENESIS.
 * 
 *    Rev 1.4   10 Jun 1997 13:29:54   tbb
 * Added ANSI_C function defs.
 * 
 *    Rev 1.3   10 Apr 1997 12:11:26   tbb
 * Corrected computation of symbol set number. Corrected some stuff where long
 * tables are read (for TrueType Asian fonts). 
 * 
 *    Rev 1.2   10 Jan 1997 13:29:38   tbb
 * Did fscanf to an integer variable for typeface ID (f switch on tt.txt line)
 * so that face name did not get wiped out.
 * 
 *    Rev 1.1   07 Nov 1996 16:38:34   tbb
 * Corrected logic for getting tapeface ID, etc. and logic for getting name
 * out of the name table. Before if we came to the end of the list of known
 * faces, data already stored in p_master was used. In the name table we
 * now accept Microsoft Unicode names.

*/

/*1 $Header: /usr/local/cvsrep/Genesis/tools/ttgetinf.c,v 1.2 2003/05/07 23:02:24 Lann Exp $

Module: TTGETINF.C

Copyright 1994, 1995 A.J. Lincoln & Co.,Inc., Concord, Mass.


1*/

/*<f><s>Description of the Module*/

/**************************************************************************

Functions herein get selected tagged data from TrueType files. (Refer to
the Microsoft Development Library, Chapter 4, for details of the TT structure).
This data is transformed into a form suitable for insertion into the MASTER_CG
structure (hopefully).

***************************************************************************/

#include <stdlib.h>
#ifndef GENESIS
#include <string.h>
#endif
#include <sys/types.h>
#include <ctype.h>

#include <stdio.h>
#include <setjmp.h>
#include "lpstd.h"
#include "ttdef.h"
#include "scaly_ff.h"
#include "symassoc.h"	
#ifndef GENESIS
#include "lib.h"
#endif


#ifdef __unix
extern int strnicmp( const char * str1, const char * str2, size_t count );
#endif
#ifdef LOW2HI
#define GET_WORD(A)  ((UNS16)((UNS16)A[0] << 8) | ((UNS8)A[1]))
#define SWAPW(a)		(short)(((unsigned char)((a) >> 8)) | ((unsigned char)(a) << 8))
#else
#define GET_WORD(A)   ( *((UNS16*)(A)))
#define SWAPW(a)		(a)
#endif

#ifdef GENESIS
FILE * info_file;
char field1[50];
char field2[50];
char tfn[128];
int f_PCLT_present;
#else
extern FILE * info_file;
extern char field1[];
extern char field2[];
extern char tfn[];
extern int f_PCLT_present;
#endif

static void Read_OffsetTable(FILE * fdes, OFFSET_TABLE * offset_table);
static int Read_DirectoryTable(FILE * fdes, OFFSET_TABLE * sfntdir);
static int Get_TableOffsetLength(OFFSET_TABLE * sfntdir, char *dir_name, UNS32 * offset, UNS32 * length);
static int Read_OneMappingTable(FILE *fdes, long offset, INDEX_DIRECTORY *idxdir);
static int Read_MappingTable(FILE *fdes, long offset, long length);
static int Read_HeadTable(FILE *fdes, long offset, long length);
static int Read_NameTable(FILE *fdes, long offset, long length);
static int Read_OS2Table(FILE *fdes, long offset, long length);
static int Read_PCLTTable(FILE *fdes, long offset, long length);
static int Read_postTable(FILE *fdes, long offset, long length);
void tt_release_font_fragment(void *ptr);
void *tt_get_font_fragment(INT32 clientID, INT32 offset, INT32 length);
static int setbits(int m, int n);
static void try_field(char * field);
static void substitute(void);
static int install_cc_etc(void);

typedef struct
{
	UNS16	format;
	UNS16	length;
	UNS16	version;
	UNS16	segCountX2;
	UNS16	searchRange;
	UNS16	entrySelector;
	UNS16	rangeShift;
	UNS16	endCount;
}	segMappingTable  ;

typedef struct
{
	UNS16	format;
	UNS16	length;
	UNS16	version;
	UNS16*	subHeaderKey;
}	HighByteMappingTable;

typedef struct
{
	UNS16 startCount ;
	UNS16 endCount ;
}	rangeCount;

typedef struct
{
	UNS16 firstCode ;
	UNS16 entryCount ;
	INT16  idDelta ;
	UNS16  idRangeOffset ;
}	subHeaders;

static rangeCount *charRange ;
static int nr_charRange ;
static OS2 os2;
typedef struct
	{
	char name[50];
	int length;
	}
	FF_NAME;
static FF_NAME ff_name; 

static MASTER_CG * p_master;

int get_tt_info(FILE * fdes, MASTER_CG * p_mast)
	{
	OFFSET_TABLE  offset_table;
	UNS32 offset, length;
	UNS16 ntbl;

	p_master = p_mast;

/*< Default the character complement array in case it is not given. Also sym_set. >*/

	memset(p_master->char_comp, -1, 8);
	p_master->char_comp[0] = 14; /* 1 1 1 0 */
	p_master->char_comp[3] = 12; /* 1 1 0 0 */
	p_master->sym_set = 0;	/* means unbound */

	Read_OffsetTable(fdes, &offset_table);
	ntbl = offset_table.numTables;
	Read_DirectoryTable(fdes, &offset_table);  /* reads all the tables */
/* head */
	Get_TableOffsetLength (&offset_table, "head", &offset, &length);
	Read_HeadTable (fdes, offset, length);
/* cmap */
	Get_TableOffsetLength (&offset_table, "cmap", &offset, &length);
	Read_MappingTable (fdes, offset, length);
/* name */
	Get_TableOffsetLength (&offset_table, "name", &offset, &length);
	Read_NameTable (fdes, offset, length);
/* OS/2 */
	Get_TableOffsetLength (&offset_table, "OS/2", &offset, &length);
	Read_OS2Table (fdes, offset, length);
/* post */
	Get_TableOffsetLength (&offset_table, "post", &offset, &length);
	Read_postTable (fdes, offset, length);
/* PCLT - optional */
	Get_TableOffsetLength (&offset_table, "PCLT", &offset, &length);

/*< If the PCLT table is missing we get some data from the required tables. Also, any data
    read from tt.txt for this font overrides any table data. >*/

	if ( f_PCLT_present = length )
		Read_PCLTTable (fdes, offset, length);
	else
		substitute();

/*< The name given in PCLT is not reliable; e.g. Microsoft calls "Arial", "M Arial". >*/

	memset(p_master->face_name, 0, sizeof(p_master->face_name));
	memcpy(p_master->face_name, ff_name.name, (ff_name.length <= 15) ? ff_name.length : 15); 

	try_field(field1);
	try_field(field2);

	
	tt_release_font_fragment (offset_table.table);
	return 0;
}

/* read first 12 bytes of Offset Table */

static void Read_OffsetTable(FILE * fdes, OFFSET_TABLE * offset_table)
	{
	UNS8 *ptr;

	ptr=(UNS8 *)tt_get_font_fragment ((INT32)fdes,(INT32)0,(INT32)12);
	offset_table->version = (SCALED)(GET_WORD(ptr)) << 16;
	offset_table->version += GET_WORD((ptr+2));
	offset_table->numTables = GET_WORD((ptr+4));
	offset_table->searchRange = GET_WORD((ptr+6));
	offset_table->entrySelector = GET_WORD((ptr+8));
	offset_table->rangeShift = GET_WORD((ptr+10));
	tt_release_font_fragment (ptr);

	return ;
	}

/* read directory of Tables */

static int Read_DirectoryTable(FILE * fdes, OFFSET_TABLE * sfntdir)
	{
	int     ii,i, size;
	UNS8  *pf;
	char name_tag[5], rev_tag[5] ;
	UNS8 *ptr ;

   size = sfntdir->numTables * sizeof(TAG_TABLE);
	sfntdir->table = (TAG_TABLE *)malloc((long)size);
	if (sfntdir->table == NULL)
		{
		fprintf(stderr, "*** Read_DirectoryTable(): malloc(sfntdir->table) fails\n\n");
   		return (1);
		}
	ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes,(INT32)12,(INT32)size);
	if ( ptr == NULL )
		{
		fprintf(stderr, "*** Read_DirectoryTable(): tt_get_font_fragment() fails\n\n");
   		return(1);
		}
	for (pf = ptr, ii=0; ii < sfntdir->numTables; ii++)
		{
		sfntdir->table[ii].tag = (SCALED)(GET_WORD(pf)) << 16;
		sfntdir->table[ii].tag += GET_WORD((pf+2));

		name_tag[4] = '\0' ;
		memcpy(name_tag,&(sfntdir->table[ii].tag),4);
#ifndef HI2LOW
		for ( i = 0 ; i < 4; i++)
			rev_tag[3-i]=name_tag[i];
#else
		for ( i = 0 ; i < 4; i++)
			rev_tag[i]=name_tag[i];
#endif
		rev_tag[4] = '\0' ;

		pf += 4;
		sfntdir->table[ii].checkSum = (SCALED)(GET_WORD(pf)) << 16;
		sfntdir->table[ii].checkSum += GET_WORD((pf+2));
		pf += 4;
		sfntdir->table[ii].offset = (SCALED)(GET_WORD(pf)) << 16;
		sfntdir->table[ii].offset += GET_WORD((pf+2));
		pf += 4;
		sfntdir->table[ii].length = (SCALED)(GET_WORD(pf)) << 16;
		sfntdir->table[ii].length += GET_WORD((pf+2));
		pf += 4;
	}
	tt_release_font_fragment (ptr);

	return(0);
	}

static int Get_TableOffsetLength(OFFSET_TABLE * sfntdir, char *dir_name, UNS32 * offset, UNS32 * length)
	{
	int i, j;
	char *name_tag, rev_tag[5] ;

	*offset = *length = 0L ;
	for (i = 0; i < sfntdir->numTables; i++)
		{
		rev_tag[4] = '\0' ;
		name_tag = (char*)&(sfntdir->table[i].tag);
#ifndef HI2LOW
		for ( j = 0 ; j < 4; j++)
			rev_tag[3-j]=name_tag[j];
#else
		for ( j = 0 ; j < 4; j++)
			rev_tag[j]=name_tag[j];
#endif
		if ( strcmp(rev_tag,dir_name) == 0 )
			{
			* offset = sfntdir->table[i].offset;
			* length = sfntdir->table[i].length;
			}
		}
	return 0 ;
	}

static int Read_OneMappingTable(FILE *fdes, long offset, INDEX_DIRECTORY *idxdir)
	{
	int i, j, nr, key, index;
	UNS8 *ptr, *array ;
	MAPPING_TABLE mappingTable ;
	segMappingTable *seg_table ;
	HighByteMappingTable highbyte_table ;
	subHeaders *sub_header ;
	UNS16 *endCount, *startCount, *idDelta, *idOffset ;				 

		ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes,
		   (INT32)idxdir->platform[0].offset + offset, (INT32)6);
		mappingTable.format = GET_WORD((ptr));
		mappingTable.length = GET_WORD((ptr+2));
		mappingTable.version = GET_WORD((ptr+4));
		tt_release_font_fragment (ptr);
		if ( mappingTable.length > 256 )
			return 0;
/* #ifdef NEVER	*/
		ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes,(INT32)idxdir->platform[0].offset+offset,(INT32)mappingTable.length);
		if ( ptr == NULL )
		{
			fprintf(stderr, "*** Read_OneMappingTable(): tt_get_font_fragment() fails\n\n");
   			exit(1);
		}
		if (mappingTable.format == 0 )
		{
			array = ptr + 6 ;
#if DEBUG_CMAP
			for ( j = 0 ; j < 256 ; j++ )
			{
				printf("index %d glyph %d\n",j,array[j]);
				/*
				if ( ((j+1) %24) == 0 )
					getchar();
				*/
			}
#endif
		}
		else if (mappingTable.format == 4 )
		{
			UNS16 Ro, t1, t2;
			INT16 Del;
			seg_table = (segMappingTable*) ptr;
			nr=SWAPW(seg_table->segCountX2);
#if DEBUG_CMAP
			printf("format %d segCountX2 %d\n",SWAPW(seg_table->format),nr);
#endif
			nr = nr>>1 ;
			endCount = (UNS16*)&(seg_table->endCount) ;
			startCount = &endCount[nr+1] ;
			idDelta = &endCount[2*nr+1] ;
			idOffset = &endCount[3*nr+1] ;
			charRange = (rangeCount *)malloc(sizeof(rangeCount)*nr);
			if (charRange == NULL )
				{
				return(1);
				}
			nr_charRange = nr ;
#if DEBUG_CMAP
			/*
			getchar();
			*/
#endif
			if ( info_file != NULL )
				{
				UNS16 * p;
				UNS16 temp;
				for ( j = 0 ; j < nr - 1 ; j++ )
					{
					t1 = charRange[j].startCount = SWAPW(startCount[j]);
					t2 = charRange[j].endCount = SWAPW(endCount[j]);
					Ro = SWAPW(idOffset[j]);
					Del = (INT16)SWAPW(idDelta[j]);
					if ( j > 98)
						fprintf(info_file, "\n  %d.", j + 1);
					else if ( j > 8)
						fprintf(info_file, "\n   %d.", j + 1);
					else
						fprintf(info_file, "\n    %d.", j + 1);
				   	for ( i = t1; i <= t2; ++ i )
						{
						if ( Ro == 0 )
							index = (int)(i + Del);
						else
							{
							p = (UNS16 *)(Ro / 2 + (i - t1) + &idOffset[j]);
							temp = * p;
							index = (int)SWAPW(temp);
							}
						while ( index > 65536 )
							index -= 65536;
						if ( i == t1 )
							fprintf(info_file, "    Char 0x%04x -> Index %d", i, index);
						else
							fprintf(info_file, "\n          Char 0x%04x -> Index %d", i, index);
						}
					}
				}
#if DEBUG_CMAP
				printf("%d startCount %4x endCount %4x ",j,charRange[j].startCount,charRange[j].endCount);
				printf("idDelta %d idOffset %u\n",SWAPW(idDelta[j]),SWAPW(idOffset[j]));
				/*
				if ( ((j+1) %24) == 0 )
	 				getchar();
				*/
#endif
			
			/**********
			if ( JIS_code && idxdir->platform[0].specificID == 1 )
			{
				printf("Please wait when reading JIS-Unicode table\n");
				ReadJIS_UnicodeTable("unicode.txt");
			}
			else
				JIS_code = 0 ;
			***********/
			free(charRange);
		}
		else if (mappingTable.format == 2 )
		{
			highbyte_table.format = GET_WORD((ptr));
			highbyte_table.length = GET_WORD((ptr+2));
			highbyte_table.version = GET_WORD((ptr+4));
			highbyte_table.subHeaderKey = (UNS16*)(ptr+6);
#if DEBUG_CMAP
			printf("format %d length %d version %d\n",
				highbyte_table.format,highbyte_table.length,highbyte_table.version );
#endif
			charRange = (rangeCount *)malloc(sizeof(rangeCount)*mappingTable.length);
			if (charRange == NULL )
			{
			return(1);
			
			}
			nr_charRange = 0 ;
		   ptr += 518 ;
			for ( j = 0 ; j < 256 ; j++ )
			{
				key=SWAPW(highbyte_table.subHeaderKey[j]);
#if DEBUG_CMAP
				printf("index %d Key %u\n",j,key=SWAPW(highbyte_table.subHeaderKey[j]));
#endif
				if (key)
				{
					sub_header = (subHeaders*)(ptr+key);
#if DEBUG_CMAP
					printf("%d firstCode %x entryCount %d idDelta %d idRangeOffset %d\n",j,
					     SWAPW(sub_header->firstCode),SWAPW(sub_header->entryCount),
						  SWAPW(sub_header->idDelta),SWAPW(sub_header->idRangeOffset));
#endif
					charRange[nr_charRange].startCount =( (j << 8 ) | SWAPW(sub_header->firstCode) );
					charRange[nr_charRange].endCount = charRange[nr_charRange].startCount +SWAPW(sub_header->entryCount)-1;
#if DEBUG_CMAP
					printf("%x firstCode %x lastCode %x\n",j,
						charRange[nr_charRange].startCount,charRange[nr_charRange].endCount );
#endif
					nr_charRange++ ;
				}
#if DEBUG_CMAP
				/*
				if ( ((j+1) %24) == 0 )
					getchar();
				*/
#endif
			}
			free(charRange);
		}
		tt_release_font_fragment (ptr);
/*	} */
#if DEBUG_CMAP
	/*
	getchar();
	*/
#endif
/* #endif   NEVER */
	return 0 ;

}

static int Read_MappingTable(FILE *fdes, long offset, long length)
	{
	int     ii, size,nr;
	UNS8  *pf;
	UNS8 *ptr ;
	INDEX_DIRECTORY idxdir;


	ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset, (INT32)length);
	idxdir.tableVersion = GET_WORD((ptr));
	idxdir.numEncoding = GET_WORD((ptr+2));
	tt_release_font_fragment (ptr);

    size = idxdir.numEncoding * sizeof(MAP_DIR);
	ptr = tt_get_font_fragment ((INT32)fdes, (INT32)(offset+4), (INT32)size);
	if ( ptr == NULL )
 		{
		fprintf(stderr, "*** Read_MappingTable(): tt_get_font_fragment() fails\n\n");
   		return(1);
		}
/*	nr = idxdir.numTables < 2 ? idxdir.numTables : 2 ; */
	nr = idxdir.numEncoding ;
	for (pf = ptr, ii=0; ii<nr; ii++)
   		{
		idxdir.platform[0].platformID = (GET_WORD(pf)) ;
		pf += 2;
		idxdir.platform[0].specificEncoding = (GET_WORD(pf)) ;
		pf += 2;
		idxdir.platform[0].offset = (SCALED)(GET_WORD(pf)) << 16;
		idxdir.platform[0].offset += GET_WORD((pf+2));
		pf += 4;
		Read_OneMappingTable (fdes, offset, &idxdir);
		}
#if DEBUG_CMAP
	/*
	getchar();
	*/
#endif 
	tt_release_font_fragment (ptr);
/*	Read_OneMappingTable (fdes,offset,&idxdir); */

	return(0);
	}

static int Read_NameTable(FILE *fdes, long offset, long length)
	{
	int i, j;
	UNS8 *ptr, * p, * p_start;
	NAME_DIR idxdir;
	UNS8 * p_storage;
	NAME_RECORD n_r;

	ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset, (INT32)length);
	idxdir.	formatSelector = GET_WORD((ptr));
	idxdir.numName = GET_WORD((ptr+2));
	idxdir.offset = GET_WORD((ptr + 4));
	tt_release_font_fragment (ptr);
	p_storage = (UNS8 *)tt_get_font_fragment ((INT32)fdes,
		 (INT32)(offset + idxdir.offset), (INT32)1000);
	p_start = ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes,
	   (INT32)offset + sizeof(NAME_DIR), (INT32)idxdir.numName * sizeof(NAME_RECORD));
	for ( i = 0; i < idxdir.numName; ++ i, ptr += sizeof(NAME_RECORD) )
		{
		n_r.platformID = GET_WORD((ptr));
		n_r.specificID = GET_WORD((ptr + 2));
		n_r.languageID = GET_WORD((ptr + 4));
		n_r.nameID = GET_WORD((ptr + 6));
		n_r.length = GET_WORD((ptr + 8));
		n_r.offset = GET_WORD((ptr + 10));
		if ( n_r.platformID == 1  /* Mac */
			 && n_r.languageID == 0 /* english */
			 && n_r.nameID == 1 ) /* font family name */
			 {
			 memset(ff_name.name, 0, sizeof(ff_name.name));
			 memcpy(ff_name.name, p_storage + n_r.offset, n_r.length);
			 ff_name.length = n_r.length;
			 }
		else if (n_r.platformID == 3 /* Microsoft */
			 && n_r.languageID == 1033 /* english */   /* Unicode */
			 && n_r.nameID == 1 ) /* font family name */
			 {
			 int i;
			 char * p_s = p_storage + n_r.offset + 1;
			 char * p_d = ff_name.name;
			 for ( i = 0; i < n_r.length; i += 2, p_s += 2, ++ p_d )
				* p_d = * p_s;
			 ff_name.length = n_r.length/2;
			 }


#ifdef NEVER
		fprintf(fout, "\n  %d   %d   %d ",n_r.platformID, n_r.specificID, n_r.nameID);
		if ( n_r.platformID != 3 )
			fwrite(p_storage + n_r.offset, 1, n_r.length, fout);

/*< For platform ID = 3 (Microsoft) each character requires two bytes according
    to the Unicode standard. Here we just take every other byte. >*/
    
    	else
    		{
			for ( j = 0, p = p_storage + n_r.offset + 1; j < n_r.length; j += 2, p += 2 )
				fwrite(p, 1, 1, fout);
    		}
#endif    				
		}
	tt_release_font_fragment (p_storage);
	tt_release_font_fragment (p_start);
	return 0;
	}
	
static int Read_OS2Table(FILE *fdes, long offset, long length)
	{
	int i = 0, j;
	UNS8 *ptr, * p_start ;

/*	fprintf(fout, "\n\n OS/2 and Windows Metrics "); */
	p_start = ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset,
	 (INT32)length);
	os2.version = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.version); */
	ptr += 2;
	os2.xAvgCharWidth = GET_WORD((ptr));
	p_master->space = os2.xAvgCharWidth;  /* must be changed for proportional spaced font */
/*	fprintf(fout, OS2strs[i ++], os2.xAvgCharWidth); */
	ptr += 2;
	os2.usWeightClass = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usWeightClass); */
	ptr += 2;
	os2.usWidthClass = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usWidthClass); */
	ptr += 2;
	os2.fsType = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.fsType); */
	ptr += 2;
	os2.ySubscriptXSize = GET_WORD((ptr));
	ptr += 2;
	os2.ySubscriptYSize = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.ySubscriptXSize, os2.ySubscriptYSize ); */
	ptr += 2;
	os2.ySubscriptXOffset = GET_WORD((ptr));
	ptr += 2;
	os2.ySubscriptYOffset = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.ySubscriptXOffset, os2.ySubscriptYOffset ); */
	ptr += 2;
	os2.ySuperscriptXSize = GET_WORD((ptr));
	ptr += 2;
	os2.ySuperscriptYSize = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.ySuperscriptXSize, os2.ySubscriptYSize ); */
	ptr += 2;
	os2.ySuperscriptXOffset = GET_WORD((ptr));
	ptr += 2;
	os2.ySuperscriptYOffset = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.ySuperscriptXOffset, os2.ySuperscriptYOffset ); */
	ptr += 2;
	os2.yStrikeoutSize = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.yStrikeoutSize); */
	ptr += 2;
	os2.yStrikeoutPosition = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.yStrikeoutPosition); */
	ptr += 2;
	os2.sFamilyClass = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.sFamilyClass >> 8, */
/*	 os2.sFamilyClass & 0x00FF); */
	ptr += 2;
	for ( j = 0; j < 10; ++ j )
		os2.panose.ar[j] = * ptr ++;
/*	fprintf(fout, OS2strs[i ++], os2.panose.ar[0], os2.panose.ar[1], os2.panose.ar[2], */
/*	os2.panose.ar[3], os2.panose.ar[4], os2.panose.ar[5], os2.panose.ar[6], os2.panose.ar[7], */
/*	os2.panose.ar[8], os2.panose.ar[9]); */
	for ( j = 0; j < 4; j ++, ptr += 4 )
		{
		os2.ulCharRange[j] = (SCALED)(GET_WORD(ptr)) << 16;
		os2.ulCharRange[j] += GET_WORD((ptr + 2));
		}
/*	fprintf(fout, OS2strs[i ++], os2.ulCharRange[0], os2.ulCharRange[1], */
/*	 os2.ulCharRange[2], os2.ulCharRange[3]); */
	for ( j = 0; j < 4; ++ j )
		os2.achVendID[j] = * ptr ++;
/*	fprintf(fout, OS2strs[i ++], os2.achVendID[0], os2.achVendID[1], */
/*	 os2.achVendID[2], os2.achVendID[3]); */
	
	os2.fsSelection = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.fsSelection); */
	os2.usFirstCharIndex = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usFirstCharIndex); */
	ptr += 2;
	os2.usLastCharIndex = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usLastCharIndex); */
	ptr += 2;
	os2.sTypoAscender = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.sTypoAscender); */
	ptr += 2;
	os2.sTypoDescender = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.sTypoDescender); */
	ptr += 2;
	os2.sTypoLineGap = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.sTypoLineGap); */
	ptr += 2;
	os2.usWinAscent = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usWinAscent); */
	ptr += 2;
	os2.usWinDescent = GET_WORD((ptr));
/*	fprintf(fout, OS2strs[i ++], os2.usWinDescent); */
	tt_release_font_fragment (p_start);
	return 0;
	}

static int Read_PCLTTable(FILE *fdes, long offset, long length)
	{
	int i = 0, j;
	UNS8 *ptr, * p_start ;
	PCLT pclt;

/*	fprintf(fout, "\n\n PCL5 Table "); */
	p_start = ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset,
	 (INT32)length);
	pclt.version = (SCALED)(GET_WORD(ptr)) << 16;
	pclt.version += GET_WORD((ptr + 2));
/*	fprintf(fout, PCLTstrs[i ++], pclt.version); */
	ptr += 4;
	pclt.fontNumber = (SCALED)(GET_WORD(ptr)) << 16;
	pclt.fontNumber += GET_WORD((ptr + 2));
/*	fprintf(fout, PCLTstrs[i ++], (pclt.fontNumber >> 24) & 0x007F, */
/*		pclt.fontNumber & 0x0FFF); */
	ptr += 4;
	pclt.pitch = GET_WORD((ptr));
	p_master->space = pclt.pitch;
/*	fprintf(fout, PCLTstrs[i ++], pclt.pitch); */
	ptr += 2;
	pclt.xHeight = GET_WORD((ptr));
/*	fprintf(fout, PCLTstrs[i ++], pclt.xHeight); */
	ptr += 2;
	pclt.style = GET_WORD((ptr));
/*	fprintf(fout, PCLTstrs[i ++], pclt.style); */
	p_master->style_SWP = pclt.style;
	ptr += 2;
	pclt.typeFamily = GET_WORD((ptr));
/*	fprintf(fout, PCLTstrs[i ++], pclt.typeFamily >> 12, pclt.typeFamily & 0x3FF); */
	p_master->typeface = pclt.typeFamily;
	ptr += 2;
	pclt.capHeight = GET_WORD((ptr));
/*	fprintf(fout, PCLTstrs[i ++], pclt.capHeight); */
	ptr += 2;
	pclt.symbolSet = GET_WORD((ptr));
/*	fprintf(fout, PCLTstrs[i ++], pclt.symbolSet); */
	p_master->sym_set = pclt.symbolSet;
	ptr += 2;
	for ( j = 0; j < 16; ++ j )
		pclt.typeface[j] = p_master->face_name[j] = * ptr ++;
/*	fprintf(fout, PCLTstrs[i ++], pclt.typeface[0], pclt.typeface[1], */
/*	 pclt.typeface[2], pclt.typeface[3], pclt.typeface[4], pclt.typeface[5], */
/*	 pclt.typeface[6], pclt.typeface[7], pclt.typeface[8], pclt.typeface[9], */
/*	 pclt.typeface[10], pclt.typeface[11], pclt.typeface[12], pclt.typeface[13], */
/*	 pclt.typeface[14], pclt.typeface[15]); */


	for ( j = 0; j < 8; ++ j )
		pclt.charComplement[j] = p_master->char_comp[j] = * ptr ++;

/*	fprintf(fout, PCLTstrs[i ++], pclt.charComplement[0], pclt.charComplement[1], */
/*	 pclt.charComplement[2], pclt.charComplement[3], pclt.charComplement[4], */
/*	 pclt.charComplement[5], pclt.charComplement[6], pclt.charComplement[7]); */
	for ( j = 0; j < 6; ++ j )
		pclt.fileName[j] = * ptr ++;
/*	fprintf(fout, PCLTstrs[i ++], pclt.fileName[0], pclt.fileName[1], */
/*	 pclt.fileName[2], pclt.fileName[3], pclt.fileName[4], pclt.fileName[5]); */

	pclt.strokeWeight = * ptr ++;
/*	fprintf(fout, PCLTstrs[i ++], pclt.strokeWeight); */
	p_master->stroke_wt = pclt.strokeWeight;
	pclt.widthType = * ptr ++;
/*	fprintf(fout, PCLTstrs[i ++], pclt.widthType); */
	pclt.serifStyle = * ptr ++;
/*	fprintf(fout, PCLTstrs[i ++], pclt.serifStyle); */
	tt_release_font_fragment (p_start);
	return 0;
	}
	

static int Read_postTable(FILE *fdes, long offset, long length)
	{
	int i = 0;
	UNS8 *ptr, * p_start ;
	POST post;

/*	fprintf(fout, "\n\n POST Table "); */
	p_start = ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset,
	 (INT32)length);
	post.formatType = (SCALED)(GET_WORD(ptr)) << 16;
	post.formatType += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.formatType); */
	ptr += 4;
	post.italicAngle = (SCALED)(GET_WORD(ptr)) << 16;
	post.italicAngle += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.italicAngle); */
	ptr += 4;
	post.underlinePosition = GET_WORD((ptr));
/*	fprintf(fout, POSTstrs[i ++], post.underlinePosition); */
	p_master->u_depth = post.underlinePosition;
	ptr += 2;
	post.underlineThickness = GET_WORD((ptr));
/*	fprintf(fout, POSTstrs[i ++], post.underlineThickness); */
	ptr += 2;
	post.isFixedPitch = (SCALED)(GET_WORD(ptr)) << 16;
	post.isFixedPitch += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.isFixedPitch); */
	p_master->spacing = post.isFixedPitch ? 0 : 1;
	ptr += 4;
	post.minMem42 = (SCALED)(GET_WORD(ptr)) << 16;
	post.minMem42 += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.minMem42); */
	ptr += 4;
	post.maxMem42 = (SCALED)(GET_WORD(ptr)) << 16;
	post.maxMem42 += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.maxMem42); */
	ptr += 4;
	post.minMem1 = (SCALED)(GET_WORD(ptr)) << 16;
	post.minMem1 += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.minMem1); */
	ptr += 4;
	post.maxMem1 = (SCALED)(GET_WORD(ptr)) << 16;
	post.maxMem1 += GET_WORD((ptr + 2));
/*	fprintf(fout, POSTstrs[i ++], post.maxMem1); */
	ptr += 4;
	return 0;
	}

static int Read_HeadTable(FILE *fdes, long offset, long length)
	{
	int i = 0, j;
	UNS8 *ptr, * p_start ;
	HEAD head;

/*	fprintf(fout, "\n\n head Table "); */
	p_start = ptr = (UNS8 *)tt_get_font_fragment ((INT32)fdes, (INT32)offset,
	 (INT32)length);
	head.version = (SCALED)(GET_WORD(ptr)) << 16;
	head.version += GET_WORD((ptr + 2));
/*	fprintf(fout, headstrs[i ++], head.version); */
	ptr += 4;
	head.fontRevision = (SCALED)(GET_WORD(ptr)) << 16;
	head.fontRevision += GET_WORD((ptr + 2));
/*	fprintf(fout, headstrs[i ++], head.fontRevision); */
	ptr += 4;
	head.checkSumAdjustment = (SCALED)(GET_WORD(ptr)) << 16;
	head.checkSumAdjustment += GET_WORD((ptr + 2));
/*	fprintf(fout, headstrs[i ++], head.checkSumAdjustment); */
	ptr += 4;
	head.magicNumber = (SCALED)(GET_WORD(ptr)) << 16;
	head.magicNumber += GET_WORD((ptr + 2));
/*	fprintf(fout, headstrs[i ++], head.magicNumber); */
	ptr += 4;
	head.flags = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.flags); */
	ptr += 2;
	head.unitsPerEm = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.unitsPerEm); */
	p_master->scale_factor = head.unitsPerEm;
	ptr += 2;
	for ( j = 0; j < 8; ++ j )
		head.created[j] = * ptr ++;

/*	fprintf(fout, headstrs[i ++], head.created[0], head.created[1], */
/*	 head.created[2], head.created[3], head.created[4], */
/*	 head.created[5], head.created[6], head.created[7]); */
	for ( j = 0; j < 8; ++ j )
		head.modified[j] = * ptr ++;

/*	fprintf(fout, headstrs[i ++], head.modified[0], head.modified[1], */
/*	 head.modified[2], head.modified[3], head.modified[4], */
/*	 head.modified[5], head.modified[6], head.modified[7]); */

	head.xMin = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.xMin); */
	ptr += 2;
	head.yMin = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.yMin); */
	ptr += 2;
	head.xMax = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.xMax); */
	ptr += 2;
	head.yMax = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.yMax); */
	ptr += 2;
	head.macStyle = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.macStyle); */
	ptr += 2;
	head.lowestRecPPEM = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.lowestRecPPEM); */
	ptr += 2;
	head.fontDirectionHint = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.fontDirectionHint); */
	ptr += 2;
	head.indexToLocFormat = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.indexToLocFormat); */
	ptr += 2;
	head.glyphDataFormat = GET_WORD((ptr));
/*	fprintf(fout, headstrs[i ++], head.glyphDataFormat); */
	ptr += 2;

	return 0;
	}

void *tt_get_font_fragment(INT32 clientID, INT32 offset, INT32 length)

/* Returns a pointer to data extracted from the sfnt. */

	{
	FILE  *fdes;
	UNS8 *ptr;
	int len;

	fdes = (FILE *)clientID;
	ptr = (UNS8 *)malloc((size_t)length);

	if (ptr == NULL)
		return ((void *)NULL);

	if (fseek(fdes, offset, 0) != 0)
		return ((void *)NULL);

	len = fread(ptr, 1, (size_t)length, fdes);
	

	return((void *)ptr);
}

void tt_release_font_fragment(void *ptr)

/* Returns the memory at pointer to the heap
  */
{
	free(ptr);
}

/*2 substitute()

Return values:

None

Description:

Gets "substitute" values for the following items which are normally obtained from the PCLT:
style_SWP (mainly for posture), face_name, stroke_wt.

Note that typeface family code and character complement values have to be obtained from the
user if no PCLT is in the font.

2*/  /*<f>*/

static void substitute(void)
	{
	p_master->style_SWP = os2.fsSelection & 1;	 /* normal or italic */
	switch (os2.usWidthClass)
		{
		case 1:
			p_master->style_SWP |= 3 << 1;
			break;
		case 2:
			p_master->style_SWP |= 2 << 1;
			break;
		case 3:
		case 4:
			p_master->style_SWP |= 1 << 1;
			break;
		case 5:
			break;
		case 6:
		case 7:
			p_master->style_SWP |= 6 << 1;
			break;
		case 8:
		case 9:
			p_master->style_SWP |= 7 << 1;
			break;
		default:;			 
		}
	switch (os2.usWeightClass)
		{
		case 100:
			p_master->stroke_wt = -7;
			break;
		case 200:
			p_master->stroke_wt = -5;
			break;
		case 300:
			p_master->stroke_wt = -3;
			break;
		case 400:
			p_master->stroke_wt = 0;
			break;
		case 500:
			p_master->stroke_wt = 1;
			break;
		case 600:
			p_master->stroke_wt = 2;
			break;
		case 700:
			p_master->stroke_wt = 3;
			break;
		case 800:
			p_master->stroke_wt = 5;
			break;
		case 900:
			p_master->stroke_wt = 7;
			break;
		default:
			p_master->stroke_wt = 0;
		}
	strcpy(p_master->face_name, "               ");
	memcpy(p_master->face_name, ff_name.name, (ff_name.length <= 15) ? ff_name.length : 15); 
	install_cc_etc();

/*< If the font is proportional the value of space is incorrect and must be obtained later. >*/

	if ( p_master->spacing )
		p_master->space = 0;
	}

/*2 setbits()

Return values:

FALSE if no error, else TRUE

Description:

Given a range of bits to be set in the character complements array, this function tests to
make sure the bits are between 0 - 63 and then sets bits according to array element 
i = 7 - bit/8, bit_in_byte_i = bit % 8.	Note that here, "set" means set to 0.

2*/  /*<f>*/

static int setbits(int m, int n)
	{
	UNS8 mask;
	if ( m > 63 || n > 63 || m < 0 || n < 0 ) 
		return TRUE;
	while ( m ++ <= n )
		{
		mask = 1 << (m % 8);
		p_master->char_comp[7 - m / 8] ^= mask;
		}
	}

/*2 try_field()

Return values:

None

Description:

This function parses a string for a typeface family ID, a character complements array or a
symbol set value (e.g. 10U). These are introduced by 'f', 'c' or 's' respectively. The
cc array is actually in the form of a string of bit numbers separated by commas (with no
spaces) which specify which bits are to be turned off. (It is not necessary to set bit 0
to off since this is implied for Unicode. A range of bits can also be specified using "-".
for example 23,27,28-31 can be used to turn off bits 23, 27, 28, 29, 30, 31. 

2*/  /*<f>*/

static void try_field(char * field)
	{
	int l, n, f_error = TRUE, f_gotstart = FALSE;
	char c, temps[5], tempe[5];
	if ( (l = strlen(field)) == 0 )
		return;
	switch ( tolower(field[0]) )
		{
		case 'f':	/* should be a family ID integer */
			if ( sscanf(&field[1], "%d", &n ) != 1 )
				fprintf(stderr, "Unable to interpret family ID %s for TT %s", &field[0], tfn);
			else
				p_master->typeface = (UNS16)n;
			break;
		case 's':	/* should be symbol set code */
			if ( isupper(c = * (field + l - 1) ) )
				{
				c -= 64;
				* (field + 1 - 1) = 0;
				if ( sscanf(&field[1], "%d", &n) == 1 )
					{
					p_master->sym_set = n * 32 + c;
					f_error = FALSE;
					}
				}
			if ( f_error )
				fprintf(stderr, "Unable to interpret sym. set ID %s for TT %s", &field[0], tfn);
			break;
		case 'c': /* should be char. complement unset bits */
			{
			int f_gotstart = FALSE;
			char * p_s = &field[1];
			char * p_t = p_s;
			int m;
			f_error = FALSE;
			while ( * p_s )
				{
				c = * p_t;
				if ( c == ',' || c == '-' || c == '\0' )
					{
					* p_t = '\0';
					if ( c != '\0' )
						++ p_t;
					if ( sscanf(p_s, "%d", &n) != 1 )
						f_error = TRUE;
					if ( c != '-' )
						{
						if ( f_gotstart )
							{
							f_gotstart = FALSE;
							f_error = setbits(m, n);
							}
						else
							f_error = setbits(n,n);
						}
					else
						{
						f_gotstart = TRUE;
						m = n;
						}
					}
				else
					f_error = TRUE;
				if ( f_error )
					break;
				}
			if ( f_error )
				fprintf(stderr, "Unable to interpret char. comp. bits %s for TT %s", &field[0], tfn);
			}
		}
	}

/*2  ()

Return values:


Description:


2*/  /*<f>*/

static int install_cc_etc()
	{
	int num = sizeof(name_char_tt) / sizeof(NAME_CHAR_TT);
	int i;

	for ( i = 0 ; i < num ; ++ i )
		if (strnicmp(name_char_tt[i].name, p_master->face_name, 4) == 0	)
			break;
	if ( i == num )
		{
		p_master->typeface = 4095;    /* maximum value */
		p_master->sym_set = 0;
		memset(p_master->char_comp, (int)0xff, 8);
		return(1);
		}
	p_master->typeface = name_char_tt[i].typeface;
	p_master->sym_set = name_char_tt[i].sym_set;
	memcpy(p_master->char_comp, char_comp_tt[name_char_tt[i].cc_index].cc, 8);
	return(0);
	}




