/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/bmp/bmputl.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: bmputl.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.8   06 Apr 1995 15:08:24   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.7   03 Jun 1994 08:46:24   JOE
 * In char_left_ref(), when calculating "orig_black_width", check if
 * character is a contourless character. If so, set to 0 rather than
 * using the "bound_box" data in the event that it is invalid.
 * 
 *    Rev 1.6   22 Apr 1994 09:20:48   LISA
 * 
 * Modified copyright/disclaimer notice.
 * 
 *    Rev 1.5   18 Apr 1994 08:17:16   JOE
 * Added comment in char_left_ref() describing fix for multi-part
 * characters with nonzero offset in first part.
 * 
 *    Rev 1.4   15 Apr 1994 16:23:14   JOE
 * 
 * Added new function char_left_ref().
 * Added arg to find_part2whole() function.
 * Made changes to fix floating accent problem.
 * 
 *    Rev 1.3   12 Feb 1993 11:04:22   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   05 Jan 1993 15:58:18   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:28:36   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:33:04   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* bmputl.c */

/* History                                                           */
/*    03-Oct-91 jfd  Removed #if TILE conditional around print_bm(). */
/*    10-Oct-91 rs   Modify 'translate_coords()' for bezier data &   */
/*                   Type 1 input.                                              */
/*    25-Oct-91 jfd  Instead of conditionally compiling entire file  */
/*                   based on CGBITMAP, conditionally compile based  */
/*                   on (CACHE || CHAR_SIZE || CHAR_HANDLE) to get   */
/*                   rid of link errors that occur if CGBITMAP is not*/
/*                   enabled.                                        */
/*    16-Jan-92 rs   Fix 'translate_coords() for PS type 1.          */
/*    23-Jan-92 awr  Removed translate_coords()                      */
/*    05-Feb-92 jfd  In print_bm(), changed format for x and y origins */
/*                   from "%d" to "%ld".                             */
/*    22-Mar-92 awr  Changed bitmap_dimensions() to take LONGs       */
/*                   Removed unused global tiling variables          */
/*    30-Mar-92 jfd  In bitmap_dimensions(), fixed DBG2 statement      */
/*                   to print "x" and "y" instead of "ur.x" and "ur.y" */
/*    03-Apr-92 rs   Portability cleanup (see port.h).                */
/*    04-Apr-92 awr  Changed calling sequence of find_part2whole()   */
/*    12-Apr-92 awr  Moved translation calcs from manipulate() to    */
/*                   make_gaso()                                     */
/*    18-Apr-92 awr  Removed clip_tile()                             */
/*     2-May-92 awr  Moved if_state.tt initialization from           */
/*                   manipualte() to make_gaso()                     */
/*    14-Jun-92 awr  Changed types of xlate and tbound to INTRs      */
/*    3-Jul-92  awr  Moved structs_are_equal() to comp_pix.c         */
/*                   Fixed test in make_gaso()... was big slow down  */
/*   11-Aug-92  jfd  Trying macro MDES2BM for optimization.          */
/*                   Included imath.h.                                          */
/*   16-Aug-92  rs   Fix nested comment.                                    */
/*   19-Aug-92  awr  In make_gaso(), only return prematurely if      */
/*                   processing a simple character.                  */
/*   15-Sep-92  jfd  Conditionally compile the following routines    */
/*                   based on IF_RDR:                                */
/*                   union_bound_boxes(), find_part2whole(),         */
/*                   make_gaso().                                    */
/*   16-Sep-92  jfd  Removed GLOBAL declarations for "entire_bm" and */
/*                   "entire_tt". They are now being declared GLOBAL */
/*                   in maker.c.                                     */
/*                   Changed declaration for "gaso_pn" from GLOBAL   */
/*                   to EXTERN. It is now declared GLOBAL in         */
/*                   comp_pix.                                       */
/*   15-Nov-92  rs   Port to Macintosh - use ANSI_DEFS on functions  */
/*   05-Jan-93  jfd  ANSI C function declaration changes             */
/*   08-Feb-93  jfd  VXWorks support.                                */
/*   13-Apr-94  awr  Added new function char_left_ref().             */
/*                   Added arg to find_part2whole() function.        */
/*                   Made changes to fixfloating accent problem.     */
/*   15-Apr-94  jfd  Added arg go char_left_ref() - y offset.        */
/*   27-May-94  jfd  In char_left_ref(), when calculating            */
/*                   "orig_black_width", check if character is a     */
/*                   contourless character. If so, set to 0 rather   */
/*                   than using the "bound_box" information in the   */
/*                   event that it is invalid.                       */
/*-------------------------------------------------------------------*/


#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>

#include "cgconfig.h"
#include "port.h"
#include "debug.h"
#include "profile.h"
#include "ifmem.h"
#include "cgif.h"
#include "cache.h"
#include "segments.h"
#include "ix.h"
#include "if_type.h"
#include "bitmap.h"
#include "imath.h"


#ifdef LINTARGS
EXTERN UW16   DArd_char(PCHR_DEF);

VOID       union_bound_boxes(PBOX, PBOX, SW16, SW16);

#else /* no LINTARGS */
EXTERN UW16   DArd_char();

VOID       union_bound_boxes();
#endif /* LINTARGS */

/* #if CGBITMAP*/     /*  Entire module is conditionally compiled  */
#if (CACHE || CHAR_SIZE || CHAR_HANDLE)


/*-------------------------------------------------*/
EXTERN CHR_DEF_HDR chr_def_hdr;
EXTERN CHR_DEF chr_def[];
EXTERN IF_STATE if_state;
EXTERN int gaso_pn;   /* last made gaso */
/*-------------------------------------------------*/


#if IF_RDR
/*----------------------*/
/*      make_gaso       */
/*----------------------*/
/*  Make a Grid aligned, Scaled Outline. Keep track of the current
 *  gaso and don't remake if not necessary.
 */

#if defined (ANSI_DEFS)
GLOBAL UW16  make_gaso(int pn, PCHR_DEF cd)
#else
GLOBAL UW16
make_gaso(pn, cd)
    int pn;
    PCHR_DEF    cd;
#endif
{
    UW16 status;

    DBG("make_gaso()\n");

    if(pn == gaso_pn && pn == 0)  /* 8-19-92 */
        return SUCCESS;


    gaso_pn = -1;   /* make invalid in case we fail */

    status = DArd_char(cd);
    if (status)
        return status;

  /*  Set translation for scaling to working fop space. Resulting outline
   *  can end up in any quadrant
   */

    if_state.tt.x = if_state.tt.y = 0;

    status = intel_char();
    if (status)
       return status;

  /*  Compute the translation
   *  from working space to working bitmap space for cc base.
   */

#if INTR_SIZE == 16
    DBG4("work buffer bounds:  (%d, %d)   (%d, %d)\n",
                                        if_state.tbound.ll.x,
                                        if_state.tbound.ll.y,
                                        if_state.tbound.ur.x,
                                        if_state.tbound.ur.y);
#else
    DBG4("work buffer bounds:  (%ld, %ld)   (%ld, %ld)\n",
                                        if_state.tbound.ll.x,
                                        if_state.tbound.ll.y,
                                        if_state.tbound.ur.x,
                                        if_state.tbound.ur.y);
#endif

  /*  NOTE: the grid alignment is correct whether x_min and y_min
   *  are positive or negative becase "&" always truncates to the left.
   *  The reason for rounding the two different ways is that after
   *  translating by tt, we want
   *      XFLOOR  (tt.x + tbound.ll.x + x.half_pixel) = 0
   *      YCEILING(tt.y + tbound.ll.y + y.half_pixel) = 1
   *
   *  This requires that:
   *
   *      -x.half_pixel <= tt.x + tbound.ll.x <  x.half_pixel
   *      -y.half_pixel <  tt.y + tbound.ll.y <= y.half_pixel
   */

    if_state.xlate.x =   -((if_state.tbound.ll.x + if_state.x.half_pixel)
                                      & if_state.x.grid_align);
    if_state.xlate.y =   ((-if_state.tbound.ll.y + if_state.y.half_pixel)
                                      & if_state.y.grid_align);

  /*  Compute working bitmap bounding box: assumes we will
   *  translate by if_state.tt
   */

    if_state.tbound.ll.x += if_state.xlate.x;
    if_state.tbound.ll.y += if_state.xlate.y;
    if_state.tbound.ur.x += if_state.xlate.x;
    if_state.tbound.ur.y += if_state.xlate.y;

  /* tt must by SW16 and is used for Intellifont translation before
   * resolution is increased to INTR
   */
    if_state.tt.x = (SW16)if_state.xlate.x;
    if_state.tt.y = (SW16)if_state.xlate.y;
    DBG2("    if_state.tt.x = %d   if_state.tt.y = %d\n",
                    if_state.tt.x, if_state.tt.y);

    if_state.tbound.ll.y += (INTR)if_state.y.half_pixel;
    if_state.tbound.ur.y += (INTR)if_state.y.half_pixel;

#if INTR_SIZE == 16
    DBG4(" bm fops bounding box (%d,%d)   (%d,%d)\n\n",
                                                    if_state.tbound.ll.x,
                                                    if_state.tbound.ll.y,
                                                    if_state.tbound.ur.x,
                                                    if_state.tbound.ur.y);
#else
    DBG4(" bm fops bounding box (%ld,%ld)   (%ld,%ld)\n\n",
                                                    if_state.tbound.ll.x,
                                                    if_state.tbound.ll.y,
                                                    if_state.tbound.ur.x,
                                                    if_state.tbound.ur.y);
#endif

    gaso_pn = pn;    /* current grid aligned scaled outline */
    return SUCCESS;
}
#endif  /* IF_RDR */

#if IF_RDR
/*----------------------*/
/*   union_bound_boxes  */
/*----------------------*/
/*  Translate the src bounding box by (dx, dy) and then merge it
 *  with the destination bounding box. Return result in the
 *  destination bounding box.
 */
#if defined (ANSI_DEFS)
GLOBAL VOID union_bound_boxes(PBOX dst, PBOX src, SW16 dx, SW16 dy)
#else
GLOBAL VOID
union_bound_boxes(dst, src, dx, dy)
    PBOX dst, src;
    SW16 dx, dy;
#endif /* ANSI_DEFS */
{
    dst->ll.x = MIN(dst->ll.x, dx + src->ll.x);
    dst->ll.y = MIN(dst->ll.y, dy + src->ll.y);
    dst->ur.x = MAX(dst->ur.x, dx + src->ur.x);
    dst->ur.y = MAX(dst->ur.y, dy + src->ur.y);
}



/*----------------------*/
/* bitmap_dimensions    */
/*----------------------*/
/*
Calculate the width and depth of a bit map in (bytes x pixels)
given the upper right corner in fops. The lower left corner is
assumed to be (1,1) in fops.


    The "final" outline that is rasterized is still in "fops"- fractional
    output pixels. The outline has been translated so as to be close to the
    the axis of the first quadrant. It is then translated 1/2 pixel up
    to make protect againts "widows" and "flats" in curves. The result
    of all of this is that if SW16DVECTOR ll is the lower left corner of
    the resulting fop bounding box, then

            -0.5 pixel <= ll.x <  0.5 pixel
               0       <  ll.y <= 1


    The width calculated is the resulting black width. The charcter
    transition array will be one bit wider to include the rightmost
    off transition.

*/



/*----------------------*/
/*    chr_left_ref      */
/*----------------------*/
/*  This function may only be called for quadrant rotations.
 *  Returned value is not really a box, but two vectors.  The vector ll of
 *  of the box is the new location in design space of the left reference
 *  baseline.  This new location takes into account chages in black width
 *  that cause the left side bearing to change.  It is calculated off the
 *  inverse image of the bitmap so that des2bm() maps it to it's bitmap
 *  location in the bitmap.
 *
 *  The vector ur is the vector from this new origin in design space to the
 *  pre-image in design space of the lower left pixel of the bitmap.
 */
#if defined (ANSI_DEFS)
GLOBAL BOX char_left_ref(SW16VECTOR org, BOX pixBound, SW16 yoff)
#else
GLOBAL BOX char_left_ref(org, pixBound, yoff)
    SW16VECTOR org;
    BOX pixBound;
    SW16 yoff;
#endif
{
    BOX retVal;
    SW16VECTOR prell, preur, dll, dur;
    SW16 orig_lsb, new_lsb;
    SW16VECTOR new_org;    /* character origin (design units)        */

    DBG("char_left_ref()\n");
    DBG2("org %d %d\n", org.x, org.y);
    DBG2("pixBound %d %d\n", pixBound.ll.x, pixBound.ll.y);
    DBG2("          %d %d\n", pixBound.ur.x, pixBound.ur.y);

    pixBound.ur.x <<= log_xpix; 
    pixBound.ur.y <<= log_ypix; 
    /*  assume pixBound.ll is (0,0) */

    prell = inv_des2bm(pixBound.ll);
    preur = inv_des2bm(pixBound.ur);
    DBG2("prell %d %d\n", prell.x, prell.y);
    DBG2("preur %d %d\n", preur.x, preur.y);

    if(prell.x < preur.x)
    {
        dll.x = prell.x;
        dur.x = preur.x;
    }
    else
    {
        dll.x = preur.x;
        dur.x = prell.x;
    }
    if(prell.y < preur.y)
    {
        dll.y = prell.y;
        dur.y = preur.y;
    }
    else
    {
        dll.y = preur.y;
        dur.y = prell.y;
    }

    DBG2("dll = (%d, %d)\n", dll.x, dll.y);
    DBG2("dur = (%d, %d)\n", dur.x, dur.y);

    orig_lsb = if_state.bound_box.ll.x - org.x;

    if(if_state.ConnectingChar)
    {
#if ((ELASTIC_Y) | (ELASTIC_X))
        pixel_align (org.x, &if_state.x, R_TWO_I, 0);
#else
        pixel_align (org.x, &if_state.x, R_TWO_I);
#endif
        new_lsb = dll.x - if_state.value;
        DBG("     ConnectingChar:\n");
        DBG2("    orig_lsb %d    new_lsb %d\n", orig_lsb, new_lsb);
    }
    else
    {
        SW16 orig_black_width, new_black_width;

        if (if_state.num_loops)
           orig_black_width = if_state.bound_box.ur.x
                            - if_state.bound_box.ll.x;
        else
           orig_black_width = 0;

        new_black_width  = dur.x - dll.x;

        new_lsb = orig_lsb + (orig_black_width - new_black_width)/2;
        DBG2("    orig_lsb %d    new_lsb %d\n", orig_lsb, new_lsb);
    }

    new_org.x = dll.x - new_lsb;
    new_org.y = if_state.aBaselnVal;

    /*  Fix for one part compound char with nonzero offset.           */
    /*  There may still be a bug for multiple part chars with nonzero */
    /*  offset in first part.                        -ss,bjg 7/23/91  */
    /*  Using "yoff" to adjust new_org should fix bug for multipart   */
    /*  characters with nonzero offset in first part.   -jfd 4/15/94  */      
#if 0
    if( chr_def->offset.y )
       new_org.y -= chr_def->offset.y;
#endif
    new_org.y -= yoff;


    DBG2("new_org = (%d, %d)\n", new_org.x, new_org.y);

    retVal.ll = new_org;

    retVal.ur.x = prell.x - new_org.x;
    retVal.ur.y = prell.y - new_org.y;

    DBG2("retVal %d %d\n", retVal.ll.x, retVal.ll.y);
    DBG2("       %d %d\n", retVal.ur.x, retVal.ur.y);

    return retVal;
}
/*----------------------*/
/*   find_part2whole    */
/*----------------------*/
/* calculate pixel address in destination bitmap of pixel
 * origin (0,0) in compound part bitmap
 */
#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR find_part2whole(SW16VECTOR wholeOrg,
                                          SW16VECTOR tt, PCHR_DEF cd)
#else
GLOBAL SW16VECTOR
find_part2whole(wholeOrg, tt, cd)
    SW16VECTOR wholeOrg;
    SW16VECTOR tt;       /* translation for part  */
    PCHR_DEF   cd;
#endif
{
    SW16VECTOR ctt;       /* translation for part  */
    SW16VECTOR des_tran;
    SW16VECTOR p2w;

    DBG("find_part2whole()\n");
    DBG2("wholeOrg %d %d\n", wholeOrg.x, wholeOrg.y);
    DBG2("pix_bound %d %d\n", cd->pix_bound.ll.x, cd->pix_bound.ll.y);
    DBG2("          %d %d\n", cd->pix_bound.ur.x, cd->pix_bound.ur.y);

  /* des_tran = design unit translation to move compound part so
   * in design space so that it is positioned correctly with respect
   * to the other parts.
   */
    if(!if_state.quadrant)  /* arbitrary rotation */
    {
    des_tran.x = chr_def_hdr.origin.x-(if_state.escape_box.ll.x - cd->offset.x);
    des_tran.y = chr_def_hdr.origin.y-(if_state.escape_box.ll.y - cd->offset.y);
        ctt = if_state.tt;       /* save translation of part */
    }
    else
    {
        BOX foo;
        foo = char_left_ref(if_state.escape_box.ll, cd->pix_bound, (SW16)0);
        des_tran.x = wholeOrg.x + cd->offset.x + foo.ur.x;
        des_tran.y = wholeOrg.y + cd->offset.y + foo.ur.y;
        ctt.x = ctt.y = 0;
    }
    DBG2("des_tran = %d %d\n", des_tran.x, des_tran.y);


  /*  Must grid align des_tran. */

#if ((ELASTIC_Y) | (ELASTIC_X))
    pixel_align (des_tran.x, &if_state.x, R_TWO_I, 0);
#else
    pixel_align (des_tran.x, &if_state.x, R_TWO_I);
#endif
    des_tran.x = if_state.value;
#if ((ELASTIC_Y) | (ELASTIC_X))
    pixel_align (des_tran.y, &if_state.y, R_TWO_I, 0);
#else
    pixel_align (des_tran.y, &if_state.y, R_TWO_I);
#endif
    des_tran.y = if_state.value;
    DBG2("grid aligned des_tran = %d %d\n", des_tran.x, des_tran.y);

    if_state.tt = tt;        /* restore translation for whole */
#if defined MDES2BM
    p2w = MDES2BM(des_tran);  /* part to whole translate */
#else
    p2w = des2bm(des_tran);  /* part to whole translate */
#endif /* defined MDES2BM */
    p2w.x -= ctt.x;
    p2w.y -= ctt.y;
    DBG2("p2w in fops = %d %d\n", p2w.x, p2w.y);

    p2w.x = XFLOOR(p2w.x);
    p2w.y = YFLOOR(p2w.y);
    DBG2("p2w in pixels = %d %d\n", p2w.x, p2w.y);

    return p2w;
}
#endif  /* IF_RDR */



#ifdef DEBUG
#if defined (ANSI_DEFS)
GLOBAL VOID  print_bm(PIFBITMAP bm)
#else
GLOBAL VOID
print_bm(bm)
    PIFBITMAP bm;
#endif
{
    DBG3("    width(bytes) x depth:     %d x %d = %ld bytes\n",
                 bm->width, bm->depth, (SL32)bm->width * (SL32)bm->depth);
    DBG2("    left_indent, top_indent   %d %d\n", bm->left_indent,
                                                  bm->top_indent);
    DBG2("    black_width, black_depth  %d %d\n", bm->black_width,
                                                  bm->black_depth);
    DBG2("    xorigin, yorigin          %ld %ld\n", bm->xorigin,
                                                    bm->yorigin);
}
#endif

#endif /* (CACHE || CHAR_SIZE || CHAR_HANDLE) */

