/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/bmp/merge.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: merge.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.9   06 Apr 1995 15:08:26   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.8   22 Apr 1994 09:20:44   LISA
 * 
 * Modified copyright/disclaimer notice.
 * 
 *    Rev 1.7   07 Jan 1994 10:57:28   JOE
 * In merge(), fixed shifting process if !LEFT_TO_RIGHT_BIT_ORDER.
 * 
 *    Rev 1.6   19 May 1993 14:53:56   JOE
 * More changes for RASTER_ORG support.
 * 
 *    Rev 1.5   18 May 1993 17:50:34   JOE
 * Added support for RASTER_ORG != EIGHT_BIT_CHUNK.
 * 
 *    Rev 1.4   17 May 1993 18:03:30   JOE
 * In merge(), when calculating "ps" and "pd", cast "src->bm" and "dst->bm"
 * as LPUB8 due to change made in IFBITMAP structure.
 * 
 *    Rev 1.3   12 Feb 1993 11:14:22   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   05 Jan 1993 16:06:42   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:30:10   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:33:08   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* merge.c */


/*---------------------------------------------------------------------
     11-22-90   awr   Fixed bug in merge, rewrote bitblt portion based
                      on bg's if2.2 version.
     4-Feb-91   awr   Split off from maker.c
     9-Jun-91   awr   changed merge() parameter types to WORDVECTOR
     25-Jun-91   awr   changed merge() parameters
    17-Jun-91   jfd   Moved "debug.h" after "port.h".
                      Moved "cgconfig.h" before "port.h".
     1-Sep-91   awr   Included "bitmap.h"
    25-Oct-91   jfd   Conditionally compiling entire module based on
                      (CACHE | CHAR_HANDLE | CHAR_SIZE) instead of CGBITMAP.
    03-Apr-92   rs    Portability cleanup (see port.h).
    17-Apr-92   awr   Improved comments
    21-Jun-92   awr   Conditionally compile on IF_RDR
    05-Jan-93   jfd   ANSI C function declaration changes.
    08-Feb-93   jfd   VXWorks support.
    17-May-93   jfd   In merge(), when calculating "ps" and "pd", cast
                      "src->bm" and "dst->bm" as LPUB8 due to change made 
                      in IFBITMAP structure.
    18-May-93   jfd   Added support for RASTER_ORG !== EIGHT_BIT_CHUNK.
    19-May-93   jfd   More changes for RASTER_ORG support.
    07-Jan-94   jfd   In merge(), fixed shifting process if 
                      !LEFT_TO_RIGHT_BIT_ORDER.
-----------------------------------------------------------------------*/


#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"

#if IF_RDR && (CACHE || CHAR_HANDLE || CHAR_SIZE)

#include "port.h"
#include "debug.h"
#include "ifmem.h"                     /* 17-Feb-91 awr */
#include "cgif.h"
#include "segments.h"
#include "ix.h"
#include "if_type.h"
#include "bitmap.h"


#define half_xpix  if_state.x.half_pixel
#define half_ypix  if_state.y.half_pixel

/*-------------------------------------------------*/
EXTERN IF_STATE if_state;
/*-------------------------------------------------*/

/*----------------------*/
/*   merge              */
/*----------------------*/
/*  Merge the bitmap in the source buffer into the destination buffer
 *  so that the two origins coincide.  It is assumed that the src bitmap
 *  will fit in the destination.  There is no error checking.
 *
 *  Fields in the IFBITMAP structure that must be valid:
 *
 *    src                dst
 *    _________________________
 *    width              width
 *                       depth
 *    left_indent
 *    top_indent
 *    black_width
 *    black_depth
 *
 */
#if defined (ANSI_DEFS)
GLOBAL VOID  merge(PIFBITMAP src, PIFBITMAP dst, SW16VECTOR org)
#else
GLOBAL VOID
merge(src, dst, org)
    PIFBITMAP  src;  /* Source      buffer                                */
    PIFBITMAP  dst;  /* Destination buffer                                */
    SW16VECTOR org;  /* source pixel origin (0,0) in destination bitmap   */
#endif
{
    SW16   ddx, ddy;  /* pixel address of blackbox in destination bitmap */
    LPCHUNK ps, pd;    /* pointers used to merge the bitmaps.             */
    SW16   sstart;    /* SB8 offset to starting SB8 of src blackbox    */
    SW16   dstart;    /* SB8 offset to starting SB8 of dst blackbox    */
    SW16   sw;        /* SB8 width of src blackbox                      */
    SW16   dw;        /* SB8 width of dst blackbox                      */
    UW16  i,j;
    UW16  mid_src_ct;
    UW16  sbit0, dbit0;
    UW16  sbit1, dbit1;
    SW16   n, nb;

    SW16   num_bytes;

    DBG2("\nmerge(org in destination = (%d, %d)", org.x, org.y);


  /*  Compute pixel address of blackbox in destination space. Address
   *  is upper left corner of pixel with origin at upper left of bitmap.
   */
    ddx = org.x;
    ddy = dst->depth - (org.y + src->black_depth);
    DBG2("    ddx, ddy  %d, %d\n", ddx, ddy);

  /*  Compute overhangs. Sbit0 and dbit0 are the number of white chunks
   *  to the left of the black box in the left most chunks that intersect
   *  the black box.
   *  Sbit1 and dbit1 are the number of black bits in the right most
   *  chunks that intersect the black box.
   */
    sbit0 = src->left_indent & (UW16)CHUNK_MASK;
    dbit0 = ddx & (UW16)CHUNK_MASK;
    sbit1 = (sbit0 + src->black_width) & (UW16)CHUNK_MASK;  if(!sbit1) sbit1 = (UW16)CHUNK_MASK + 1;
    dbit1 = (dbit0 + src->black_width) & (UW16)CHUNK_MASK;  if(!dbit1) dbit1 = (UW16)CHUNK_MASK + 1;
    DBG2("    sbit0 = %d    dbit0 = %d\n", sbit0, dbit0);
    DBG2("    sbit1 = %d    dbit1 = %d\n", sbit1, dbit1);

  /*  Compute starting chunk offset along a raster and the number
   *  of chunks that covers the black width in a raster. Note that
   *  sw and dw can vary from each other by as much as 1 depending
   *  on overhangs.
   */

    sstart = src->left_indent >> CHUNK_SHIFT;
    dstart = ddx >> CHUNK_SHIFT;
    sw    = ((src->left_indent + src->black_width + (UW16)CHUNK_MASK) >> CHUNK_SHIFT) - sstart;
    dw    = ((ddx + src->black_width + (UW16)CHUNK_MASK) >> CHUNK_SHIFT) - dstart;

  /*  Set pointers */

    ps = (LPCHUNK)((LPUB8)src->bm + (src->top_indent * src->width)) + sstart;  /* 5-17-93 */
    pd = (LPCHUNK)((LPUB8)dst->bm + (ddy * dst->width)) + dstart; /* 5-17-93 */

  /*  Special case if no shifting is required */

    if(sbit0 == dbit0)
    {
        for(i=0; i<src->black_depth; i++)
        {
            for(j=0; j<sw; j++)
                *pd++ |= *ps++;

            ps += (src->width >> (CHUNK_SHIFT - 3)) - sw;
            pd += (dst->width >> (CHUNK_SHIFT - 3)) - dw;
        }
        return;
    }

  /*  Shifting case. Transfer each raster in (up to) three steps:
   *    1. If src overhangs on left, then shift src and or into dst.
   *    2. Transfer all "middle" src chunks that overlap two dst
   *       chunks (if any).
   *    3. If src overhangs on right, then shift src and or into dst.
   */

  /* number of source chunks to transfer in the middle loop, step 2 */

    mid_src_ct = sw;  /* Assume no overhangs, all src chunks are in step 2 */
    if(sbit0 > dbit0) /* Source overhangs on left  */
        mid_src_ct--;
    if(sbit1 < dbit1) /* Source overhangs on right */
        mid_src_ct--;

    DBG1("    mid_src_ct = %u\n", mid_src_ct);

  /*  Shifting values. Each source overlaps two dst chunks.
   *  Shift src chunk >> n  before oring into left  dst chunk
   *  Shift src chunk << nb before oring into right dst chunk
   */

    n  = dbit0 - sbit0;   /* calculate n. Assume no left overhang        */
    if(n<0)               /* If n < 0, then left overhnag and we really  */
        n = (SW16)CHUNK_MASK + 1 + n;        /* calculated -nb, so convert to n.            */

    nb = (SW16)CHUNK_MASK + 1 - n;
    DBG2("    n = %u   nb = %u\n", n, nb);

  /*  For each raster in the black box shift and transfer the data from
   *  the source to the destination
   */

    for(i=0; i<src->black_depth; i++)
    {

      /*  Step 1. Take care of the case that the source overhangs the
       *  destination on the left. In this case, this first source chunk only
       *  contributes to one destination chunk.
       */
        if(sbit0 > dbit0)
#if LEFT_TO_RIGHT_BIT_ORDER
            *pd |= *ps++ << nb;
#else
            *pd |= *ps++ >> nb;
#endif

      /*  Step 2. Transfer the middle data. For each source chunk, put
       *  a piece in each of the two destination chunk that it overlaps.
       */
        for(j=0; j<mid_src_ct; j++)
        {
#if LEFT_TO_RIGHT_BIT_ORDER
            *pd++ |= *ps   >> n;
            *pd   |= *ps++ << nb;
#else
            *pd++ |= *ps   << n;
            *pd   |= *ps++ >> nb;
#endif
        }

              /*  Rule for pointers after each of these three steps:
               *      ps now points to the next src chunk to be processed.
               *      pd now points to the last dest chunk processed.
               */

      /*  Step 3. Take care of the case that the source overhangs the
       *  destination on the right. In this case, this last source chunk only
       *  contributes to one destination chunk.
       */

        if(sbit1 < dbit1)
#if LEFT_TO_RIGHT_BIT_ORDER
            *pd |= *ps++ >> n;
#else
            *pd |= *ps++ << n;
#endif

      /* On to the next raster */

        num_bytes = src->width - (sw * sizeof(CHUNK));
        ps += (num_bytes / sizeof(CHUNK));

        num_bytes = sizeof(CHUNK) + dst->width - (dw * sizeof(CHUNK));
        pd += (num_bytes / sizeof(CHUNK));
    }
}
#endif


