/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/bmp/metrics.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: metrics.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.17   06 Apr 1995 15:08:16   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.16   11 Dec 1994 17:31:26   MIKE
 * 
 * Changed if_state.cs.origin to "origin_cs".
 * 
 *    Rev 1.15   02 May 1994 11:49:04   MIKE
 * Conditionally compile 1.14 change based on SLIM_FONTS.
 * 
 *    Rev 1.14   27 Apr 1994 17:08:04   MIKE
 * Add new special case for mirror, oblique for rotation 90.
 * 
 *    Rev 1.13   27 Apr 1994 10:58:02   MIKE
 * In metrics(), handle special case of mirrored, obliqued characters.
 * 
 *    Rev 1.12   22 Apr 1994 10:31:14   JOE
 * Added missing comment delimiter at start of line prior to "History".
 * 
 *    Rev 1.11   22 Apr 1994 09:20:44   LISA
 * 
 * Modified copyright/disclaimer notice.
 * 
 *    Rev 1.10   08 Mar 1994 15:52:06   JOE
 * Corrected external declaration for variable "special_case".
 * Was missing type "SW16".
 * 
 *    Rev 1.9   06 Mar 1994 14:11:18   MIKE
 * 
 * Adjust character's horizontal position only if CONVERGENT_FONTS == 1.
 * 
 *    Rev 1.8   03 Feb 1994 14:22:48   MIKE
 * Adjust character horizontal position for Convergent Fonts.
 * 
 *    Rev 1.7   22 Sep 1993 16:06:10   JOE
 * In metrics(), conditionally compiled block of code which checks for
 * mirrored character before distributing alternate width based on
 * SLIM_FONTS.
 * 
 *    Rev 1.6   07 Sep 1993 17:15:34   MIKE
 * 
 * Fix bug in x-origin for mirrored characters using alternate widths.
 * 
 *    Rev 1.5   14 Jun 1993 09:17:54   JOE
 * In metrics(), in calls to fp_to_16() for calculating bm->yorigin, cast each
 * member of expression for first argument to SL32 to prevent SW16 calculation.
 * 
 *    Rev 1.4   12 Feb 1993 11:16:34   JOE
 * VXWorks support.
 * 
 *    Rev 1.3   29 Jan 1993 09:25:24   JOE
 * In metrics(), saved em box size in IFBITMAP structure for IF, PS and TT.
 * 
 *    Rev 1.2   05 Jan 1993 16:10:00   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:29:58   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:32:56   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* metrics.c */
/*
 * History
 *
 *    03-27-90   jfd     In "metrics()", instead of splitting the difference
 *                       between the alternate width and the real width
 *                       and applying that value to both "x0" and "x1",
 *                       apply it only to "x0" and set "x1" equal to "x0"
 *                       plus the alternate width. This will ensure the
 *                       proper escapement. It was off by 1 design unit
 *                       in some cases.
 *    05-22-90   awr     Corrected yorigin calculation to use the grid
 *                       alligned baseline (d.aBaselnVal) instead of the
 *                       the unaligned version.
 *    05-22-90   awr     Added metric calculations for quadrant rotations
 *    09-04-90   awr     Declared des2wrkbm() to fix compiler warning,
 *                       removed rot90() declaration.
 *    05-Oct-90  awr     Added connecting character side bearing calculation
 *    28-Jan-91  jfd     Conditionally compiling fp_to_16() based on
 *                       (MAX_BM_BITS > 16)
 *                       Including "cgconfig.h".
 *    30-Jan-91  dET     Added function prototyping for MSC
 *    31-Jan-91  dET     Added support for multi-model MSC compilation
 *     3-Feb-91  awr     Removed include imath.h- not needed,
 *                       updated copyright.
 *     8-Feb-91  awr     Moved inv_des2wrkbm() to scale.c
 *                       Changed name des2wrkbm() to des2bm()
 *                       and inv_des2wrkbm() to inv_des2bm().
 *    11-Mar-91  awr     Changed return value of des2bm() and inv_des2bm
 *                       to PWORDVECTOR
 *    05-Jun-91  jfd     In metrics(), in case of 90 or 270 degree rotation,
 *                       when calculatin lb.y, use bm->depth instead of
 *                       d.bmdepth (FIX FROM EARLIER VERSION)
 *     9-Jun-91  awr     changed metrics() parameters
 *    17-Jun-91  jfd     Moved "debug.h" after "port.h".
 *     4-Jul-91  awr     Removed SCALE fields "px" and "py".
 *     8-Jul-91  bjg     Added Elasticity.
 *     5-Aug-91  tbh     Cleaned up Elasticity.
 *    10 Aug 91  ss      Added code to support mirrored chars and -ROTs.
 *    24-Aug-91  awr     Changed calling sequence of des2bm()
 *    27-Aug-91  bjg     Add special case handling for mirrored chars 
 *     1-Sep-91  awr     Included "bitmap.h"
 *    17-Sep-91  jfd     Changed fp_to_16() to LONG instead of WORD.
 *                       In metrics(), calling DBG2() unconditionally now.
 *    20-Sep-91  jfd     In metrics(), when calculating xorigin and yorigin,
 *                       cast operands as LONG due to change in type to LONG
 *                       for xorigin and yorigin.
 *    30-Oct-91  awr     Corrected lsb calculations for connecting chars.
 *     5-Feb-92  awr     In short_metrics(), changed the way x and y origins
 *                       are calculated.
 *     7-Mar-92  awr     Changed metrics() to read design bounding box from
 *                       if_state
 *    03-Apr-92  rs      Portability cleanup (see port.h).
 *    10 Apr 92  ss      Don't reset special_case flag here any more.
 *    09-Jun-92  jfd     In metrics(), load "org" with contents of
 *                       "if_state.cs.dorigin", not "if_state.cs.origin".
 *    21-Jun-92  awr     Conditionally compiled
 *    28-Jun-92  awr     Changed TT & PS for 16bit scan conversion
 *    08-Jul-92  rs      Code cleanup.
 *    21-Jul-92  awr     Conditional compile changes
 *    11-Aug-92  jfd     Trying macro MDES2BM for optimization.
 *                       Included imath.h.
 *    05-Jan-93  jfd     ANSI C function declaration changes.
 *    25-Jan-93  jfd     In metrics(), store size of em box ("du_emx" and 
 *                       "du_emy") into IFBITMAP structure for IF, PS and TT.
 *    08-Feb-93  jfd     VXWorrks support.
 *    14-Jun-93  jfd     In metrics(), in calls to fp_to_16() for calculating
 *                       bm->yorigin, cast each member of expression for first
 *                       argument to SL32 to prevent truncation to SW16.
 *    01-Sep-03  mby     In metrics() if if_state.alt_width is on and the character
 *                       is mirrored in X, adjust the x-origin in the opposite
 *                       direction, because of a previous reflection.
 *    22-Sep-93  jfd     In metrics(), conditionally compiled block of code
 *                       which checks for mirrored character before distribu-
 *                       ting alternate width based on SLIM_FONTS.
 *    25-Jan-94  mby     Adjust character position by lsb_adj in metrics()
 *                       for Convergent Fonts.
 *    06-Mar-94  mby     Use lsb_adj only if CONVERGENT_FONTS == 1.
 *    08-Mar-94  jfd     Corrected external declaration for variable 
 *                       "special_case". Was missing type "SW16".
 *    22-Apr-94  jfd     Added missing comment delimiter at start of line
 *                       prior to "History".
 *    26-Apr-94  mby/    In metrics() for special_case of mirrored, obliqued
 *               bjg     characters, (1) set origin to character baseline;
 *                       (2) adjust bm->xorigin,yorigin.
 *    27-Apr-94  mby/    In metrics() add new special case for p0 for
 *               bjg     rotation of 0 and rotation of 90.
 *    02-May-94  mby     Conditionally compile above change to SLIM_FONTS.
 *    11-Dec-94  mby     Change if_state.cs.origin to "origin_cs".
 */

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>

#include "cgconfig.h"
#include "port.h"
#include "debug.h"
#include "ifmem.h"
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "bitmap.h"
#include "imath.h"

/*-------------------------------------------------*/
EXTERN CHR_DEF chr_def[];
EXTERN CHR_DEF_HDR chr_def_hdr;
EXTERN IF_STATE if_state;
/*-------------------------------------------------*/

#ifdef LINT_ARGS
MLOCAL SL32 fp_to_16(SL32, PCOORD_DATA);
#else /* no LINT_ARGS */
MLOCAL SL32 fp_to_16();
#endif /* LINT_ARGS */





#if ((ELASTIC_Y) | (ELASTIC_X))
EXTERN UW16 isElastic;
#endif
#if SLIM_FONTS
EXTERN SW16 special_case;
#endif


/*----------------------*/
/*    fp_to_16()        */
/*----------------------*/
/*  Convert working bitmap space fractional pixels to 16ths of a pixel */
#if defined (ANSI_DEFS)
MLOCAL SL32  fp_to_16(SL32 v, PCOORD_DATA xy)
#else
MLOCAL SL32
fp_to_16(v, xy)
    SL32 v;
    PCOORD_DATA xy;
#endif
{
    SL32 val;

    val = LABS(v);
    val = (val * 16 + xy->half_pixel) >> xy->grid_shift;

    if(v<0)
        return -val;
    else
        return val;
}



/*----------------------*/
/*    metrics           */
/*----------------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID  metrics(PIFBITMAP bm)
#else
GLOBAL VOID
metrics(bm)
    PIFBITMAP  bm;
#endif
{
    SW16VECTOR org;
    SW16       escx;

    SW16 orig_black_width;
    SW16 new_black_width;
    SW16 orig_lsb;
    SW16 new_lsb;
    SW16 adj_width;

    SW16VECTOR lb, rb;     /* left & right bound (fractional pixels) */
    SW16VECTOR dlb, drb;   /* left & right bound (design units)      */
    SW16VECTOR new_org;    /* character origin (design units)        */
    SW16VECTOR p0;         /* character origin (fractional pixels)   */
    SL32VECTOR p;

    SW16 tmp;

#if ((ELASTIC_Y) | (ELASTIC_X))
    SW16 pixelwidth,disc;
#endif

    DBG("metrics()\n");
    org  = if_state.cs.dorigin;  /* 6-9-92 */
    escx = if_state.cs.escapement;

  /*  Adjust escape box for alternate width (if any) */

    if(if_state.alt_width > 0)  /* not supposed to be negative, ignore if < 0 */
    {
      /*  Distribute the alternate width change equally (not proportionately)
       *  over each side bearing.
       */

        DBG1("    alternate width = %d\n", if_state.alt_width);
        adj_width = (if_state.alt_width - escx) / 2;
#if SLIM_FONTS
        if (chr_def_hdr.mirror_flags & MIRROR_X_MASK)   /* mby 9-01-93 */
            org.x += adj_width;
        else
#endif
            org.x -= adj_width;
        escx = if_state.alt_width;
        DBG3("    alt origin (%d, %d)  escapement %d\n",
                                          org.x, org.y, escx);
    }

  /*  Adjust character position by lsb_adj for convergent fonts */
#if (CONVERGENT_FONTS == 1)     /* mby 3-06-94 */
#if SLIM_FONTS
    if (if_state.awtsub_on)
    {
      if (chr_def_hdr.mirror_flags & MIRROR_X_MASK)
            org.x += if_state.lsb_adj;
      else
#endif
        org.x -= if_state.lsb_adj;
    }
#endif  /* CONVERGENT_FONTS */

    bm->escapement   = escx;
#if NON_IF_FONT
if(if_state.fst_type != FC_IF_TYPE)
{
    DBG("short_metrics()\n");
    p.x = ((SL32)if_state.cs.origin_cs.x + (SL32)if_state.xlate.x)
                                                  >>if_state.right_shift;
    p.y = ((SL32)if_state.cs.origin_cs.y + (SL32)if_state.xlate.y)
                                                   >>if_state.right_shift;
    bm->xorigin = - fp_to_16(p.x, &if_state.x);
    bm->yorigin =   fp_to_16((SL32)((SL32)bm->depth << (SL32)log_ypix)  - p.y, &if_state.y);

    /* Store size of em box */

    bm->du_emx  =   if_state.cs.du_emx;
    bm->du_emy  =   if_state.cs.du_emy;

    return;
}
#endif
#if IF_RDR
  /*  Compute vector (xorigin, yorign) in 16ths of an output pixel.
   *      step 1. Compute (p0.x, p0.y) the character origin in working
   *              bitmap space.
   *      step 2. Compute (xorigin, yorigin).
   */

/*---------step 1-------------------------*/

  /*  If arbitrary rotation, simply transfer the design origin to
   *   working bitmap space. If special 90 degree rotation, then we
   *   take into account the change in the character's black width
   *   and distribute this change equally over both side bearings.
   */

    if(!if_state.quadrant)      /* arbitrary rotation */
    {
        DBG("    Arbitrary rotate: ");

      /* Transform character origin to working bitmap space */

#if defined MDES2BM
        p0 = MDES2BM(org);
#else
        p0 = des2bm(org);
#endif /* defined MDES2BM */
#if SLIM_FONTS          /* mby, 4/26 - 5/1/94 */
        if (special_case && if_state.shear == ROT0)
        {                   /* rotation is 0/180 and shear is nonzero */
            SW16 savex = p0.x;
            org.y = if_state.escape_box.ll.y;
#if defined MDES2BM
            p0 = MDES2BM(org);  /* p0 is the transform of the character    */
#else                           /* origin in X, and the character baseline */
            p0 = des2bm(org);   /* in Y. */
#endif /* defined MDES2BM */
            p0.x = savex;
        }
        else if (special_case && if_state.shear == ROT90)
        {
            SW16 savey = p0.y;
            org = if_state.escape_box.ll;    /* x and y */
#if defined MDES2BM
            p0 = MDES2BM(org);
#else
            p0 = des2bm(org);
#endif /* defined MDES2BM */
            p0.y = savey;
        }
#endif  /* SLIM_FONTS */
    }
    else                /* special 0, 90, 180, or 270 degree rotation */
    {
      /*  Intellifont is on.
       */

        if(if_state.quadrant==ROT0 || if_state.quadrant==ROT180
                        /* added -ss */
         || if_state.quadrant == -ROT0 || if_state.quadrant == -ROT180)
        {
            lb.y = rb.y = 0;

          /* assume ROT0, swap if ROT180 */

            lb.x = bm->left_indent << log_xpix;
            rb.x = lb.x + (bm->black_width << log_xpix);
            if( if_state.quadrant == ROT180
             || if_state.quadrant == -ROT0 )
            {
                tmp = lb.x;
                lb.x = rb.x;
                rb.x = tmp;
            }
        }
        else
        {
            lb.x = rb.x = 0;

          /* Assume ROT270, swap if ROT90 */

            lb.y = (bm->depth - bm->top_indent) << log_ypix;
            rb.y = lb.y - (bm->black_depth << log_ypix);
            if( if_state.quadrant == ROT90
             || if_state.quadrant == -ROT270 )
            {
                tmp = lb.y;
                lb.y = rb.y;
                rb.y = tmp;
            }
        }

        dlb = inv_des2bm(lb);
        drb = inv_des2bm(rb);

        DBG2("lb = (%d, %d)\n", lb.x, lb.y);
        DBG2("rb = (%d, %d)\n", rb.x, rb.y);
        DBG2("dlb = (%d, %d)\n", dlb.x, dlb.y);
        DBG2("drb = (%d, %d)\n", drb.x, drb.y);

        orig_lsb = if_state.cs.dxmin - org.x;

        if(if_state.ConnectingChar)
        {
#if ((ELASTIC_Y) | (ELASTIC_X))
            pixel_align (org.x, &if_state.x, R_TWO_I, 0);
#else
            pixel_align (org.x, &if_state.x, R_TWO_I);
#endif
            new_lsb = dlb.x - if_state.value;
            DBG("     ConnectingChar:\n");
            DBG2("    orig_lsb %d    new_lsb %d\n", orig_lsb, new_lsb);
        }
        else
        {
            orig_black_width = if_state.cs.dxmax - if_state.cs.dxmin;
            new_black_width  = drb.x - dlb.x;

            new_lsb = orig_lsb + (orig_black_width - new_black_width)/2;
            DBG2("    orig_lsb %d    new_lsb %d\n", orig_lsb, new_lsb);
        }

        new_org.x = dlb.x - new_lsb;
        new_org.y = if_state.aBaselnVal;

        /*  Fix for one part compound char with nonzero offset.           */
        /*  There may still be a bug for multiple part chars with nonzero */
        /*  offset in first part.                        -ss,bjg 7/23/91  */
        if( chr_def->offset.y )
           new_org.y -= chr_def->offset.y;

        DBG2("new_org = (%d, %d)\n", new_org.x, new_org.y);

#if defined MDES2BM
        p0 = MDES2BM(new_org);
#else
        p0 = des2bm(new_org);
#endif /* defined MDES2BM */
    }    

    DBG2("    (p0.x, p0.y) = (%d, %d)\n", p0.x, p0.y);

/*---------step 2-------------------------*/




#if ((ELASTIC_Y) | (ELASTIC_X))
    if (isElastic && if_state.alt_black_width)
      {
      bm->xorigin = 0L;
      pixelwidth = if_state.alt_black_width/if_state.x.orig_pixel;
      disc = bm->black_width - pixelwidth;
      if(disc > 0)
        bm->xorigin = -16L;
      if(disc < 0)
        bm->xorigin = 16L;
      }
    else
      {
      isElastic = NO;  /* this is specifically for communication to ifdos.c */
      bm->xorigin = - fp_to_16((SL32)p0.x, &if_state.x);
      }
#else
      bm->xorigin = - fp_to_16((SL32)p0.x, &if_state.x);
#endif

    bm->yorigin =   fp_to_16((SL32)(((SL32)bm->depth << (SL32)log_ypix)  - p0.y),
                             &if_state.y);

#if SLIM_FONTS
    /*  handle special case of mirrored characters */
    if( special_case &&
        (if_state.quadrant==ROT0 || if_state.quadrant==ROT180
          || if_state.quadrant == -ROT0 || if_state.quadrant == -ROT180
          || if_state.shear == ROT0) )       /* test for shear -mby 4-26-94 */
       bm->yorigin = (SL32)((SL32)bm->top_indent<<5L) +
                     (SL32)((SL32)bm->black_depth<<4L) -
                     bm->yorigin;

    if( special_case &&
        (if_state.quadrant==ROT90 || if_state.quadrant==ROT270
          || if_state.quadrant == -ROT90 || if_state.quadrant == -ROT270
          || if_state.shear == ROT90) )      /* test for shear */
       bm->xorigin = - (SL32)((SL32)bm->left_indent<<5L) -
                     (SL32)((SL32)bm->black_width<<4L) -
                     bm->xorigin;

    /**  don't reset this here, its used again later. -ss 4/10/92 **
    special_case = 0;
    **/
#endif

    /* Store size of em box */

    bm->du_emx  =   if_state.cs.du_emx;
    bm->du_emy  =   if_state.cs.du_emy;

#endif /* IF_RDR  */
}
