/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/cor/des2bm.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: des2bm.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   06 Apr 1995 15:11:02   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   22 Apr 1994 09:28:24   LISA
 * 
 * Modified copyright/disclaimer notice for 1994.
 * 
 *    Rev 1.3   12 Feb 1993 11:25:54   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   07 Jan 1993 10:42:38   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:35:02   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:37:50   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* des2bm.c */



/*---------------------------------------------------------------------
     05-05-90   awr   Added landscape. New function rot90(), changed
                      des2wrkbm() to rotate coords. Changed bitmpa_size()
                      to swap bmwidth and bmdepth and compute d.max_x
                      and d.max.y.

    11-Dec-90   awr   Changed inv_des2wrkbm() to account for the new
                      tx() and ty()- see imath.c.
     8-Feb-91   awr   Split off from maker.c and shortened names to
                      des2bm() nad inv_des2bm().
    11-Mar-91   awr   Changed return value of des2bm() and inv_des2bm
                      to PWORDVECTOR
    06-Apr-91   awr   Changed rotate calculations to floating point
    17-Jun-91   jfd   Moved "debug.h" after "port.h".
                      Moved "profile.h" after "port.h".
    29-Jun-91   awr   added tt0 to make des2bm() = matrix o translate
                      and other 4 transforms for mirroring
    14 Aug 91   ss    Fixed inv_des2bm() to use correct vars for -ROT cases.
    24-Aug-91   awr   Moved function decls to include file
                      Moved tx() and ty() inline in des2bm()
    11-Nov-91   rs    Broke 'cg_scale()' off from des2bm() - for performance.
    15-Nov-91   jfd   In des2bm(), removed local variables "xx" and "yy"
                      which were never referenced.
     7-Mar-92   awr   Reversed order in inv_des2bm(): rotate 1st, scale 2nd.
    03-Apr-92   rs    Portability cleanup (see port.h).
    11-Aug-92   jfd   Trying new version of des2bm() for optimization.
    15-Sep-92   jfd   Conditionally compiling entire module based on
                      IF_RDR.
    07-Jan-93   jfd   ANSI C function declaration changes.
    08-Feb-93   jfd   VXWorks support.
-----------------------------------------------------------------------*/


#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"
#include "port.h"

#if IF_RDR

#include "debug.h"
#include "profile.h"
#include "imath.h"
#include "ifmem.h"
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "fpmath.h"


EXTERN IF_STATE if_state;

#if defined MDES2BM
/*----------------------*/
/*      des2bm          */
/*----------------------*/
#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR des2bm(SW16VECTOR v)
#else
GLOBAL SW16VECTOR
des2bm(v)
    SW16VECTOR v;
#endif
{
    SW16VECTOR new;

    /* Always arbitrary rotation */

    new.x = ifpmul(v.x, if_state.m[0])
          + ifpmul(v.y, if_state.m[2]) + if_state.tt.x;
    new.y = ifpmul(v.x, if_state.m[1])
          + ifpmul(v.y, if_state.m[3]) + if_state.tt.y;
    return (new);
}
#else
/*----------------------*/
/*      des2bm          */
/*----------------------*/
#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR des2bm(SW16VECTOR v)
#else
GLOBAL SW16VECTOR
des2bm(v)
    SW16VECTOR v;
#endif
{
    SW16VECTOR new;

    if(!if_state.quadrant)    /* arbitrary rotation */
    {
        new.x = ifpmul(v.x, if_state.m[0])
              + ifpmul(v.y, if_state.m[2]) + if_state.tt.x;
        new.y = ifpmul(v.x, if_state.m[1])
              + ifpmul(v.y, if_state.m[3]) + if_state.tt.y;
		  return (new);
    }
    else return (cg_scale (v));
}
#endif /* defined MDES2BM */

/*----------------------*/
/*    inv_des2bm        */
/*----------------------*/
/*  DOES NOT WORK FOR ARBITRARY ROTATION. Don't fix it either because
*  we don't need this function for arbitrary rotation. Fix it if
*  we ever do and save the space in the meantime.
*/
#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR inv_des2bm(SW16VECTOR v)
#else
GLOBAL SW16VECTOR
inv_des2bm(v)
   SW16VECTOR v;
#endif
{
   SW16 xx, yy;
   SW16VECTOR new;


   DBG2("    %d = inv_tx(%d)\n", xx, v.x);
   DBG2("    %d = inv_ty(%d)\n", v.y, v.y);

   xx = v.x - if_state.tt.x;
   yy = v.y - if_state.tt.y;

   if(if_state.quadrant == ROT0)
   {
      new.x = xx;
      new.y = yy;
   }
   else if(if_state.quadrant == ROT270)
   {
      new.x = - yy;
      new.y =   xx;
   }
   else if(if_state.quadrant == ROT180)
   {
      new.x =  - xx;
      new.y =  - yy;
   }
   else if(if_state.quadrant == ROT90)
   {
      new.x =   yy;
      new.y = - xx;
   }
   else if(if_state.quadrant == -ROT0)    /* Next 4: mirror about y axis first */
   {
      new.x = -xx;
      new.y =  yy;
   }
   else if(if_state.quadrant == -ROT90)
   {
      new.x = -yy;
      new.y = -xx;
   }
   else if(if_state.quadrant == -ROT180)
   {
      new.x =  xx;
      new.y = -yy;
   }
   else if(if_state.quadrant == -ROT270)
   {
      new.x = yy;
      new.y = xx;
   }


    if(new.x < 0)
        new.x = -(SW16)(  ( (   ((UL32)(-new.x)) << if_state.x.shift
                            )     * (UL32)if_state.x.p_pixel
                          ) >> 16
                       );
    else
        new.x =  (SW16)(  ( (   ((UL32)(new.x)) << if_state.x.shift
                            )     * (UL32)if_state.x.p_pixel
                          ) >> 16
                       );

    if(new.y < 0)
        new.y = -(SW16)(  (  (  ((UL32)(-new.y)) << if_state.y.shift
                             )    * (UL32)if_state.y.p_pixel
                          ) >> 16
                       );
    else
        new.y =  (SW16)(  (  (  ((UL32)(new.y)) << if_state.y.shift
                             )    * (UL32)if_state.y.p_pixel
                          ) >> 16
                       );

   return new;
}
#endif  /* IF_RDR */
