/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/cor/imath.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: imath.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   06 Apr 1995 15:10:20   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   22 Apr 1994 09:28:16   LISA
 * 
 * Modified copyright/disclaimer notice for 1994.
 * 
 *    Rev 1.3   12 Feb 1993 11:20:38   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   07 Jan 1993 10:33:38   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:36:28   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:37:26   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/************************************************************************
**    imath.c                                                          **
**    -------------------------------------------------------------    **
 *
*************************************************************************/
/*
 *  10-Dec-90 awr  changed tx() and t() to account for the changed method
 *                 of computing output pixel size. See comp_pix().
 *  18-Jan-91 jfd  Trying to get rid of bogus value returned by "tx()"
 *                 and "ty()" when argument passed to it is a negative
 *                 number.
 *  30-Jan-91 dET  Added function prototyping for MSC
 *   3-Feb-91 awr  Removed scale_rem() and multiply_i_i - no longer used.
 *                 Moved function declarations to imath.h, updated copyright.
 *  20-Jun-91 bjg  Make rounding in nx() and ny() depend on sign.
 *  24-Jun-91 jfd  Moved "debug.h" after "port.h".
 *  26-Aug-91 awr  Moved tx() and ty() inline in des2bm().
 *                 Merged nx() and ny() into nxy()
 *  11-Nov-91 rs   Add 'cg_scale()' which was broken off from 'des2bm()'.
 *   7-Mar-92 awr  Reversed order in cg_scale(): scale 1st, rotate 2nd.
 *  03-Apr-92 rs   Portability cleanup (see port.h).
 *  15-Sep-92 jfd  Conditionally compiling entire module based on IF_RDR.
 *  22-Sep-92 jfd  Unconditionally compiling scale_iii().
 *  07-Jan-93 jfd  ANSI C function declaration changes.
 *  08-Feb-93 jfd  VXWorks support.
 */

#include "cgconfig.h"
#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "port.h"

#include "debug.h"
#include "ifmem.h"
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "segments.h"
#include "adj_skel.h"
#include "tr_type.h"
#include "imath.h"    /* function declarations for this module */


#ifdef LINT_ARGS
GLOBAL SW16VECTOR cg_scale (SW16VECTOR);
#else
GLOBAL SW16VECTOR cg_scale ();
#endif /* LINT_ARGS */


EXTERN IF_STATE if_state;
EXTERN TRAN    x_tran,y_tran;



/*-------------------------------------------------------------------*/
/*        scale_iii()                                                */
/*-------------------------------------------------------------------*/

#if defined (ANSI_DEFS)
SW16 scale_iii (SW16 x, SW16 y, SW16 z)
#else
SW16
scale_iii (x, y, z)

        SW16 x,y,z;
#endif
{
        return (
                  (SW16)(
                           ( (SL32)x  * (SL32)y ) / (SL32)z
                        )
               );
}

#if IF_RDR
/*------------------*/
/*    nxy           */
/*------------------*/
#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR nxy(SW16VECTOR v)
#else
GLOBAL SW16VECTOR
nxy(v)
    SW16VECTOR v;
#endif
{
SW16VECTOR new;
SL32  x,y;

	new.x = v.x - x_tran.old0;
	new.y = v.y - y_tran.old0;

	if(new.x == 0)
    	new.x = x_tran.new0;
	else
	{
    	x = (SL32)(new.x) * (SL32)x_tran.num;
    	if(x > 0)
        	new.x = (SW16)((x + x_tran.half_den) / x_tran.den) + x_tran.new0;
    	else
        	new.x = (SW16)((x - x_tran.half_den) / x_tran.den) + x_tran.new0;
	}

	if(new.y == 0)
    	new.y = y_tran.new0;
	else
	{
    	y = (SL32)(new.y) * (SL32)y_tran.num;
    	if(y > 0)
        	new.y = (SW16)((y + y_tran.half_den) / y_tran.den) + y_tran.new0;
    	else
        	new.y = (SW16)((y - y_tran.half_den) / y_tran.den) + y_tran.new0;
	}

    return new;
}


#if defined (ANSI_DEFS)
GLOBAL SW16VECTOR cg_scale (SW16VECTOR v)
#else
GLOBAL SW16VECTOR cg_scale (v)
SW16VECTOR v;
#endif
{
SW16	xx, yy;
SW16VECTOR	new;

   if (v.x < 0)
   {
      xx = -((  (UW16)(  ((UL32)-v.x << 16)  / if_state.x.p_pixel
                     ) + if_state.x.shift_rnd
            ) >> if_state.x.shift);
   }
   else
   {
      xx = (  (UW16)(  ((UL32)v.x << 16)  / if_state.x.p_pixel
                     ) + if_state.x.shift_rnd
            ) >> if_state.x.shift;
   }

   if (v.y < 0)
   {
      yy = -((  (UW16)(  ((UL32)-v.y << 16)  / if_state.y.p_pixel
                     ) + if_state.y.shift_rnd
            ) >> if_state.y.shift);
   }
   else
   {
      yy = (  (UW16)(  ((UL32)v.y << 16)  / if_state.y.p_pixel
                     ) + if_state.y.shift_rnd
            ) >> if_state.y.shift;
   }

   if(if_state.quadrant == ROT0)
   {
      new.x =  xx;
      new.y =  yy;
   }
   else if(if_state.quadrant == ROT90)
   {
      new.x = -yy;
      new.y =  xx;
   }
   else if(if_state.quadrant == ROT180)
   {
      new.x =  -xx;
      new.y =  -yy;
   }
   else if(if_state.quadrant == ROT270)
   {
      new.x =  yy;
      new.y = -xx;
   }
   else if(if_state.quadrant == -ROT0)    /* Next 4: mirror about y axis first */
   {
      new.x = -xx;
      new.y =  yy;
   }
   else if(if_state.quadrant == -ROT90)
   {
      new.x = -yy;
      new.y = -xx;
   }
   else if(if_state.quadrant == -ROT180)
   {
      new.x =   xx;
      new.y =  -yy;
   }
   else if(if_state.quadrant == -ROT270)
   {
      new.x = yy;
      new.y = xx;
   }

   new.x += if_state.tt.x;
   new.y += if_state.tt.y;

	return new;
}
#endif  /* IF_RDR */
