/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/cor/manipula.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: manipula.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.8   06 Apr 1995 15:11:04   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.7   22 Jun 1994 16:39:30   JOE
 * In manipulate(), if processing a space character, test if the
 * bounding box ur.x exceeds MAX_DESIGN, and if it does, set the upper
 * right bounding box in working bitmap space "tbound.ur" to "tbound.ll".
 * This resolves the space-compound character offset bug.
 * 
 *    Rev 1.6   22 Apr 1994 09:28:16   LISA
 * 
 * Modified copyright/disclaimer notice for 1994.
 * 
 *    Rev 1.5   18 Mar 1994 08:31:20   JOE
 * Fix for font cartridge space-compound character bug in manipulate().
 * 
 *    Rev 1.4   29 Jul 1993 12:31:46   AL
 * 
 * Non-zero winding changes
 * 
 *    Rev 1.3   12 Feb 1993 11:23:24   JOE
 * 
 * VXWorks support.
 * 
 *    Rev 1.2   07 Jan 1993 10:38:32   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:36:14   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:37:28   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* maniplua.c */


/*------------------------------------------------------------------------
   3-Feb-91  awr  Split off from nxt_pt.c
   8-Feb-91  awr  Changed name des2wrkbm() to des2bm()
                  Removed EXTERN ADJUTED_SKEL x_skel[], y_skel[] and
                  made parameters of manipulate() and init_xy_tran()
                  Moved GLOBAL WORD work_buffer[WORK_BUFFER_SIZE] to
                  to if.c and passed as parameter.
   8-Mar-91  jfd  For MULTI-MODEL, change references to LOOP* to PLOOP.
  10-Mar-91  awr  work_buffer is now array of WORDVECTOR. Changed call to
                  manipulate().
  11-Mar-91  awr  Changed return value of des2bm() to PWORDVECTOR
   6-Apr-91  awr  values of ttx and tty are converted now from FPNUMs
   2-Jun-91  awr  rename some variables
  17-Jun-91  jfd  Moved "debug.h" after "port.h".
                  Moved "profile.h" after "port.h".
                  Added "cgconfig.h" before "port.h".
  21-Jun-91  awr  make lower corner of tbound precise
  24 Jun 91  ss   Added prototpye for manipulate() for not LINT_ARGS.
  24-Aug-91  awr  Moved function decls to include file
  03-Apr-92  rs   Portability cleanup (see port.h).
  12-Apr-92  awr  Moved translation calcs to a higher level.
   2-May-92  awr  Moved if_state.tt initialization to higher level.
  14-Jun-92  awr  Changed tbound to INTR
  11-Aug-92  jfd  Trying macro MDES2BM for optimization.
  19-Aug-92  jfd  In manipulate(), if processing a space character,
                  set bounding box values to 0 and return.
  15-Sep-92  jfd  Conditionally compile entire module based on IF_RDR.
  07-Jan-93  jfd  ANSI C function declaration changes.
  08-Feb-93  jfd  VXWorks support.
  27-Jul-93  awr  Count transition runs for non-zero winding support
  11-Mar-94  jfd  In manipulate(), if processing a space character,
                  scale the character bounding box (in design units)
                  "if_state.bound_box" to calculate the bounding box 
                  in working bitmap space (if_state.tbound). This
                  resolves the space-compound character offset bug.
                  Previously, we were setting "if_state.bound" to 0.
  22-Jun-94  jfd  In manipulate(), if processing a space character,
                  test if the bounding box ur.x exceeds MAX_DESIGN
                  (we have encountered a pcleo where this occurs) and
                  if it does, set the upper right bounding box in working 
                  bitmap space "tbound.ur" to "tbound.ll". This is
                  another attempt to resolve the space-compound character
                  offset bug.
------------------------------------------------------------------------*/

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"
#include "port.h"

#if IF_RDR

#include "debug.h"
#include "profile.h"
#include "imath.h"
#include "fpmath.h"
#include "ifmem.h"
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "segments.h"
#include "adj_skel.h"
#include "tr_type.h"



/*------- External Data --------*/

EXTERN IF_STATE if_state;


#ifdef LINT_ARGS
MLOCAL VOID   set_transform(PTRAN);
MLOCAL VOID   init_transform(PTRAN, SW16);
#endif


/* dET initialize the variables "x_tran, y_tran".  This is a trick to force 
       the variables into the data segment rather than the BSS segment.
       NOTE the changes in "imath.asm"
*/
/*-------------------------*/
GLOBAL BOOLEAN no_skel;
GLOBAL TRAN x_tran = {1,0}, y_tran = {1,0};
/*-------------------------*/




/*-------------------*/
/*  set_transform    */
/*-------------------*/
/*   Sets a new x or y scaling transform that is used by nx() and ny()
 *   to scale countour coordinate data.
 *
 *    The trasforms are specified by:
 *        old0
 *        num
 *        den
 *        rnd
 *
 *    The transform rule is:
 *
 *        x -> rnd + ((x - old0) * num) / den;   if den != 0 
 *        x -> rnd + (x - old0);                 if den == 0
 *
 *    The four transform parameters are returned in the structure
 *
 *        TRAN *td;
 *
 *    provided by the caller. There are two TRAN structures, one for x
 *    coordinate scaling and one for y coordinate scaling:
 *
 *        TRAN x_tran, y_tran;
 *
 *    In addition to the 4 transform parameters, this function returns
 *    the contour coordinate index of the first point in the next scaling
 *    segment. This is returned in
 *
 *        td->end
 *
 *
 */

#if defined (ANSI_DEFS)
MLOCAL VOID  set_transform(PTRAN td)
#else
MLOCAL VOID
set_transform(td)
      PTRAN td;     /* ptr to x or y transform data  */
#endif
{
    PADJUSTED_SKEL skel;

    DBG("\n\n====set_transform=======================================\n");
    DBG1("     td->sct %d\n", td->sct);

    skel = td->skel;   /* adjusted skel data */

    /* have we processed all sk pts ?  */
    if (td->sct == td->nsk)
    {
        /* last part of loop, scaling parameters should */
        /* be same as in first scaling segment.         */

        td->num      = td->num0;
        td->den      = td->den0;
        td->half_den = td->half_den0;
        td->end      = td->end0;
        td->old0     = td->old00;
        td->new0     = td->new00;
    }
    else
    {
 	   
      /* Vector index at which to start the next scaling segment */

        td->end   = *(td->skel_to_contr++);

      /*  Step to next pair of new pts and compute numerator  */

        td->new0 = td->new1;
        td->new1 = skel->adjusted;
        td->num  = td->new1 - td->new0;

      /* Step to next pair of original pts and compute denominator */

        td->old0 = td->old1;
        td->old1 = skel->original;
        td->den  =  td->old1 - td->old0;

        if(!td->den) td->num = td->den = 1;
        if(td->den < 0)    /* denominator will always be positive */
        {
            td->den = - td->den;
            td->num = - td->num;
        }
        td->half_den = td->den / 2;

      /* set for next scaling segment  */

        td->skel++;
        td->sct += 1;

        DBG1("end %d\n", td->end);
        DBG2("new0  %d   new1  %d\n", td->new0, td->new1);
        DBG2("old0  %d   old1  %d\n", td->old0, td->old1);
        DBG2("num %d  den %d\n\n\n",td->num,td->den);
    } 

}




/*---------------------*/
/*   init_transform    */
/*---------------------*/
/*  Initialize transform data for the loop */

#if defined (ANSI_DEFS)
MLOCAL VOID  init_transform(PTRAN td, SW16  ncoords)
#else
MLOCAL VOID
init_transform(td, ncoords)
    PTRAN td;  
    SW16  ncoords;
#endif
{
    td->nsk = *(td->num_skel_loop++);
    if (td->nsk < 2)
    {
      /*  0 skeletal points can
       *  occur in un-Intellifonted characters. 1 skel
       *  point can never happen- we ignore if it happens.
       */

        no_skel = TRUE;  

        td->end      = ncoords;

        td->old0     = 0;
        td->new0     = 0;
        td->num      = 1;
        td->den      = 1;
        td->half_den = 0;
    }
    else
    {
        no_skel = FALSE;

        td->sct = 0;    /* Number of scale segments processed           */

        td->old1 = (td->skel + td->nsk - 1)->original;
        td->new1 = (td->skel + td->nsk - 1)->adjusted;


        set_transform(td);

      /*  Remember data from first scaling segment
       *  because we'll need it again when we come
       *  around the loop.
       */

        td->end0     = ncoords;     /*  just bigger than the last
                                     *  vector number so we'll never
                                     *  make a new scaling segment
                                     */
        td->num0      = td->num;
        td->den0      = td->den;
        td->half_den0 = td->half_den;
        td->old00     = td->old0;
        td->new00     = td->new0;
    }

}



/*---------------------*/
/*   init_xy_tran      */
/*---------------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID  init_xy_tran(PADJUSTED_SKEL x_skel, PADJUSTED_SKEL y_skel)
#else
GLOBAL VOID
init_xy_tran(x_skel, y_skel)
    PADJUSTED_SKEL x_skel, y_skel;
#endif
{
  /* Set ptrs to arrays of adjusted skel data */

    x_tran.skel = x_skel;
    y_tran.skel = y_skel;

  /* set ptrs to arrays of number of skel pts per loop */

    x_tran.num_skel_loop = if_state.xskel.num_skel_loop;
    y_tran.num_skel_loop = if_state.yskel.num_skel_loop;

  /* set pointers to loop order lists of skel pts (contour indices)  */

    x_tran.skel_to_contr = if_state.xskel.skel_to_contr;
    y_tran.skel_to_contr = if_state.yskel.skel_to_contr;
}




/*---------------------*/
/*   manipulate        */
/*---------------------*/
#if defined (ANSI_DEFS)
GLOBAL UW16  manipulate(PADJUSTED_SKEL x_skel, PADJUSTED_SKEL y_skel)
#else
GLOBAL UW16
manipulate(x_skel, y_skel)
    PADJUSTED_SKEL x_skel, y_skel;
#endif
{
    LOOP    loop;           /* current loop                               */
    UW16   loopct;         /* current loop number                        */
    SW16    end_scale_seg;  /* last contour (index) in scale segment      */
    SW16    next_vec_num;
    SW16    ncoords;
    PSW16VECTOR cv;         /* Pointer to coordinate & contact points     */
    SW16VECTOR wcv;         /* Used for handling space character          */
    SW16        x, y;
    SW16        x_min, x_max;
    SW16        y_min, y_max;

    DBG("\n\n\n\n  m a n i p u l a t e ( )\n\n");
    DBG1("number of loops = %d\n", if_state.num_loops);

    /* If processing a space, reset bounding box */

    if(!if_state.num_loops)
    {
#if defined MDES2BM
       wcv = MDES2BM(if_state.bound_box.ll);
#else
       wcv = des2bm(if_state.bound_box.ll);
#endif /* defined MDES2BM */
       if_state.tbound.ll.x = (INTR)wcv.x;
       if_state.tbound.ll.y = (INTR)wcv.y;

       if (if_state.bound_box.ur.x > MAX_DESIGN)
       {
          if_state.tbound.ur.x = if_state.tbound.ll.x;
          if_state.tbound.ur.y = if_state.tbound.ll.y;
       }
       else
       {
#if defined MDES2BM
          wcv = MDES2BM(if_state.bound_box.ur);
#else
          wcv = des2bm(if_state.bound_box.ur);
#endif /* defined MDES2BM */
          if_state.tbound.ur.x = (INTR)wcv.x;
          if_state.tbound.ur.y = (INTR)wcv.y;
       }
       return SUCCESS;
    }

  /*--------------------------------------------------------------------*/
  /*                     I n i t i a l i z e                            */

    init_xy_tran(x_skel, y_skel); /* initialize for skeletal interpolation */

    x_min = y_min =  MAX_DESIGN;
    x_max = y_max = -MAX_DESIGN;

  /*                     I n i t i a l i z e                            */
  /*--------------------------------------------------------------------*/
  /*          F o r    e a c h    c o n t o u r    l o o p              */

    first_loop();
    for(loopct = 0; loopct < if_state.num_loops; loopct++)
    {
        DBG("\n\n\nN e x t    L o o p\n");

      /*----------------------------------------------------------------*/
      /*       I n i t i a l i z e    c o n t o u r    l o o p          */

        next_loop(&loop);
        cv = loop.cv;                               /* coordinate points */
        ncoords = loop.ncoords;  /* num coords: not counting first twice */
        DBG1("ncoords  = %u\n", ncoords);

      /*  Initialize skeletal transforms */

        DBG("x init_transform()\n");
        init_transform(&x_tran, ncoords);

        DBG("y init_transform()\n");
        init_transform(&y_tran, ncoords);

        if (x_tran.end < y_tran.end) end_scale_seg = x_tran.end;
                                else end_scale_seg = y_tran.end;

      /*       I n i t i a l i z e    c o n t o u r    l o o p          */
      /*----------------------------------------------------------------*/
      /*  F o r    e a c h    c u r v e    p o i n t    i n    l o o p  */

        for(next_vec_num = 0; next_vec_num < ncoords; next_vec_num++)
        {

          /*  Set up next scaling segment  */

            if (next_vec_num == end_scale_seg)
            {
                DBG1("new scale segment: next_vec_num %d\n", next_vec_num);
                if (next_vec_num == x_tran.end)   set_transform(&x_tran);
                if (next_vec_num == y_tran.end)   set_transform(&y_tran);

                if (x_tran.end < y_tran.end) end_scale_seg = x_tran.end;
                                        else end_scale_seg = y_tran.end;

                DBG1("end_scale_seg %d\n",end_scale_seg);
            }

            if(no_skel)
#if defined MDES2BM
                *cv = MDES2BM(*cv);
#else
                *cv = des2bm(*cv);
#endif /* defined MDES2BM */
            else 
#if defined MDES2BM
                *cv = MDES2BM(nxy(*cv));
#else
                *cv = des2bm(nxy(*cv));
#endif /* defined MDES2BM */

            x = cv->x;
            y = cv->y;
            if(x < x_min) x_min = x;
            if(x > x_max) x_max = x;
            if(y < y_min) y_min = y;
            if(y > y_max) y_max = y;
            DBG3("%d: %d %d\n", next_vec_num,cv->x, cv->y);

            cv++;

        } /*  for(next_vec_num...) */

      /*  F o r    e a c h    c u r v e    p o i n t    i n    l o o p  */
      /*----------------------------------------------------------------*/

        *cv = *loop.cv;   /* close the loop */

#if NON_Z_WIND
        if(if_state.non_z_wind)
        {
            int     y_goes_up;
            SW16    yprev;

            DBG("nzw processing\n");
            cv = loop.cv;
            yprev = cv->y;
            ++if_state.cs.tran_run_ct;   /*  bump for first tran run;
                                          *  assumes this contour is not empty
                                          */
            y_goes_up = (yprev <= (cv+1)->y);

            for(next_vec_num = 0; next_vec_num <= ncoords; next_vec_num++)
            {
                y = cv->y;
                DBG1("y = %d\n", y);
                if((y_goes_up && (yprev > y)) || (!y_goes_up && (yprev < y)))
                {
                    ++if_state.cs.tran_run_ct;
                    y_goes_up = !y_goes_up;
                }
                yprev = y;
                cv++;
            }
            DBG1("nzw  tran_run_ct %d\n", if_state.cs.tran_run_ct);
        }
#endif
    } /* for(loopct...) */

  /*          F o r    e a c h    c o n t o u r    l o o p              */
  /*--------------------------------------------------------------------*/

    if_state.tbound.ll.x = (INTR)x_min;
    if_state.tbound.ll.y = (INTR)y_min;
    if_state.tbound.ur.x = (INTR)x_max;
    if_state.tbound.ur.y = (INTR)y_max;

    return SUCCESS;
}
#endif  /* IF_RDR */
