/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/cor/q3.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: q3.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.7   06 Jul 1995 08:20:42   JOE
 * 
 * Basic 1200 dpi support.
 * 
 *    Rev 1.6   06 Apr 1995 15:11:00   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.5   22 Apr 1994 09:28:14   LISA
 * 
 * Modified copyright/disclaimer notice for 1994.
 * 
 *    Rev 1.4   29 Jun 1993 13:16:18   JOE
 * Removed ROB_OPTIMIZATION stuff, per AWR.
 * Installed fix in arc() for bounding box clipping problem.
 * 
 *    Rev 1.3   12 Feb 1993 11:33:48   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   07 Jan 1993 11:02:26   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:36:52   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 08:37:40   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */

/* q3.c */


/*---------------------------------------------------------------------
   6-26-90  bjg  Added arc_bitmap, arc_outline and arc_function for
                 outline processors.  Also added control0 and control1
                 for quadratic.
   1-31-91  dET  Add function prototyping for MSC and mix model compilation.
   3-10-91  awr  Added control_points() for quadratic outlines
                 and cleaned up. Added #include q3.h
   3-19-91  jfd  In curve_pt(), fix bug in debug statements for printing
                 a->x and a->y. 
  17-Jun-91 jfd  Moved "debug.h" after "port.h".
  22-Aug-91 jfd  Conditionally compiling control_points() based on QUADRA.
  14-Sep-91 awr  Use LONGs in flatten(), changed recursion.
   9-Nov-91 awr  Moved if_long_mult() from if2.31
  11-Nov-91 jfd  Changed if_long_mult() from EXTERN to GLOBAL.
  25-Nov-91 mac  Changed stopping criterion in arc() to produce the optimal
                 number of vectors for better quality and performance
                 throughout the 6-999 pt. at 300 dpi range. Given that we are
                 not increasing resolution nor subdividing by multiples of 3.
                 AREA >>= ( d.x.grid_shift + 5 );
  15-Jan-92 jfd  Added function prototype for lineto().
  22-Jan-92 rs   Change 'lineto()' to 'bmras_lineto()'.
  24-Feb-92 jfd  In q3(), removed declaration for unreferenced variable "i".
  03-Apr-92 rs   Portability cleanup (see port.h).
  21-May-92 jfd  In flatten() and arc(), replaced all direct outline function
                 calls with calls via the function table.
                 Added pointer to font function table "pras".
  26-Jun-92 bjg  Removed if_long_mult (except for cubics) and replaced with
                 fp routines.
   3-Jul-92 awr  #define RECURSIVE 1
  14-Jul-92 jfd  In projection(), removed unreferenced local variables
                 "mult", "den" and "shift".
  18-Jul-92 awr  Simplified fp calls
  31-Jul-92 rs   Performance enhancements - call lineto() through register
                 variable instead of multiple layers of indirection.
  04-Aug-92 jfd  Performance enhancements - call lineto() through global
                 pointer instead of register variable.
  16-Aug-92 awr  Corrected projection()
  07-Jan-93 jfd  ANSI C function declaration changes.
  08-Feb-93 jfd  VXWorks support.
  19-Jun-93 jwd  Removed ROB_OPTIMIZATION stuff, per AWR.
  22-Jun-93 jwd  Installed fix in arc() for bounding box clipping problem.
  06-Jul-95 mac  Basic 1200 dpi support.
*/

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"

#if IF_RDR               /* Conditionally compile whole module */

#include "port.h"
#include "debug.h"
#include "ifmem.h"
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "q3.h"
#include "fpmath.h"

#define dot(A,B) ( ((SL32)(A.x)*(SL32)(B.x)) + ((SL32)(A.y)*(SL32)(B.y)) )
#define crs(A,B) ( ((SL32)(A.x)*(SL32)(B.y)) - ((SL32)(A.y)*(SL32)(B.x)) )

EXTERN IF_STATE if_state;

#ifdef LINT_ARGS
MLOCAL UW16       arc(INTRVECTOR, INTRVECTOR, SW16VECTOR, BOOLEAN);
MLOCAL UW16       flatten(INTRVECTOR, INTRVECTOR, INTRVECTOR, UW16);
MLOCAL SW16VECTOR projection(SW16VECTOR, SW16VECTOR);
GLOBAL VOID       if_long_mult(SL32, SW16, LPSL32, LPSW16);
#else
GLOBAL VOID       if_long_mult();
MLOCAL UW16       flatten();
#endif

#if (CUBIC)
/* ----------------------------------------------------------------------- */
/* Routine: if_long_mult                                                   */
/* Purpose: Multiplies long by word and computes full result in two long   */
/*          words. However, it returns a result that fits in a             */
/*          long word by shifting down the full shift number of times      */
/* Input  : signed long (32 bit) and signed word (16 bit) integers.        */
/* Output : result of multiplication scaled down by 2 to the shift power   */
/* Method : Straight forward implementation of long multiply.              */
/* History: 19-Dec-88 MAC Original routine                                 */
/*          12-Apr-89 BJG Changed the algorithm                            */
/*          25-Mar-91 MAC Fixed constant conditional expressions           */
/*                        And special case when either input is minimum    */
/*          17-Jun-91 MAC FIxed bug in if ( Ch & 0x00800000 ) {            */
/* Notes  : The requirements for this routine are specific of Intellifont. */
/*          This is not intended to be a general purpose mult function.    */
/* ----------------------------------------------------------------------- */
#if defined (ANSI_DEFS)
GLOBAL VOID if_long_mult ( SL32 LA, SW16 WB, LPSL32 LC, LPSW16 shift)
#else
GLOBAL VOID if_long_mult ( LA, WB, LC, shift )
SL32  LA;               /* Long word A */
SW16  WB;               /* Short word B */
LPSL32  LC;              /* Long result C shifted down shift times to fit   */
LPSW16  shift;           /* Number of shifts to fit result in longword      */
#endif
{
	SL32  Al, Ah, Bl, Ch, Cl;
	SL32 sign;
	sign = 1;
	if ( LA == MIN_LONG ) LA = LA + 1;
	if ( WB == MIN_WORD ) WB = WB + 1;
	if ( LA < (SL32)0 ) {
		sign *= -1;
		LA *= (SL32)-1;
	}
	if ( WB < (SW16)0 ) {
		sign *= -1;
		WB *= -1;
	}
	Al = (0xffff)&LA; Bl = WB;
	Ah = LA>>16; 
	Cl = Al*Bl;
	Ch = Ah*Bl;
	Ch = Ch + (Cl >> 16);
	Cl = Cl&0xffff;

	if ( (Ch & 0xffff8000L) == 0 ) {
		*shift = 0;
		*LC = sign * (Cl + (Ch<<16));
	} else if ( (Ch & 0xffff0000L) == 0 ) {
		*shift = 1;
		*LC = sign * ( (Cl>>1) + (Ch<<15));
	} else if ( Ch & 0xff000000L) {
		if ( Ch & 0xf0000000L) {
			if ( Ch & 0xc0000000L) {
				*shift = 16;
				*LC = sign * Ch;
			} else if ( Ch & 0x20000000L) {
				*shift = 15;
				*LC = sign * ( (Cl>>15) + (Ch<<1));
			} else {
				*shift = 14;
				*LC = sign * ( (Cl>>14) + (Ch<<2));
			}
		} else {
			if ( Ch & 0x0c000000L) {
				if ( Ch & 0x08000000L) {
					*shift = 13;
					*LC = sign * ( (Cl>>13) + (Ch<<3));
				} else {
					*shift = 12;
					*LC = sign * ( (Cl>>12) + (Ch<<4));
				}
			} else if ( Ch & 0x02000000L) {
				*shift = 11;
				*LC = sign * ( (Cl>>11) + (Ch<<5));
			} else {
				*shift = 10;
				*LC = sign * ( (Cl>>10) + (Ch<<6));
			}
		}
	} else {
		if ( Ch & 0x00f00000L) {
			if ( Ch & 0x00c00000L) {
				if ( Ch & 0x00800000L) {
					*shift = 9;
					*LC = sign * ( (Cl>>9) + (Ch<<7));
				} else {
					*shift = 8;
					*LC = sign * ( (Cl>>8) + (Ch<<8));
				}
			} else if ( Ch & 0x00200000L) {
				*shift = 7;
				*LC = sign * ( (Cl>>7) + (Ch<<9));
			} else {
				*shift = 6;
				*LC = sign * ( (Cl>>6) + (Ch<<10));
			}
		} else {
			if ( Ch & 0x000c0000L) {
				if ( Ch & 0x00080000L) {
					*shift = 5;
					*LC = sign * ( (Cl>>5) + (Ch<<11));
				} else {
					*shift = 4;
					*LC = sign * ( (Cl>>4) + (Ch<<12));
				}
			} else if ( Ch & 0x00020000L) {
				*shift = 3;
				*LC = sign * ( (Cl>>3) + (Ch<<13));
			} else {
				*shift = 2;
				*LC = sign * ( (Cl>>2) + (Ch<<14));
			}
		}
	}
}
#endif

#define RECURSIVE 1
#if RECURSIVE
/*-----------------*/
/*     flatten     */
/*-----------------*/
#if defined (ANSI_DEFS)
MLOCAL UW16  flatten(INTRVECTOR P0, INTRVECTOR CNTRL, INTRVECTOR P1, UW16 n)
#else
MLOCAL UW16
flatten(P0, CNTRL, P1, n)
    INTRVECTOR P0, CNTRL, P1;
    UW16 n;
#endif
{
    UW16 status;
    INTRVECTOR LC, RC, C;
    INTRVECTOR p;

    LC.x  = (P0.x + CNTRL.x) >> 1;   /* next control point on left  */
    LC.y  = (P0.y + CNTRL.y) >> 1;
    RC.x  = (CNTRL.x + P1.x) >> 1;   /* next control point on right */
    RC.y  = (CNTRL.y + P1.y) >> 1;
    C.x   = (LC.x + RC.x) >> 1;   /* curve point */
    C.y   = (LC.y + RC.y) >> 1;

    n--;
    if(n)
    {
        status = flatten(P0, LC, C, n);
        if(status)
            return status;
    }
    p = C;
    if(p.x < if_state.tbound.ll.x) p.x = if_state.tbound.ll.x;
    if(p.y < if_state.tbound.ll.y) p.y = if_state.tbound.ll.y;
    if(p.x > if_state.tbound.ur.x) p.x = if_state.tbound.ur.x;
    if(p.y > if_state.tbound.ur.y) p.y = if_state.tbound.ur.y;

    status = if_state.out.lineto (p.x, p.y);
    if(status)
        return status;

    if(n)
    {
        status = flatten(C, RC, P1, n);
        if(status)
            return status;
    }
    return SUCCESS;
}

#else  /* not RECURSIVE */
/*  This is a non-recursive version of the flatten() function.
 *  It is slower than the recursive version because it pushes and
 *  immediately pops before outputting- an extra push/pop per point.
 */

#define STKDEPTH (1<<MAX_SD_LEVS)
MLOCAL INTRVECTOR stack[2*STKDEPTH];
MLOCAL UW16  level[STKDEPTH];

/*-----------------*/
/*     flatten     */
/*-----------------*/
#if defined (ANSI_DEFS)
MLOCAL UW16  flatten(INTRVECTOR P0, INTRVECTOR CNTRL, INTRVECTOR P1, UW16 n)
#else
MLOCAL UW16
flatten(P0, CNTRL, P1, n)
    INTRVECTOR P0, CNTRL, P1;
    UW16 n;
#endif
{
    UW16 status;
    INTRVECTOR RC;
    INTRVECTOR p;
    int stackcount;
    INTRVECTOR *s;
    UW16  lev;

    lev =0;
    stackcount = 0;
    s = stack;

loop:
    if(lev < n)
    {
        RC.x  = (CNTRL.x + P1.x) >> 1;      /* next control point on right */
        RC.y  = (CNTRL.y + P1.y) >> 1;
        CNTRL.x  = (P0.x + CNTRL.x) >> 1;   /* next control point on left  */
        CNTRL.y  = (P0.y + CNTRL.y) >> 1;
        *s++ = RC;
        *s++ = P1;
        level[stackcount++] = ++lev;
        P1.x   = (CNTRL.x + RC.x) >> 1;   /* curve point */
        P1.y   = (CNTRL.y + RC.y) >> 1;
    }
    else
    {
        if(!stackcount) return SUCCESS;
        P0 = P1;                  /* pop curve to right */
        P1    = *--s;
        CNTRL = *--s;
        lev = level[--stackcount];

        p = P0;
        if(p.x < if_state.tbound.ll.x) p.x = if_state.tbound.ll.x;
        if(p.y < if_state.tbound.ll.y) p.y = if_state.tbound.ll.y;
        if(p.x > if_state.tbound.ur.x) p.x = if_state.tbound.ur.x;
        if(p.y > if_state.tbound.ur.y) p.y = if_state.tbound.ur.y;

        status = if_state.out.lineto (p.x, p.y);
        if(status)
            return status;
    }
    goto loop;

    return status;    /* for brain dead compilers */
}
#endif


/*-----------------*/
/*   projection    */
/*-----------------*/
/*  Calculate and return the projection of C (perpendicular to C)
 *  onto A.
 */
#if defined (ANSI_DEFS)
MLOCAL SW16VECTOR projection(SW16VECTOR C, SW16VECTOR A)
#else
MLOCAL SW16VECTOR
projection(C, A)
    SW16VECTOR C, A;
#endif
{
    SW16VECTOR   B;
    SW16VECTOR   E;
    SL32  div, fac;
    FPNUM fdiv, ffac;

    DBG2("    projection  C = (%d,%d)\n", C.x, C.y);

    fac = dot(C,C);
    div = dot(A,C);
    DBG2("    fac = %ld     div = %ld\n", fac, div);

    B.x = A.x << powmins;
    B.y = A.y << powmins;

    if ( div == 0 )
    {
        E.x = 0;
        E.y = 0;
    }
    else
    {
#ifdef NOT1200DPI
        if ( log_xpix > powplus )
        {
            E.x = (SW16)((fac*(SL32)B.x)/div);
            E.y = (SW16)((fac*(SL32)B.y)/div);
        }
        else
#endif  /* NOT1200DPI */
        {
            ffac = fplongexp(fac,0);
            fdiv = fplongexp(div,0);
            E.x = fp2word(fpdiv(fpmul( ffac, fplongexp((SL32)B.x,0)), fdiv));
            E.y = fp2word(fpdiv(fpmul( ffac, fplongexp((SL32)B.y,0)), fdiv));
        }
    }
    DBG2("    E = (%d,%d)\n", E.x, E.y);

    return E;
}




/*-----------------*/
/*   arc           */
/*-----------------*/
#if defined (ANSI_DEFS)
MLOCAL UW16  arc(INTRVECTOR p0, INTRVECTOR p1, SW16VECTOR A, BOOLEAN neg)
#else
MLOCAL UW16
arc(p0, p1, A, neg)
    INTRVECTOR p0, p1;
    SW16VECTOR A;
    BOOLEAN  neg;
#endif
{
    SW16 status;
    SW16VECTOR   C, D, E;
    INTRVECTOR   BB, DD, EE, CNTRL;
    int    indx;
    int    area, powr;
    SL32   AREA;

#if INTR_SIZE == 16
    DBG6("arc(x0=%d, y0=%d, x1=%d, y1=%d, A.x=%d, A.y=%d,\n",
                   p0.x, p0.y, p1.x, p1.y, A.x, A.y);
#else
    DBG6("arc(x0=%ld, y0=%ld, x1=%ld, y1=%ld, A.x=%d, A.y=%d,\n",
                   p0.x, p0.y, p1.x, p1.y, A.x, A.y);
#endif
    DBG1("        neg=%u)\n", neg);

    C.x = decr(p1.x - p0.x);
    C.y = decr(p1.y - p0.y);
    E   = projection(C, A);

    if(neg)
    {
        CNTRL.x = p1.x - E.x;
        CNTRL.y = p1.y - E.y;
    }
    else
    {
        CNTRL.x = p0.x + E.x;
        CNTRL.y = p0.y + E.y;
    }

#if INTR_SIZE == 16
    DBG2("    CNTRL = (%d, %d)\n", CNTRL.x, CNTRL.y);
#else
    DBG2("    CNTRL = (%ld, %ld)\n", CNTRL.x, CNTRL.y);
#endif
  /*  Compute number of points to generate */

    D.x = (SW16)incr(C.x);
    D.y = (SW16)incr(C.y);
    AREA =  crs(D,E);
    if(AREA<0) AREA = -AREA;
    AREA = ( AREA >> areadiv );
    DBG1("  AREA %ld\n", AREA);
    area = (int)AREA;

    powr = 1;
    for ( indx = 0; indx < MAX_SD_LEVS; indx++ )
    {
        if ( area <= powr ) break; 
        powr = powr << 3;
    }
    DBG1( "  area %d\n", area);
    DBG2( "  indx %d  powr %d\n", indx, powr);

    if ( indx > 1 && area < (powr>>1) )
    {
      /* subdivide into thirds, then divide each third indx-2 levels */

        indx -= 2;

#if LINEAR
        if(if_state.est_outline)
        {
            if_state.cs.ct_quadto--;
            if_state.cs.ct_lineto += 3*(1<<indx);
            if_state.p0 = p1;
            return SUCCESS;
        }
        else
#endif
        {
            BB.x     = CNTRL.x + (p1.x   - CNTRL.x)/3;
            BB.y     = CNTRL.y + (p1.y   - CNTRL.y)/3;
            CNTRL.x  = p0.x    + (CNTRL.x- p0.x   )/3;
            CNTRL.y  = p0.y    + (CNTRL.y- p0.y   )/3;
            DD.x     = CNTRL.x + (BB.x   - CNTRL.x)/3;
            DD.y     = CNTRL.y + (BB.y   - CNTRL.y)/3;

            if(indx)
            {
                status = flatten(p0, CNTRL, DD, indx);
                if(status)
                    return status;
            }

            /* Check returned values versus bounding box,  clip if necessary */
            
            if(DD.x < if_state.tbound.ll.x) DD.x = if_state.tbound.ll.x;
            if(DD.y < if_state.tbound.ll.y) DD.y = if_state.tbound.ll.y;
            if(DD.x > if_state.tbound.ur.x) DD.x = if_state.tbound.ur.x;
            if(DD.y > if_state.tbound.ur.y) DD.y = if_state.tbound.ur.y;

            status = if_state.out.lineto (DD.x, DD.y);
            if(status)
                return status;

            CNTRL.x  = DD.x    + ((BB.x - DD.x   )>>1);
            CNTRL.y  = DD.y    + ((BB.y - DD.y   )>>1);
            BB.x     = BB.x    + ((p1.x - BB.x   )>>1);
            BB.y     = BB.y    + ((p1.y - BB.y   )>>1);
            EE.x     = CNTRL.x + ((BB.x - CNTRL.x)>>1);
            EE.y     = CNTRL.y + ((BB.y - CNTRL.y)>>1);

            if(indx)
            {
                status = flatten(DD, CNTRL, EE, indx);
                if(status)
                    return status;
            }

            /* Check returned values versus bounding box,  clip if necessary */
            
            if(EE.x < if_state.tbound.ll.x) EE.x = if_state.tbound.ll.x;
            if(EE.y < if_state.tbound.ll.y) EE.y = if_state.tbound.ll.y;
            if(EE.x > if_state.tbound.ur.x) EE.x = if_state.tbound.ur.x;
            if(EE.y > if_state.tbound.ur.y) EE.y = if_state.tbound.ur.y;

            status = if_state.out.lineto (EE.x, EE.y);
            if(status)
                return status;

            if(indx)
            {
                status = flatten(EE, BB, p1, indx);
                if(status)
                    return status;
            }
        }
    }
    else   /*  not doing thirds  */
    {
#if LINEAR
        if(if_state.est_outline)
        {
            if_state.cs.ct_quadto--;
            if_state.cs.ct_lineto += 1<<indx;
            if_state.p0 = p1;
            return SUCCESS;
        }
        else
#endif
        {
            if(indx)
            {
                status = flatten(p0, CNTRL, p1, indx);
                if(status)
                    return status;
            }
        }
    }

    return (if_state.out.lineto (p1.x, p1.y));
}



/*-----------------*/
/*   q3            */
/*-----------------*/
#if defined (ANSI_DEFS)
GLOBAL UW16  q3(INTRVECTOR s, INTRVECTOR c, INTRVECTOR e)
#else
GLOBAL UW16
q3(s, c, e)
    INTRVECTOR  s, c, e;   /* start, contact, end */
#endif
{
    UW16 status;
    SW16VECTOR   A;

#if INTR_SIZE == 16
    DBG6("q3(%d, %d, %d, %d, %d, %d)\n", s.x, s.y, c.x, c.y, e.x, e.y);
#else
    DBG6("q3(%ld, %ld, %ld, %ld, %ld, %ld)\n", s.x, s.y, c.x, c.y, e.x, e.y);
#endif

    A.x = decr(e.x - s.x);
    A.y = decr(e.y - s.y);

    status = arc(s, c, A, TRUE);
    if(status)
        return status;

    return arc(c, e, A, FALSE);
}





/*-----------------*/
/*  control_points */
/*-----------------*/
/*  Calculate the two control points for a bi-arc */
#if defined (ANSI_DEFS)
GLOBAL VOID  control_points(INTRVECTOR s, INTRVECTOR c, INTRVECTOR e,
   PINTRVECTOR pcon0, PINTRVECTOR pcon1)
#else
GLOBAL VOID
control_points(s, c, e, pcon0, pcon1)
    INTRVECTOR s, c, e;
    PINTRVECTOR pcon0, pcon1;
#endif
{
    SW16VECTOR A, B, E;

    A.x = decr(e.x - s.x);
    A.y = decr(e.y - s.y);

    B.x = decr(c.x - s.x);
    B.y = decr(c.y - s.y);

    E = projection(B, A);
    pcon0->x = c.x - E.x;
    pcon0->y = c.y - E.y;

    B.x = decr(e.x - c.x);
    B.y = decr(e.y - c.y);

    E = projection(B, A);
    pcon1->x = c.x + E.x;
    pcon1->y = c.y + E.y;
}
#endif  /* IF_RDR */
