/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/dep/fpmath.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: fpmath.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.8   06 Jul 1995 08:19:04   JOE
 * 
 * Basic 1200 dpi support.
 * 
 *    Rev 1.7   07 Apr 1995 08:37:08   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.6   30 Nov 1994 18:51:08   MIKE
 * 
 * Added fpsqrt() function.
 * 
 *    Rev 1.5   21 Apr 1994 15:43:08   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.4   12 Feb 1993 13:59:44   JOE
 * VXWorks support.
 * 
 *    Rev 1.3   14 Jan 1993 13:54:42   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.2   17 Dec 1992 17:20:02   ROB
 * Add ANSI function declarations based on ANSI_DEFS.
 * 
 *    Rev 1.1   11 Dec 1992 15:30:02   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 15:26:00   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* fpmath.c */
/*
 *
 *  Floating point math
 *
 *  HISTORY:
 *
 *  29-Mar-91 awr created
 *  17-Jun-91 jfd Moved "debug.h" after "port.h".
 *   4-Jul-91 awr fixed 
 *  15-Aug-91 awr fixed fplt() returned TRUE if args were equal.
 *  25-Oct-91 jfd In fp2long(f) (!INT_FP version), changed MAX_LONG to
 *                EXT_FP_MAX_LONG because of redefinition warning
 *                due to MAX_LONG in port.h .
 *  12-Nov-91 awr In fplongexp(), cast "1" as LONG before shifting to
 *                prevent divide by 0 (e == 16) or reverse character
 *                (e == 15) errors.
 *   8-Mar-92 awr Added fpatof().
 *  14-Mar-92 awr Fixed rounding in fp2long(), simplified fpabs() and
 *                fp2word().
 *  03-Apr-92 rs  Portability cleanup (see port.h).
 *  17-Dec-92 rs  Add function declarations based on ANSI_DEFS.
 *  08-Feb-93 jfd VXWorks support.
 *  28-Nov-94 mby fpsqrt() function for INT_FP=1.
 *  06-Jul-95 mac Basic 1200 dpi support.
 */


#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"
#include "port.h"
#include "debug.h"
#include "fpmath.h"

GLOBAL UW16  fpmath_error;




#if !INT_FP

/*             External floating point function implementation    */

#include <math.h>

#if LINTARGS
#ifdef DEBUG
GLOBAL VOID fpprint(FPNUM);
#endif
GLOBAL FPNUM fplongexp(SL32, SW16);
GLOBAL SW16 fp2word(FPNUM);
GLOBAL SL32 fp2long(FPNUM);
GLOBAL FPNUM fpdiv(FPNUM, FPNUM);
GLOBAL SW16 fp2intel(FPNUM, LPSW16);

#else /* not LINTARGS */
#ifdef DEBUG
GLOBAL VOID fpprint();
#endif
GLOBAL FPNUM fplongexp();
GLOBAL SW16 fp2word();
GLOBAL SL32 fp2long();
GLOBAL FPNUM fpdiv();
GLOBAL SW16 fp2intel();

#endif /* LINTARGS */

#ifdef DEBUG
/*  Print an FPNUM when DEBUG is defined */
#if defined (ANSI_DEFS)
GLOBAL VOID fpprint(FPNUM x)
#else
GLOBAL VOID fpprint(x)
FPNUM x;
#endif /* (ANSI_DEFS) */
{
    DBG1("%f  ", x);
}
#endif /* DEBUG */

#if defined (ANSI_DEFS)
GLOBAL FPNUM fplongexp(SL32 n, SW16 e)
#else
GLOBAL FPNUM fplongexp(n, e)
    SL32 n;
    SW16 e;
#endif /* (ANSI_DEFS) */
{
    FPNUM f;

    f = (FPNUM)n;

    if(e>0)
        return f / (FPNUM)(1L<<e);
    else if(e<0)
        return f * (FPNUM)(1L<<-(e));
    else
        return f;
}

#if defined (ANSI_DEFS)
GLOBAL SW16 fp2word(FPNUM f)
#else
GLOBAL SW16 fp2word(f)
FPNUM f;
#endif /* (ANSI_DEFS) */
{
    if(f<-32767.0 || 32767.0<f)
    {
        fpmath_error = FPconversion;
        return 0;
    }
    return (SW16)f;
}

#if defined (ANSI_DEFS)
GLOBAL SL32 fp2long(FPNUM f)
#else
GLOBAL SL32 fp2long(f)
FPNUM f;
#endif /* (ANSI_DEFS) */
{
#define EXT_FP_MAX_LONG 2147483647.0
    if(f<-EXT_FP_MAX_LONG || EXT_FP_MAX_LONG<f)
        fpmath_error = FPconversion;   /* too big for a SL32 */

    return (SL32)f;
}

#if defined (ANSI_DEFS)
GLOBAL FPNUM fpdiv(FPNUM x, FPNUM y)
#else
GLOBAL FPNUM fpdiv(x, y)
FPNUM x, y;
#endif /* (ANSI_DEFS) */
{
    if(y == 0.0)
    {
        fpmath_error = FPdivide_by_zero;
        return 0.0;
    }
    return x / y;
}

#if defined (ANSI_DEFS)
GLOBAL SW16 fp2intel(FPNUM x, LPSW16 pexp)
#else
GLOBAL SW16 fp2intel(x, pexp)
    FPNUM x;
    LPSW16 pexp;
#endif /* (ANSI_DEFS) */
{
    FPNUM f;
    int i;

    DBG1("fp2intel(%f)\n", x);

    f = frexp(x, &i);
    DBG2("%f %d\n", f, i);

    *pexp = 15 - i;
    return (SW16)(f * 32768.0);
}


#else  /* INT_FP ------------------------------------------------------ */

/*             Internal floating point function implementation
*/


#if LINTARGS
GLOBAL VOID  fpprint(FPNUM);
#ifdef DEBUG
#endif
MLOCAL FPNUM normalize (FPNUM);
GLOBAL FPNUM fplongexp (SL32, SW16);
GLOBAL SW16  fp2word (FPNUM);
GLOBAL SL32  fp2long (FPNUM);
MLOCAL FPNUM scale (FPNUM, int);
FPNUM  fpatof(LPSB8);
GLOBAL FPNUM fpneg (FPNUM);
GLOBAL FPNUM fpadd (FPNUM, FPNUM);
GLOBAL FPNUM fpsub (FPNUM, FPNUM);
GLOBAL FPNUM fpmul (FPNUM, FPNUM);
GLOBAL FPNUM fpdiv (FPNUM, FPNUM);
GLOBAL BOOLEAN fplt (FPNUM, FPNUM);
GLOBAL FPNUM fpabs (FPNUM);
GLOBAL SW16  ifpmul (SW16, FPNUM);
GLOBAL SW16  fp2intel (FPNUM, LPSW16);
GLOBAL FPNUM fpsqrt(FPNUM);
#else /* not LINTARGS */

GLOBAL VOID  fpprint();
#ifdef DEBUG
#endif
MLOCAL FPNUM normalize ();
GLOBAL FPNUM fplongexp ();
GLOBAL SW16  fp2word ();
GLOBAL SL32  fp2long ();
MLOCAL FPNUM scale ();
FPNUM  fpatof();
GLOBAL FPNUM fpneg ();
GLOBAL FPNUM fpadd ();
GLOBAL FPNUM fpsub ();
GLOBAL FPNUM fpmul ();
GLOBAL FPNUM fpdiv ();
GLOBAL BOOLEAN fplt ();
GLOBAL FPNUM fpabs ();
GLOBAL SW16  ifpmul ();
GLOBAL SW16  fp2intel ();
GLOBAL FPNUM fpsqrt();

#endif /* LINTARGS */



#ifdef DEBUG
/*  Print an FPNUM in the format 1.xxxx e xx */
#if defined (ANSI_DEFS)
GLOBAL VOID fpprint(FPNUM x)
#else
GLOBAL VOID
fpprint(x)
    FPNUM x;
#endif /* ANSI_DEFS */
{
    double f;

    if(!x.n)
    {
        DBG("0.0  ");
    }
    else
    {
        if(x.e < MIN_EXPONENT || MAX_EXPONENT < x.e)
        {
            DBG("(float out of range) ");
            return;
        }

        if(x.n<0)
        {
            DBG("-");
            x.n = -x.n;
        }
        if(!x.n & 0x40000000)
        {
            DBG("not normalized\n");
        }
        else
        {
            f = (double)x.n/(double)0x40000000;

            if(x.e>0)
            {
                DBG1("%f  ", f*(double)((SL32)1<<x.e));
            }
            else
            {
                DBG1("%f  ", f/(double)((SL32)1<<(-x.e)));
            }
        }
    }
}
#endif /* DEBUG */

/*----------------*/
/*    normalize   */
/*----------------*/
/*  Return normalized version of x */
#if defined (ANSI_DEFS)
MLOCAL FPNUM normalize(FPNUM x)
#else
MLOCAL FPNUM
normalize(x)
    FPNUM x;
#endif /* ANSI_DEFS */
{
    if(!x.n)
    {
        x.e = 0;
        return x;
    }
    if(x.n > 0)
    {
        while(x.n < 0x40000000L)
        {
            x.n <<= 1;
            x.e--;
        }
    }
    else
    {
        x.n = -x.n;
        while(x.n < 0x40000000L)
        {
            x.n <<= 1;
            x.e--;
        }
        x.n = -x.n;
    }
    if(x.e < MIN_EXPONENT)
    {
        x.n = 0;
        x.e = 0;
    }
    return x;
}


/*
               C o n v e r s i o n    F u n c t i o n s
*/


/*----------------*/
/*   fplongexp    */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL FPNUM fplongexp(SL32 n, SW16 e)
#else
GLOBAL FPNUM
fplongexp(n, e)
    SL32 n;
    SW16 e;
#endif /* ANSI_DEFS */
{
    FPNUM f;

    f.n = n;
    f.e = 30 - e;
    return normalize(f);
}




/*----------------*/
/*    fp2word     */
/*----------------*/
/*  Convert the FPNUM f to a SW16 */
#if defined (ANSI_DEFS)
GLOBAL SW16 fp2word(FPNUM f)
#else
GLOBAL SW16
fp2word(f)
    FPNUM f;
#endif /* ANSI_DEFS */
{
    SL32 r;

    r = fp2long(f);
    if((r < -32767L) || (32767L < r)) 
    {
        fpmath_error = FPconversion;
        return 0;
    }
    return (SW16)r;
}




/*----------------*/
/*    fp2long     */
/*----------------*/
/*  Convert the FPNUM f to a SL32 */
#if defined (ANSI_DEFS)
GLOBAL SL32 fp2long (FPNUM f)
#else
GLOBAL SL32
fp2long(f)
    FPNUM f;
#endif /* ANSI_DEFS */
{
    int shift;
    SL32 r;

    if(!f.n) return 0;
    shift = 30 - f.e;          /* ammount to shift right */
    if(shift > 31 ) return 0;  /* result is < 0.5 in abs value */
    if(shift <  0)             /* overflow ? */
    {
        fpmath_error = FPconversion;
        return 0;
    }
    if(shift == 0) return (SL32)f.n;

    r = f.n >> shift;
    if(f.n & (1L<<(shift-1)))   /* If bit just to right of binary point */
        r++;                    /* then round up result */
    return r;
}


/*----------------*/
/*   scale        */
/*----------------*/
#if defined (ANSI_DEFS)
MLOCAL FPNUM scale(FPNUM x, int e)
#else
MLOCAL FPNUM
scale(x, e)
    FPNUM x;
    int   e;
#endif /* ANSI_DEFS */
{
    FPNUM pow;

    pow = fpint2fp(10L);

    if(e > 0)
    {
        while(e)
        {
            if(e&1) x = fpmul(x, pow);
            e >>= 1;
            pow = fpmul(pow, pow);
        }
    }
    else
    {
        e = -e;
        while(e)
        {
            if(e&1) x = fpdiv(x, pow);
            e >>= 1;
            pow = fpmul(pow, pow);
        }
    }
    return x;
}

/*----------------*/
/*    fpatof      */
/*----------------*/

#define convert_digit(r) r = fpadd(fpmul(r, ten), fpint2fp((SL32)(*p++ - '0')))
#define isdigit(c)  (c == '0' || c == '1' || c == '2' || c == '3' \
   || c == '4' ||  c == '5' || c == '6' || c == '7' || c == '8' || c == '9')

#if defined (ANSI_DEFS)
FPNUM fpatof(LPSB8 p)
#else
FPNUM fpatof(p)
    LPSB8 p;
#endif /* ANSI_DEFS */
{
    FPNUM ten;
    FPNUM x;
    int    sign;
    FPNUM  frac;
    int    fcnt;
    int    exp;
    int    esign;

    ten = fpint2fp(10L);
    x = frac = fpint2fp(0L);

    sign = 0;
    fcnt = 0;
    exp = 0;
    esign = 0;
    while(*p == ' ' || *p == '\t') p++;
    if(*p == '-')
    {
        p++;
        sign = 1;
    }
    while(isdigit(*p)) convert_digit(x);

    if(*p == '.')
    {
        p++;
        while(isdigit(*p))
        {
            convert_digit(frac);
            fcnt++;
        }
    }
    if(*p == 'e' || *p == 'E')
    {
        p++;
        if(*p == '+') p++;
        else if(*p == '-')
        {
            p++;
            esign = 1;
        }
        while(isdigit(*p)) exp = 10*exp + (*p++ - '0');
        if(esign) exp = -exp;
    }
    x = scale(fpadd(x, scale(frac, -fcnt)), exp);
    if(sign) x = fpneg(x);
    return x;
}




/*
             A r i t h m e t i c    F u n c t i o n s
*/



/*----------------*/
/*    fpneg       */
/*----------------*/
/*  Return the negative of the FPNUM x  */
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpneg (FPNUM x)
#else
GLOBAL FPNUM
fpneg(x)
    FPNUM x;
#endif /* ANSI_DEFS */
{
    x.n = -x.n;
    return x;
}



#define SIGN(f)   (f.n & 0x80000000L)

/*----------------*/
/*   fpadd        */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpadd(FPNUM x, FPNUM y)
#else
GLOBAL FPNUM
fpadd(x, y)
    FPNUM x, y;
#endif /* ANSI_DEFS */
{
    FPNUM r;
    int diff;

    if(!x.n) return y;
    if(!y.n) return x;

    diff = x.e - y.e;
    if(diff > 0)
    {
      /* shift y right the difference  */
        if(diff > 30)
            return x;
        else
            y.n >>= diff;

        r.e = x.e;
    }
    else
    {
      /* shift x right the difference  */
        diff = - diff;
        if(diff > 30)
            return y;
        else
            x.n >>= diff;

        r.e = y.e;
    }

    r.n = x.n + y.n;

  /*  overflow check  */

    if(SIGN(x) != SIGN(y))      /* no overflow if opposite signs */
        return normalize(r);    /* so return normalized result   */

  /* signs of x and y are same. Overflow if sign of result is different */

    if(SIGN(x) != SIGN(r))
    {
        r.e++;
        if(r.e > MAX_EXPONENT)
        {
            fpmath_error = FPadd_overflow;
            return r;
        }

        if(SIGN(r))  /* overflow sign is backwards */
          /* Result should be positive */
            r.n = (r.n >> 1) & 0x7fffffffL;
        else
          /* Result should be negative */
            r.n = (r.n >> 1) | 0x80000000L;
    }

    return r;
}




/*----------------*/
/*   fpsub        */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpsub(FPNUM x, FPNUM y)
#else
GLOBAL FPNUM
fpsub(x, y)
    FPNUM x, y;
#endif /* ANSI_DEFS */
{
    return fpadd(x, fpneg(y));
}



#define FLO(x)  ((x) & 0xffff)
#define FHI(x)  ((x) >> 16)    /* Assumes x is UL32 */

/*----------------*/
/*   fpmul        */
/*----------------*/
/*  Return x*y  */
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpmul(FPNUM x, FPNUM y)
#else
GLOBAL FPNUM
fpmul(x, y)
    FPNUM x, y;
#endif /* ANSI_DEFS */
{
    int sign;
    FPNUM p;
    UL32 xx, yy;
    UL32 t0, t1, t2, t3;
    UL32 m0, m1;

  /* special case x or y == 0  */

    if(!x.n) return x;
    if(!y.n) return y;

  /* Convert to positive numbers */

    sign = 1;
    if(x.n < 0)
    {
        x.n = -x.n;
        sign = - sign;
    }
    if(y.n < 0)
    {
        y.n = -y.n;
        sign = - sign;
    }
    xx = (UL32)x.n;
    yy = (UL32)y.n;

    t0 = FLO(xx) * FLO(yy);
    t1 = FLO(xx) * FHI(yy);
    t2 = FHI(xx) * FLO(yy);
    t3 = FHI(xx) * FHI(yy);

    m0 = FHI(t0) + FLO(t1) + FLO(t2);
    m1 = FHI(t1) + FHI(t2) + t3 + FHI(m0);
    m0 = FLO(m0);

    p.e = x.e + y.e;
    if(0x20000000L & m1)
    {
      /* only need one shift left */
        m1 <<= 1;
        m1 |= (m0>>15);
        p.e++;
    }
    else
    {
      /* shift left two */
        m1 <<= 2;
        m1 |= (m0>>14);
    }

    p.n = (SL32)m1;
    if(sign<0) p.n = -p.n;

    if(p.e < MIN_EXPONENT)
    {
        p.n = 0;
        p.e = 0;
    }
    if(p.e > MAX_EXPONENT)
        fpmath_error = FPmultiply_overflow;

    return p;
}




/*----------------*/
/*   fpdiv        */
/*----------------*/
/*  Return x/y  */
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpdiv(FPNUM x, FPNUM y)
#else
GLOBAL FPNUM
fpdiv(x, y)
    FPNUM x, y;
#endif /* ANSI_DEFS */
{

    int    sign, i;
    FPNUM  q;
    UL32  xx, yy;

    if(!y.n)
    {
        fpmath_error = FPdivide_by_zero;
        return y;
    }
    if(!x.n) return x;

  /* convert to positive numbers */

    sign = 1;
    if(x.n < 0)
    {
        x.n = -x.n;
        sign = - sign;
    }
    if(y.n < 0)
    {
        y.n = -y.n;
        sign = - sign;
    }

    xx = (UL32)x.n;
    yy = (UL32)y.n;

    q.n = 0;
    q.e = x.e - y.e;
    if(yy > xx)
    {
        xx <<= 1;
        q.e--;
    }

  /*  Generate 31 bits of quotient  */

    for(i=0; i<31; i++)
    {
        q.n <<= 1;
        if(yy <= xx)
        {
            q.n |= 1;
            xx -= yy;
        }
        xx <<= 1;
    }

    if(sign<0) q.n = - q.n;

    if(q.e < MIN_EXPONENT || q.e > MAX_EXPONENT)
        fpmath_error = FPdivide_overflow;

    return q;
}


/*----------------*/
/*   fplt         */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL BOOLEAN fplt(FPNUM x, FPNUM y)
#else
GLOBAL BOOLEAN
fplt(x, y)
    FPNUM x, y;
#endif /* ANSI_DEFS */
{
    if(x.n<0 && y.n>=0) return TRUE;
    else if(x.n>=0 && y.n<=0) return FALSE;
    else if(x.n<0 && y.n<0)
    {
        if(x.e < y.e) return FALSE;
        else if(x.e > y.e) return TRUE;
    }
    else
    {
        if(x.e < y.e) return TRUE;
        else if(x.e > y.e) return FALSE;
    }

  /* exponents are equal */

    if(x.n < y.n) return TRUE;
    else          return FALSE;
}




/*----------------*/
/*   fpabs        */
/*----------------*/
/*  Return absolute value of x  */
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpabs (FPNUM x)
#else
GLOBAL FPNUM
fpabs(x)
    FPNUM x;
#endif /* ANSI_DEFS */
{
    if(x.n < 0)
        x.n = - x.n;
    return x;
}




/*----------------*/
/*   ifpmul       */
/*----------------*/
/*  Multiply a SW16 by a FPNUM and return a rounded SW16 */
#if defined (ANSI_DEFS)
GLOBAL SW16 ifpmul(SW16 x, FPNUM y)
#else
GLOBAL SW16
ifpmul(x, y)
    SW16  x;
    FPNUM y;
#endif /* ANSI_DEFS */
{
    int sign;
    UW16 xx;
    UL32 yy;
    UL32 t0, t1;
    UL32 m;
    SW16  p;
    int   right_shift;

  /* special case x or y == 0  */

    if(!x)   return 0;
    if(!y.n) return 0;

  /* Convert to positive numbers */

    sign = 1;
    if(x < 0)
    {
        x = -x;
        sign = - sign;
    }
    if(y.n < 0)
    {
        y.n = -y.n;
        sign = - sign;
    }
    xx = (UW16)x;
    yy = (UL32)y.n;

    t0 = xx * FLO(yy);
    t1 = xx * FHI(yy);

    m = FHI(t0) + t1;
    if(y.e>14)
        fpmath_error = FPmultiply_overflow;  /* result longer than a SW16 */

    right_shift = 14 - y.e;
    if(right_shift == 16)
        p = (SW16)(FHI((m + 0x8000)));    /* 0x8000 rounds */
    if(right_shift > 16)
    {
        right_shift -= 16;
        p = (SW16)  (FHI(m + (1<<(right_shift-1))) >> right_shift);
    }
    else
    {
        m = (m + (1<<(right_shift-1))) >> right_shift;
        p = (SW16)(FLO(m));
        if(p<0 || FHI(m))
            fpmath_error = FPmultiply_overflow;
    }

    if(sign<0) p = -p;
    return p;
}


/*----------------*/
/*   fp2intel     */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL SW16 fp2intel (FPNUM x, LPSW16 pexp)
#else
GLOBAL SW16
fp2intel(x, pexp)
    FPNUM x;
    LPSW16 pexp;
#endif /* ANSI_DEFS */
{
    *pexp = 14 - x.e;
    return (SW16)(x.n >> 16);
}


/*------------*/
/*   fpsqrt   */
/*------------*/
#if defined (ANSI_DEFS)
GLOBAL FPNUM fpsqrt (FPNUM ff)
#else
GLOBAL FPNUM fpsqrt(ff)
    FPNUM ff;
#endif /* ANSI_DEFS */
{
  /* Newton's Method (2 iterations) for values between 0.5 and 2.0
   *  x0 = x/(x+1) + (x+1)/4;
   *  x1 = (x/x0 + x0)/2
   */
    FPNUM tmp;
    SW16 expadj;
    if (fpiszero(ff))
        return(ff);
    if (fpisneg(ff))
    {
        fpmath_error = FPsqrt_neg;
        return(ff);
    }
    ff = normalize(ff);
    expadj = (ff.e + 1) >> 1;
    ff.e -= (expadj << 1);
    tmp = fpadd(ff, fpint2fp(1L));
    tmp = fpadd(fpdiv(ff, tmp), fpdiv(tmp, fpint2fp(4L)));
    tmp = (fpdiv(fpadd(ff, fpmul(tmp, tmp)), fpmul(tmp, fpint2fp(2L))));
    tmp.e += expadj;
    return (tmp);
}

#endif  /* internal fp implementation */




/*=====================================================================*/
/*                        Test Main                                    */

#define TEST_FPNUM  0
#if TEST_FPNUM

GLOBAL VOID  main(int, char**);

int trace_sw = 1;

GLOBAL VOID
main(argc, argv)
    int  argc;
    char *argv[];
{
    FPNUM a, b, c;

#if 0
    if(argc == 1)
    {
        printf("Usage:\n");
        printf("  fpmath <float point string>\n\n");
        printf("Where the string is of the form for conversion by atof()\n");
        exit(1);
    }

    c = fpatof(argv[1]);
#endif


    c = fpdiv(fplongexp(2, 0), fplongexp(3, 0));
    fpprint(c);
    printf("%ld\n", fp2long(c));

    c = fpdiv(fplongexp(-2, 0), fplongexp(3, 0));
    fpprint(c);
    printf("%ld\n", fp2long(c));
}
#endif  /* TEST_FPNUM */

