/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/dep/instance.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: instance.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.13   14 Jun 1995 17:47:26   MIKE
 * Fixed argument to macro FC_ISAPPGEN
 * 
 *    Rev 1.12   07 Apr 1995 08:36:58   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.11   02 Feb 1995 13:20:40   JOE
 * In INSTsetState(), check "useMapping" before calling SYMnew().
 * 
 *    Rev 1.10   21 Apr 1994 15:43:14   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.9   15 Nov 1993 09:16:52   MAIB
 * 
 * Change optimization for multi-threading and cache access
 * 
 *    Rev 1.8   01 Nov 1993 13:10:56   MAIB
 * Additional data required when switching instances
 * 
 *    Rev 1.7   25 Oct 1993 10:58:28   MAIB
 * Optimized instance switching for multi-threading
 * 
 *    Rev 1.6   03 Aug 1993 14:44:36   MAIB
 * Changed interface to resolve naming conflicts
 * 
 *    Rev 1.5   02 Aug 1993 14:21:18   MAIB
 * Changes to support multi-tasking
 * 
 *    Rev 1.4   15 Jun 1993 09:59:52   JOE
 * Moved "vxWorks.h" before "port.h" to resolve conflicts.
 * 
 *    Rev 1.3   12 Feb 1993 14:12:16   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   17 Dec 1992 17:45:14   ROB
 * Add ANSI function declarations based on ANSI_DEFS.
 * 
 *    Rev 1.1   11 Dec 1992 15:31:30   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 15:25:36   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* instance.c */
/*
 *
 *  History:
 *  original  awr 
 *  18-May-90 dET add instance handling for MS-WINDOWS
 *  27-Nov-90 bg  In INSTinit(), changed line "pInst->status == FREE;" to
 *                "pInst->status = FREE;".
 *   5-Feb-91 awr Change error return in INSTfree() from MAXCALLERS to
 *                MAX_UWORD to decouple cgif.c (MAXCALLERS had to be
 *                defined in cgif.c also).
 *   8-Feb-91 awr Removed FCcopy_fc(), coded in line.
 *                Added #include "instance.h" to declare
 *                multicaller functions.
 *                Removed #include "cache.h"- not needed.
 *  23-Feb-91 awr Conditionally compiled MEMinstFree() in INSTexit()
 *                on INT_MEM_MGT & DEFUND
 *  17-Jun-91 jfd Moved "debug.h" after "port.h".
 *                Moved "cgconfig.h" before "port.h".
 *  19-Jul-91 jfd Changed "MSWINDOWS" to "_WINDOWS".
 *  03-Apr-92 rs  Portability cleanup (see port.h).
 *   4-Jul-92 awr Conditionally compiled cc_buf on CHAR_SIZE.
 *  15-Nov-92 rs  Port to Mac -> move cgconfig.h & port.h to beginning of includes.
 *                Include <stdlib.h> for malloc().
 *  17-Dec-92 rs  Add ANSI function declarations based on ANSI_DEFS.
 *  08-Feb-93 jfd VXWorks support.
 *  15-Jun-93 jfd Moved "vxWorks.h" before "port.h" to resolve conflicts.
 *  02-Aug-93 maib Changed interface to prevent CGIFxxx() routines from calling
 *                 other CGIFxxx() routines, required for multi-tasking
 *  03-Aug-93 maib changed definitions for cgifXXXX() routines to resolve
 *                 naming conflicts
 *  25-Oct-93 maib Optimized INSTsetState for multi-threading
 *  01-Nov-93 maib Additional data required when switching instances
 *  15-Nov-93 maib Preventing cache from being updated for threads which
 *                 do not access the cache. Also determine if the symbol
 *                 set has changed btwn threads and update accordingly.
 *  02-Feb-95 jfd  In INSTsetState(), added check for "useMapping". If
 *                 "useMapping" is 0 call SYMnoMap() instead of SYMnew(). 
 *                 For Application Generated Bitmaps, do not call SYMnew 
 *                 or SYMnoMap. Set appropriate "if_state" values too.
 *                 Included "cgmacros.h", "ix.h", "if_type.h" and 
 *                 "ttroment.h".
 *  14-Jun-95 mby  In INSTsetState(), fixed argument to macro FC_ISAPPGEN().
 */

#include "cgconfig.h"

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include "port.h"

#include <stdio.h>
#include <string.h>

#if defined (MACINTOSH) || defined (VXWORKS)
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include "debug.h"

#if MULTICALLER    /*  Entire module conditionally compiled */

#include "mixmodel.h"
#include "ifmem.h"
#include "cgif.h"
#include "instance.h"     /* function declarations */
#include "sym.h"
#include "ix.h"

#include "if_type.h"
#include "cgmacros.h"
#if TT_ROM1
#include "ttroment.h"
#endif 

typedef struct
{
    UW16       status;
    FONTCONTEXT fc;
    IFCONFIG    cfg;
#if TILE
    IFBITMAP entire_bm;
    SW16VECTOR entire_tt;
#endif
} INSTANCE;

typedef INSTANCE * PINSTANCE;
typedef INSTANCE FARPTR * FPINSTANCE;


#ifdef LINT_ARGS
GLOBAL VOID INSTinit (VOID);
MLOCAL PINSTANCE INSTfind(UW16);
GLOBAL UW16 INSTfree (VOID);
MLOCAL PINSTANCE INSTfind (UW16);
GLOBAL UW16 INSTactive (VOID);
GLOBAL UW16 INSTexit (UW16);
GLOBAL UW16 INSTsaveConfig (UW16, PIFCONFIG);
GLOBAL UW16 INSTsaveState (UW16, PFONTCONTEXT);
GLOBAL UW16 INSTsetState (UW16);
EXTERN VOID  FCcopy(PFONTCONTEXT, PFONTCONTEXT);
EXTERN BOOLEAN SYMnew(UW16, UW16);
EXTERN UW16 CACfont(PFONTCONTEXT);
#else /* no LINTARGS */
GLOBAL VOID INSTinit ();
MLOCAL PINSTANCE INSTfind();
GLOBAL UW16 INSTfree ();
MLOCAL PINSTANCE INSTfind ();
GLOBAL UW16 INSTactive ();
GLOBAL UW16 INSTexit ();
GLOBAL UW16 INSTsaveConfig ();
GLOBAL UW16 INSTsaveState ();
GLOBAL UW16 INSTsetState ();
EXTERN VOID    FCcopy();
EXTERN BOOLEAN SYMnew();
EXTERN UW16 CACfont();
#endif /* LINTARGS */

#if TILE
EXTERN IFBITMAP entire_bm;
EXTERN SW16VECTOR entire_tt;
#endif


/*-----------------------------------------*/
#define MAXCALLERS 10   /* arbitrary */
#define FIRSTTIME  ~0   /* a large value */

MLOCAL  UW16  hCompCharBuf = 0;   /* handle for windows, flag for others */
GLOBAL  SB8  *ccBuffer;           /* buffer pointer for comp-char buffer */
GLOBAL  UW16  uLastCallerId = FIRSTTIME;  /* an initial value */

#define FREE  0
#define INUSE 1

INSTANCE aInstance[MAXCALLERS];        /* array of instances  */


#if SEM_ACCESS
/*
 * 11/15/93 maib
 * this array identifies which instances access the
 * cache and is examined when switching instances
 * to determine whether or not to update the cache
 */
UW16 InstCACAccess[MAXCALLERS];
#endif /* SEM_ACCESS */

/*
 * 11/1/93
 * update cgif.c's copy of fontcontext whenever an
 * instance is switched
 */
EXTERN FONTCONTEXT fcCur;

/*
 * 11/15/93 maib
 * current symbol set data which must be updated
 * when an instance is switched
 */
EXTERN SYMBOLSET symbolset;

/*
 * 02/02/95 jfd
 * current system state data which must be updated
 * when an instance is switched
 */
EXTERN IF_STATE if_state;

/*-----------------------------------------*/

/* initialize all handles to instances by marking status FREE */

/*----------------*/
/*  INSTinit      */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID INSTinit ()
#else
GLOBAL VOID
INSTinit()
#endif /* (ANSI_DEFS) */
{
int   i; /* change SW16 to int - 12/17/92 - rs */
PINSTANCE pInst = aInstance;

    for(i = 0; i < MAXCALLERS; i++, pInst++)   /* all instances FREE */
        pInst->status = FREE;      /* 11-27-90 bg */
}




/*----------------*/
/*  INSTfree      */
/*----------------*/
/* return the first free handle to an instance.  Return MAX_UWORD if
** no free instances available.
*/

#if defined (ANSI_DEFS)
GLOBAL UW16 INSTfree ()
#else
GLOBAL UW16
INSTfree()
#endif /* (ANSI_DEFS) */
{
int   i; /* change from SW16 to int - 12/17/92 - rs */
PINSTANCE pInst = aInstance;

    for (i = 0; i < MAXCALLERS; i++, pInst++) /* find first free instance */
        if(pInst->status == FREE)
        {
            pInst->status = INUSE;         /* mark this one INUSE */
            return i;
        }
    return (MAX_UWORD);
}




/*----------------*/
/*  INSTfind      */
/*----------------*/
/* return a pointer to the instance data of caller 'id'.  Return a
 * NULL pointer upon error.
*/

#if defined (ANSI_DEFS)
MLOCAL PINSTANCE INSTfind (UW16 id)
#else
MLOCAL PINSTANCE
INSTfind(id)
    UW16 id;
#endif /* (ANSI_DEFS) */
{
    if(id > MAXCALLERS)
      return ((PINSTANCE)NULL);

    if(aInstance[id].status == INUSE)
    {
        uLastCallerId = id;                   /* remember current caller */
        return (PINSTANCE)&aInstance[id];
    }
    else
      return ((PINSTANCE)NULL);
}




/*----------------*/
/*  INSTactive    */
/*----------------*/
/* return INUSE if any caller instances are active, otherwise return FREE */

#if defined (ANSI_DEFS)
GLOBAL UW16 INSTactive ()
#else
GLOBAL UW16
INSTactive()
#endif /* (ANSI_DEFS) */
{
int i; /* change from SW16 to int - 12/17/92 - rs */
PINSTANCE pInst = aInstance;

    for(i = 0; i < MAXCALLERS; i++, pInst++) /* find first inuse instance */
        if(pInst->status == INUSE) 
            return (INUSE);
    return (FREE);                           /* no instances inuse */
}




/*----------------*/
/*  INSTexit      */
/*----------------*/
/*  If an application exits the subsystem we must check to see if the 
 *  application requested the sub-system to allocate a comp-char buffer.
 *  If so then we free it.  It may be allocated again.
 *  Mark the caller instance 'id' as being free and defund all memory
 *  funded by this caller.
*/

#if defined (ANSI_DEFS)
GLOBAL UW16 INSTexit (UW16 id)
#else
GLOBAL UW16
INSTexit(id)
    UW16 id;
#endif /* (ANSI_DEFS) */
{
#ifdef _WINDOWS
/* I do not want to include WINDOWS.H, so here are some definitions */
    EXTERN SW16   far pascal GlobalUnlock(SW16);
    EXTERN SW16   far pascal GlobalFree(SW16);
#endif

/*
 * 11/15/93 maib, added to allow SPARC compilation
 */
#ifndef SPARC
    EXTERN VOID   free();
#endif
    PINSTANCE pInst;

    pInst = INSTfind(id);
    if(pInst == (PINSTANCE)NULL)
      return (ERR_bad_ID);

#if CHAR_SIZE
#ifdef _WINDOWS
    if(pInst->cfg.cc_buf_ptr == NULL && hCompCharBuf)
    {
        GlobalUnlock(hCompCharBuf);   /* unlock and free comp-char buffer */
        GlobalFree(hCompCharBuf);
        hCompCharBuf = 0; 
    }
#else
    if(pInst->cfg.cc_buf_ptr == NULL && hCompCharBuf)
    {
        free(ccBuffer);       /* free comp-char buffer */
        hCompCharBuf = 0;
    }
#endif /* _WINDOWS */
#endif /* CHAR_SIZE */

#if INT_MEM_MGT && DEFUND
    MEMinstFree(id);                /* free memory funded by this instance */
#endif

    pInst->status = FREE;
    return (SUCCESS);
}




/*----------------*/
/* INSTsaveConfig */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL UW16 INSTsaveConfig (UW16 id, PIFCONFIG cfg)
#else
GLOBAL UW16
INSTsaveConfig(id, cfg)
    UW16     id;
    PIFCONFIG cfg;
#endif /* (ANSI_DEFS) */
{
    PINSTANCE pInst;
#if CHAR_SIZE
    MLOCAL SL32  ccBufferSize = 0L;
#endif

    if( (pInst = INSTfind(id)) == (PINSTANCE)NULL) return ERR_bad_ID;

#if CHAR_SIZE
    if(cfg->cc_buf_ptr == NULL)  /* must allocate or reallocate space */
    {
        /* buffer must be re-sized or allocated */
        if(cfg->cc_buf_size != ccBufferSize || !hCompCharBuf)
        {
#ifdef  _WINDOWS
#define GMEM_FIXED        0x0000
#define GMEM_MOVEABLE     0x0002

/* I do not want to include WINDOWS.H so here are some definitions */

            EXTERN SW16  far pascal       GlobalAlloc(SW16, UL32);
            EXTERN UB8 far * far pascal GlobalLock(SW16);
            EXTERN SW16  far pascal       GlobalReAlloc(SW16, UL32, UW16);

            if(hCompCharBuf) 
                hCompCharBuf = GlobalReAlloc(hCompCharBuf,
                                            (UL32)cfg->cc_buf_size,
                                             GMEM_MOVEABLE);
            else 
                hCompCharBuf = GlobalAlloc(GMEM_MOVEABLE,
                                         (UL32)cfg->cc_buf_size);

            if(hCompCharBuf == NULL) return ERR_ccBuff_alloc;
            ccBuffer = GlobalLock(hCompCharBuf);
            if(!ccBuffer) return ERR_ccBuff_alloc;
#else
/*            EXTERN SB8 *realloc();
            EXTERN SB8 *malloc();
*/

            if(hCompCharBuf)
                ccBuffer = realloc(ccBuffer, (UW16)cfg->cc_buf_size); /*!!!*/
            else
            {
                ccBuffer = malloc((UW16)cfg->cc_buf_size); /*!!!!*/
                hCompCharBuf = 1;           /* used as a flag in this case */
            }
            if(!ccBuffer)  return ERR_ccBuff_alloc;
#endif

            ccBufferSize = cfg->cc_buf_size;
        }
    }
#endif
    pInst->cfg = *cfg;                        /* save config in state data */  

    return SUCCESS;
}




/*----------------*/
/*  INSTsaveState */
/*----------------*/
/* save the current FONTCONTEXT of the caller 'id' */

#if defined (ANSI_DEFS)
GLOBAL UW16 INSTsaveState (UW16 id, PFONTCONTEXT fc)
#else
GLOBAL UW16
INSTsaveState(id, fc)
    UW16        id;
    PFONTCONTEXT fc;
#endif /* (ANSI_DEFS) */
{
    PINSTANCE pInst;

    pInst = INSTfind (id);
    if(pInst == (PINSTANCE)NULL)
      return (ERR_bad_ID);

    pInst->fc = *fc;       /* save FONTCONTEXT */
#if TILE
    pInst->entire_bm = entire_bm;
    pInst->entire_tt = entire_tt;
#endif

    return (SUCCESS);
}


/*----------------*/
/*  INSTsetState  */
/*----------------*/
/* reset the system from the state data of caller 'id' */

#define DL_SYMSET  0x8000

#if defined (ANSI_DEFS)
GLOBAL UW16 INSTsetState (UW16 id)
#else
GLOBAL UW16
INSTsetState(id)
    UW16   id;
#endif /* (ANSI_DEFS) */
{
    PINSTANCE pInst;
    BOOLEAN useMapping;

    pInst = INSTfind (id);
    if(pInst == (PINSTANCE)NULL)
      return (ERR_bad_ID);

/* 10/25/93 maib, separate config not needed for multithreading */
#if !SEM_ACCESS
    cgif_Config(&pInst->cfg);    /* reconfigure buffers and others  */
#endif /* !SEM_ACCESS */

    /*
     * 11/1/93 maib
     * cgif.c's copy of the fontcontext must be updated whenever
     * an instance is switched. also, setup symbol set data for
     * the new fontcontext if required
     */
    fcCur = pInst->fc;

    useMapping = (fcCur.ssnum != DL_SYMSET);
#if TT_ROM1
    if_state.rom_thing = 0;
    if ( FC_ISTT(&fcCur) && (fcCur.format & FC_ROM_TYPE) )
    {
        if (GET_xLONG(fcCur.font_hdr) == ENTITY1)
        {
            useMapping = FALSE;
            if_state.rom_thing = ENTITY1;
        }
    }
#endif  /* TT_ROM1 */

    if_state.usePlugins = (useMapping ? TRUE : FALSE);

#if ALIEN_BITS
   /* For Application Generated Bitmaps, DO NOT call SYMnew or SYMnoMap */
    if (FC_ISAPPGEN(&fcCur))
        return SUCCESS;
#endif

    if (fcCur.ssnum != symbolset.curssnum)
    {
#if (PCLEO_RDR || TT_ROM1)
      if (!useMapping)
      {
          if (!SYMnoMap(fcCur.font_hdr, fcCur.format))
              return ERR_IXnew_ss;
      }
      else
#endif
      {
      if (!SYMnew(fcCur.ssnum, fcCur.format))
        return ERR_IXnew_ss;
      }
    }

#if CACHE

#if SEM_ACCESS
    /*
     * 11/15/93 maib
     * for performance reasons, only update the cache 
     * if this thread actually uses it
     */
    if (IS_CAC_THREAD(id))
#endif
      CACfont(&pInst->fc);

#endif /* CACHE */

#if TILE
    entire_bm = pInst->entire_bm;
    entire_tt = pInst->entire_tt;
#endif

    return (SUCCESS);
}

#endif    /* MULTICALLER */
