/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/dep/intmem.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: intmem.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.12   07 Apr 1995 08:37:12   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.11   13 Jan 1995 18:06:08   ROB
 * Support for > 64KB memory blocks in DOS.
 * 
 *    Rev 1.10   29 Sep 1994 14:44:40   MARTIN
 * 
 * Removed line in MEMalloc that forced the handle to the requested size. It w
 * 
 *    Rev 1.9   31 Aug 1994 14:20:30   JOE
 * 
 * Added MEMgethandlesize() routine and modified MEMalloc() so that it
 * returns the amount of memory requested, not more.
 * 
 *    Rev 1.8   03 Jun 1994 10:02:08   JOE
 * 
 * Changed all occurrences of ENTRY to CGENTRY to resolve conflict on
 * some compilers.
 * 
 *    Rev 1.7   21 Apr 1994 15:43:12   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.6   22 Feb 1994 16:31:36   JOE
 * 
 * In MEMalloc() and MEMfree(), if running on _AM29K system under UNIX,
 * must 8-byte align due to Type 1 structures containing "double" fields.
 * 
 *    Rev 1.5   03 Aug 1993 14:44:50   MAIB
 * Changed interface to resolve naming conflicts
 * 
 *    Rev 1.4   02 Aug 1993 14:21:00   MAIB
 * Changes to support multi-tasking
 * 
 *    Rev 1.3   12 Feb 1993 14:15:10   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   22 Dec 1992 13:38:38   ROB
 * Added ANSI function declarations based on ANSI_DEFS. Added prototypes.
 * 
 *    Rev 1.1   11 Dec 1992 15:29:30   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 15:25:56   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/*  intmem.c  */
/*
 *
 * History
 *
 *    04-Sep-91  jfd   Internal memory management routines extraced from
 *                     mem.c.
 *    29-Sep-91  awr   Start MEMinstFree() at page 1 instead of page 0
 *                     Simplified CGIFdefund() for buffer memory.
 *    26-Nov-91  jfd   In CGIFdefund(), conditionally declared 
 *                     "status" and "reinit_cache" based on CACHE
 *                     to get rid of compiler warnings.
 *    27-Feb-92  jfd   Moved "#if INT_MEM_MGT" statement to top of file.
 *    03-Apr-92  rs    Portability cleanup (see port.h).
 *    16 Apr 92  ss    Moved hfontCur, pfontCur defs out of CGIFmove_block()
 *                     to make the available to CGIFdefund().
 *                     In CGIFdefund(), added call to CGIFfont() to reset
 *                     fontcontext info after it got wiped out by CACdefund().
 *    02-Sep-92  jfd   In CGIFdefund(), before calling IXinit(), save
 *                     current symbol set number from "symbolset" structure
 *                     and restore it after call to IXinit() because IXinit()
 *                     calls SYMinit() which zeroes out this value.
 *                     Included "sym.h".
 *    22-Dec-92  rs    Add ANSI function declarations based upon ANSI_DEFS.
 *                     Add function prototypes.
 *    08-Feb-93  jfd   VXWorks support.
 *    02-Aug-93 maib Changed interface to prevent CGIFxxx() routines from 
 *		     calling other CGIFxxx() routines, required for 
 *		     multi-tasking. also added multicaller support code to 
 *		     CGIFxxx() routines
 *    03-Aug-93 maib changed definitions for cgifXXXX() routines to resolve
 *		     naming conflicts
 *    17-Feb-94  jfd   In MEMalloc() and MEMfree(), if running on _AM29K
 *                     system under UNIX, must 8-byte align due to PS
 *                     structures containing "double" fields.
 *    03-Jun-94  jfd   Changed all occurrences of ENTRY to CGENTRY to resolve
 *                     conflict on some compilers.
 *    09-Aug-94  sbm   Added MEMgethandlesize routine, and modified MEMalloc
 *                     so that it returns the amount of memory requested, not more.
 *    28-Sep-94  sbm   Removed the above mentioned modification to MEMalloc,
 *                     it was causing a fragmentation of memory.
 *    09-Jan-95  rob   Modify internal memory mgt to support > 64KB blocks.
 */
#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"

#if INT_MEM_MGT  /*  internal memory manager  */

#include "port.h"
#include "debug.h"
#include "ifmem.h"                     /* 17-Feb-91 awr */
#include "cgif.h"

#include "sym.h"                       /* 02-Sep-92 jfd */

#define MEM_TRACE    0      /* set to 1 to trace MEMalloc() & MEMfree() */
#define MEM_DUMP     0      /* set to 1 to trace memory alloc and free  */


#if defined (LINT_ARGS)
#if MULTICALLER
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFfund (UW16, UW16, HPSB8, UL32, LPUW16);
UW16 CGENTRY CGIFmove_block (UW16, UW16, HPSB8);
#else
UW16 CGENTRY CGIFfund (UW16, UW16, LPSB8, UL32, LPUW16);
UW16 CGENTRY CGIFmove_block (UW16, UW16, LPSB8);
#endif

#if DEFUND
GLOBAL UW16 cgif_Defund (UW16, UW16);
UW16 CGENTRY CGIFdefund (UW16, UW16);
GLOBAL VOID MEMinstFree (UW16);
#endif /* DEFUND */

#else /* not MULTICALLER */
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFfund (UW16, HPSB8, UL32, LPUW16);
UW16 CGENTRY CGIFmove_block (UW16, HPSB8);
#else
UW16 CGENTRY CGIFfund (UW16, LPSB8, UL32, LPUW16);
UW16 CGENTRY CGIFmove_block (UW16, LPSB8);
#endif

#if DEFUND
GLOBAL UW16 cgif_Defund (UW16);
UW16 CGENTRY CGIFdefund (UW16);
#endif /* DEFUND */
#endif /* MULTICALLER */

#if (HUGE_PTR_SUPPORT)
GLOBAL HPSB8 hMEMptr (MEM_HANDLE);
#else
GLOBAL LPSB8 MEMptr (MEM_HANDLE);
#endif
GLOBAL VOID MEMinit (VOID);
GLOBAL MEM_HANDLE MEMalloc (UW16, SL32);
GLOBAL VOID MEMfree (UW16, MEM_HANDLE);
GLOBAL SL32 MEMgethandlesize(UW16,MEM_HANDLE);
#if MEM_DUMP
GLOBAL VOID MEMdump_init (VOID);
GLOBAL VOID MEMdump (VOID);
EXTERN SL32 CACdump (UW16);
EXTERN SL32 BUFdump(UW16);
#endif /* MEM_DUMP */
#if MEM_TRACE
GLOBAL VOID MEMstat (UW16);
#endif /* MEM_TRACE */

#else /* no LINT_ARGS */
UW16 CGENTRY CGIFfund ();
UW16 CGENTRY CGIFmove_block ();

#if DEFUND
GLOBAL UW16 cgif_Defund ();
UW16 CGENTRY CGIFdefund ();
#if MULTICALLER
GLOBAL VOID MEMinstFree ();
#endif /* MULTICALLER */
#endif /* DEFUND */

#if (HUGE_PTR_SUPPORT)
GLOBAL HPSB8 hMEMptr ();
#else
GLOBAL LPSB8 MEMptr ();
#endif
GLOBAL VOID MEMinit ();
GLOBAL MEM_HANDLE MEMalloc ();
GLOBAL VOID MEMfree ();
#if MEM_DUMP
GLOBAL VOID MEMdump_init ();
GLOBAL VOID MEMdump ();
EXTERN SL32 CACdump ();
EXTERN SL32 BUFdump(UW16);
#endif /* MEM_DUMP */
#if MEM_TRACE
GLOBAL VOID MEMstat ();
#endif /* MEM_TRACE */

#endif /* LINT_ARGS */



/*
The memory manager must handle two memory pools. The pools are referenced
by the defines in cgif.h:

        #define CACHE_POOL   0
        #define BUFFER_POOL  1

Each pool is funded with a number of bytes of memory stored in mem_fund[].
During operation, some of this memory will be given to clients via
MEMalloc() to be later returned via MEMfree(). The number of bytes that
are still availble is stored in mem_avail[]. Thus calls to MEMalloc()
decrease mem_avail[] while calls to MEMfree() increase mem_avail[].
If the number of bytes requested through MEMalloc() is more than
mem_avail[], the request fails and NIL_MH is returned.

*/

#define NP    2                /* number of memory pools      */
GLOBAL SL32   mem_fund[NP];    /* Total memory in each pool   */
MLOCAL SL32   mem_avail[NP];   /* Available memory in pool    */


/*
The memory management option supported below is compiled depending on the
setting in cgconfig.h of INT_MEM_MGT.

The internal memory manager directly provides dynamic memory allocation.

Either the external or internal memory manager may by used. The external
memory manager may be modified to use other external allocation functions.

The two memory managers differ in the way that their pool memory is
funded. The internal manager is funded via a function CGIFfund().
The external manager's fund sizes are simply set in MEMinit().
*/


/*--------------------------------------------------------------------*/
/*       I n t e r n a l    M e m o r y    m a n a g e r              */

#include "cache.h"
#include "ix.h"


#define FRAG_MIN    64
#define NH           9      /* number of handles */

typedef struct
{
    SL32        size;    /* size of free block */
    MEM_HANDLE  next;    /* next free block */
} MEMHEAD;
typedef MEMHEAD FARPTR * PMEMHEAD;
MLOCAL MEM_HANDLE avail[NP];    /* circular linked list of free mem blocks */


typedef struct
{
    MEMHEAD  nullblock[NP];    /* empty mem blocks for initialization      */
} PAGE0;

MLOCAL PAGE0 page0;


typedef struct
{
    UW16   pool;        /* memory pool that block belongs to */
    SL32    size;        /*  size of block given to us thru CGIFfund() */
#if (HUGE_PTR_SUPPORT)
    HPSB8   ptr;         /* current address of this block */
#else
    LPSB8   ptr;         /* current address of this block */
#endif
#if MULTICALLER
    UW16   id;          /* ID of calling process returned by CGIFinit() */
#endif
} MEM;
typedef MEM FARPTR * PMEM;

MLOCAL MEM mem[NH];

#if DEFUND
EXTERN SYMBOLSET symbolset;
#endif

#if MULTICALLER
EXTERN  SB8  *ccBuffer;           /* buffer pointer for comp-char buffer */
EXTERN  UW16  uLastCallerId;              /* an initial value */
#endif  

/*----------------*/
/*   MEMptr       */
/*----------------*/
/*  Convert a memory handle to a pointer */

#if defined (ANSI_DEFS)
#if (HUGE_PTR_SUPPORT)
GLOBAL HPSB8 hMEMptr (MEM_HANDLE mem_handle)
#else
GLOBAL LPSB8 MEMptr (MEM_HANDLE mem_handle)
#endif
#else
#if (HUGE_PTR_SUPPORT)
GLOBAL HPSB8 hMEMptr(mem_handle)
#else
GLOBAL LPSB8 MEMptr(mem_handle)
#endif
    MEM_HANDLE mem_handle;
#endif /* ANSI_DEFS */
{
    return mem[PAGE(mem_handle)].ptr + OFF(mem_handle);
}




/*----------------*/
/*   MEMinit      */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID MEMinit ()
#else
GLOBAL VOID MEMinit ()
#endif /* ANSI_DEFS */
{
    UW16 i, j;
    PMEM  mp;
    MEM_HANDLE handle;

    for(i=0, handle = 0; i<NP; i++, handle += sizeof(MEMHEAD))
    {

      /* Initialize nullblock. This is the start of a circular linked
       * list of available memory blocks.
       */

        page0.nullblock[i].size   = 0L;        /* I'm nothing but overhead */
        page0.nullblock[i].next   = handle;    /* point to me              */

        avail[i]      = handle;          /* point to nullblock       */
        mem_avail[i]  = 0L;
        mem_fund[i]   = 0L;
    }

  /* Initialize page table. */

    mp = mem;

  /* Set page 0 */

    mp->size = sizeof(PAGE0);
#if (HUGE_PTR_SUPPORT)
    mp->ptr  = (HPSB8)&page0;
#else
    mp->ptr  = (LPSB8)&page0;
#endif
       
  /*  Zero out all other block sizes. A zero block size means
   *  the handle is not in use.
   */

    for(j=1; j<NH; j++)
    {
        mp++;
        mp->size = 0L;
    }
}






/*----------------*/
/*  MEMalloc      */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL MEM_HANDLE MEMalloc (UW16 i, SL32 size)
#else
GLOBAL MEM_HANDLE MEMalloc (i, size)
    UW16 i;
    SL32  size;
#endif /* ANSI_DEFS */
{
    PMEMHEAD    prev, p;
    MEM_HANDLE hprev, hp;
    SL32      frag_size;

#if MEM_TRACE
    DBG2("\n\nMEMmalloc(%u, %ld)\n", i, size);
#endif

#if defined(_AM29K) && defined(UNIX)

  /*  Make room for "size" and round up for 8-byte alignment. This is
   *  necessary for _AM29K systems running under UNIX. - jfd 2-17-94
   */

    size += ( sizeof( SL32 ) + sizeof( SL32 ) + (MEM_ALIGN + 4));  /* MEMHEAD.size is long -ss */
    size &= ~(MEM_ALIGN + 4);
#else

  /*  Make room for "size" and round up for alignment */

    size += sizeof( SL32 ) + MEM_ALIGN;    /* MEMHEAD.size is long -ss */
    size &= ~MEM_ALIGN;
#endif
    size = MAX(size, (SL32)sizeof(MEMHEAD)); /* 9-apr-91 dET minimum size */

#if MEM_TRACE
    DBG1("    adjusted size %ld\n", size);
#endif

  /* First fit search for big enough block */

    hprev = avail[i];
#if (HUGE_PTR_SUPPORT)
    prev  = (PMEMHEAD)hMEMptr(hprev);
    hp    = prev->next;
    p     = (PMEMHEAD)hMEMptr(hp);
#else
    prev  = (PMEMHEAD)MEMptr(hprev);
    hp    = prev->next;
    p     = (PMEMHEAD)MEMptr(hp);
#endif
    for( ; ; )
    {
        if(p->size >= size) break;
        if(hp == avail[i])
            return NIL_MH;

        hprev = hp;
        prev  = p;
        hp    = p->next;
#if (HUGE_PTR_SUPPORT)
        p     = (PMEMHEAD)hMEMptr(hp);
#else
        p     = (PMEMHEAD)MEMptr(hp);
#endif
    }

  /* If p is large enough, cut off the end and return it */

    frag_size = p->size - size;

    if(frag_size >= FRAG_MIN)
    {
        p->size = frag_size;                   /* shorten p                */
        hp = hp + frag_size;                   /* new handle for tail      */
#if (HUGE_PTR_SUPPORT)
        p = (PMEMHEAD)((HPSB8)p + frag_size);  /* point p at tail of block */
#else
        p = (PMEMHEAD)((LPSB8)p + frag_size);  /* point p at tail of block */
#endif
        p->size = size;
    }
    else
        prev->next = p->next;                  /* unlink p */

    avail[i] = hprev;

    mem_avail[i] -= p->size;

#if MEM_TRACE
    DBG("-------after------------\n");
    MEMstat(i);
    DBG1("    %lx\n", hp + sizeof(SL32));
#endif

#if defined(_AM29K) && defined(UNIX)
    return hp + sizeof(SL32) + sizeof(SL32);  /* skip over ("size" + 4) -jfd 2-18-94 */
#else
    return hp + sizeof(SL32);    /* skip over "size"  - was '+ 4' -ss  */
#endif
}




/*----------------*/
/*  MEMfree       */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID MEMfree (UW16 i, MEM_HANDLE hm)
#else
GLOBAL VOID MEMfree (i, hm)
    UW16 i;
    MEM_HANDLE hm;
#endif /* ANSI_DEFS */
{
    PMEMHEAD    p,  prev,  next;
    MEM_HANDLE  hp, hprev, hnext, t;


#if MEM_TRACE
    DBG2("MEMfree(%u, %lx)  ", i, hm);
#endif

    if(!hm) return;

#if defined(_AM29K) && defined(UNIX)
    hp = hm - (sizeof(SL32) + sizeof(SL32)); /* back up over "size + 4"  -jfd 2-18-94 */
#else
    hp = hm - sizeof(SL32);      /* back up over "size"  - was '- 4'  -ss  */
#endif
#if (HUGE_PTR_SUPPORT)
    p = (PMEMHEAD)hMEMptr(hp);
#else
    p = (PMEMHEAD)MEMptr(hp);
#endif

    mem_avail[i] += p->size;                    /* checkbook deposit       */

#if MEM_TRACE
    DBG1("    size %ld\n", p->size);
#endif


  /* Find the right place to insert p in avail ring */

    hprev = avail[i];
#if (HUGE_PTR_SUPPORT)
    prev  = (PMEMHEAD)hMEMptr(hprev);
    hnext = prev->next;
    next  = (PMEMHEAD)hMEMptr(hnext);
#else
    prev  = (PMEMHEAD)MEMptr(hprev);
    hnext = prev->next;
    next  = (PMEMHEAD)MEMptr(hnext);
#endif
    while(  !(hprev < hp && hp < hnext)  )  /*  stop when hp is between
                                             *  hprev and hnext
                                             */
    {
      /*  Also stop when hprev is the largest address and hnext is
       *  the smallest address and hp is the new largest or the
       *  new smallest
       */

        if (hprev >= hnext && (hp > hprev || hp < hnext) ) break;
        hprev = hnext;
        prev  = next;
        hnext = next->next;
#if (HUGE_PTR_SUPPORT)
        next  = (PMEMHEAD)hMEMptr(hnext);
#else
        next  = (PMEMHEAD)MEMptr(hnext);
#endif
    }

  /*  p should now be linked between prev and next. Here we check
   *  whether any or all of the three blocks need to be merged.
   */

    t = hp + p->size;             /* handle of block just after p     */
    if(t == hnext && PAGE(hp) == PAGE(hnext))
    {
      /* merge p and next */

        p->size += next->size;
        p->next  = next->next;
    }
    else
        p->next = hnext;   /* link p before next */

    t = hprev + prev->size;
    if(t == hp && PAGE(hprev) == PAGE(hp))
    {
      /* merge prev and p */

        prev->size += p->size;
        prev->next  = p->next;
    }
    else
        prev->next = hp;

    avail[i] = hprev;

#if MEM_TRACE
    DBG("-------after------------\n");
    MEMstat(i);
#endif
}

/*----------------*/
/* MEMgethandlesize */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL SL32 MEMgethandlesize (UW16 i, MEM_HANDLE handle)
#else
GLOBAL SL32 MEMgethandlesize (i, handle)
		UW16 i;
		MEM_HANDLE handle;
#endif /* ANSI_DEFS */
{
    PMEMHEAD    prev, p;
		MEM_HANDLE  hprev, hp;
		SL32        size;

	/* Initialize size */
		size = 0;

		if(!handle) return size;

    hp = handle - sizeof(SL32);      /* back up over "size"  - was '- 4'  -ss  */
#if (HUGE_PTR_SUPPORT)
    p = (PMEMHEAD)hMEMptr(hp);
#else
    p = (PMEMHEAD)MEMptr(hp);
#endif

		size = p->size - sizeof(MEM_HANDLE);   /* subtract size of handle itself */

#if MEM_TRACE
		DBG2("\n\nMEMgethandlesize(%u, %ld)\n", i, size);
#endif


		return size;
}




/*----------------*/
/*  CGIFfund      */
/*----------------*/

#if MULTICALLER
#if defined (ANSI_DEFS)
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFfund (UW16 uCallerId, UW16 pool, HPSB8 memptr, UL32 size,
                        LPUW16 handle)
#else
UW16 CGENTRY CGIFfund (UW16 uCallerId, UW16 pool, LPSB8 memptr, UL32 size,
                        LPUW16 handle)
#endif
#else
UW16 CGENTRY CGIFfund (uCallerId, pool, memptr, size, handle)
    UW16  uCallerId;
    UW16  pool;
#if (HUGE_PTR_SUPPORT)
    HPSB8  memptr;
#else
    LPSB8  memptr;
#endif
    UL32  size;
    LPUW16 handle;
#endif /* ANSI_DEFS */
#else /* not MULTICALLER */
#if defined (ANSI_DEFS)
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFfund (UW16 pool, HPSB8 memptr, UL32 size, LPUW16 handle)
#else
UW16 CGENTRY CGIFfund (UW16 pool, LPSB8 memptr, UL32 size, LPUW16 handle)
#endif
#else
UW16 CGENTRY CGIFfund (pool, memptr, size, handle)
    UW16  pool;
#if (HUGE_PTR_SUPPORT)
    HPSB8  memptr;
#else
    LPSB8  memptr;
#endif
    UL32  size;
    LPUW16 handle;
#endif /* ANSI_DEFS */
#endif /* MULTICALLER */
{
    UW16 status;
    SL32   lsize;   /* internal sizes are SL32  */
    PMEM   mp;
    UW16  h;

#if MULTICALLER
#if SEM_ACCESS
  if (status = acquire_semaphore())
    return status;
#endif /* SEM_ACCESS */

    if(uCallerId != uLastCallerId)          /* use correct fontcontext */
    {
      status = INSTsetState(uCallerId);
      if (status)
      {
#if SEM_ACCESS
        release_semaphore();
#endif /* SEM_ACCESS */
        return status;
      }
    }
#endif /* MULTICALLER */

    if(pool > BUFFER_POOL) 
    {
#if (MULTICALLER && SEM_ACCESS)
      release_semaphore();
#endif /* (MULTICALLER && SEM_ACCESS) */
      return ERR_bad_pool;
    }

  /* Take care of alignment. */

    lsize = (SL32)size & ~MEM_ALIGN;  /* round size down to aligned value */

    /* get a page handle */
    /* Find first available handle */
    mp = mem;               /* get beginning of handle array       */
    for( h = 0; h < NH; h++ )
    {
        if ( !mp->size )    /* Look for an empty slot (size == 0)   */
        {
          /* fill in size and memory pointer */
          mp->pool = pool;
          mp->size = lsize;
          mp->ptr  = memptr;
#if MULTICALLER
          mp->id   = uCallerId;
#endif
          break;
        }
        mp++;
    }

    if(h >= NH)
    {
#if (MULTICALLER && SEM_ACCESS)
      release_semaphore();
#endif /* (MULTICALLER && SEM_ACCESS) */
      return ERR_out_of_handles;
    }

    *handle = h;

  /* Free the page */

    ((PMEMHEAD)memptr)->size = lsize;
                                              /*  vvvv was '+ 4'  -ss  */
    MEMfree(pool, ((UL32)h << OFFBITS) + sizeof(SL32));
    mem_fund[pool] += lsize;

#if (MULTICALLER && SEM_ACCESS)
    release_semaphore();
#endif /* (MULTICALLER && SEM_ACCESS) */
    return SUCCESS;
}


#if CACHE
EXTERN HIFFONT    hfontCur;           /* handle of current FONT         */
EXTERN PFONT      pfontCur;           /* ptr to current FONT            */
#endif


/*----------------*/
/* CGIFmove_block */
/*----------------*/
#if MULTICALLER
#if defined (ANSI_DEFS)
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFmove_block (UW16 uCallerId, UW16 page, HPSB8 memptr)
#else
UW16 CGENTRY CGIFmove_block (UW16 uCallerId, UW16 page, LPSB8 memptr)
#endif
#else
UW16 CGENTRY CGIFmove_block (uCallerId, page, memptr)
    UW16  uCallerId;
    UW16 page;
#if (HUGE_PTR_SUPPORT)
    HPSB8 memptr;
#else
    LPSB8 memptr;
#endif
#endif /* ANSI_DEFS */
#else /* not MULTICALLER */
#if defined (ANSI_DEFS)
#if (HUGE_PTR_SUPPORT)
UW16 CGENTRY CGIFmove_block (UW16 page, HPSB8 memptr)
#else
UW16 CGENTRY CGIFmove_block (UW16 page, LPSB8 memptr)
#endif
#else
UW16 CGENTRY CGIFmove_block (page, memptr)
    UW16 page;
#if (HUGE_PTR_SUPPORT)
    HPSB8 memptr;
#else
    LPSB8 memptr;
#endif
#endif /* ANSI_DEFS */
#endif /* MULTICALLER */
{
    UW16 status;
    PMEM  mp;

#if MULTICALLER
#if SEM_ACCESS
  if (status = acquire_semaphore())
    return status;
#endif /* SEM_ACCESS */

    if(uCallerId != uLastCallerId)          /* use correct fontcontext */
    {
      status = INSTsetState(uCallerId);
      if (status)
      {
#if SEM_ACCESS
        release_semaphore();
#endif /* SEM_ACCESS */
        return status;
      }
    }
#endif /* MULTICALLER */

    if(page > NH) return ERR_bad_handle;
    mp = mem + page;
    if(!mp->size) return ERR_bad_handle;    /* page isn't active */
    mp->ptr = memptr;

#if CACHE
  /*  See if the page that moved contains the current FONT. If it does,
   *  recover the pointer now. This is done here so that we don't have
   *  to recover the FONT pointer everytine CGIFchar() is called. This
   *  makes a noticable difference in the cache speed.
   */
    {
         if(PAGE(hfontCur) == page)
#if (HUGE_PTR_SUPPORT)
           pfontCur = (PFONT)hMEMptr(hfontCur);
#else
           pfontCur = (PFONT)MEMptr(hfontCur);
#endif
    }
#endif

#if (MULTICALLER && SEM_ACCESS)
    release_semaphore();
#endif /* SEM_ACCESS */
    return SUCCESS;
}


/*---------------------------------------------------------------------*/
/*                           D E F U N D                               */
#if DEFUND

/*----------------*/
/*  cgif_Defund    */
/*----------------*/

#if MULTICALLER
#if defined (ANSI_DEFS)
GLOBAL UW16 cgif_Defund (UW16 id, UW16 page)
#else
GLOBAL UW16 cgif_Defund (id, page)
    UW16  id;
    UW16  page;
#endif /* ANSI_DEFS */
#else /* not MULTICALLER */
#if defined (ANSI_DEFS)
GLOBAL UW16 cgif_Defund (UW16 page)
#else
GLOBAL UW16 cgif_Defund (page)
    UW16  page;
#endif /* ANSI_DEFS */
#endif /* MULTICALLER */
{
    PMEM          mp;
    SL32          size;
    MEM_HANDLE    hp, hprev;
    PMEMHEAD      p, prev;
    UW16         pool;
#if CACHE
    BOOLEAN       reinit_cache;
    PFONT         sav_font;
#endif
#if DEFUND
    UW16         sav_curssnum;
#endif
    UW16         status;

    if(page > NH)        return ERR_bad_handle;

    mp = mem + page;
    if(!(size = mp->size)) return ERR_bad_handle;    /* page isn't active */
#if MULTICALLER
    if(mp->id != id) return ERR_bad_handle;        /* id doesn't own page */
#endif
    pool = mp->pool;

  /*================ C A C H E _ P O O L ======================*/
#if CACHE
    if(pool == CACHE_POOL)
    {
        sav_font = pfontCur;   /* save copy to get at FC later -ss 4/16/92 */
      /* Free FONTs and BITMAPs using this page */
        reinit_cache = CACdefund(page);
    }
    else
#endif    

  /*================ B U F F E R _ P O O L ======================*/

    if(pool == BUFFER_POOL)
        IXexit();

  /*================ Remove memory block from avail ring =============*/

    hprev = avail[pool];
#if (HUGE_PTR_SUPPORT)
    prev  = (PMEMHEAD)hMEMptr(hprev);
    hp    = prev->next;
    p     = (PMEMHEAD)hMEMptr(hp);
#else
    prev  = (PMEMHEAD)MEMptr(hprev);
    hp    = prev->next;
    p     = (PMEMHEAD)MEMptr(hp);
#endif
    for( ; ; )
    {
        if(PAGE(hp) == page) break;
        if(hp == avail[pool])
        {
            DBG("    failed\n");
            return ERR_missing_block;
        }
        hprev = hp;
        prev  = p;
        hp    = p->next;
#if (HUGE_PTR_SUPPORT)
        p     = (PMEMHEAD)hMEMptr(hp);
#else
        p     = (PMEMHEAD)MEMptr(hp);
#endif
    }
    avail[pool] = hprev;

  /* Check that it's all there */

    if(p->size != size)
        return ERR_lost_mem;
    prev->next = p->next;                    /* unlink p */
    mem_avail[pool] -= size;

    mp->size = 0L;                        /* this page is now available */

  /*================ C A C H E _ P O O L ======================*/
#if CACHE
    if(pool == CACHE_POOL)
    {
        if(reinit_cache)
        {
            status = CACinit();
            if (status)
                return status;
            if( sav_font )   /* if have valid pointer reinit fontcontext. */
            {                /*                              -ss 4/16/92  */
               status = cgif_Font( &sav_font->fc );
               if (status)
               {
                   DBG1("(Reinit cache) cgif_Font error %d\n", status);
                   return status;
               }
            }  /* sav_font */
        }  /* reinit */
    }
    else
#endif
  /*================ B U F F E R _ P O O L ======================*/
    if(pool == BUFFER_POOL)
    {
        sav_curssnum = symbolset.curssnum; /* save ss number */
        status = IXinit();
        if (status)
           return (status);
        symbolset.curssnum = sav_curssnum; /* restore ss number */
    }

    return SUCCESS;  /* to keep a compiler quiet */
}

/*----------------*/
/*  CGIFdefund    */
/*----------------*/

#if MULTICALLER
#if defined (ANSI_DEFS)
UW16 CGENTRY CGIFdefund (UW16 uCallerId, UW16 page)
#else
UW16 CGENTRY CGIFdefund (uCallerId, page)
    UW16  uCallerId;
    UW16  page;
#endif /* ANSI_DEFS */
#else /* not MULTICALLER */
#if defined (ANSI_DEFS)
UW16 CGENTRY CGIFdefund (UW16 page)
#else
UW16 CGENTRY CGIFdefund (page)
    UW16  page;
#endif /* ANSI_DEFS */
#endif /* MULTICALLER */
{
  UW16 status;

#if MULTICALLER
#if SEM_ACCESS
  if (status = acquire_semaphore())
    return status;
#endif /* SEM_ACCESS */

    DBG2("CGIFdefund(uCallerId == %d, page == %d)\n", uCallerId, page);
    if(uCallerId != uLastCallerId)          /* use correct fontcontext */
    {
      status = INSTsetState(uCallerId);
      if (status)
      {
#if SEM_ACCESS
        release_semaphore();
#endif /* SEM_ACCESS */
        return status;
      }
    }
    status = cgif_Defund(uCallerId, page);
#if SEM_ACCESS
    release_semaphore();
#endif /* SEM_ACCESS */
    return status;
#else /* !MULTICALLER */
    DBG1("CGIFdefund(page == %d)\n", page);
    return cgif_Defund(page);
#endif /* MULTICALLER */
}


#if MULTICALLER
/*----------------*/
/*  MEMinstFree   */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID MEMinstFree (UW16 id)
#else
GLOBAL VOID MEMinstFree(id)
    UW16 id;
#endif /* ANSI_DEFS */
{
    PMEM  pMem = mem;
    UW16 page;

    for(page=1;page<NH;page++)
    {
        pMem++;
        if((pMem->id == id) && pMem->size)
            cgif_Defund(id, page);
    }
}
#endif /* MULTICALLER */


#endif   /* DEFUND */
/*                           D E F U N D                               */
/*---------------------------------------------------------------------*/




#if MEM_DUMP

MLOCAL SL32 cache_tot;
MLOCAL SL32 buf_tot;

#if defined (ANSI_DEFS)
GLOBAL VOID MEMdump_init ()
#else
GLOBAL VOID MEMdump_init ()
#endif /* ANSI_DEFS */
{
    cache_tot = mem_avail[CACHE_POOL];
    buf_tot   = mem_avail[BUFFER_POOL];
}

/*----------------*/
/*  MEMdump       */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID MEMdump ()
#else
GLOBAL VOID MEMdump ()
#endif /* ANSI_DEFS */
{
    MEM_HANDLE hp;
    PMEMHEAD  p;

    SL32 cacTotal = 0L;
    SL32 bufTotal = 0L;
    static SW16 count=0;

    hp = page0.nullblock[0].next;
#if (HUGE_PTR_SUPPORT)
    p  = (PMEMHEAD)hMEMptr(hp);
#else
    p  = (PMEMHEAD)MEMptr(hp);
#endif

    DBG1("\n\t\t\tCACHE POOL %d\n", ++count);

    while(p->size)
    {
      cacTotal += p->size;
      hp = p->next;
#if (HUGE_PTR_SUPPORT)
      p  = (PMEMHEAD)hMEMptr(hp);
#else
      p  = (PMEMHEAD)MEMptr(hp);
#endif
    }

    cacTotal += CACdump(0);

    if(cacTotal != cache_tot)
    {
    
        DBG2("Missing memory: %ld out of %ld\n", cacTotal, mem_fund[0]);
        while(p->size)
        {
          cacTotal += p->size;
          DBG4("Handle %6lx  Pointer %Fp  Size %5ld  Next %6lx\n", hp, p, p->size, p->next);
          hp = p->next;
#if (HUGE_PTR_SUPPORT)
          p  = (PMEMHEAD)hMEMptr(hp);
#else
          p  = (PMEMHEAD)MEMptr(hp);
#endif
        }
        CACdump(1);
    }

    hp = page0.nullblock[1].next;
#if (HUGE_PTR_SUPPORT)
    p  = (PMEMHEAD)hMEMptr(hp);
#else
    p  = (PMEMHEAD)MEMptr(hp);
#endif

    DBG1("\n\t\t\tBUFFER POOL %d\n", count);

    while(p->size)
    {
      bufTotal += p->size;
      hp = p->next;
#if (HUGE_PTR_SUPPORT)
      p  = (PMEMHEAD)hMEMptr(hp);
#else
      p  = (PMEMHEAD)MEMptr(hp);
#endif
    }

    bufTotal += BUFdump(0);

    if(bufTotal != buf_tot)
    {
    
        DBG2("Missing memory: %ld out of %ld\n", bufTotal, mem_fund[1]);
        while(p->size)
        {
          bufTotal += p->size;
          DBG4("Handle %6lx  Pointer %Fp  Size %5ld  Next %6lx\n", hp, p, p->size, p->next);
          hp = p->next;
#if (HUGE_PTR_SUPPORT)
          p  = (PMEMHEAD)hMEMptr(hp);
#else
          p  = (PMEMHEAD)MEMptr(hp);
#endif
        }
        BUFdump(1);
    }
}
#endif /* MEM_DUMP */





#if MEM_TRACE

/*----------------*/
/*  MEMstat       */
/*----------------*/
#if defined (ANSI_DEFS)
GLOBAL VOID MEMstat (UW16 i)
#else
GLOBAL VOID MEMstat (i)
    UW16  i;
#endif /* ANSI_DEFS */
{
    SL32  total, max_size;
    MEM_HANDLE hp;
    PMEMHEAD  p;


    total    = 0;
    max_size = 0;

    hp = avail[i];
#if (HUGE_PTR_SUPPORT)
    p  = (PMEMHEAD)hMEMptr(hp);
#else
    p  = (PMEMHEAD)MEMptr(hp);
#endif
    for( ; ; )
    {
        if(p)
        {
          DBG4("pointer == %Fp  handle == %Fp size==%5ld  next== %Fp\n",
                                p, hp, p->size, p->next);
        }
        else
        {
          DBG("NULL POINTER \n");
        }

        if(p->size > max_size)  max_size = p->size;
        total += p->size;

        hp = p->next;
#if (HUGE_PTR_SUPPORT)
        p  = (PMEMHEAD)hMEMptr(hp);
#else
        p  = (PMEMHEAD)MEMptr(hp);
#endif

        if(hp == avail[i])  break;
    }

    DBG1("MEMstat pool %u\n",i);
    DBG1("             Largest free block  %ld\n", max_size);
    DBG1("             Total available     %ld\n", total);
    DBG1("             Total in fund       %ld\n", mem_fund[i]);
    if(mem_avail[i] != total)
        DBG("?????????? does not compute ?????????\n");
}

#endif /* MEM_TRACE */

/*       I n t e r n a l    M e m o r y    m a n a g e r              */
/*--------------------------------------------------------------------*/

#endif   /*  INT_MEM_MGT  */
