/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/gen/makechar.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: makechar.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.9   07 Apr 1995 08:42:26   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.8   22 Apr 1994 14:08:46   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.7   26 Feb 1993 16:08:54   MIKE
 * Changed argument in app_make_character() from SW16 to UW16.
 * 
 *    Rev 1.6   12 Feb 1993 19:27:26   MIKE
 * For foreign bitmap support: in makechar(), call app_make_character()
 * 
 *    Rev 1.5   12 Feb 1993 14:22:22   JOE
 * VXWorks support.
 * 
 *    Rev 1.4   29 Jan 1993 09:16:42   JOE
 * In load_space_metrics(), save em box size for bitmaps or outlines.
 * 
 *    Rev 1.3   22 Jan 1993 09:03:14   JOE
 * Removed escapement adjustment code from makechar() and moved it back
 * to MAKifbmheader() because adjustments were not being made if CHAR_SIZE
 * was enabled.
 * 
 *    Rev 1.2   05 Jan 1993 14:02:04   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:23:36   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 16:03:10   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* makechar.c */
/*
 *
 *
 * History
 *
 *   17-Jun-91  jfd  Moved "debug.h" after "port.h".
 *                   Moved "cgconfig.h" before "port.h".
 *    6-Sep-91  awr  call to MAKifbmheader()
 *   20-Sep-91  jfd  In makechar(), conditionally declaring "bm" to get
 *                   rid of compiler warning.
 *   20-Jan-92  jfd  In makechar(), after calls to bitmap() and outdata(),
 *                   storing xorigin, yorigin and escapement into *phbm
 *                   from bm.
 *   25-Feb-92  jfd  Disabled storing of origin and escapement if processing
 *                   outline.
 *   03-Apr-92  rs   Portability cleanup (see port.h).
 *    8 Mar 92  ss   Moved lines to copy metrics returned from MAKifbmheader
 *                   to (bitmap.c)bitmap_ufst() since they are generated
 *                   correctly if IF-type chars in bitmap() already.
 *   19- Jun-92 jfd  In makechar(), if call to MAKifbmheader() returns
 *                   ERR_fixed_space() (space character not found in font),
 *                   set "num_loops" to 0.
 *                   Later, if bitmap() or outdata() returns ERR_fixed_space,
 *                   call new routine load_space_metrics() to set metric
 *                   information for space.
 *                   Added new routine "load_space_metrics()".
 *                   Included "ix.h" and "if_type.h".
 *    2 Sep 92  ss   Moved escapement adjustment code from MAKifbmheader() to
 *                   makechar(). Worked it around fixed space handling.
 *    9-Sep-92  jfd  In makechar(), corrected placement of NON_IF_FONT 
 *                   conditional compile statements.
 *   05-Jan-93  jfd  ANSI C function declaration changes.
 *   21-Jan-93  jfd  Moved escapement adjustment code back to MAKifbmheader()
 *                   (maker.c) from makechar() because adjustments were not
 *                   being made if CHAR_SIZE were enabled.
 *                   Removed function prototype for fgseg().
 *   25-Jan-93  jfd  In load_space_metrics(), save em box size if PIFBITMAP/
 *                   PIFOUTLINE structure.
 *   08-Feb-93  jfd  VXWorks support.
 *   08-Feb-93  mby  In makechar(), call app_make_character() if ALIEN_BITS
 *                   is turned on.
 *   26-Feb-93  mby  Changed arg in app_make_character() from SW16 to UW16.
 */

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"                  /* 11-20-90 jfd */
#include "port.h"
#include "debug.h"
#include "ifmem.h"                     /* 17-Feb-91 awr */
#include "cgif.h"
#include "ix.h"
#include "if_type.h"
#include "segments.h"
#include "fpmath.h"

#ifdef LINT_ARGS
GLOBAL UW16    makechar(PFONTCONTEXT, SW16, PHIFBITMAP);
EXTERN UW16    MAKifbmheader(PFONTCONTEXT, SW16, PIFBITMAP);
EXTERN UW16    bitmap(PHIFBITMAP);
EXTERN UW16    outdata(PHIFOUTLINE);
MLOCAL VOID    load_space_metrics(PFONTCONTEXT, PHIFBITMAP);
EXTERN UW16    app_make_character( PFONTCONTEXT, UW16, PHIFBITMAP );
#else
EXTERN UW16    MAKifbmheader();
EXTERN UW16    bitmap();
EXTERN UW16    outdata();
MLOCAL VOID    load_space_metrics();
EXTERN UW16    app_make_character();
#endif

EXTERN IF_STATE if_state;

/*----------------------*/
/*  load_space_metrics  */
/*----------------------*/
#if defined (ANSI_DEFS)
MLOCAL VOID  load_space_metrics(PFONTCONTEXT fc, PHIFBITMAP phbm)
#else
MLOCAL VOID
load_space_metrics(fc, phbm)
    PFONTCONTEXT fc;
    PHIFBITMAP   phbm;
#endif
{
    PIFOUTLINE   pif;
    PIFBITMAP    pbm;

    if (FC_ISBITMAP(fc))
    {
       pbm = (PIFBITMAP)MEMptr(*phbm);
       pbm->xorigin    = 0L;
       pbm->yorigin    = 0L;
       pbm->width      = 0;
       pbm->depth      = 0;
       pbm->escapement = if_state.cs.escapement;
       pbm->du_emx     = if_state.cs.du_emx;
       pbm->du_emy     = if_state.cs.du_emy;
    }
    else if (FC_ISOUTLINE(fc))
    {
       pif = (PIFOUTLINE)MEMptr(*phbm);
       pif->left       = 0;
       pif->top        = 0;
       pif->right      = 0;
       pif->bottom     = 0;
       pif->xscale     = 0;
       pif->yscale     = 0;
       pif->escapement = if_state.cs.escapement;
       pif->du_emx     = if_state.cs.du_emx;
       pif->du_emy     = if_state.cs.du_emy;
    }
    return;
}


/*-------------------*/
/*    makechar       */
/*-------------------*/
/*  Make a character bitmap or outline according to the given FONTCONTEXT */
#if defined (ANSI_DEFS)
GLOBAL UW16  makechar(PFONTCONTEXT fc, SW16 chId, PHIFBITMAP phbm)
#else
GLOBAL UW16
makechar(fc, chId, phbm)
  PFONTCONTEXT  fc;                             /* FONTCONTEXT            */
  SW16          chId;                           /* character ID           */
  PHIFBITMAP    phbm;                           /* returned bitmap handle */
#endif
{
  UW16    status;
  IFBITMAP     bm;

#if ALIEN_BITS
    if (FC_ISAPPGEN(fc))
    {
        return (app_make_character(fc, (UW16)chId, phbm));
    }
#endif  /* ALIEN_BITS */

    status = MAKifbmheader(fc, chId, &bm);
    if (status == ERR_fixed_space)
       if_state.num_loops = 0;
    else if (status)
       return status;

#if CGBITMAP
    if(FC_ISBITMAP(fc))
       status = bitmap(phbm);
    else
#endif
#if OUTLINE
    if(FC_ISOUTLINE(fc))
       status = outdata(phbm);
    else
#endif
        return ERRinvalid_outproc_type;

    if( status )
    {
       if (status == ERR_fixed_space)
          load_space_metrics(fc, phbm);
       else return status;
    }

    return status;
}
