/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/inc/fpmath.h,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: fpmath.h,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   07 Apr 1995 08:45:20   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   30 Nov 1994 18:44:54   MIKE
 * 
 * Added fpsqrt() prototype and macro; error 9007.
 * 
 *    Rev 1.3   21 Apr 1994 15:48:42   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.2   15 Jan 1993 09:35:18   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.1   14 Dec 1992 09:40:38   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:11:26   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* fpmath.h */
/*
 *
 *
 * HISTORY:
 *  29-Mar-91 awr created
 *  20-Aug-91 ss  Added prototype for fpabs() in !LINTARGS case.
 *   8-Mar-92 awr Added fpatof().
 *  03-Apr-92 rs  Portability cleanup (see port.h).
 *  28-Nov-94 mby Add fpsqrt() macro; function prototype. Add error 9007.
 */
/*        F l o a t i n g    p o i n t    m a t h                  */
/*
Objective:

The objective of this package is to provide floating point math
computation to some portions of Intellifont that require large precision
or dynamic range. These functions should not be used in time critical
areas of the software. Simetimes floating point is not available. This
package is small and can be used in those instances. If floating point is
available, an alternative version of the floating point functions use the
extarnal fp math instead.

Overview

A small, special purpose floating point math package is included. The
package is object oriented and so it's data type and math functions can
be replaced by external functions should space or performance be an issue.


The package is special purpose in that it only contains those functions
that the client software of this package needed.

The objects of this set of methods is the floating point number class. A
floating point structure is declared in port.h. A set of functions
that operate on floating point objects is declared in fpmath.h.
The functions are implemented in fpmath.c

Client software is only allowed to reference floating point objects by using

    1. macros and functions in fmath.h
    2. direct assignment of the floating point structure.

A floating point object is a c typedef'd structure named FPNUM.

Should floating point arithmetic already be available within the rest of
the client system, FPNUM may be typedef'd to whatever that data type
happens to be. Than short glue functions that call the external floating
point functions must be provided in fpmath.c to implement the
internal floating point routines.

Two versions of the floating point math package are supplied. They are
called the "internal" and "external" floating point math packages. The
internal package implements the floating point functions using integer
arithmetic. The external package defines an FPNUM as the c "double"
and uses the compiler to generate calls to either the floating
c runtime floating point emulator or hardware functions.

Either version may be conditioanly compiled by setting the define
in cgconfig.h

    #define INT_FP    1     0 external, 1 internal

The functions store a return status:

    fpmath_error is set to 0 for normal returns
                           error otherwise:

Error Returns are
*/

EXTERN UW16  fpmath_error;          /* Non zero for errors */
#define FPconversion         9000
#define FPadd_overflow       9001
#define FPmultiply_overflow  9003
#define FPdivide_by_zero     9005
#define FPdivide_overflow    9006
#define FPsqrt_neg           9007


#if INT_FP

/*        Internal Floating Point Object Definition

An FPNUM is a typedef'd structure that represents a floating point number
defined in port.h as

           typedef struct
           {
               SL32 n;     normalized, signed mantissa 
               SW16 e;     exponent                    
           } FPNUM;

The interpretation of an FPNUM is that
The mantissa represents a signed
32 bit integer. The integer is considered fixed point with 30 bits after
the binary point. Thus

       -2 < mantissa < 2;
The exponent is a signed integer representing the power of 2 that the
mantissa should by multiplied by. Thus The value of an FPNUM is

    value = mantissa * (2**exponent)
The range of an FPNUM is determined by the limits imposed on the exponent

    MIN_EXPONENT <= exponent <= MAX_EXPONENT

where:
*/

#define MIN_EXPONENT   -64
#define MAX_EXPONENT    64

/* The functions in fpmath.c provide integer and FPNUM conversion,
 * arithmetic functions and print function.
 *
 */


        /*  Macro definitions  */


#define fpiszero(x)   (!x.n)             /* TRUE if x is 0         */
#define fpisneg(x)    (x.n<0)            /* TRUE if x < 0          */
#define fpint2fp(x)   fplongexp(x, 0)    /* Convert SL32 to FPNUM  */
#define fpfixed2fp(x) fplongexp(x, 16)   /*         FIXED to FPNUM */

        /*  Function definitions  */

#ifdef LINTARGS
EXTERN FPNUM   fplongexp(SL32, SW16); /* Convert SL32 to FPNUM  */
EXTERN SW16    fp2word(FPNUM);        /*         FPNUM to SW16  */
EXTERN SL32    fp2long(FPNUM);        /*         FPNUM to SL32  */
EXTERN FPNUM   fpatof(LPSB8);         /*         ASCII to FPNUM */
EXTERN SW16    fp2intel(FPNUM, LPSW16);/* Convert to Intellifont */
EXTERN FPNUM   fpneg(FPNUM);          /* Negate FPNUMs          */
EXTERN FPNUM   fpadd(FPNUM, FPNUM);   /* Add                    */
EXTERN FPNUM   fpsub(FPNUM, FPNUM);   /* Subtract               */
EXTERN FPNUM   fpmul(FPNUM, FPNUM);   /* Multiply               */
EXTERN FPNUM   fpdiv(FPNUM, FPNUM);   /* Divide                 */
EXTERN BOOLEAN fplt(FPNUM, FPNUM);    /* Compare Less Than      */
EXTERN FPNUM   fpabs(FPNUM);          /* Absolute value         */
EXTERN SW16    ifpmul(SW16, FPNUM);   /* Multiply SW16 * FPNUM  */
EXTERN FPNUM   fpsqrt(FPNUM);         /* sqrt function          */
#ifdef DEBUG
EXTERN VOID    fpprint(FPNUM);        /* print to stdout fpnum  */
#endif
#else
EXTERN FPNUM   fplongexp();
EXTERN SW16    fp2word();
EXTERN SL32    fp2long();
EXTERN FPNUM   fpatof();
EXTERN SW16    fp2intel();
EXTERN FPNUM   fpneg();
EXTERN FPNUM   fpadd();
EXTERN FPNUM   fpsub();
EXTERN FPNUM   fpmul();
EXTERN FPNUM   fpdiv();
EXTERN BOOLEAN fplt();
EXTERN FPNUM   fpabs();
EXTERN SW16    ifpmul();
EXTERN FPNUM   fpsqrt();
#ifdef DEBUG
EXTERN VOID    fpprint();
#endif
#endif

#else   /*  not INT_FP   */

/*  Define an FPNUM to be a double in port.h  */


        /*  Macro definitions  */

#define fpiszero(x)   ((x) == 0.0)              /* TRUE if x is 0         */
#define fpisneg(x)    ((x) < 0.0)               /* TRUE if x < 0.0        */
#define fpint2fp(x)   ((FPNUM)(x))              /* Convert SL32 to FPNUM  */
#define fpfixed2fp(x) ((FPNUM)(x)/65536.0)      /*         FIXED to FPNUM */
#define fpatof(x)     (atof(x))                 /*         ASCII to FPNUM */
#define fpneg(x)      (-(x))                    /* Negate FPNUMs          */
#define fpadd(x, y)   ((x)+(y))                 /* Add                    */
#define fpsub(x, y)   ((x)-(y))                 /* Subtract               */
#define fpmul(x, y)   ((x)*(y))                 /* Multiply               */
#define fplt(x, y)    ((x)<(y))                 /* Compare Less Than      */
#define fpabs(x)      (((x)<0.0) ? (-(x)):(x))  /* Absolute value         */
#define ifpmul(x, y)  fp2word((FPNUM)x*y)       /* Multiply SW16 * FPNUM  */
#define fpsqrt(x)     sqrt(x)                   /* sqrt(FPNUM) [math.h]   */

        /*  Function definitions  */

#ifdef LINTARGS
EXTERN FPNUM   fplongexp(SL32, SW16); /* Convert SL32 to FPNUM  */
EXTERN SW16    fp2word(FPNUM);        /*         FPNUM to SW16  */
EXTERN SL32    fp2long(FPNUM);        /*         FPNUM to SL32  */
EXTERN FPNUM   fpdiv(FPNUM, FPNUM);   /* Divide                 */
EXTERN SW16    fp2intel(FPNUM, LPSW16);
#ifdef DEBUG
EXTERN VOID    fpprint(FPNUM);        /* print to stdout fpnum  */
#endif
#else
EXTERN FPNUM   fplongexp();
EXTERN SW16    fp2word();
EXTERN SL32    fp2long();
EXTERN FPNUM   fpdiv();
EXTERN SW16    fp2intel();
#ifdef DEBUG
EXTERN VOID    fpprint();
#endif
#endif

#endif /* external fp */

