/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/met/kern.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: kern.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   07 Apr 1995 09:13:02   LISA
 * 
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   22 Apr 1994 14:20:36   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.3   12 Feb 1993 14:39:30   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   06 Jan 1993 14:30:00   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:16:34   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 15:59:08   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/* kern.c */


/*---------------------------------------------------------------------
     07-28-90   awr   Added check in set_kern() for existence of kerning
                      segments
     01-30-91   dET  function prototypes and modification to support 
                     MSC multi-model compilation.
     4-Feb-91   awr  Split off from maker.c
    23-Feb-91   awr  Changed bucket sub-segment handles to offsets.
    17-Jun-91   jfd  Moved "debug.h" after "port.h".
                     Moved "profile.h" after "port.h".
                     Moved "cgconfig.h" before "port.h".
    26 Jun 91   ss   Added cast for kern.
     1 Jul 91   ss   Added prototypes for fgseg().
    03-Apr-92   rs   Portability cleanup (see port.h)
    17-Sep-92   jfd  Changed conditional compile of entire module to include
                     IF_RDR.
    06-Jan-93   jfd  ANSI C function declaration changes.
    08-Feb-93   jfd  VXWorks support.
-----------------------------------------------------------------------*/


#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include "cgconfig.h"
#include "port.h"
#include "debug.h"
#include "profile.h"
#include "ifmem.h"                     /* 17-Feb-91 awr */
#include "cgif.h"
#include "segments.h"
#include "ix.h"
#include "if_type.h"

#if KERN && IF_RDR


#ifdef LINT_ARGS
GLOBAL SW16    kern(UW16, UW16, UW16);
MLOCAL BOOLEAN set_kern(UW16, UW16, LPSW16, LPSW16);
EXTERN UW16   IXsetup_chr_def(UW16, PCHR_DEF, PPBUCKET);
/* fm.c */
EXTERN LPSB8   fgseg(PBUCKET);

#else
MLOCAL BOOLEAN set_kern();
EXTERN UW16   IXsetup_chr_def();

/* fm.c */
EXTERN LPSB8   fgseg();
#endif




/*............Data from compound character segment.........................*/

EXTERN CHR_DEF chr_def[];
EXTERN CHR_DEF_HDR chr_def_hdr;



/*-------------------*/
/*  set_kern         */
/*-------------------*/
#if defined (ANSI_DEFS)
MLOCAL BOOLEAN set_kern(UW16 code, UW16 chId, LPSW16 left, LPSW16 right)
#else
MLOCAL BOOLEAN
set_kern(code, chId, left, right)
    UW16 code;
    UW16 chId;
    LPSW16 left;
    LPSW16 right;
#endif
{
    PBUCKET     bucket;
    SW16        index;
    UW16       okern;
    LPUB8      kern;
    UW16       i;

    DBG2("set_kern(code=%d, chId=%d)\n", code, chId);

    if(IXsetup_chr_def(chId, chr_def, &bucket)) /* 11-14-90 jfd */
       return FALSE;

    if(chr_def_hdr.is_compound) return FALSE;

    index = chr_def[0].index;

  /* Point kern to start of sector info; skip 6 byte header */


    if(code == TEXT_KERN)
        okern = bucket->p.in.otext_kern_seg; /* jfd */
    else /* must be DESIGN_KERN */
        okern = bucket->p.in.odesign_kern_seg; /* jfd*/

    if(okern != MAX_UWORD)
        kern = (LPUB8)fgseg(bucket) + okern;     /* added cast -ss 6/26/91 */
    else
    {
        DBG("    missing kerning segment\n");
        return FALSE;
    }

    DBG3("    sign %d   unit %d    num sectors %d\n",
                     *(LPSW16)kern, *(LPSW16)(kern+2), *(LPSW16)(kern+4));

    kern += 6 + 4*index;

    for(i=0; i<4; i++)
    {
        *left++ = *kern >> 4;
        *right++ = *kern++ & 0xf;
    }

    return TRUE;
}

/*-------------------*/
/*       kern        */
/*-------------------*/
#if defined (ANSI_DEFS)
GLOBAL SW16  kern(UW16 code, UW16 chId0, UW16 chId1)
#else
GLOBAL SW16
kern(code, chId0, chId1)
    UW16 code;
    UW16 chId0, chId1;
#endif
{
    SW16  left0[4], right0[4],
          left1[4], right1[4];
    LPSW16 r, l;
    SW16  min, sum;
    UW16 i;

    DBG("kern()\n");

    if(!set_kern(code, chId0, &left0[0], &right0[0])) return 0;
    if(!set_kern(code, chId1, &left1[0], &right1[0])) return 0;

    min = 256;
    r = right0;
    l = left1;
    for(i=0; i<4; i++)
    {
        DBG2("    %d %d\n", *r, *l);
        sum = *r++ + *l++;
        if(sum < min) min = sum;
    }
    DBG1("    min = %d\n", min);

  /* Convert from 54 units per em to 8782 units per em */

    return (SW16)(((SL32)min * 8782L + 27L) / 54L);
}
#endif  /* KERN && IF_RDR */


