/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/ps/ift1hint.c,v 1.2 2003/05/07 23:02:25 Lann Exp $ */
/* $Log: ift1hint.c,v $
/* Revision 1.2  2003/05/07 23:02:25  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.8   07 Apr 1995 10:15:38   LISA
 * Changed copyright from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.7   04 May 1994 14:29:46   JOE
 * In add_stem(), changed MAXSTEMS to MAXCHARSTEMS due to naming change
 * made in include file to resolve naming conflict.
 * 
 *    Rev 1.6   22 Apr 1994 14:34:30   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.5   03 Aug 1993 14:47:26   MAIB
 * Changed interface to resolve naming conflicts
 * 
 *    Rev 1.4   02 Aug 1993 14:22:42   MAIB
 * Changes to support multi-tasking
 * 
 *    Rev 1.3   12 Feb 1993 15:32:02   JOE
 * VXWorks support.
 * 
 *    Rev 1.2   06 Jan 1993 10:22:16   JOE
 * ANSI C function declaration changes.
 * 
 *    Rev 1.1   14 Dec 1992 09:15:20   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   09 Dec 1992 15:54:12   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:25 $ */
/*
 * IFT1HINT.C - Marked-up version of ATECH's T1HINTS.C
 *
 *
 * History
 *
 *  22-May-91  jfd  Original Program
 *  10-Jun-91  jfd  Included cache.h
 *  16-Jun-91  jfd  Moved "debug.h" after "port.h".
 *  25-Jun-91  rs   Add 'DBG' statements, remove 'error_msg()'.
 *						  Conditionalize entire file w/'PST1'.
 *  02-Jul-91  jfd  Added EXTERN variable "uCallerID" in case of MULTICALLER.
 *  05-Jul-91  rs   Clean up error messages & return values.
 *  19-Jul-91  rs   Fix problems in writeBlueValues() -
 *                  change 'ND' to 'def', write empty array if baseline
 *                  condition not met, make sure each pair is at least 3
 *                  away from other pairs (default BlueFuzz).
 *  24-Jul-91  jfd  Windowize the file as follows:
 *                    sprintf --> SPRINTF
 *                  Added mixmodel.h.
 *  01-Aug-91  jfd  In getCharYvals(), when resetting "hol", set it to NIL_MH
 *                  instead of "0'.
 *  23-Aug-91  rs   Begin adding phase 1 hinting code for stems (lite).
 *  03-Sep-91  rs   Subtract left sidebearing from X (H) stems.
 *  04-Sep-91  rs	  Begin 2nd level of hints (lite+) using multiple
 *                  dimensions.
 *                  Use global text buffer 'sbuff' instead of CHARalloc-ing.
 *  09-Sep-91  rs   Add 'ghost_stems()'.
 *                  Call 'odifbb()' instead of make_char_part().
 *  11-Sep-91  rs   Add error messages (move to cgif.h at release).
 *                  Small changes to 'ghost_stems()' to fix 'hyphen' problem.
 *  13-Sep-91  rs   Change to filter_stems() when both points have same
 *                  top/bot, std dim, rep - used to discard both, now keep
 *                  2nd point - it should have a fair chance.
 *  20-Sep-91  jfd  In writeStdWidths(), changed "sprintf" to "SPRINTF"
 *                  to get rid of compile error when building under WINDOWS.
 *  26-Sep-91  jfd  Conditionally compiling "sbuff" and "tpb" based on
 *                  _WINDOWS.
 *  16-Oct-91  rs   In writeStdWidths(), changed line "stdVW = 33" to
 *                  "stdHW = 33";
 *  21-Nov-91  rs   Fix for symbol sets w/o X,x,O,o - BlueValues.
 *  06-Jan-92  jfd  Changed IF_CHAR structure reference to IF_STATE.
 *                  Changed "c" references to "if_state".
 *                  Changed "d" references to "if_state".
 *  7-Mar-92   awr  removed unused "EXTERN IXsetup_chr_def();"
 * 03-Apr-92   rs   Portability cleanup (see port.h).
 * 18-Jul-92   awr  changed parameters of odifbb() to account for INTRs
 * 17-Sep-92   jfd  Changed conditional compile of entire module to include
 *                  IF_RDR.
 * 03-Oct-92   rs   Changes toward ANSI C - if file not compiled, must
 *                  declared  at least 1 external - int iIFT1HINT = 0.
 * 15-Oct-92   mby  SYMmap() called with two arguments.
 * 06-Jan-93   jfd  ANSI C function declaration changes.
 * 08-Feb-93   jfd  VXWorks support.
 * 02-Aug-93   maib Changed calls to CGIFxxx() routines to cgifXXX() routines,
 *		    required for multi-tasking
 * 03-Aug-93 maib changed definitions for cgifXXXX() routines to resolve
 *                  naming conflicts
 * 04-May-94 jfd    In add_stem(), changed MAXSTEMS to MAXCHARSTEMS
 *                  due to naming change made in include file to
 *                  resolve conflict.
 */

#include "cgconfig.h"
#if PST1 && IF_RDR			/* enable PostScript Type 1 writer */

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include <stdio.h>
#include <string.h>

#include "port.h"
#include "debug.h"
#include "ps_type1.h"

#include "ifmem.h"
#include "mixmodel.h"  /* 07-23-91 jfd */
#include "cgif.h"
#include "cache.h"
#include "ix.h"

#if T1HINTS
#include "if_type.h"	/* for hinting - 8/23/91 - rs */
#endif /* T1HINTS */


/*
Definitions - move these to cgif.h when ready
*/
#define ERR_ps_toomanystems	208	/* no room to add stem to StemStr */
#define ERR_ps_toomanystdwid	209	/* no room to add std dim to array */


#ifdef LINT_ARGS
GLOBAL UW16      writeBlueValues(VOID);
MLOCAL UW16      getCharYvals(UW16, LPSW16, LPSW16);
#if T1HINTS
GLOBAL UW16		StemGen (VOID);
GLOBAL UW16		writeStdWidths(VOID);
GLOBAL UW16		ghost_stems (VOID);
MLOCAL UW16		add_stem (SW16, SW16, PSTEMSTRUC);
MLOCAL UW16		filter_stems (PSTEMSTRUC, SW16);
MLOCAL SW16			get_point (SW16, UB8);
MLOCAL UW16		add_stdwidth (SW16, SW16);
#endif /* T1HINTS */

EXTERN VOID       encWrite(LPUB8);
EXTERN UW16      outline(PHIFOUTLINE, UW16);
EXTERN UW16      SYMmap(SW16, int);
EXTERN SW16 		SYMmapR (UW16);

#if T1HINTS	/* add 8/23/91 - rs */
EXTERN UW16      odifbb(PCHR_DEF, PINTRBOX, PINTRVECTOR);
EXTERN VOID			first_loop(VOID);
EXTERN VOID			next_loop(PLOOP);
#endif /* T1HINTS */

#else
GLOBAL UW16      writeBlueValues();
MLOCAL UW16      getCharYvals();
#if T1HINTS
GLOBAL UW16		StemGen ();
GLOBAL UW16		writeStdWidths();
GLOBAL UW16		ghost_stems ();
MLOCAL UW16		add_stem ();
MLOCAL UW16		filter_stems ();
MLOCAL SW16			get_point ();
MLOCAL UW16		add_stdwidth ();
#endif /* T1HINTS */

EXTERN VOID       encWrite();
EXTERN UW16      outline();
EXTERN UW16      SYMmap();
EXTERN SW16 	  SYMmapR ();

#if T1HINTS	/* add 8/23/91 - rs */
EXTERN UW16      odifbb();
EXTERN VOID			first_loop();
EXTERN VOID			next_loop();
#endif /* T1HINTS */

#endif	/* LINT_ARGS */

/*
External variable references
*/
#ifdef _WINDOWS
EXTERN UB8 far sbuff[];	/* add 9/26/91 - jfd */
#else
EXTERN UB8 sbuff[];	/* add 9/25/91 - rs */
#endif


#if MULTICALLER
EXTERN UW16 uCallerId;
#endif /* MULTICALLER */

#if T1HINTS	/* add 8/23/91 - rs */
EXTERN CHR_DEF_HDR chr_def_hdr;
EXTERN CHR_DEF chr_def[];
EXTERN IF_STATE if_state;
EXTERN SW16	StdHW, StdVW, StemSnapH[], StemSnapV[];
EXTERN SW16 glob_x_dim, glob_y_dim;
EXTERN SW16			cg_italicAngle;		/* italic angle (degrees) */
#endif /* T1HINTS */


/*
Create our global variables
*/

#if T1HINTS
GLOBAL STEMSTRUC	XStemStr, YStemStr, WStemStr;
#endif /* T1HINTS */


/*
Local variables mostly for efficiency
*/
#if T1HINTS
MLOCAL SW16			xscale, yscale;	/* current scale factors log base 2 */
MLOCAL INTRVECTOR	trans;				/* translation for points to origin */
MLOCAL PSKEL		skel_ptr;			/* ptr to X or Y skel */
MLOCAL SW16			current_loop;		/* current loop number */
MLOCAL PSW16VECTOR cv;					/* -> coordinates */
MLOCAL UB8		num_skel;			/* # skels in character (X or Y) */
MLOCAL LPUB8		num_skel_loop;		/* -> # skels per loop (X or Y) */
MLOCAL LOOP			loop;					/* current loop */
MLOCAL INTRBOX		bbox;					/* bounding box for hints */
MLOCAL SW16			tlsb, right; 		/* left & right sidebearing for hinting */
MLOCAL SW16			top, bot;			/* top & bottom of char for hinting */

/*
Array of chars for std width (hints)...
Make sure that 'H' and 'l' are near the beginning. The numerals should be
near the end. We depend upon 'l' and 'H' for StdVW & StdHW, respectively.
Once the StemSnap arrays are full, we can accept no more stems.
*/
MLOCAL UW16		cgnumstdx[] =
		{CG_l, CG_H, CG_O, CG_o, CG_0, CG_1, CG_7};

#endif /* T1HINTS */


#if T1HINTS
#if defined (ANSI_DEFS)
GLOBAL UW16 StemGen ()
#else
GLOBAL UW16 StemGen ()
#endif
/*
This function will attempt to find all stems in the current character
as specified in chr_def[].
The stem info will be written to the global structures XStemStr & YStemStr.
No attempt is made to remove either bogus or overlapping stems. This function
is handled by 'FilterStems()'.
Ghost stems are handled separately.

This version only processes the base character in compound characters.

CUBIC outlines must be enabled
map_and_set_scale() must have been called for this character prior to
arriving here. The chr_def[] will have been set up.

return (SUCCESS) if no errors, else a standard Intellifont error code.
*/
{
UW16			status;
SW16			point1, point2;
PCHR_DEF    cd;
SW16			j, k, num_assoc;
UB8			sk_indx, skel_i, child_i, as_indx;

	DBG("\n\n\n\nS t e m G e n()\n");

	cd = chr_def;	/* point to base character */

	status = odifbb(cd, &bbox, &trans);
    if (status)
		return status;

	XStemStr.numStems = YStemStr.numStems = 0;
	/*
		If this is not an Intellifont character, go no further
	*/
	if (!if_state.char_is_if)
		return (SUCCESS);

	/*
		Process the X skeletal data
		Determine shift count for scale factors ->
		log base 2 of if_state.x.pixel_size, if_state.y.pixel_size
	*/
	for (j = 1, xscale = 0; j < if_state.x.pixel_size; j *= 2, xscale++)
		;
	for (j = 1, yscale = 0; j < if_state.y.pixel_size; j *= 2, yscale++)
		;

	skel_ptr = &if_state.xskel;
	WStemStr.numStems = 0;

	first_loop();				/* init the vector generator */
	next_loop (&loop);		/* set up this loop */
	current_loop = 0;
	cv = loop.cv;				/* -> ptr to interleaved coords */
	num_skel = skel_ptr->num_skel;
	num_skel_loop = skel_ptr->num_skel_loop;

/*
	Process the X skeletal data...

	for each X skeletal point in the tree
*/
	top = (bbox.ur.y >> yscale);	/* calc top of char */
	tlsb = (bbox.ll.x >> xscale);	/* calc left side bearing */
	bot = (bbox.ll.y >> yscale);	/* calc bottom of char */
	right = (bbox.ur.x >> xscale);/* calc right of char */

	as_indx = 0;
	for (sk_indx = 0; sk_indx < skel_ptr->num_nodes; sk_indx++)
	{
   	if (!(num_assoc = skel_ptr->tree_attrib[sk_indx] >> 5))
			continue;

   	skel_i = skel_ptr->tree_skel_i[sk_indx];
 		point1 = get_point (0, skel_i);

		/*
			process each of this parent skel's associations...
			If the child's association to its parent is across black space,
			we will treat it as a stem.
		*/
		for (k = 0; k < num_assoc; k++)
		{
			/*
				Look for an association across black space
			*/
			as_indx++;
			if (as_indx <= sk_indx)
				as_indx = sk_indx + 1;				/* reached a new root */
			if (skel_ptr->tree_attrib[as_indx] & 0x03)
				continue;

	   	child_i = skel_ptr->tree_skel_i[as_indx];
	   	point2 = get_point (0, child_i);

 			/*
				Process stem between point 1 & 2
			*/
			status = add_stem (point1, point2, &WStemStr);
            if (status)
				return (status);
		} /* for each association to parent */
	}   /* for each X skeletal point in tree */

	/*
		We now have a working structure containing all discreet X stems
		in this character. Filter out all overlapping stems and copy
		the remaining ones to the X stem structure.
	*/
	if (WStemStr.numStems)
	{
		status = filter_stems (&XStemStr, 0);
        if (status)
			return (status);
	}

	/*
		Process the Y skeletal data
	*/
	skel_ptr = &if_state.yskel;
	WStemStr.numStems = 0;
	first_loop();	/* init the vector generator */
	next_loop (&loop);		/* set up this loop */
	current_loop = 0;
	cv = loop.cv;				/* -> ptr to interleaved coords */
	num_skel = skel_ptr->num_skel;
	num_skel_loop = skel_ptr->num_skel_loop;

	/*
		For each contour of the character outline do:

		for each Y skeletal point in the tree
	*/
	as_indx = 0;
   for (sk_indx = 0; sk_indx < skel_ptr->num_nodes; sk_indx++)
   {
      if (!(num_assoc = skel_ptr->tree_attrib[sk_indx] >> 5))
			continue;

      skel_i = skel_ptr->tree_skel_i[sk_indx];
     	point1 = get_point (1, skel_i);

		/*
			process each of this parent skel's associations...
			If the child's association to its parent is across black space,
			we will treat it as a stem.
		*/
		for (k = 0; k < num_assoc; k++)
		{
			/*
				Look for an association across black space
			*/
			as_indx++;
			if (as_indx <= sk_indx)
				as_indx = sk_indx + 1;				/* reached a new root */
			if (skel_ptr->tree_attrib[as_indx] & 0x03)
				continue;

	      child_i = skel_ptr->tree_skel_i[as_indx];
	     	point2 = get_point (1, child_i);

 			/*
				Process stem between point 1 & 2
			*/
			status = add_stem (point1, point2, &WStemStr);
            if (status)
				return (status);
		} /* for each association to parent */
   }   /* for each Y skeletal point in tree */

	/*
		We now have a working structure containing all discreet Y stems
		in this character. Filter out all overlapping stems and copy
		the remaining ones to the Y stem structure.
	*/
	if (WStemStr.numStems)
		status = filter_stems (&YStemStr, 1);
        if (status)
			return (status);

	return SUCCESS;
}

#if defined (ANSI_DEFS)
MLOCAL UW16 add_stem (SW16 p1, SW16 p2, PSTEMSTRUC pss)
#else
MLOCAL UW16 add_stem (p1, p2, pss)
SW16	p1, p2;
PSTEMSTRUC	pss;
#endif
/*
Add a stem to a 'ASTEMSTRUC'

return SUCCESS if no error, else error code
*/
{
SW16	k, count;
SW16	point, width;
PASTEMSTRUC	pas;

	point = MIN (p1, p2);
	width = MAX (p1, p2) - point;
	count = pss->numStems;
	pas = pss->a;
	for (k = 0; k < count; k++, pas++)
	{
		if (pas->loc == point && pas->width == width)
		{
			pas->rep++;
			return (SUCCESS);
		}
	}
	/*
		Did not find an exact match in the stem structure, add a new
		entry if there is room
	*/
	if (pss->numStems >= MAXCHARSTEMS)
		return (ERR_ps_toomanystems);

	pss->numStems++;
	pas->rep = 1;
	pas->loc = point;
	pas->width = width;
	return (SUCCESS);
}

#if defined (ANSI_DEFS)
MLOCAL UW16 filter_stems (PSTEMSTRUC pss, SW16 xyflg)
#else
MLOCAL UW16 filter_stems (pss, xyflg)
PSTEMSTRUC	pss;
SW16			xyflg;	/* 0 == X, 1 == Y */
#endif
/*
Filter out overlapping stems from working stem structure 'WStemStr' and
copy the remaining stems to the specified stem structure (x or y)
uses LOCALs:
	top (top of character bb)
	bot (bottom of character bb)
	tlsb (left of character bb)
	right (right of character bb)

return (SUCCESS) if no error, else error code
*/
{
SW16 j, k, m, numStems, xflg, p1std, p2std, p1top, p2top, p1bot, p2bot;
LPSW16	par;
PSTEMSTRUC	wpss; /* ptr to working stem structure */
PASTEMSTRUC	pass, wp1ass, wp2ass; /* ptrs to src & wkg stem struc arrays */

	if (!xyflg)
		par = StemSnapV;
	else par = StemSnapH;
	wpss = &WStemStr;
	/*
		handle case of 1 stem by setting wp2ass -> 1st point & catch at end
	*/
	wp1ass = wp2ass = wpss->a;	
	pass = pss->a;
	xflg = 0;

	numStems = wpss->numStems;
	pss->numStems = 0;
	p1std = p2std = 0;
	for (j = 0; j < numStems-1; j++, wp1ass++, wp2ass = wp1ass)
	{
		xflg = 0;	/* not discarded yet */
		if (!wp1ass->rep)	/* this stem was discarded already */
			continue;
		/*
			See if this stem is 'near' a std dimension
		*/
		p1std = 0;
		for (m = 0; m < par[0]; m++)
		{
#define BIGFILTER	/* capture 'almost' stems */
#ifdef BIGFILTER /* filtering confuses 'H' VSTEM w/'i' dot */
			if ((wp1ass->width >= par[m+1]-1) && (wp1ass->width <= par[m+1]+1))
#else
			if (wp1ass->width == par[m+1])
#endif /* BIGFILTER */
			{
				p1std = 1;	/* point 1 is 'near' a std dim */
				break;
			}
			if (wp1ass->width < par[m+1])
				break;
		}

		/*
			If 'Y', see if this stem is near the top of a char
		*/
		p1top = 0;
		if (xyflg)
			if ((wp1ass->loc + wp1ass->width) > (top - 5))
				p1top = 1;

		/*
			If 'Y', see if this stem is near the bottom of a char
		*/
		p1bot = 0;
		if (xyflg)
			if (wp1ass->loc < (bot + 5))
				p1bot = 1;

		for (k = j+1; k < numStems; k++)
		{
			wp2ass = &wpss->a[k];
			if (!wp2ass->rep)	/* this stem was discarded already */
				continue;
			/*
				See if this stem is 'near' a std dimension
			*/
			p2std = 0;
			for (m = 0; m < par[0]; m++)
			{
#ifdef BIGFILTER /* filtering confuses 'H' VSTEM w/'i' dot */
				if ((wp2ass->width >= par[m+1]-1) && (wp2ass->width <= par[m+1]+1))
#else
				if (wp2ass->width == par[m+1])
#endif /* BIGFILTER */
				{
					p2std = 1;	/* point 2 is 'near' a std dim */
					break;
				}
				if (wp2ass->width < par[m+1])
					break;
			}

			/*
				If 'Y', see if this stem is near the top of a char
			*/
			p2top = 0;
			if (xyflg)
				if ((wp2ass->loc + wp2ass->width) > (top - 5))
					p2top = 1;

			/*
				If 'Y', see if this stem is near the bottom of a char
			*/
			p2bot = 0;
			if (xyflg)
				if (wp2ass->loc < (bot + 5))
					p2bot = 1;

			/*
				Make sure that the the start of P1 is not (>= the start
				of P2 AND start of P1 is not <= the end of P2) AND
				the end of P1 is not (>= the start of P2 AND
				the end of P1 is not <= the end of P2) AND
				the the start of P2 is not (>= the start of P1 AND
				the end of P2 is not <= the end of P1) AND
				the end of P2 is not (>= the start of P1 AND
				the end of P2 is not <= the end of P1)...
			*/
			if ( ((wp1ass->loc >= wp2ass->loc) &&
				   (wp1ass->loc <= (wp2ass->loc+wp2ass->width)) ) ||
				  (((wp1ass->loc+wp1ass->width) >= wp2ass->loc) &&
				   ((wp1ass->loc+wp1ass->width) <= (wp2ass->loc+wp2ass->width))) ||
				  ((wp2ass->loc >= wp1ass->loc) &&
				   (wp2ass->loc <= (wp1ass->loc+wp1ass->width))) ||
				  (((wp2ass->loc+wp2ass->width) >= wp1ass->loc) &&
				   ((wp2ass->loc+wp2ass->width) <= (wp1ass->loc+wp1ass->width)))
				)
			{
				/*
					Overlapping point -
						If one of the stems is 'near' the top and the other is not,
						keep it. If both are 'near' the top, keep the smaller one.
						If one of the stems is 'near' the bottom and the other is not,
						keep it. If both stems are 'near' the bottom, keep the smaller.
						If one of the stems is 'close' to std dimension, use it
						else if one of the stems had a higher repeat count than the
						other, take it. Otherwise, throw them both away in this rev.
				*/

				/*
					If a horizontal stem (Y) is near both the top and bottom,
					forget about it. - add 9/11/91 - rs
				*/
				if (xyflg)
				{
					if (p1top && p1bot)
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
				} /* if (xyflg) - horizontal stem */
				else
				{
				/*
					If a vertical stem (X) is near both the left and right,
					forget about it. - add 9/11/91 - rs
				*/
					if ((wp1ass->loc + wp1ass->width) > (right - tlsb - 10))
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
				} /* else (xyflg) - vertical stem */

				if (p1top && !p2top)
				{
					/*
						point 1 is near the top of the char and point 2 is not,
						keep considering point 1 & discard point 2
					*/
					wp2ass->rep = 0;	/* discard point 2 */
					continue;
				}
				else if (!p1top && p2top)
				{
					/*
						point 2 is near the top of the char & point 1 is not,
						keep considering point 2 (next loop) & discard point 1
					*/
					wp1ass->rep = 0;	/* discard point 1 */
					xflg = 1;
					break;	/* forget point 1, compare point 2 */
				}
				else if (p1top && p2top)
				{
					/*
						point 1 AND point 2 are near the top of the char.
						If 1 is a std dim, keep it. Else, keep the
						point which is higher in consideration, discard the other.
					*/
					if (p1std && !p2std)
					{
						wp2ass->rep = 0;	/* discard point 2 */
						continue;
					}
					else if (!p1std && p2std)
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
					else if (wp1ass->loc > wp2ass->loc)
					{
						wp2ass->rep = 0;	/* discard point 2 */
						continue;
					}
					else
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
				}
				else if (p1bot && !p2bot)
				{
					/*
						point 1 is near the bottom of the char and point 2 is not,
						keep considering point 1 & discard point 2
					*/
					wp2ass->rep = 0;	/* discard point 2 */
					continue;
				}
				else if (!p1bot && p2bot)
				{
					/*
						point 2 is near the bottom of the char & point 1 is not,
						keep considering point 2 (next loop) & discard point 1
					*/
					wp1ass->rep = 0;	/* discard point 1 */
					xflg = 1;
					break;	/* forget point 1, compare point 2 */
				}
				else if (p1bot && p2bot)
				{
					/*
						point 1 AND point 2 are near the bottom of the char.
						If one of them is a std dim, keep it. Else keep the
						point which is lower in consideration, discard the other.
					*/
					if (p1std && !p2std)
					{
						wp2ass->rep = 0;	/* discard point 2 */
						continue;
					}
					else if (!p1std && p2std)
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
					else if (wp1ass->loc < wp2ass->loc)
					{
						wp2ass->rep = 0;	/* discard point 2 */
						continue;
					}
					else
					{
						wp1ass->rep = 0;	/* discard point 1 */
						xflg = 1;
						break;	/* forget point 1, compare point 2 */
					}
				}
				/*
					The point is not near the top or bottom. See if it is
					a std dim.
				*/
				else if (p1std && !p2std)
				{
					/*
						point 1 is near a std dim & point 2 is not, keep considering
						point 1 & discard point 2
					*/
					wp2ass->rep = 0;	/* discard point 2 */
					continue;
				}
				else if (!p1std && p2std)
				{
					/*
						point 2 is near a std dim & point 1 is not, keep considering
						point 2 (next loop) & discard point 1
					*/
					wp1ass->rep = 0;	/* discard point 1 */
					xflg = 1;
					break;	/* forget point 1, compare point 2 */
				}
				/*
					else, either both points are near std dims or neither one is
					use the larger stem.
				*/
				else if (wp1ass->width > wp2ass->width)
				{
					wp2ass->rep = 0;	/* discard point 2 */
					continue;
				}
				else if (wp2ass->width > wp1ass->width)
				{
					wp1ass->rep = 0;	/* discard point 1 */
					xflg = 1;			/* get next point to start comparisons */
					break;
				}
				/*
					Both width are the same, if one has a higher repeat count,
					keep it
				*/
				else if (wp1ass->rep > wp2ass->rep)
				{
					wp2ass->rep = 0;	/* discard point 2 */
					continue;
				}
				else if (wp2ass->rep > wp1ass->rep)
				{
					wp1ass->rep = 0;	/* discard point 1 */
					xflg = 1;			/* get next point to start comparisons */
					break;
				}
				else
				/*
					both points have same width, rep count, std dim relation
					and top relation, discard the 1st because if it hasn't won
					so far, we may as well let the 2nd point have a chance.
				*/
				{
					wp1ass->rep = 0;
					xflg = 1;
					break;
				}
			} /* big IF for overlapping test */
		}	/* search next point for overlap with 1st point */
		if (xflg)	/* already handled this set of points */
			continue;
		/*
			This stem has been resolved, just copy it
		*/
		pss->numStems++;
		pass->rep = wp1ass->rep;
		pass->loc = wp1ass->loc;
		pass->width = wp1ass->width;
		pass++;
		xflg = 1;
	}
	/*
		Don't forget the last point (or only point if 1 stem)
	*/
	if (wp2ass->rep)
	{
		pss->numStems++;
		pass->rep = wp2ass->rep;
		pass->loc = wp2ass->loc;
		pass->width = wp2ass->width;
	}

	return (SUCCESS);
}

#if defined (ANSI_DEFS)
GLOBAL UW16 ghost_stems ()
#else
GLOBAL UW16 ghost_stems ()
#endif
/*
Generate any necessary 'ghost' stems for the current character
There must be a HSTEM or 'ghost' HSTEM at the top and bottom of each char.
There must be a VSTEM or 'ghost' VSTEM at the left and right of each char.

return (SUCCESS) if no error, else standard error code
*/
{
SW16	j, needYmin, needYmax, needXmin, needXmax, dim_w, dim_1, dim_2;

#define YMINGHOST	1
#define YMAXGHOST	1

	needYmin = needYmax = needXmin = needXmax = 1;

	dim_w = right - tlsb;
	dim_1 = 40;	/* relative to tlsb */
	dim_2 = dim_w - 40;
	for (j = 0; j < XStemStr.numStems; j++)
	{
		if (XStemStr.a[j].loc < dim_1)
			needXmin = 0;
		if (XStemStr.a[j].loc + XStemStr.a[j].width > dim_2)
			needXmax = 0;
	}

	dim_w = top - bot;
	dim_1 = bot + 40;
	dim_2 = top - 40;
	for (j = 0; j < YStemStr.numStems; j++)
	{
		if (YStemStr.a[j].loc < dim_1)
			needYmin = 0;
		if (YStemStr.a[j].loc + YStemStr.a[j].width > dim_2)
			needYmax = 0;
	}

#ifdef XMINGHOST
	if (needXmin)
	{
		j = XStemStr.numStems;
		XStemStr.a[j].rep	= 1;
		XStemStr.a[j].loc	= 0;
		XStemStr.a[j].width = 20;
		XStemStr.numStems++;
	}
#endif /* XMINGHOST */
#ifdef XMAXGHOST
	if (needXmax)
	{
		j = XStemStr.numStems;
		XStemStr.a[j].rep	= 1;
		XStemStr.a[j].loc	= right - tlsb;
		XStemStr.a[j].width = -20;
		XStemStr.numStems++;
	}
#endif /* XMAXGHOST */

#ifdef YMINGHOST
	if (needYmin)
	{
		j = YStemStr.numStems;
		YStemStr.a[j].rep	= 1;
		YStemStr.a[j].loc	= bot;
		YStemStr.a[j].width = 20;
		YStemStr.numStems++;
	}
#endif /* YMINGHOST */
#ifdef YMAXGHOST
	if (needYmax)
	{
		j = YStemStr.numStems;
		YStemStr.a[j].rep	= 1;
		YStemStr.a[j].loc	= top;
		YStemStr.a[j].width = -20;
		YStemStr.numStems++;
	}
#endif /* YMAXGHOST */

	return (SUCCESS);
}


#if defined (ANSI_DEFS)
MLOCAL SW16 get_point (SW16 flg, UB8 skel_i)
#else
MLOCAL SW16 get_point (flg, skel_i)
SW16	flg;
UB8	skel_i;
#endif
/*
Get a scaled skeletal point from the contour array 'cv' based upon its
index in the skeletal array 'tree_skel_i'.

For X - subtract lsb - 9/3/91 - rs

entry:
	flg = 0 for X, 1 for Y
	skel_i is index into 'tree_skel_i'
	tlsb = left sidebearing

exit:
	uses MLOCAL skel_ptr, xscale or yscale
	uses MLOCAL 'trans' to translate outline to origin (0,0)
	return X or Y scaled point
*/
{
UW16			tmpuw, k;
SW16VECTOR	tmpwv;


	/*
		Make sure we have the right contour set up

		First find the desired contour number
	*/
	for (tmpuw = k = 0; tmpuw < if_state.num_loops; tmpuw++)
	{
		k += num_skel_loop[tmpuw];	/* calc 1 past this loop */
		if ((UW16)skel_i < k)
			break;
	}

	if (tmpuw < current_loop)
	{
		first_loop ();
		next_loop (&loop);					/* set up this loop */
		current_loop = 0;
	}

	while (current_loop < tmpuw)
	{
		next_loop (&loop);					/* set up this loop */
		current_loop++;
	}

	cv = loop.cv;					/* -> coordinates */

	tmpuw = skel_ptr->skel_to_contr[skel_i];
   tmpwv = *(cv + tmpuw);
	if (!flg)	/* X */
	   return ( ((tmpwv.x + trans.x) >> xscale) - tlsb);
	else 			/* Y */
	   return ((tmpwv.y + trans.y) >> yscale);
}

#if defined (ANSI_DEFS)
GLOBAL UW16 writeStdWidths()
#else
GLOBAL UW16 writeStdWidths()
#endif
/*
Determine standard stem widths for key characters.
The first StemSnap entry will always be the global std dim from the library.
Write StdHW & StdVW.
Write StemSnapH array.
Write StemSnapV array if not italic typeface.
StemSnap arrays contain the number of entries as the first value. There are
a maximum of MAXSTDWIDTHS in the array.

return (SUCCESS) if no errors, else standard error code
*/
{
SW16			chId, j, k, m;
HIFOUTLINE	hol;
UW16			ures;
#ifdef _WINDOWS
UB8	  far *tpb;
#else
UB8		   *tpb;
#endif

	StemSnapH[0] = StemSnapV[0] = 0;	/* assume no stems */
	StdHW = StdVW = 0;

	for (k = 0; k < (sizeof (cgnumstdx) / sizeof (cgnumstdx[0])); k++)
	{
		if ((chId = SYMmapR (cgnumstdx[k])) == (SW16)MAX_UWORD)
			continue;

		/*
			Note: We should not have to actually generate the outline here.
					StemGen() will just have to generate it again. However,
					in the interest of time, we will use this as a quick way to
					get everything set up properly.
		*/
		hol = NIL_MH;
		ures = cgif_Char_Handle(chId, (PHIFBITMAP)&hol, 0);
		if (ures)	/* should not get an error here, but continue for now */
		{
			if (hol)
				CHARfree (hol);
			continue;
		}
		if (hol)
			CHARfree (hol);

		if (StemGen ())
			continue;

		/*
			Get the std dims 'StdVW' & 'StdHW' if this is the right char
		*/
		if (cgnumstdx[k] == CG_l)
		{	/* use VSTEM from 'l' for std dim */
			if (XStemStr.numStems)
				StdVW = XStemStr.a[0].width;
			else
			{
				StdVW = 19;
				ures = add_stdwidth (0, 19);
                if (ures)
					return (ures);
			}
		}

		if (cgnumstdx[k] == CG_o)
		{	/* use HSTEM from 'o' for std dim */
			if (YStemStr.numStems)
				StdHW = YStemStr.a[0].width;
			else
			{
				StdHW = 33;
				ures = add_stdwidth (1, 33);
                if (ures)
					return (ures);
			}
		}

		/*
			Process the stems to create the V & H arrays
		*/
		for (j = 0; j < XStemStr.numStems; j++)
		{
			if (StemSnapV[0] >= MAXSTDWIDTHS) /* better not be > */
				break;
			ures = add_stdwidth (0, XStemStr.a[j].width);
            if (ures)
				return (ures);
		} /* for (j = ...) */

		for (j = 0; j < YStemStr.numStems; j++)
		{
			if (StemSnapH[0] >= MAXSTDWIDTHS) /* better not be > */
				break;
			ures = add_stdwidth (1, YStemStr.a[j].width);
            if (ures)
				return (ures);
		} /* for (j = ...) */
	} /* for (k = ... ) */

	/*
		Sort the std widths in ascending order
	*/
	for (j = 1; j < StemSnapH[0]; j++)
		for (k = j+1; k <= StemSnapH[0]; k++)
			if (StemSnapH[k] < StemSnapH[j])
			{
				m = StemSnapH[j];
				StemSnapH[j] = StemSnapH[k];
				StemSnapH[k] = m;
			}

	for (j = 1; j < StemSnapV[0]; j++)
		for (k = j+1; k <= StemSnapV[0]; k++)
			if (StemSnapV[k] < StemSnapV[j])
			{
				m = StemSnapV[j];
				StemSnapV[j] = StemSnapV[k];
				StemSnapV[k] = m;
			}

	if (StdHW)
	{
		SPRINTF (sbuff, "/StdHW [%d] def\r\n", StdHW);
		encWrite(sbuff);
		SPRINTF (sbuff, "/StemSnapH [");
		for (k = 1; k <= StemSnapH[0]; k++)
		{
			tpb = &sbuff[STRLEN(sbuff)];
			SPRINTF (tpb, "%d ", StemSnapH[k]);
		}
		tpb = &sbuff[STRLEN(sbuff)];
		SPRINTF (tpb, "] def\r\n");
		encWrite(sbuff);
	} /* if (StdHW) */

	if (StdVW)
	{
		SPRINTF (sbuff, "/StdVW [%d] def\r\n", StdVW);
		encWrite(sbuff);
		if (!cg_italicAngle)
		{	/* empty array for italic font */
			SPRINTF (sbuff, "/StemSnapV [");
			for (k = 1; k <= StemSnapV[0]; k++)
			{
				tpb = &sbuff[STRLEN(sbuff)];
				SPRINTF (tpb, "%d ", StemSnapV[k]);
			}
			tpb = &sbuff[STRLEN(sbuff)];
			SPRINTF (tpb, "] def\r\n");
			encWrite(sbuff);
		}
	} /* if (StdVW) */

	return(SUCCESS);
}

#if defined (ANSI_DEFS)
MLOCAL UW16 add_stdwidth (SW16 xyflg, SW16 width)
#else
MLOCAL UW16 add_stdwidth (xyflg, width)
SW16	xyflg;	/* 0 = X, 1 = Y */
SW16	width;	/* width of stem to add */
#endif
/*
Add an entry to the StemSnapH or StemSnapV array
Because there is only room for a limited number of entries, we will
filter within +/- 1 pixel. The V array seems to fill up on about the
8th or 9th character, otherwise. We may want to filter within 2 or even
3 to get more different stems if necessary.

return (SUCCESS) if added, else error code
*/
{
LPSW16			parr;
SW16			m, count, dup;

	if (!xyflg)
		parr = StemSnapV;
	else
		parr = StemSnapH;

	count = parr[0];
	for (m = 1, dup = 0; m <= count; m++)
		if ((parr[m] >= width - 1) && (parr[m] <= width + 1))
			return (SUCCESS);	/* duplicate, don't add */

	if (count >= MAXSTDWIDTHS) /* better not be > */
		return (ERR_ps_toomanystdwid);

	parr[count+1] = width;
	parr[0]++;

	return (SUCCESS);
}

#endif /* T1HINTS */


#if 0
#if defined (ANSI_DEFS)
void writeHints(struct ruleHdrStr *ruleHdr, struct olCharStr *olChar)
#else
void
writeHints(ruleHdr,olChar)
struct ruleHdrStr *ruleHdr;
struct olCharStr *olChar;
#endif
{
	int j;
	int needXmax=TRUE,needYmax=TRUE,needXmin=TRUE,needYmin=TRUE;

	/* X Rules */
	for (j = 0 ; j < ruleHdr->xRules ; j++) {
		/* Coordinates are relative to sidebearing point */
		encPSputCoord(xRules[j].startPos - olChar->xmin);
		encPSputCoord(SGN(xRules[j].endPos-xRules[j].startPos)*xRules[j].width);
		encodeVec[nEncode++] = 3;  /* vstem */

		if (xRules[j].startPos == olChar->xmax)needXmax = FALSE;
		if (xRules[j].startPos == olChar->xmin)needXmin = FALSE;
	}

	/* Y Rules */
	for (j = 0 ; j < ruleHdr->yRules ; j++) {
		encPSputCoord(-yRules[j].startPos);
		encPSputCoord(SGN(yRules[j].startPos-xRules[j].endPos)*yRules[j].width);
		encodeVec[nEncode++] = 1;  /* hstem */

		if (yRules[j].startPos == olChar->ymax)needYmax = FALSE;
		if (yRules[j].startPos == olChar->ymin)needYmin = FALSE;
	}

	/* Write Ghost Stems, if necessary */
	if (needXmin)
	{
		encPSputCoord(0);
		encPSputCoord(stdVW);
		encodeVec[nEncode++] = 3;  /* vstem */
	}
	if (needXmax)
	{
		encPSputCoord(olChar->xmax - olChar->xmin);
		encPSputCoord(-stdVW);
		encodeVec[nEncode++] = 3;  /* vstem */
	}
	if (needYmin)
	{
		encPSputCoord(-olChar->ymin);
		encPSputCoord(-stdHW);
		encodeVec[nEncode++] = 1;  /* hstem */
	}
	if (needYmax)
	{
		encPSputCoord(-olChar->ymax);
		encPSputCoord(stdHW);
		encodeVec[nEncode++] = 1;  /* hstem */
	}
}
#endif /* 0 */

#if defined (ANSI_DEFS)
GLOBAL UW16  writeBlueValues()
#else
GLOBAL UW16
writeBlueValues()
#endif
{
SW16  capOymax, capOymin, lowOymin, lowOymax;
SW16  capXymax, capXymin, lowXymin, lowXymax;
UW16 chId;
UW16 chId_O, chId_X, chId_o, chId_x;
static UB8 blue_txt2[] = "] def\r\n";

   chId_O = chId_X = chId_o = chId_x = 0; /* what if chars not in ss? */

   for (chId = 0; chId < MAXPS_SYMBOLS; chId++)
   {
      if (SYMmap(chId, FC_IF_TYPE) == CG_O)
         chId_O = chId;
      else if (SYMmap(chId, FC_IF_TYPE) == CG_X)
         chId_X = chId;
      else if (SYMmap(chId, FC_IF_TYPE) == CG_o)
         chId_o = chId;
      else if (SYMmap(chId, FC_IF_TYPE) == CG_x)
         chId_x = chId;
   }

   /* Have to retrieve y-bounding box values for the following
    * characters: "O", "o", "X" and "x"
	 * It is possible that some of these characters may be in the symbol
	 *	set but not in the library? 11/21/91 - rs
    */
   if (chId_O)
      if (getCharYvals(chId_O, &capOymin, &capOymax))
			chId_O = 0;
   if (chId_X)
      if (getCharYvals(chId_X, &capXymin, &capXymax))
			chId_X = 0;

   if (chId_o)
   	if (getCharYvals(chId_o, &lowOymin, &lowOymax))
			chId_o = 0;
   if (chId_x)
   	if (getCharYvals(chId_x, &lowXymin, &lowXymax))
			chId_x = 0;

   encWrite((LPUB8)"/BlueValues [");
	/*
		We must follow the official Adobe Blue Value rules. A maximum
		of 7 pairs may be written. The first pair must be the baseline
		overshoot/baseline position. All other pairs are for top-zones.
		Since this is a required member of the Private dictionary, if
		no values are to be written, an empty array must be used.
	*/

	/* bottom zone : base line */
	if (chId_O && (capOymin <= 0))	/* use upper case for bottom zone */
	{
		SPRINTF(sbuff,"%d %d ",capOymin, (SW16)0);
		encWrite(sbuff);
	}
	else if (chId_o && (lowOymin <= 0))	/* use lower case for bottom zone */
	{
		SPRINTF(sbuff,"%d %d ",lowOymin, (SW16)0);
		encWrite(sbuff);
	}
	else /* no bottom zone - just skip blue values */
	{
		encWrite(blue_txt2); /* no blues - 11/21/91 - rs */
		return(SUCCESS);
	}

	/* X-height */
	if (chId_o && chId_x && (lowOymax > lowXymax))
	{
		SPRINTF(sbuff,"%d %d ", lowXymax, lowOymax);
		encWrite(sbuff);
	}

	/*
		Cap height
		add test for >= 3 units apart from other pairs - 7/19/91 - rs
	*/
	if (chId_O && chId_X && (capOymax > capXymax) && (capOymax >= lowOymax + 3))
	{
		SPRINTF(sbuff, "%d %d ", capXymax, capOymax);
		encWrite(sbuff);
	}

	encWrite(blue_txt2);
	return(SUCCESS);
}

#if defined (ANSI_DEFS)
MLOCAL UW16  getCharYvals(UW16 chId, LPSW16 ymin, LPSW16 ymax)
#else
MLOCAL UW16 
getCharYvals(chId, ymin, ymax)
   UW16        chId;
   LPSW16        ymin;
   LPSW16        ymax;
#endif
/*
Calculates shift factor for scaling FOPS, stores in xscale & yscale.
*/
{
   HIFOUTLINE   hol;
   PIFOUTLINE   pol;
   UW16 ures;
	SW16	j;

		/*
		 * Create the character outline
		 */
		hol = NIL_MH;
		ures = cgif_Char_Handle(chId, (PHIFBITMAP)&hol, 0);
		if (ures == ERR_fixed_space)
      {
         hol = NIL_MH;
         return (ures);
      }

		if (ures)
		{
			if (hol)
				CHARfree (hol);
         hol = NIL_MH;
			DBG1("getCharYvals() - Error creating character outline %d\n",
				ures);
			return(ures);
		}

		/*
		 * Process outline data
		 */
		pol = (PIFOUTLINE)MEMptr (hol);	/* deref outline handle */


	/*
		NOTE: There are differences in values obtained for negative points
				between shifting and dividing. Be consistent (always shift).
		Determine shift count for scale factors - log base 2 of pol->#scale
	*/
	for (j = 1, xscale = 0; j < pol->xscale; j *= 2, xscale++)
		;
	for (j = 1, yscale = 0; j < pol->yscale; j *= 2, yscale++)
		;

      *ymin = (pol->bottom >> yscale);
      *ymax = (pol->top >> yscale);

		if (hol)
			CHARfree (hol);
      hol = NIL_MH;
      return (SUCCESS);
}
#else
#if ((defined (__STDC__) && (__STDC__  > 0) && !defined(__i960)) || \
    (defined(__i960) && defined(ANSI_960))) || \
    ((defined (__STDC__) && (__STDC__  > 0) && !defined(VXWORKS)) || \
      (defined(VXWORKS) && defined(ANSI_960)))
int iIFT1HINT = 0;
#endif /* __STDC__ */
#endif /* PST1 && IF_RDR */

