/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/tt/fontmath.c,v 1.2 2003/05/07 23:02:30 Lann Exp $ */
/* $Log: fontmath.c,v $
/* Revision 1.2  2003/05/07 23:02:30  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.6   07 Apr 1995 11:18:00   LISA
 * Changed copyright information from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.5   13 Jan 1995 18:09:36   ROB
 * Add support for > 64KB memory blocks in DOS. Remove code which is contained
 * in ifmem.h.
 * 
 *    Rev 1.4   22 Apr 1994 14:55:26   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.3   06 Aug 1993 17:25:20   MIKE
 * 
 * Fixed CHKOVF() macro.
 * 
 *    Rev 1.2   15 Jan 1993 09:50:14   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.1   14 Dec 1992 09:48:34   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:16:40   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:30 $ */
/*
   File:    FontMath.c

 
 

   Copyright:  1990 by Apple Computer, Inc., all rights reserved.



    Change History (most recent first):

    AGFA changes:

       06-Aug-93  mby   Fixed CHKOVF() macro to return NEGINFINITY when BIT(31)
                        is set; POSINFINITY when BIT(31) is clear.
       30-Nov-92  jfd   Port to Mac - ANSI function definitions using ANSI_DEFS
       22-Nov-92  rs    Port to Macintosh - FixMul().
       11-Nov-92  mby   Cleaned up some warnings - CompMul(), CompDiv()
       25-Jun-92  jfd   Moved "cgconfig.h" ahead of "port.h".
       27-Apr-92  mby   Dropped in new versions of FixMul(), FixDiv(),
                          FracMul(), FracDiv(), FracSqrt(). No FP.
                        NOT_ON_THE_MAC replaces NOT_ON_MOTOROLA
                        Conditionally compile function definitions with
                        LINT_ARGS. Sun doesn't like in line prototypes.
       16-Apr-92  mby   Took care of MSC 6.0 level 3 warnings.
       07-Feb-92  rs    Conditionalize entire file.
       27-Sep-91  mby   Renamed #include files to 8 chars or less.
                        Added Macintosh Toolbox functions:
                        FixMul(), FracMul(), FixDiv(), FracDiv(), FracSqrt()

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		 <3>	 11/9/90	MR		Fix CompDiv when numer and denom have zero hi longs. [rb]
		 <2>	 11/5/90	MR		Remove Types.h from include list, rename FixMulDiv to LongMulDiv
									[rb]
		 <1>	10/20/90	MR		Math routines for font scaler. [rj]

*/

#include "cgconfig.h"
#include "port.h"

#if   TT_RDR   /* conditionalize entire file */

#include "ifmem.h"   /* add 1/11/95 - rob */
#include "fscdefs.h"
#include "fontmath.h"


#ifdef LINT_ARGS
static void CompMul(long src1, long src2, long* dst);
static long CompDiv(long src1, long* src2);

#else /* !LINT_ARGS */
static void CompMul();
static long CompDiv();
#endif /* LINT_ARGS */


#if defined (ANSI_DEFS)
static void CompMul(long src1, long src2, long* dst)

#else 
static void 
CompMul(src1, src2, dst)
   long src1;
   long src2;
   long *dst;
#endif
{
	int negative = (src1 ^ src2) < 0;
	register unsigned long dsthi, dstlo;

	if (src1 < 0)
		src1 = -src1;
	if (src2 < 0)
		src2 = -src2;
	{	unsigned short src1hi, src1lo;
		register unsigned short src2hi, src2lo;
		register unsigned long temp;
		src1hi = (unsigned short)(src1 >> 16);
		src1lo = (unsigned short)(src1);
		src2hi = (unsigned short)(src2 >> 16);
		src2lo = (unsigned short)(src2);
		temp = (unsigned long)src1hi * src2lo + (unsigned long)src1lo * src2hi;
		dsthi = (unsigned long)src1hi * src2hi + (temp >> 16);
		dstlo = (unsigned long)src1lo * src2lo;
		temp <<= 16;
		dsthi += (dstlo += temp) < temp;
		dst[0] = dsthi;
		dst[1] = dstlo;
	}
	if (negative)
        {
        if ((dstlo = -((long)dstlo)) != 0)
			dsthi = ~dsthi;
		else
			dsthi = -((long)dsthi);
        }
	dst[0] = dsthi;
	dst[1] = dstlo;
}

#if defined (ANSI_DEFS)
static long CompDiv(long src1, long* src2)

#else 
static long 
CompDiv(src1, src2)
   long src1;
   long *src2;
#endif
{
	register unsigned long src2hi = src2[0], src2lo = src2[1];
	int negative = (long)(src2hi ^ src1) < 0;

	if ((long)src2hi < 0)
        {
        if ((src2lo = -((long)src2lo)) != 0)
			src2hi = ~src2hi;
		else
			src2hi = -((long)src2hi);
        }
	if (src1 < 0)
		src1 = -src1;
	{	register unsigned long src1hi, src1lo;
		unsigned long result = 0, place = 0x40000000L;

		if ((src1hi = src1) & 1)
			src1lo = 0x80000000L;
		else
			src1lo = 0;

		src1hi >>= 1;
		src2hi += (src2lo += src1hi) < src1hi;		/* round the result */

		if (src2hi > src1hi || src2hi == src1hi && src2lo >= src1lo)
			if (negative)
				return NEGINFINITY;
			else
				return POSINFINITY;
		while (place && src2hi)
		{	src1lo >>= 1;
			if (src1hi & 1)
				src1lo += 0x80000000L;
			src1hi >>= 1;
			if (src1hi < src2hi)
			{	src2hi -= src1hi;
				src2hi -= src1lo > src2lo;
				src2lo -= src1lo;
				result += place;
			}
			else if (src1hi == src2hi && src1lo <= src2lo)
			{	src2hi = 0;
				src2lo -= src1lo;
				result += place;
			}
			place >>= 1;
		}
        if (src2lo >= (unsigned long)src1)
			result += src2lo/src1;
		if (negative)
			return -((long)result);
		else
			return result;
	}
}

/*
 *	a*b/c
 */
#if defined (ANSI_DEFS)
long LongMulDiv (long a, long b, long c)

#else 
long 
LongMulDiv (a, b, c)
   long a;
   long b;
   long c;
#endif
{
	long temp[2];

	CompMul(a, b, temp);
	return CompDiv(c, temp);
}

#ifdef NOT_ON_THE_MAC

#if defined (ANSI_DEFS)
F26Dot6 ShortFracMul (F26Dot6 a, ShortFrac b)

#else 
F26Dot6 
ShortFracMul (a, b)
   F26Dot6   a;
   ShortFrac b;
#endif
{
	int negative = false;
	uint16 al, ah;
	uint32 lowlong, midlong, hilong;

	if (a < 0) { a = -a; negative = true; }
	if (b < 0) { b = -b; negative ^= true; }

	al = LOWORD(a); ah = HIWORD(a);

	midlong = USHORTMUL(ah, b);
	hilong = midlong & 0xFFFF0000L;
	midlong <<= 16;
	midlong += 1 << 13;
	lowlong = USHORTMUL(al, b) + midlong;
	if (lowlong < midlong)
		hilong += ONEFIX;

	midlong = (lowlong >> 14) | (hilong << 2);
	return negative ? -((int32)midlong) : midlong;
}

#if defined (ANSI_DEFS)
int32 ShortMulDiv(int32 a, int16 b, int16 c)

#else 
int32 
ShortMulDiv (a, b, c)
   int32 a;
   int16 b;
   int16 c;
#endif
{
	return LongMulDiv(a, (long)b, (long)c);
}

#endif      /* NOT_ON_THE_MAC */


#if defined (ANSI_DEFS)
ShortFrac ShortFracDot(ShortFrac a, ShortFrac b)

#else 
ShortFrac 
ShortFracDot (a, b)
   ShortFrac a;
   ShortFrac b;
#endif
{
	return (ShortFrac)((SHORTMUL(a,b) + (1 << 13)) >> 14);
}

#define FASTMUL26LIMIT		46340L
/*
 *	Total precision routine to multiply two 26.6 numbers		<3>
 */
#if defined (ANSI_DEFS)
F26Dot6 Mul26Dot6 (F26Dot6 a, F26Dot6 b)

#else 
F26Dot6 
Mul26Dot6 (a, b)
   F26Dot6 a;
   F26Dot6 b;
#endif
{
	int negative = false;
	uint16 al, bl, ah, bh;
	uint32 lowlong, midlong, hilong;

	if ((a <= FASTMUL26LIMIT) && (b <= FASTMUL26LIMIT) && (a >= -FASTMUL26LIMIT) && (b >= -FASTMUL26LIMIT))
        return (a * b + (1 << 5)) >> 6;     /* fast case */

	if (a < 0) { a = -a; negative = true; }
	if (b < 0) { b = -b; negative ^= true; }

	al = LOWORD(a); ah = HIWORD(a);
	bl = LOWORD(b); bh = HIWORD(b);

	midlong = USHORTMUL(al, bh) + USHORTMUL(ah, bl);
	hilong = USHORTMUL(ah, bh) + HIWORD(midlong);
	midlong <<= 16;
	midlong += 1 << 5;
	lowlong = USHORTMUL(al, bl) + midlong;
	hilong += lowlong < midlong;

	midlong = (lowlong >> 6) | (hilong << 26);
	return negative ? -((int32)midlong) : midlong;
}

#define FASTDIV26LIMIT	(1L << 25)
/*
 *	Total precision routine to divide two 26.6 numbers			<3>
 */
#if defined (ANSI_DEFS)
F26Dot6 Div26Dot6 (F26Dot6 num, F26Dot6 den)

#else 
F26Dot6 
Div26Dot6 (num, den)
   F26Dot6 num;
   F26Dot6 den;
#endif
{
	int negative = false;
	register uint32 hinum, lownum, hiden, lowden, result, place;

	if (den == 0) return (num < 0 ) ? NEGINFINITY : POSINFINITY;

	if ( (num <= FASTDIV26LIMIT) && (num >= -FASTDIV26LIMIT) )			/* fast case */
		return (num << 6) / den;

	if (num < 0) { num = -num; negative = true; }
	if (den < 0) { den = -den; negative ^= true; }

	hinum = ((uint32)num >> 26);
	lownum = ((uint32)num << 6);
	hiden = den;
	lowden = 0;
	result = 0;
	place = HIBITSET;

	if (hinum >= hiden) return negative ? NEGINFINITY : POSINFINITY;

	while (place)
	{
		lowden >>= 1;
		if (hiden & 1) lowden += HIBITSET;
		hiden >>= 1;
		if (hiden < hinum)
		{
			hinum -= hiden;
			hinum -= lowden > lownum;
			lownum -= lowden;
			result += place;
		}
		else if (hiden == hinum && lowden <= lownum)
		{
			hinum = 0;
			lownum -= lowden;
			result += place;
		}
		place >>= 1;
	}

	return negative ? -((int32)result) : result;
}




#ifndef USE_EXTERNAL_FIXMATH

/*************************************************************
 *
 *  Here we define Fixed [16.16] and Fract [2.30] precision 
 *  multiplication and division functions and a Fract square root 
 *  function which are compatible with those in the Macintosh toolbox.
 *
 *  The division functions load the 32-bit numerator into the "middle"
 *  bits of a 64-bit numerator, then call the 64-bit by 32-bit CompDiv()
 *  function defined above, which can return a NEGINFINITY or POSINFINITY
 *  overflow return code.
 *
 *  The multiply functions call the 32-bit by 32-bit CompMul() function
 *  defined above which produces a 64-bit result, then they extract the
 *  "interesting" 32-bits from the middle of the 64-bit result and test 
 *  for overflow.
 *
 *  The GET32(a,i) macro defined below extracts a 32-bit value with "i" 
 *  bits of fractional precision from the 64-bit value in "a", a 2-element
 *  array of longs.
 *
 *  The CHKOVF(a,i,v) macro tests the most significant bits of the 
 *  64-bit value in "a", a 2-element array of longs, and tests the 
 *  32-bit result "v" for overflow.  "v" is defined as having "i" bits
 *  of fractional precision.
 *
 *  BIT() and OVFMASK() are "helper" macros used by GET32() and CHKOVF().
 *
 *  BIT(i) returns a mask with the "i"-th bit set.
 *  OVFMASK(i) returns a mask with the most-significant "32-i" bits set.
 *
 *  Macintosh Toolbox Utility Routines:
 *    Fixed  FixMul( Fixed a, Fixed b );
 *    Fixed  FixDiv( Fixed a, Fixed b );
 *    Fract  FracMul( Fract a, Fract b );
 *    Fract  FracDiv( Fract a, Fract b );
 *    Fract  FracSqrt( Fract a, Fract b );
 *                                                           */
/*************************************************************/


#define BIT(i)          (1L<<(i))
#define OVFMASK(i)      ( ~0L ^ ( ((unsigned long)BIT(i)) - 1 ) )
#define CHKOVF(a,i,v)   (\
        ( ((a)[0] & OVFMASK(i))==0)          ? ( (v)>=0 ?(v) :POSINFINITY) : \
        ( ((a)[0] & OVFMASK(i))==OVFMASK(i)) ? ( (v)<=0 ?(v) :NEGINFINITY) : \
        ( ((a)[0] & BIT(31))                 ? NEGINFINITY   :POSINFINITY)   \
    )
#define GET32(a,i) \
((((a)[0]<<(32-(i))) | ((unsigned long)((a)[1])>>(i))) + !!((a)[1] & BIT((i)-1)))



#if !defined (MACINTOSH)
#if defined (ANSI_DEFS)
Fixed FIXMATH_ENTRY FixMul (Fixed fxA, Fixed fxB)

#else 
Fixed FIXMATH_ENTRY 
FixMul (fxA, fxB)
   Fixed fxA;
   Fixed fxB;
#endif
{
    long alCompProd[2];
    Fixed fxProd;

    if  (fxA == 0 || fxB == 0)
        return 0;

    CompMul (fxA, fxB, alCompProd);
    fxProd = GET32 (alCompProd,16);
    return CHKOVF(alCompProd,16,fxProd);
}
#endif /* MACINTOSH */



#if defined (ANSI_DEFS)
Fixed FIXMATH_ENTRY FixDiv (Fixed fxA, Fixed fxB)

#else 
Fixed FIXMATH_ENTRY 
FixDiv (fxA, fxB)
   Fixed fxA;
   Fixed fxB;
#endif
{
    long alCompProd[2];
    
    alCompProd[0] = fxA >> 16;
    alCompProd[1] = fxA << 16;
    return CompDiv (fxB, alCompProd);
}




#if defined (ANSI_DEFS)
Fract FIXMATH_ENTRY FracMul (Fract frA, Fract frB)

#else 
Fract FIXMATH_ENTRY 
FracMul (frA, frB)
   Fract frA;
   Fract frB;
#endif
{
    long alCompProd[2];
    Fract frProd;

    if  (frA == 0 || frB == 0)
        return 0;

    CompMul (frA,frB,alCompProd);
    frProd = GET32 (alCompProd,30);
    return CHKOVF(alCompProd,30,frProd);
}




#if defined (ANSI_DEFS)
Fract FIXMATH_ENTRY FracDiv (Fract frA, Fract frB)

#else 
Fract FIXMATH_ENTRY 
FracDiv (frA, frB)
   Fract frA;
   Fract frB;
#endif
{
    long alCompProd[2];

    alCompProd[0] = frA >> 2;
    alCompProd[1] = frA << 30;
    return CompDiv (frB, alCompProd);
}




/* 
   Fract FracSqrt (Fract xf)
   Input:  xf           2.30 fixed point value
   Return: sqrt(xf)     2.30 fixed point value
*/

#if defined (ANSI_DEFS)
Fract FIXMATH_ENTRY FracSqrt (Fract xf)

#else 
Fract FIXMATH_ENTRY 
FracSqrt (xf)
   Fract xf;
#endif
{
    Fract b = 0L;
    unsigned long c, d, x = xf;
    
    if (xf < 0) return (NEGINFINITY);

    /*
    The algorithm extracts one bit at a time, starting from the
    left, and accumulates the square root in b.  The algorithm 
    takes advantage of the fact that non-negative input values
    range from zero to just under two, and corresponding output
    ranges from zero to just under sqrt(2).  Input is assigned
    to temporary value x (unsigned) so we can use the sign bit
    for more precision.
    */
    
    if (x >= 0x40000000L)
    {
        x -= 0x40000000L;
        b  = 0x40000000L;
    }

    /*
    This is the main loop.  If we had more precision, we could 
    do everything here, but the lines above perform the first
    iteration (to align the 2.30 radix properly in b, and to 
    preserve full precision in x without overflow), and afterward 
    we do two more iterations.
    */
    
    for (c = 0x10000000L; c; c >>= 1)
    {
        d = b + c;
        if (x >= d)
        {
            x -= d; 
            b += (c<<1); 
        }
        x <<= 1;
    }

    /*
    Iteration to get last significant bit.
    
    This code has been reduced beyond recognition, but basically,
    at this point c == 1L>>1 (phantom bit on right).  We would
    like to shift x and d left 1 bit when we enter this iteration,
    instead of at the end.  That way we could get phantom bit in
    d back into the word.  Unfortunately, that may cause overflow
    in x.  The solution is to break d into b+c, subtract b from x,
    then shift x left, then subtract c<<1 (1L).
    */
    
    if (x > (unsigned long)b)   /* if (x == b) then (x < d). */
    {                           /* We want to test (x >= d). */
        x -= b;
        x <<= 1;
        x -= 1L;
        b += 1L; /* b += (c<<1) */
    }
    else
    {
        x <<= 1;
    }

    /* 
    Final iteration is simple, since we don't have to maintain x.
    We just need to calculate the bit to the right of the least
    significant bit in b, and use the result to round our final answer.
    */
    
    return ( b + (x > (unsigned long)b) );
}

#endif   /* USE_EXTERNAL_FIXMATH */

#endif   /* TT_RDR  conditionalize entire file */


