/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/tt/fontscal.c,v 1.2 2003/05/07 23:02:30 Lann Exp $ */
/* $Log: fontscal.c,v $
/* Revision 1.2  2003/05/07 23:02:30  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.18   07 Apr 1995 11:18:04   LISA
 * Changed copyright information from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.17   13 Jan 1995 18:09:46   ROB
 * Add support for > 64KB memory blocks in DOS. Remove code which is contained
 * in ifmem.h.
 * 
 *    Rev 1.16   11 Dec 1994 17:12:30   MIKE
 * 
 * Fixed 16-bit overflow in fs_FindCharBearings()
 * 
 *    Rev 1.15   22 Apr 1994 14:55:36   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.14   03 Jan 1994 12:42:54   MIKE
 * 
 * Added debugging code (#define UDEBUG).
 * 
 *    Rev 1.13   07 Oct 1993 19:10:08   MIKE
 * Got rid of fs_GetFontBBox() (data is read out from within fs_NewSfnt(). )
 * 
 *    Rev 1.12   06 Aug 1993 17:22:24   MIKE
 * jwerhviu4eaf uy-9
 * Fixed overflow problem for key->mirror_char in fs_NewTransformation().
 * 
 *    Rev 1.11   16 Jul 1993 14:53:02   MAIB
 * added function fs_FindCharBearings
 * 
 *    Rev 1.10   15 Jun 1993 10:56:36   JOE
 * Moved "vxWorks.h" before "port.h" to resolve conflicts.
 * 
 *    Rev 1.9   10 Jun 1993 08:59:02   JOE
 * Changed all references to "char" to "SB8" for portability.
 * 
 *    Rev 1.8   24 May 1993 13:33:04   JOE
 * Initialize key->glyphLength to 0 in fs__Contour().
 * 
 *    Rev 1.7   31 Mar 1993 11:17:04   MIKE
 * Conditionally compile fs_GetFontBBox(). Remove fs_GetCharBBox().d
 * 
 *    Rev 1.6   12 Feb 1993 16:24:18   JOE
 * VXWorks support.
 * 
 *    Rev 1.5   08 Feb 1993 12:12:10   MIKE
 * Add fs_GetFontBox() & fs_GetCharBBox() interface functions.
 * Change fs_GetDesignWidth() from int32 to int.
 * 
 *    Rev 1.4   14 Jan 1993 13:33:40   MIKE
 * Cleaned up some compiler warnings.
 * 
 *    Rev 1.3   15 Dec 1992 19:04:58   MIKE
 * Cleanup
 * 
 *    Rev 1.2   14 Dec 1992 18:33:20   ROB
 * Fix compile errors with casts.
 * 
 *    Rev 1.1   14 Dec 1992 09:49:34   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:16:42   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:30 $ */
/*
   File:    FontScaler.c

 
 

   Copyright:  1988-1990 by Apple Computer, Inc., all rights reserved.



    Change History (most recent first):

    AGFA changes:
    
    11-Dec-94  mby  Passed 32-bit bounding box to fs_FindCharBearings() to fix
                    16-bit overflow in calculation of "leftSideBearing" and
                    "devLeftSideBearing".
    16-Dec-93  mby  Debugging: #if defined (UDEBUG), print nonScaledAW and
                    nonScaledLSB, and compile fs_GetCharBBox()
    01-Oct-93  mby  Got rid of fs_GetFontBBox(). Instead, read out
                    bounding box from within fs_NewSfnt().
    06-Aug-93  mby  Fixed overflow problem with key->mirror_char in
                    fs_NewTransformation(). Use UFST floating point calls
                    rather than FixMul().
    16-Jul-93  maib Added function fs_FindCharBearings (based on the function
                    fs_FindBitMapSize), required to calculate leftSideBearing 
                    value later used to determine the correct character origin
    15-Jun-93  jfd  Moved "vxWorks.h" before "port.h" to resolve conflicts.
    10-Jun-93  jfd  Changed all references to "char" to "SB8" for portability.

    21-May-93  mby  Initialize key->glyphLength to 0 in fs__Contour().

    26-Mar-93  mby  Conditionally compile fs_GetFontBBox().
                    Remove fs_GetCharBBox() from compilation.

    08-Feb-93  jfd  VXWorks support.

    18-Jan-92  mby  Add fs_GetFontBBox(), fs_GetCharBBox interface functions.
                    Change fs_GetDesignWidth() from int32 to int.

    14-Jan-92  mby  Clean up some compiler warnings.

    05-Dec-92  rs   Fix compile errors - cast "outputPtr->".

    30-Nov-92  jfd  Port to Mac - ANSI function definitions using ANSI_DEFS.

    27-Oct-92  mby  Return fs_GlyphInfoType->contourPolarity in fs__Contour().
                    Check for mirror xform in fs_NewTransformation(). Set
                    "key->mirror_char".

    20-Oct-92  mby  Added casts whenever SWAPx() macro is used.

    03-Oct-92  rs   Changes toward ANSI C - pascal -> PASCAL.

    20-Jul-92  mby  Remove TT_PCLEOI conditional code in fs__Contour().
                    Now there is only 1 GetSfntGlyphPtr function.

    07-Jul-92  mby  fs_GetDesignWidth() <== fs_GetNonScaledWidth()
                    Added fastMetrics argument fs_GetNonScaledWidth().

    25-Jun-92  jfd  Moved "cgconfig.h" ahead of "port.h".

    04-Jun-92  mby  Never mind the 4/14 changes. *Remove* MEM_HANDLE
                    code. sfnt_GetTablePtr() returns a pointer again.
                    Add fs_GetTablePtr() to interface.
                    In fs__Contour() set up function pointers
                    key->GetHSfntGlyph, RelHSfntGlyph, GetMSfntGlyph

    29-Apr-92  mby  Conditionally compile function protos for LINT_ARGS.

    14-Apr-92  mby  "Depointerize" code. Make allocated memory
                    handle-based rather than pointer-based.

    09-Apr-92  mby  Until/Unless we decide to support it, I removed
                     outline caching -- #ifdef USE_OUTLINE_CACHE &
                     functions fs_SizeOfOutlines(), fs_SaveOutlines(),
                     and fs_RestoreOutlines()
                    Removed fs_SetSplineDataPtrs() & fs_SetInternalOffsets()

    24-Feb-92  jfd  Added dummy macro for Debugger().

    10-Feb-92  mby  #ifdef UFST_SUBSYSTEM ==>
                    don't need fs_FindBitmapSize(),
                    fs_ContourScan(),and fs_GetAdvanceWidth(),
                    fs_45DegreePhaseShift(), fs_dropOutVal()

    07-Feb-92  rs   Conditionalize entire file.

    31-Jan-92  mby  (UFST) fs_NewSfnt returns emResolution in outputPtr

    28-Jan-92  mby  (UFST) Added new Interface function,
                    fs_GetNonScaledWidth(). Provides escapement
                    in character design units.

    08-Oct-91  mby  Renamed #include files.
                    Client interface functions must be "int32 pascal"
                      instead of "pascal int32".

TO DO: If Intel byte-order, swap bytes in key->maxProfile table.

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		<13>	12/20/90	RB		Add error return to NewGlyph if glyph index is out of range.
									[mr]
		<12>	12/10/90	RB		Fix findbitmapsize to correctly return error when point has
									migrated outside -32768,32767 range. [cel]
		<11>	11/27/90	MR		Need two scalars: one for (possibly rounded) outlines and cvt,
									and one (always fractional) metrics. [rb]
		<10>	11/21/90	RB		Allow client to disable DropOutControl by returning a NIL
									pointer to memoryarea[7]. Also make it clear that we inhibit
									DOControl whenever we band. [This is a reversion to 8, so mr's
									initials are added by proxy]
		 <9>	11/13/90	MR		(dnf) Revert back to revision 7 to fix a memmory-trashing bug
									(we hope). Also fix signed/unsigned comparison bug in outline
									caching.
		 <8>	11/13/90	RB		Fix banding so that we can band down to one row, using only
									enough bitmap memory and auxillary memory for one row.[mr]
		 <7>	 11/9/90	MR		Add Default return to fs_dropoutval. Continue to fiddle with
									banding. [rb]
		 <6>	 11/5/90	MR		Remove FixMath.h from include list. Clean up Stamp macros. [rb]
		 <5>	10/31/90	MR		Conditionalize call to ComputeMapping (to avoid linking
									MapString) [ha]
		 <4>	10/31/90	MR		Add bit-field option for integer or fractional scaling [rb]
		 <3>	10/30/90	RB		[MR] Inhibit DropOutControl when Banding
		 <2>	10/20/90	MR		Restore changes since project died. Converting to smart math
									routines, integer ppem scaling. [rb]
		<16>	 7/26/90	MR		dont include ToolUtils.h
		<15>	 7/18/90	MR		Fix return bug in GetAdvanceWidth, internal errors are now ints.
		<14>	 7/14/90	MR		remove unused fields from FSInfo
		<13>	 7/13/90	MR		Ansi-C fixes, rev. for union in FSInput
		<11>	 6/29/90	RB		Thus endeth the too long life of encryption
		<10>	 6/21/90	MR		Add calls to ReleaseSfntFrag
		 <9>	 6/21/90	RB		add scanKind info to fs_dropoutVal
		 <8>	  6/5/90	MR		remove fs_MapCharCodes
		 <7>	  6/1/90	MR		Did someone say MVT? Yuck!!! Out, damn routine.
		 <6>	  6/1/90	RB		fixed bandingbug under dropout control
		 <4>	  5/3/90	RB		added dropoutval function.  simplified restore outlines.
									support for new scanconverter in contourscan, findbitmapsize,
									saveoutlines, restoreoutlines.
		 <3>	 3/20/90	CL		Changed to use fpem (16.16) instead of pixelsPerEm (int) Removed
									call to AdjustTransformation (not needed with fpem) Added call
									to RunXFormPgm Removed WECANNOTDOTHIS #ifdef Added
									fs_MapCharCodes
		 <2>	 2/27/90	CL		New error code for missing but needed table. (0x1409 ).  New
									CharToIndexMap Table format.
									Fixed transformed component bug.
	   <3.6>	11/15/89	CEL		Put an else for the ifdef WeCanNotDoThis so Printer compile
									could use more effecient code.
	   <3.5>	11/14/89	CEL		Left Side Bearing should work right for any transformation. The
									phantom points are in, even for components in a composite glyph.
									They should also work for transformations. Device metric are
									passed out in the output data structure. This should also work
									with transformations. Another leftsidebearing along the advance
									width vector is also passed out. whatever the metrics are for
									the component at it's level. Instructions are legal in
									components. Instructions are legal in components. The
									transformation is internally automatically normalized. This
									should also solve the overflow problem we had. Now it is legal
									to pass in zero as the address of memory when a piece of the
									sfnt is requested by the scaler. If this happens the scaler will
									simply exit with an error code ! Five unnecessary element in the
									output data structure have been deleted. (All the information is
									passed out in the bitmap data structure) fs_FindBMSize now also
									returns the bounding box.
	   <3.4>	 9/28/89	CEL		fs_newglyph did not initialize the output error. Caused routine
									to return error from previous routines.
	   <3.3>	 9/27/89	CEL		Took out devAdvanceWidth & devLeftSideBearing.
	   <3.2>	 9/25/89	CEL		Changed the NEED_PROTOTYPE ifdef to use the NOT_ON_THE_MAC flag
									that existed previously.
	   <3.1>	 9/15/89	CEL		Changed dispatch scheme Calling conventions through a trap
									needed to match Macintosh pascal. Pascal can not call C unless
									there is extra mucky glue. Bug that caused text not to appear
									The font scaler state was set up correctly but the sfnt was
									purged. It was reloaded and the clientid changed but was still
									the same font. Under the rules of the FontScaler fs_newsfnt
									should not have to be called again to reset the state. The extra
									checks sent back a BAD_CLIENTID_ERROR so QuickDraw would think
									it was a bad font and not continue to draw.
	   <3.0>	 8/28/89	sjk		Cleanup and one transformation bugfix
	   <2.4>	 8/17/89	sjk		Coded around MPW C3.0 bug
	   <2.3>	 8/14/89	sjk		1 point contours now OK
	   <2.2>	  8/8/89	sjk		Improved encryption handling
	   <2.1>	  8/2/89	sjk		Fixed outline caching bug
	   <2.0>	  8/2/89	sjk		Just fixed EASE comment
	   <1.5>	  8/1/89	sjk		Added composites and encryption. Plus some enhancements
	   <1.4>	 6/13/89	SJK		Comment
	   <1.3>	  6/2/89	CEL		16.16 scaling of metrics, minimum recommended ppem, point size 0
									bug, correct transformed integralized ppem behavior, pretty much
									so
	   <1.2>	 5/26/89	CEL		EASE messed up on c comments
	  <1.1>	 5/26/89	CEL		Integrated the new Font Scaler 1.0 into Spline Fonts
	   <1.0>	 5/25/89	CEL		Integrated 1.0 Font scaler into Bass code for the first time

	To Do:
*/
/*		<3+>	 3/20/90	mrr		Conditionalized error checking in fs_SetUpKey.
									Compiler option for stamping memmory areas for debugging
									Removed error field from FSInfo structure.
									Added call to RunFontProgram
									Added private function prototypes.
									Optimizations from diet clinic

*/

#include "cgconfig.h"

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#ifdef UDEBUG
#include <stdio.h>
#endif
#include "port.h"
#include "fpmath.h"

#if   TT_RDR   /* conditionalize entire file */

#include "setjmp.h"

/** FontScalers Includes **/
#include "ifmem.h"   /* add 1/11/95 - rob */
#include "fserror.h"
#include "fscdefs.h"
#include "fontmath.h"
#include "sfnt.h"
#include "sc.h"
#include "fnt.h"
#include "fontscal.h"
#include "fsglue.h"
#include "privsfnt.h"

#define LOOPDOWN(n)		for (--n; n >= 0; --n)

#define OUTLINEBIT    0x02

#define SETJUMP(key, error) if ( (error = setjmp(key->env)) != 0 ) return( error )

#ifdef SEGMENT_LINK
#pragma segment FONTSCALER_C
#endif


/* PRIVATE PROTOTYPES */

#ifdef LINT_ARGS
fsg_SplineKey* fs_SetUpKey(fs_GlyphInputType* inptr, unsigned stateBits, int* error);
static int32 fs__Contour(fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr, boolean useHints); 
#ifndef UFST_SUBSYSTEM
void fs_45DegreePhaseShift (sc_CharDataType *glyphPtr);
int32 fs_dropOutVal (fsg_SplineKey *key);
#endif              /* not UFST_SUBSYSTEM */

#else /* !LINT_ARGS */

fsg_SplineKey* fs_SetUpKey();
static int32 fs__Contour();
#ifndef UFST_SUBSYSTEM
void fs_45DegreePhaseShift();
int32 fs_dropOutVal();
#endif              /* not UFST_SUBSYSTEM */
#endif /* LINT_ARGS */

#ifdef	DEBUGSTAMP

#define STAMPEXTRA              4
#define STAMP                   'sfnt'

	void SETSTAMP(Ptr p)
	{
		*((int32*)((p) - STAMPEXTRA)) = STAMP;
	}

#define Debugger()

#if defined (ANSI_DEFS)
	void CHECKSTAMP(Ptr p)
#else 
    void CHECKSTAMP (p)
      Ptr p;
#endif
	{
		if (*((int32*)((p) - STAMPEXTRA)) != STAMP) Debugger();
	}

#else  /* !DEBUGSTAMP */

#define STAMPEXTRA      0
#define SETSTAMP(p)
#define CHECKSTAMP(p)

#endif  /* DEBUGSTAMP */


#ifdef LINT_ARGS
void dummyReleaseSfntFrag (void);
void dummyReleaseSfntFrag (void)

#else /* !LINT_ARGS */
void dummyReleaseSfntFrag ();
void dummyReleaseSfntFrag ()
#endif /* LINT_ARGS */
{
#ifndef NOT_ON_THE_MAC
#pragma unused(data)
#endif
}

/*
 *	Set up the key in case memmory has moved or been purged.
 */
#if defined (ANSI_DEFS)
fsg_SplineKey* fs_SetUpKey(register fs_GlyphInputType* inptr, register unsigned stateBits, int* error)

#else 
fsg_SplineKey* fs_SetUpKey(inptr, stateBits, error)
   register fs_GlyphInputType *inptr;
   register unsigned           stateBits;
   int                        *error;
#endif
{
	register fsg_SplineKey* key;

#ifdef DEBUG		/* <4> */
    if ( !(key = (fsg_SplineKey *)MEMptr(inptr->memoryBases[KEY_PTR_BASE])) )
	{
		*error = NULL_KEY_ERR;
		return 0;
	}
	if ( !(key->memoryBases = inptr->memoryBases) )
	{
		*error = NULL_MEMORY_BASES_ERR;
		return 0;
	}
	if ( !(key->GetSfntFragmentPtr = inptr->GetSfntFragmentPtr) )
	{
		*error = NULL_SFNT_FRAG_PTR_ERR;
		return 0;
	}
#else
    key = (fsg_SplineKey *)MEMptr(inptr->memoryBases[KEY_PTR_BASE]);
	key->memoryBases = inptr->memoryBases;
	key->GetSfntFragmentPtr = inptr->GetSfntFragmentPtr;
#endif

#ifdef RELEASE_MEM_FRAG
	if (inptr->ReleaseSfntFrag)
		key->ReleaseSfntFrag = inptr->ReleaseSfntFrag;
	else
		key->ReleaseSfntFrag = dummyReleaseSfntFrag;
#endif

    if ( ((unsigned)key->state & stateBits) != stateBits )
	{
		*error = OUT_OFF_SEQUENCE_CALL_ERR;
		return 0;
	}
	key->clientID = inptr->clientID;
	*error = NO_ERR;

	return key;
}

/*** FONT SCALER INTERFACE ***/

/*
 *
 */

#if defined (ANSI_DEFS)
int32 PASCAL fs_OpenFonts( fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_OpenFonts( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	if ( outputPtr ) {
		outputPtr->memorySizes[KEY_PTR_BASE]  			= fsg_KeySize( ) + STAMPEXTRA;
		outputPtr->memorySizes[VOID_FUNC_PTR_BASE]  	= fsg_InterPreterDataSize() + STAMPEXTRA;
		outputPtr->memorySizes[SCAN_PTR_BASE]   		= fsg_ScanDataSize() + STAMPEXTRA;
		outputPtr->memorySizes[WORK_SPACE_BASE]  		= 0; /* we need the sfnt for this */
		outputPtr->memorySizes[PRIVATE_FONT_SPACE_BASE]	= 0; /* we need the sfnt for this */
		outputPtr->memorySizes[BITMAP_PTR_1]  			= 0; /* we need the grid fitted outline for this */
		outputPtr->memorySizes[BITMAP_PTR_2]  			= 0; /* we need the grid fitted outline for this */
		outputPtr->memorySizes[BITMAP_PTR_3]  			= 0; /* we need the grid fitted outline for this */
	} else
		return NULL_OUTPUT_PTR_ERR;
	if ( inputPtr ) {
		inputPtr->memoryBases[KEY_PTR_BASE]  			= 0;
		inputPtr->memoryBases[VOID_FUNC_PTR_BASE]  		= 0;
		inputPtr->memoryBases[SCAN_PTR_BASE]   			= 0;
		inputPtr->memoryBases[WORK_SPACE_BASE]  		= 0;
		inputPtr->memoryBases[PRIVATE_FONT_SPACE_BASE]	= 0;
		inputPtr->memoryBases[BITMAP_PTR_1]  			= 0;
		inputPtr->memoryBases[BITMAP_PTR_2]  			= 0;
		inputPtr->memoryBases[BITMAP_PTR_3]  			= 0;
	} else
		return NULL_INPUT_PTR_ERR;
	return NO_ERR;
}

#if defined (ANSI_DEFS)
int32 PASCAL fs_Initialize( fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_Initialize( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	register fsg_SplineKey 			*key;
	fnt_GlobalGraphicStateType		tmpGS;

	if ( outputPtr )
	{
		if ( !inputPtr ) { return NULL_INPUT_PTR_ERR; }

		SETSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
		SETSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
		SETSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);

        if ( !(key = (fsg_SplineKey *)MEMptr(inputPtr->memoryBases[KEY_PTR_BASE])) )
            { return NULL_KEY_ERR; }
		key->memoryBases = inputPtr->memoryBases;
        if ( (tmpGS.function = (FntFunc*)MEMptr(key->memoryBases[VOID_FUNC_PTR_BASE])) != 0 )
        {
			fnt_Init( &tmpGS );
            key->state = (uint16)INITIALIZED;
		} else
			return VOID_FUNC_PTR_BASE_ERR;

		CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
		CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
		CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);

		return NO_ERR;
	}
	return NULL_OUTPUT_PTR_ERR;
}

/*
 * This guy asks for memory for points, instructions, fdefs and idefs.
 */
#if defined (ANSI_DEFS)
int32 PASCAL fs_NewSfnt(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_NewSfnt( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	int error;
	register fsg_SplineKey* key = fs_SetUpKey(inputPtr, INITIALIZED, &error);
	if (!key) return error;
	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);

    key->extern_font = inputPtr->param.newsfnt.extern_font;
	sfnt_DoOffsetTableMap( key );					/* Map offset and length table */

	{
		sfnt_FontHeader* fontHead = (sfnt_FontHeader*)sfnt_GetTablePtr( key, sfnt_fontHeader, true );
		sfnt_HorizontalHeader* horiHead = (sfnt_HorizontalHeader*)sfnt_GetTablePtr( key, sfnt_horiHeader, true );

        if ( fontHead->magicNumber != (uint32)SWAPL(SFNT_MAGIC) )
            return BAD_MAGIC_ERR;
        key->emResolution               = (uint16)SWAPW(fontHead->unitsPerEm);
        outputPtr->emResolution         = key->emResolution;
        key->fontFlags                  = (uint16)SWAPW(fontHead->flags);
        key->numberOf_LongHorMetrics    = (uint16)SWAPW(horiHead->numberOf_LongHorMetrics);
        key->indexToLocFormat           = (int16)SWAPW(fontHead->indexToLocFormat);
#if FONTBBOX
        outputPtr->bboxInfo.xMin = (int16)(SWAPW(fontHead->xMin));
        outputPtr->bboxInfo.yMin = (int16)(SWAPW(fontHead->yMin));
        outputPtr->bboxInfo.xMax = (int16)(SWAPW(fontHead->xMax));
        outputPtr->bboxInfo.yMax = (int16)(SWAPW(fontHead->yMax));
#endif

		RELEASESFNTFRAG(key, horiHead);
		RELEASESFNTFRAG(key, fontHead);
	}
	{
		void* p = sfnt_GetTablePtr( key, sfnt_maxProfile, true );
		key->maxProfile = *((sfnt_maxProfileTable*)p);
		RELEASESFNTFRAG(key, p);
	}

	outputPtr->memorySizes[PRIVATE_FONT_SPACE_BASE]	= fsg_PrivateFontSpaceSize( key ) + STAMPEXTRA;
	outputPtr->memorySizes[WORK_SPACE_BASE] 		= fsg_WorkSpaceSetOffsets( key ) + STAMPEXTRA;

#ifndef NEVERCOMPUTEMAPPING
    if ( (error = sfnt_ComputeMapping( key, inputPtr->param.newsfnt.platformID, inputPtr->param.newsfnt.specificID)) != 0 )
        return error;
#endif

    key->state = (uint16)(INITIALIZED | NEWSFNT);
	key->scanControl = 0;
		
	/*
	 *	Can't run font program yet, we don't have any memmory for the graphic state
	 *	Mark it to be run in NewTransformation.
	 */
	key->executeFontPgm = true;

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);

	return NO_ERR;
}

#if defined (ANSI_DEFS)
int32 PASCAL fs_NewTransformation( fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_NewTransformation( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	register fsg_SplineKey 		*key;
	int16						resolution;
	int							error;
    Fixed                       *ctm;

	if (!(key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT, &error)))
		return error;
	SETJUMP(key, error);
			
	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);

	SETSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	SETSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

    key->currentTMatrix = * ((transMatrix *) MEMptr(inputPtr->param.newtrans.transformMatrix));
    ctm = (Fixed *)key->currentTMatrix.transform;
    if (fpisneg( fpsub( fpmul( fplongexp(ctm[0],0), fplongexp(ctm[4],0) ),
                        fpmul( fplongexp(ctm[1],0), fplongexp(ctm[3],0) ) ) ))
        key->mirror_char = 1;
    else
        key->mirror_char = 0;
/*    key->mirror_char = ( FixMul(ctm[0],ctm[4]) - FixMul(ctm[1],ctm[3]) < 0 ); */
	key->fixedPointSize = inputPtr->param.newtrans.pointSize;
	key->pixelDiameter	= inputPtr->param.newtrans.pixelDiameter;

	if ( inputPtr->param.newtrans.xResolution >= inputPtr->param.newtrans.yResolution )
	{
		short yRes = inputPtr->param.newtrans.yResolution;
		resolution = inputPtr->param.newtrans.xResolution;
		if ( resolution != yRes )
		{
			key->currentTMatrix.transform[0][1] = ShortMulDiv( key->currentTMatrix.transform[0][1], yRes, resolution );
			key->currentTMatrix.transform[1][1] = ShortMulDiv( key->currentTMatrix.transform[1][1], yRes, resolution );
			key->currentTMatrix.transform[2][1] = ShortMulDiv( key->currentTMatrix.transform[2][1], yRes, resolution );
		}
	}
	else
	{
		short xRes = inputPtr->param.newtrans.xResolution;
		resolution = inputPtr->param.newtrans.yResolution;
		key->currentTMatrix.transform[0][0] = ShortMulDiv( key->currentTMatrix.transform[0][0], xRes, resolution );
		key->currentTMatrix.transform[1][0] = ShortMulDiv( key->currentTMatrix.transform[1][0], xRes, resolution );
		key->currentTMatrix.transform[2][0] = ShortMulDiv( key->currentTMatrix.transform[2][0], xRes, resolution );
	}
	key->metricScalar = ShortMulDiv( key->fixedPointSize, resolution, POINTSPERINCH );
	/*
	 *	Modifies key->fpem and key->currentTMatrix.
	 */
	fsg_ReduceMatrix( key );

	if (key->fontFlags & USE_INTEGER_SCALING)
		key->interpScalar = (key->metricScalar + 0x8000L) & 0xffff0000L;
	else
		key->interpScalar = key->metricScalar;

	/****************************************************************************
	 *	At this point, we have 													*
	 *		fixedPointSize = user defined fixed									*
	 *		metricScalar   = fixed scaler for scaling metrics					*
	 *		interpScalar   = fixed scaler for scaling outlines/CVT  			*
	 *		pixelDiameter  = user defined fixed									*
	 *		currentTMatrix = 3x3 user transform and non-squareness resolution	*
	 ****************************************************************************/

	fsg_InitInterpreterTrans( key  );		/* get premultipliers if any, also called in sfnt_ReadSFNT */

	/*
	 *	This guy defines FDEFs and IDEFs.  The table is optional
	 */
	if (key->executeFontPgm)
	{
        if ( (error = fsg_SetDefaults( key )) != 0 )
            return error;
        if ( (error = fsg_RunFontProgram(key, inputPtr->param.newtrans.traceFunc)) != 0 )
            return error;
		key->executeFontPgm = false;
	}

	if ( inputPtr->param.newtrans.traceFunc )
	{
		/* Do this now so we do not confuse font editors */
		/* Run the pre program and scale the control value table */
		/* Sets key->executePrePgm to false */
        if ( (error = fsg_NewTransformation( key, inputPtr->param.newtrans.traceFunc )) != 0 )
            return error;
	}
	else
	{
		/* otherwise postpone as long as possible */
		key->executePrePgm = true;
	}
    key->state = (uint16)(INITIALIZED | NEWSFNT | NEWTRANS);

    outputPtr->scaledCVT = (F26Dot6 *)
            (MEMptr (key->memoryBases[PRIVATE_FONT_SPACE_BASE]) + key->offset_controlValues);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	return NO_ERR;
}



/*
 * Compute the glyph index from the character code.
 */
#if defined (ANSI_DEFS)
int32 PASCAL fs_NewGlyph(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_NewGlyph( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	int error;
	register fsg_SplineKey* key = fs_SetUpKey(inputPtr, 0, &error);
	if (!key) return error;
	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

    key->state = (uint16)(INITIALIZED | NEWSFNT | NEWTRANS);  /* clear all other bits */

	if ( inputPtr->param.newglyph.characterCode != NONVALID )
	{
		uint16 glyphIndex = key->mappingF( key, inputPtr->param.newglyph.characterCode );
		outputPtr->glyphIndex = glyphIndex;
		key->glyphIndex = glyphIndex;
		outputPtr->numberOfBytesTaken = key->numberOfBytesTaken;
	}
	else
	{
		key->glyphIndex = outputPtr->glyphIndex = inputPtr->param.newglyph.glyphIndex;
		outputPtr->numberOfBytesTaken = 0;
	}

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);
    if( (uint16)key->glyphIndex > (uint16)SWAPW(key->maxProfile.numGlyphs) )
        return INVALID_GLYPH_INDEX;
	return NO_ERR;
}


#ifndef UFST_SUBSYSTEM

/*
 * this call is optional.
 *
 * can be called right after fs_NewGlyph()
 */
#if defined (ANSI_DEFS)
int32 PASCAL fs_GetAdvanceWidth(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_GetAdvanceWidth( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	int error;
	register fsg_SplineKey*	key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS, &error);
	if (!key) return error;
	SETJUMP(key, error);

	sfnt_ReadSFNTMetrics( key, key->glyphIndex );
	
	outputPtr->metricInfo.advanceWidth.x = ShortMulDiv( key->metricScalar, key->nonScaledAW, key->emResolution );
	outputPtr->metricInfo.advanceWidth.y = 0;

	if ( ! key->identityTransformation )
		fsg_FixXYMul( &outputPtr->metricInfo.advanceWidth.x, &outputPtr->metricInfo.advanceWidth.y,
				   &key->currentTMatrix );

	return NO_ERR;
}

#endif              /* not UFST_SUBSYSTEM */


/*
 *** FS_GETDESIGNWIDTH (optional) ***
 *
 * Returns the escapement in character design units (required by UFST)
 *
 * If fastMetrics is 0, we get the hmtx table and look up the width
 *    using key->glyphIndex
 * If fastMetrics is set, the width is assumed to be in key->nonScaledAW
 *
 * Calling Sequence: fs_NewGlyph has to be called first (sets key->glyphIndex)
 *                   Use fastMetrics = 1 only if fs__Contour() or
 *                   fs_GetAdvancedWidth() have been called first.
 */

#if defined (ANSI_DEFS)
int  fs_GetDesignWidth (fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr,
                        int fastMetrics)

#else 
int  fs_GetDesignWidth (inputPtr, outputPtr, fastMetrics)
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
   int  fastMetrics;
#endif
{
int  error;

register fsg_SplineKey* key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS, &error);
if (!key) return error;
SETJUMP (key, error);

if (fastMetrics)
    outputPtr->metricInfo.escapement = key->nonScaledAW;
else
    {
    register sfnt_HorizontalMetrics  *horizMetricPtr;
    register unsigned int  numberOf_LongHorMetrics = key->numberOf_LongHorMetrics;
    register unsigned int  glyphIndex = key->glyphIndex;
    horizMetricPtr = (sfnt_HorizontalMetrics *)sfnt_GetTablePtr( key, sfnt_horizontalMetrics, true );
    if (glyphIndex < numberOf_LongHorMetrics)
        outputPtr->metricInfo.escapement =
                (uint16)SWAPW(horizMetricPtr[glyphIndex].advanceWidth);
    else
        outputPtr->metricInfo.escapement =
                (uint16)SWAPW(horizMetricPtr[numberOf_LongHorMetrics-1].advanceWidth);
    }
return NO_ERR;
}




#if 0       /* This routine was reorganized out of existence - 10-01-93 mby
             * The font bbox is now read by fs_NewSfnt() and stored
             * in the font bucket by ttload_font()
             */
/* #if FONTBBOX */
/*
 *  fs_GetFontBBox (fs_GlyphInputType *inPtr, fs_GlyphInfoType  *outPtr)
 *
 *  Returns the font bounding box in design units in outPtr->bboxInfo
 */
#if defined (ANSI_DEFS)
int  fs_GetFontBBox( fs_GlyphInputType *inPtr, fs_GlyphInfoType *outPtr )

#else
int  fs_GetFontBBox( inPtr, outPtr )
   fs_GlyphInputType *inPtr;
   fs_GlyphInfoType  *outPtr;
#endif
{
    sfnt_FontHeader* fontHead;
    int  error;
    register fsg_SplineKey* key = fs_SetUpKey(inPtr, INITIALIZED | NEWSFNT, &error);
    if (!key) return error;
    SETJUMP (key, error);

    fontHead = (sfnt_FontHeader*)sfnt_GetTablePtr( key, sfnt_fontHeader, true );
    outPtr->bboxInfo.xMin = (int16)(SWAPW(fontHead->xMin));
    outPtr->bboxInfo.yMin = (int16)(SWAPW(fontHead->yMin));
    outPtr->bboxInfo.xMax = (int16)(SWAPW(fontHead->xMax));
    outPtr->bboxInfo.yMax = (int16)(SWAPW(fontHead->yMax));
    RELEASESFNTFRAG(key, fontHead);
    return NO_ERR;
}
#endif  /* FONTBBOX */




#ifdef UDEBUG   /* not in use */
/*
 *  fs_GetCharBBox( fs_GlyphInputType *inPtr,
 *                  fs_GlyphInfoType  *outPtr )
 *
 *  Returns the character bounding box in design units in outPtr->bboxInfo
 *
 *  Calling sequence: must call fs_NewGlyph() first in order to set key->glyphIndex.
 */
#if defined (ANSI_DEFS)
int  fs_GetCharBBox( fs_GlyphInputType *inPtr, fs_GlyphInfoType *outPtr )

#else
int  fs_GetCharBBox( inPtr, outPtr )
   fs_GlyphInputType *inPtr;
   fs_GlyphInfoType  *outPtr;
#endif
{
    MEM_HANDLE  gHndl;
    voidPtr glyphDataPtr;
    uint32   length, offset;
    int     result;

    int  error;
    register fsg_SplineKey* key = fs_SetUpKey(inPtr, INITIALIZED | NEWSFNT | NEWTRANS, &error);
    if (!key) return error;
    SETJUMP (key, error);

    key->GetSfntGlyphPtr = inPtr->GetSfntGlyphPtr;
    key->RelSfntGlyphPtr = inPtr->RelSfntGlyphPtr;
    if (!key->extern_font)
        sfnt_GetGlyphLocation( key, key->glyphIndex, &offset, &length );
    glyphDataPtr = key->GetSfntGlyphPtr( key->clientID, key->glyphIndex,
                                         offset, &length, &gHndl );
    if (!glyphDataPtr)
        return FSE_GET_GLYPH_FAILURE;
    result = NO_ERR;
    if ( length > 0 )
    {
        int16* shortP = (int16*)((uint8*)glyphDataPtr + 2);
        outPtr->bboxInfo.xMin  = (int16)SWAPWINC(shortP);
        outPtr->bboxInfo.yMin  = (int16)SWAPWINC(shortP);
        outPtr->bboxInfo.xMax  = (int16)SWAPWINC(shortP);
        outPtr->bboxInfo.yMax  = (int16)SWAPWINC(shortP);
    }
    else
        result = FSE_NO_CHAR_DATA;

#ifdef RELEASE_MEM_FRAG
    key->RelSfntGlyphPtr(gHndl);
#endif
    return result;
}
#endif  /* UDEBUG */



/*
 *** FS_GETTABLEPTR ***
 *
 * Allows routines outside the font scaler core to get at sfnt tables.
 * Returns pointer if table exists.
 */
#if defined (ANSI_DEFS)
void*  fs_GetTablePtr (fs_GlyphInputType *inputPtr, int sfnt_tablenum)

#else 
void*  fs_GetTablePtr (inputPtr, sfnt_tablenum)
  fs_GlyphInputType *inputPtr;
  int sfnt_tablenum;
#endif
{
int  error;
register fsg_SplineKey* key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT, &error);
if (!key) return 0;
if ( (error = setjmp(key->env)) != 0 )
    return( 0 );

return ( sfnt_GetTablePtr( key, sfnt_tablenum, false ) );
}





#if defined (ANSI_DEFS)
static int32 fs__Contour(fs_GlyphInputType* inputPtr, fs_GlyphInfoType* outputPtr, boolean useHints )

#else 
static int32 fs__Contour( inputPtr, outputPtr, useHints )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
   boolean            useHints;
#endif
{
	register fsg_SplineKey 		*key;
	int error;

	if (!(key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS, &error)))
		return error;
	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	/* potentially do delayed pre program execution */
	if ( key->executePrePgm ) {
		/* Run the pre program and scale the control value table */
        if ( (error = fsg_NewTransformation( key, 0L )) != 0 )
			return error;
	}
	
	key->outputPtr 		= outputPtr;
	key->styleFunc		= inputPtr->param.gridfit.styleFunc;
    key->GetSfntGlyphPtr = inputPtr->GetSfntGlyphPtr;
    key->RelSfntGlyphPtr = inputPtr->RelSfntGlyphPtr;
    key->glyphLength = 0;       /* initialized - mby, 5/21/93 */

    /* THE CALL */
    if ( (error = fsg_GridFit( key, inputPtr->param.gridfit.traceFunc, useHints)) != 0 )
		return error;

	{
        int8* workSpacePtr      = MEMptr(key->memoryBases[WORK_SPACE_BASE]);
		fsg_OffsetInfo* offsetPtr	= &(key->elementInfoRec.offsets[1]);

		outputPtr->xPtr		= (F26Dot6 *)(workSpacePtr + offsetPtr->newXOffset);
		outputPtr->yPtr		= (F26Dot6 *)(workSpacePtr + offsetPtr->newYOffset);
		outputPtr->startPtr	= (int16 *)(workSpacePtr + offsetPtr->startPointOffset);
		outputPtr->endPtr	= (int16 *)(workSpacePtr + offsetPtr->endPointOffset);
		outputPtr->onCurve	= (uint8 *)(workSpacePtr + offsetPtr->onCurveOffset);
        outputPtr->contourPolarity = (uint8 *)(workSpacePtr + offsetPtr->contourPolarityOffset);
	}
	outputPtr->numberOfContours	= key->elementInfoRec.interpreterElements[GLYPHELEMENT].nc;
    outputPtr->scaledCVT = (F26Dot6 *)
            (MEMptr (key->memoryBases[PRIVATE_FONT_SPACE_BASE]) + key->offset_controlValues);

	outputPtr->outlinesExist = (uint16)(key->glyphLength != 0);

    key->state = (uint16)(INITIALIZED | NEWSFNT | NEWTRANS | GOTGLYPH);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	return NO_ERR;
}

#if defined (ANSI_DEFS)
int32 PASCAL fs_ContourNoGridFit( fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_ContourNoGridFit( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	return fs__Contour( inputPtr, outputPtr, false );
}

#if defined (ANSI_DEFS)
int32 PASCAL fs_ContourGridFit(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_ContourGridFit( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	return fs__Contour( inputPtr, outputPtr, true );
}

#if defined (ANSI_DEFS)
int32 PASCAL fs_FindCharBearings( fs_GlyphInputType* inputPtr,
                    fs_GlyphInfoType* outputPtr,
                    int32 xMin, int32 yMin, int32 xMax, int32 yMax )
#else 
int32 PASCAL fs_FindCharBearings( inputPtr, outputPtr, xMin, yMin, xMax, yMax )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
   int32  xMin, yMin, xMax, yMax;
#endif
{
	register fsg_SplineKey *key;
	register fnt_ElementType *elementPtr;
	register int8	*workSpacePtr;
	register fsg_OffsetInfo	*offsetPtr;
	sc_CharDataType	charData;
	int	numPts, error;
	register int32 tmp32;

	if (!(key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS | GOTGLYPH, &error)))
		return error;

	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

#ifdef UDEBUG
printf("  nonScaledAW: %u  nonScaledLSB: %d\n", key->nonScaledAW, key->nonScaledLSB);
#endif

	key->outlineIsCached = false;

	elementPtr			= &(key->elementInfoRec.interpreterElements[GLYPHELEMENT]);
  workSpacePtr    = MEMptr(key->memoryBases[WORK_SPACE_BASE]);
	offsetPtr			= &(key->elementInfoRec.offsets[GLYPHELEMENT]);

	charData.x			= (int32 *)(workSpacePtr + offsetPtr->newXOffset);
	charData.y			= (int32 *)(workSpacePtr + offsetPtr->newYOffset);
	charData.ctrs		= elementPtr->nc;
	charData.sp			= (int16 *)(workSpacePtr + offsetPtr->startPointOffset);
	charData.ep			= (int16 *)(workSpacePtr + offsetPtr->endPointOffset);
	charData.onC		= (int8 *)(workSpacePtr + offsetPtr->onCurveOffset);
	numPts = charData.ep[charData.ctrs-1] + 1 + PHANTOMCOUNT;
	{
		register int ten = 10;
		register metricsType* metric = &outputPtr->metricInfo;
		register int index1 = numPts-PHANTOMCOUNT+RIGHTSIDEBEARING;
		register int index2 = numPts-PHANTOMCOUNT+LEFTSIDEBEARING;

		metric->devAdvanceWidth.x = (charData.x[index1] - charData.x[index2]) << ten;
		metric->devAdvanceWidth.y = (charData.y[index1] - charData.y[index2]) << ten;
		index1 = numPts-PHANTOMCOUNT+LEFTEDGEPOINT;
		metric->devLeftSideBearing.x = (charData.x[index1] - charData.x[index2])  << ten;
		metric->devLeftSideBearing.y = (charData.y[index1] - charData.y[index2])  << ten;

		metric->advanceWidth.x = ShortMulDiv( key->metricScalar, key->nonScaledAW, key->emResolution );
		metric->advanceWidth.y = 0;
		if ( ! key->identityTransformation )
			fsg_FixXYMul( &metric->advanceWidth.x, &metric->advanceWidth.y, &key->currentTMatrix );

		index2 = numPts-PHANTOMCOUNT+ORIGINPOINT;
		metric->leftSideBearing.x = (charData.x[index1] - charData.x[index2])  << ten;
		metric->leftSideBearing.y = (charData.y[index1] - charData.y[index2])  << ten;

		/* store away sidebearing along the advance width vector */
		metric->leftSideBearingLine = metric->leftSideBearing;
		metric->devLeftSideBearingLine = metric->devLeftSideBearing;

		/* Add vector to left upper edge of bitmap for ease of positioning by client */
      /* Modified to fix 16 bit overflow, mby 12-11-94 */
		tmp32 = (xMin - charData.x[index1]) << 10;
		metric->leftSideBearing.x		+= tmp32;
		metric->devLeftSideBearing.x	+= tmp32;
		tmp32 = (yMax - charData.y[index1]) << 10;
		metric->leftSideBearing.y		+= tmp32;		
		metric->devLeftSideBearing.y	+= tmp32;		
	}

	return NO_ERR;
}

#ifndef UFST_SUBSYSTEM

#if defined (ANSI_DEFS)
int32 PASCAL fs_FindBitMapSize(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_FindBitMapSize( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	register fsg_SplineKey 		*key;
	register fnt_ElementType 	*elementPtr;
	register int8				*workSpacePtr;
	register fsg_OffsetInfo		*offsetPtr;
	sc_CharDataType				charData;
	register sc_BitMapData		*bitRecPtr;
	uint16						scan, byteWidth;
	int 						numPts;
	register int32 				tmp32;
	int32						nx, size;
	int error;

	if (!(key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS | GOTGLYPH, &error)))
		return error;
	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	key->outlineIsCached = false;

	elementPtr			= &(key->elementInfoRec.interpreterElements[GLYPHELEMENT]);
	workSpacePtr		= key->memoryBases[WORK_SPACE_BASE];
	offsetPtr			= &(key->elementInfoRec.offsets[GLYPHELEMENT]);
	bitRecPtr			= &(key->bitMapInfo);

	charData.x			= (int32 *)(workSpacePtr + offsetPtr->newXOffset);
	charData.y			= (int32 *)(workSpacePtr + offsetPtr->newYOffset);
	charData.ctrs		= elementPtr->nc;
	charData.sp			= (int16 *)(workSpacePtr + offsetPtr->startPointOffset);
	charData.ep			= (int16 *)(workSpacePtr + offsetPtr->endPointOffset);
	charData.onC		= (int8 *)(workSpacePtr + offsetPtr->onCurveOffset);

	if (key->phaseShift)
		fs_45DegreePhaseShift( &charData );

    if ( (error = sc_FindExtrema( &charData,  bitRecPtr )) != 0 )
			return error;

	scan					= bitRecPtr->high;
	byteWidth				= bitRecPtr->wide >> 3;
	{
		BitMap* bm = &outputPtr->bitMapInfo;
		bm->baseAddr		= 0;
		bm->rowBytes		= byteWidth;
		{
			Rect* r = &bm->bounds;
			r->top		= bitRecPtr->yMin;
			r->left		= bitRecPtr->xMin;
			r->bottom	= bitRecPtr->yMax;
			r->right	= bitRecPtr->xMax;
		}
	}

	numPts = charData.ep[charData.ctrs-1] + 1 + PHANTOMCOUNT;
	{
		register int ten = 10;
		register metricsType* metric = &outputPtr->metricInfo;
		register int index1 = numPts-PHANTOMCOUNT+RIGHTSIDEBEARING;
		register int index2 = numPts-PHANTOMCOUNT+LEFTSIDEBEARING;

		metric->devAdvanceWidth.x 	  = (charData.x[index1] - charData.x[index2]) << ten;
		metric->devAdvanceWidth.y 	  = (charData.y[index1] - charData.y[index2]) << ten;
		index1 = numPts-PHANTOMCOUNT+LEFTEDGEPOINT;
		metric->devLeftSideBearing.x  = (charData.x[index1] - charData.x[index2])  << ten;
		metric->devLeftSideBearing.y  = (charData.y[index1] - charData.y[index2])  << ten;

		metric->advanceWidth.x 	  = ShortMulDiv( key->metricScalar, key->nonScaledAW, key->emResolution );
		metric->advanceWidth.y 	  = 0;
		if ( ! key->identityTransformation )
			fsg_FixXYMul( &metric->advanceWidth.x, &metric->advanceWidth.y, &key->currentTMatrix );

		index2 = numPts-PHANTOMCOUNT+ORIGINPOINT;
		metric->leftSideBearing.x = (charData.x[index1] - charData.x[index2])  << ten;
		metric->leftSideBearing.y = (charData.y[index1] - charData.y[index2])  << ten;

		/* store away sidebearing along the advance width vector */
		metric->leftSideBearingLine = metric->leftSideBearing;
		metric->devLeftSideBearingLine = metric->devLeftSideBearing;

		/* Add vector to left upper edge of bitmap for ease of positioning by client */
		tmp32 = ((int32)(bitRecPtr->xMin) << 16) - (charData.x[index1] << ten);
		metric->leftSideBearing.x		+= tmp32;
		metric->devLeftSideBearing.x	+= tmp32;
		tmp32 = ((int32)(bitRecPtr->yMax) << 16) - (charData.y[index1] << ten);
		metric->leftSideBearing.y		+= tmp32;		
		metric->devLeftSideBearing.y	+= tmp32;		
	}

    key->state = (uint16)(INITIALIZED | NEWSFNT | NEWTRANS | GOTGLYPH | SIZEKNOWN);

	/* get memory for bitmap in bitMapRecord */
	if( scan == 0 ) ++scan;

	outputPtr->memorySizes[BITMAP_PTR_1] = SHORTMUL(scan, byteWidth) + STAMPEXTRA;
	
	/* get memory for yLines & yBase in bitMapRecord */
	size = (int32)scan * ((bitRecPtr->nYchanges + 2) * sizeof(int16) + sizeof(int16*)); 
	outputPtr->memorySizes[BITMAP_PTR_2] = size;

	if( fs_dropOutVal( key ) )
	{
		/* get memory for xLines and xBase - used only for dropout control */			
		nx 		= bitRecPtr->xMax - bitRecPtr->xMin;
		if( nx == 0 ) ++nx;
		size 	= nx * ((bitRecPtr->nXchanges+2) * sizeof(int16) + sizeof(int16*));
	}
	else size = 0;
	
	outputPtr->memorySizes[BITMAP_PTR_3] = size;

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	return NO_ERR;
}




/* rwb - 4/21/90 - fixed to better work with caching */
#if defined (ANSI_DEFS)
int32 PASCAL fs_ContourScan(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_ContourScan( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
	register fsg_SplineKey 		*key;
	register sc_BitMapData		*bitRecPtr;
	register fnt_ElementType 	*elementPtr;
	register int8				*workSpacePtr;
	register fsg_OffsetInfo		*offsetPtr;
	sc_GlobalData				*scPtr;
	sc_CharDataType				charData;
	int32 						scanControl;
	int16						lowBand, highBand;
	uint16 						nx, ny;
	int							error;

	if (!(key = fs_SetUpKey(inputPtr, INITIALIZED | NEWSFNT | NEWTRANS | GOTGLYPH | SIZEKNOWN, &error)))
		return error;
	SETJUMP(key, error);

	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	bitRecPtr = &key->bitMapInfo;

	{
		elementPtr			= &(key->elementInfoRec.interpreterElements[GLYPHELEMENT]);
		workSpacePtr		= key->memoryBases[WORK_SPACE_BASE];
		offsetPtr			= &(key->elementInfoRec.offsets[GLYPHELEMENT]);

		
		charData.x			= (int32 *)(workSpacePtr + offsetPtr->newXOffset);
		charData.y			= (int32 *)(workSpacePtr + offsetPtr->newYOffset);
		charData.ctrs		= elementPtr->nc;
		charData.sp			= (int16 *)(workSpacePtr + offsetPtr->startPointOffset);
		charData.ep			= (int16 *)(workSpacePtr + offsetPtr->endPointOffset);
		charData.onC		= (int8 *)(workSpacePtr + offsetPtr->onCurveOffset);
	}

		scPtr				= (sc_GlobalData *)key->memoryBases[SCAN_PTR_BASE];
		
		nx 					= bitRecPtr->xMax - bitRecPtr->xMin;
		if( nx == 0 ) ++nx;
		
		scanControl = fs_dropOutVal( key );
		/* set up banding.  Assume highBand is 1 higher than highest scanrow <10>  */		
		highBand = inputPtr->param.scan.topClip;	
		lowBand  = inputPtr->param.scan.bottomClip;
		
		/* 	If topclip < bottom clip there is no banding by convention  */
		if( highBand <= lowBand )
		{
			highBand = bitRecPtr->yMax;
			lowBand = bitRecPtr->yMin;
		}
		/* check for out of bounds band request  							<10> */
		if( highBand > bitRecPtr->yMax ) highBand = bitRecPtr->yMax;
		if( lowBand  < bitRecPtr->yMin )  lowBand = bitRecPtr->yMin;
		
/* 11/16/90 rwb - We now allow the client to turn off DOControl by returning a NIL pointer
to the memory area used by DOC.  This is done so that in low memory conditions, the
client can get enough memory to print something.  We also always turn off DOC if the client
has requested banding.  Both of these conditions may change in the future.  Some versions
of TT may simply return an error condition when the NIL pointer to memoryarea 7 is
provided.  We also need to rewrite the scan converter routines that fill the bitmap 
under dropout conditions so that they use noncontiguous memory for the border scanlines
that need to be present for doing DOC.  This will allow us to do DOC even though we are
banding, providing there is enough memory.  By preflighting the fonts so that the request
for memory for areas 6 and 7 from findBitMapSize is based on actual need rather than
worse case analysis, we may also be able to reduce the memory required to do DOC in all
cases and particulary during banding.
*/
		/* inhibit DOControl if banding 									<10> */
		if( highBand < bitRecPtr->yMax || lowBand > bitRecPtr->yMin ) scanControl = 0;

		/* Allow client to turn off DOControl                               <10> */
		if( key->memoryBases[BITMAP_PTR_3] == 0 ) scanControl = 0;
		
		bitRecPtr->bitMap	= (uint32 *)key->memoryBases[BITMAP_PTR_1];		

		if( scanControl )					
		{
			SB8* memPtr		= (SB8*)key->memoryBases[BITMAP_PTR_3];
			bitRecPtr->xLines 	= (int16*)memPtr;
			
			memPtr				+= (bitRecPtr->nXchanges+2) * nx * sizeof(int16);
			bitRecPtr->xBase 	= (int16**)memPtr;
			
			ny					= bitRecPtr->yMax - bitRecPtr->yMin;
			if( ny == 0 ) ++ny;
			
			memPtr				= (SB8 *)key->memoryBases[BITMAP_PTR_2];
			bitRecPtr->yLines 	= (int16*)memPtr;
			
			memPtr 				+= (bitRecPtr->nYchanges+2) * ny * sizeof(int16);
			bitRecPtr->yBase 	= (int16**)memPtr;
		}
		else
		{
			SB8*	memPtr;
			ny					= highBand - lowBand;
			if( ny == 0 ) ++ny;
			memPtr				= (SB8 *)key->memoryBases[BITMAP_PTR_2];
			bitRecPtr->yLines 	= (int16*)memPtr;
			
			memPtr 				+= (bitRecPtr->nYchanges+2) * ny * sizeof(int16);
			bitRecPtr->yBase 	= (int16**)memPtr;
		}
    if ( (error = sc_ScanChar( &charData, scPtr, bitRecPtr, lowBand, highBand, scanControl )) != 0 )
		return error;
	{
		register BitMap* bm = &outputPtr->bitMapInfo;
		bm->baseAddr		= (int8 *)bitRecPtr->bitMap;
		bm->rowBytes		= bitRecPtr->wide >> 3;
		bm->bounds.top		= bitRecPtr->yMin;
		bm->bounds.left		= bitRecPtr->xMin;
		bm->bounds.bottom	= bitRecPtr->yMin + ny;
		bm->bounds.right	= bitRecPtr->xMin + nx;
	}
	
	CHECKSTAMP(inputPtr->memoryBases[0] + outputPtr->memorySizes[0]);
	CHECKSTAMP(inputPtr->memoryBases[1] + outputPtr->memorySizes[1]);
	CHECKSTAMP(inputPtr->memoryBases[2] + outputPtr->memorySizes[2]);
	CHECKSTAMP(inputPtr->memoryBases[3] + outputPtr->memorySizes[3]);
	CHECKSTAMP(inputPtr->memoryBases[4] + outputPtr->memorySizes[4]);

	return NO_ERR;
}

#endif              /* not UFST_SUBSYSTEM */



#if defined (ANSI_DEFS)
int32 PASCAL fs_CloseFonts(  fs_GlyphInputType *inputPtr, fs_GlyphInfoType *outputPtr )

#else 
int32 PASCAL fs_CloseFonts( inputPtr, outputPtr )
   fs_GlyphInputType *inputPtr;
   fs_GlyphInfoType  *outputPtr;
#endif
{
return NO_ERR;
}



#ifndef UFST_SUBSYSTEM

#if defined (ANSI_DEFS)
static void fs_45DegreePhaseShift(sc_CharDataType *glyphPtr)

#else 
static void fs_45DegreePhaseShift (glyphPtr)
   sc_CharDataType *glyphPtr;
#endif
{
	F26Dot6* x = glyphPtr->x;
	int16 count = glyphPtr->ep[glyphPtr->ctrs-1];
	for (; count >= 0; --count)
	{
		(*x)++;
		++x;
	}
}

/* Use various spline key values to determine if dropout control is to be activated
 * for this glyph, and if so what kind of dropout control.
 * The use of dropout control mode in the scan converter is controlled by 3 conditions.
 * The conditions are: Is the glyph rotated?, is the glyph stretched?,
 * is the current pixels per Em less than a specified threshold?
 * These conditions can be OR'd or ANDed together to determine whether the dropout control
 * mode ought to be used.

Six bits are used to specify the joint condition.  Their meanings are:

BIT		Meaning if set
8		Do dropout mode if other conditions don't block it AND
			pixels per em is less than or equal to bits 0-7
9		Do dropout mode if other conditions don't block it AND
			glyph is rotated
10		Do dropout mode if other conditions don't block it AND
			glyph is stretched
11		Do not do dropout mode unless ppem is less than or equal to bits 0-7
			A value of FF in 0-7  means all sizes
			A value of 0 in 0-7 means no sizes
12		Do not do dropout mode unless glyph is rotated	
13		Do not do dropout mode unless glyph is stretched
			
In other words, we do not do dropout control if:
No bits are set,
Bit 8 is set, but ppem is greater than threshold
Bit 9 is set, but glyph is not rotated
Bit 10 is set, but glyph is not stretched
None of the conditions specified by bits 11-13 are true.

For example, 0xA10 specifies turn dropout control on if the glyph is rotated providing
that it is also less than 0x10 pixels per em.  A glyph is considered stretched if
the X and Y resolutions are different either because of the device characteristics
or because of the transformation matrix.  If both X and Y are changed by the same factor
the glyph is not considered stretched.

 */

#if defined (ANSI_DEFS)
static int32 fs_dropOutVal( fsg_SplineKey *key )

#else 
static int32 fs_dropOutVal (key)
  fsg_SplineKey   *key;
#endif
{
register int32 condition = key->scanControl;
if( !(condition & 0x3F00) ) return 0;
if( (condition & 0xFFFF0000L) == NODOCONTROL ) return 0;
{
	register int32 imageState = key->imageState;
	if( (condition & 0x800) && ((imageState & 0xFF) > (condition & 0xFF)) ) return 0;
	if( (condition & 0x1000) && !(imageState & ROTATED) ) return 0;
	if( (condition & 0x2000) && !(imageState & STRETCHED) ) return 0;
	if( (condition & 0x100) && ((imageState & 0xFF) <= (condition & 0xFF) )) return condition;
	if( (condition & 0x100) && ((condition & 0xFF) == 0xFF )) return condition;
	if( (condition & 0x200) && (imageState & ROTATED) ) return condition;
	if( (condition & 0x400) && (imageState & STRETCHED) ) return condition;
	return 0;
}
}

#endif              /* not UFST_SUBSYSTEM */


#endif   /* TT_RDR  conditionalize entire file */

/* END OF fontscal.c */

