/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/tt/mapstrin.c,v 1.2 2003/05/07 23:02:30 Lann Exp $ */
/* $Log: mapstrin.c,v $
/* Revision 1.2  2003/05/07 23:02:30  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   07 Apr 1995 11:18:02   LISA
 * Changed copyright information from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   22 Apr 1994 14:55:36   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.3   11 Mar 1993 15:07:10   MAIB
 * Initial changes to support Asian character encoding (TrueType 1st pass)
 * 
 *    Rev 1.2   15 Jan 1993 09:50:18   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.1   14 Dec 1992 09:50:28   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:17:00   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:30 $ */
/*
   File:    MapString.c

	Contains:	Character to glyph mapping functions

	Written by:	Mike Reed

 


   Copyright:  1990 by Apple Computer, Inc., all rights reserved.



    Change History (most recent first):

    AGFA changes:

       11-Mar-93  maib  modified glyph lookup for format2 mapping
       30-Nov-92  jfd   Port to Mac - ANSI function definitions using ANSI_DEFS

       20-Oct-92  mby   Added casts whenever SWAPx() macro is used.
       25-Jun-92  jfd   Moved "cgconfig.h" ahead of "port.h".
        6 May 92  ss    Added !LINT_ARGS function definitions for porting.
       21-Apr-92  mby   Conditionalize entire file.
                        Eliminate warnings (MSC 6.0 /W3)

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		 <2>	11/16/90	MR		Fix range field of cmap-4 (thanks Eric) [rb]
		 <3>	 9/27/90	MR		Change selector in ComputeIndex4 to be log2(range) instead of
									2*log2(range), and fixed up comments
		 <2>	 8/10/90	MR		Add textLength field to MapString2
		 <1>	 7/23/90	MR		first checked in
				 7/23/90	MR		xxx put comment here xxx

	To Do:
*/


#include "cgconfig.h"
#include "port.h"

#if   TT_RDR   /* conditionalize entire file */

#include "fscdefs.h"
#include "mapstrin.h"

/*
 *	These next 4 MapString functions map the charCodes into glyphs,
 *	returning the number of glyphs processed.
 */
#if defined (ANSI_DEFS)
long MapString0(uint8* map, register uint8* charCodes, register int16* glyphs, long glyphCount)
#else
long 
MapString0(map, charCodes, glyphs, glyphCount)
    uint8*          map;
    register uint8* charCodes;
    register int16* glyphs;
    long            glyphCount;
#endif 
{
	register int16 count = (int16)glyphCount;

	for (--count; count >= 0; --count)
		*glyphs++ = map[*charCodes++];

	return glyphCount;
}

/*
 * High byte mapping through table
 *
 * Useful for the national standards for Japanese, Chinese, and Korean characters.
 *
 * Dedicated in spirit and logic to Mark Davis and the International group.
 *
 *	Algorithm: (I think)
 *		First byte indexes into KeyOffset table.  If the offset is 0, keep going, else use second byte.
 *		That offset is from beginning of data into subHeader, which has 4 words per entry.
 *			entry, extent, delta, range
 *
 *	textLength is optional.  If it is nil, it is ignored.
 */
#if defined (ANSI_DEFS)
long MapString2(register uint16* map, uint8* charCodes, int16* glyphs, long glyphCount, long* textLength)
#else
long 
MapString2(map, charCodes, glyphs, glyphCount, textLength)
   register uint16* map;
   uint8*           charCodes;
   int16*           glyphs;
   long             glyphCount;
   long*             textLength;
#endif 
{
	register short count = (short)glyphCount;
	register uint8* codeP = charCodes;
	register uint16 mapMe;
	int16* origGlyphs = glyphs;
   uint16 mapHB, mb;

	for (--count; count >= 0; --count)
	{
		uint16 highByte = *codeP++;

		if ( map[highByte] != 0 )
			mapMe = *codeP++;
		else
			mapMe = highByte;

		if (textLength && *textLength < codeP - charCodes)
		{
			--codeP;
			if ( map[highByte] )
				--codeP;
			break;
		}

      /*
       * 3/11/93 maib, corrected the access to the glyph ptr
       */  
      mapHB = map[highByte];
      mapHB = SWAPW(mapHB);
      map = ((uint16*)((int8*)map + mapHB)) + 256;	/* <4> bad mapping */
 		mapMe -= SWAPWINC(map); 		/* Subtract first code. */

      mb = SWAPWINC(map);
		if ( mapMe < mb ) {	/* See if within range. */
			uint16 idDelta;

			idDelta = SWAPWINC(map);
			mapMe += mapMe; /* turn into word offset */

      mb = SWAPW(*map);
      map = (uint16*)((int8*)map + mb + mapMe );

      mb = SWAPW(*map);
			if ( mb )
				*glyphs++ = mb + idDelta;
			else
				*glyphs++ = 0;		/* missing */
		} else
			*glyphs++ = 0;			/* missing */					/* <4> bad mapping */
	}
	if (textLength)
		*textLength = codeP - charCodes;		/* report # bytes eaten */

	return glyphs - origGlyphs;				/* return # glyphs processed */
}

#define maxLinearX2 16
#define BinaryIteration \
        newP = (uint16 *) ((int8 *)tableP + (range >>= 1)); \
        if ( charCode > (uint16)SWAPW(*newP) ) tableP = newP;

/*
 * Segment mapping to delta values, Yack.. !
 *
 * In memory of Peter Edberg. Initial code taken from code example supplied by Peter.
 *
 * Added SWAPW macros for port to DOS; mby 11/07/91
 */
#if defined (ANSI_DEFS)
static uint16 ComputeIndex4(uint16* tableP, uint16 charCode)
#else
static uint16 
ComputeIndex4(tableP, charCode)
   uint16* tableP;
   uint16  charCode;
#endif 
{
	register uint16 idDelta;
	register uint16 offset, segCountX2, index;

    index = 0; /* assume missing initially */
    segCountX2 = (uint16)SWAPWINC(tableP);

	if ( segCountX2 < maxLinearX2 ) {
		tableP += 3; /* skip binary search parameters */
	} else {
		/* start with unrolled binary search */
		register uint16 *newP;
		register int16  range; 		/* size of current search range */
		register uint16 selector; 	/* where to jump into unrolled binary search */
		register uint16 shift; 		/* for shifting of range */

        range    = (int16)SWAPWINC(tableP);   /* == 2*(2**floor(log2(segCount))) == 2*largest power of two <= segCount */
        selector = (uint16)SWAPWINC(tableP);  /* == log2(range/2) */
        shift    = (uint16)SWAPWINC(tableP);  /* == 2*segCount-range */
        /* tableP points at endCount[] */

        if ( charCode >= (uint16)SWAPW( (*((uint16 *)((int8 *)tableP + range))) ) )
            tableP = (uint16 *) ((int8 *)tableP + shift); /* range to low shift it up */
		switch ( selector )
		{
		case 15: BinaryIteration;
		case 14: BinaryIteration;
		case 13: BinaryIteration;
		case 12: BinaryIteration;
		case 11: BinaryIteration;
		case 10: BinaryIteration;
		case  9: BinaryIteration;
		case  8: BinaryIteration;
		case  7: BinaryIteration;
		case  6: BinaryIteration;
		case  5: BinaryIteration;
		case  4: BinaryIteration;
		case  3:
		case  2:  /* drop through */
		case  1:
		case  0:
			break;
		}
	}
    /* Now do linear search */
    while ( charCode > (uint16)SWAPW(*tableP) ) /* goes one to far, so we can skip the reservedPad */
        tableP++;
    tableP++;

	/* End of search, now do mapping */

	tableP = (uint16 *) ((int8 *)tableP + segCountX2); /* point at startCount[] */
    if ( charCode >= (uint16)SWAPW(*tableP) ) {
        offset = charCode - (uint16)SWAPW(*tableP);
        tableP = (uint16 *) ((int8 *)tableP + segCountX2); /* point to idDelta[] */
        idDelta = (uint16)SWAPW(*tableP);
        tableP = (uint16 *) ((int8 *)tableP + segCountX2); /* point to idRangeOffset[] */
        if ( SWAPW(*tableP) == 0 ) {
			index	= charCode + idDelta;
		} else {
			offset += offset; /* make word offset */
            tableP = (uint16 *) ((int8 *)tableP + (uint16)SWAPW(*tableP) + offset ); /* point to glyphIndexArray[] */
            if (index = (uint16)SWAPW(*tableP) )
				index += idDelta;
		}
	}
	return index;
}

#if defined (ANSI_DEFS)
long MapString4_8(uint16* map, uint8* charCodes, int16* glyphs, long glyphCount)
#else
long 
MapString4_8(map, charCodes, glyphs, glyphCount)
   uint16* map;
   uint8*  charCodes;
   int16*  glyphs;
   long    glyphCount;
#endif 
{
	register short count = (short)glyphCount;

	for (--count; count >= 0; --count)
		*glyphs++ = ComputeIndex4( map, *charCodes++ );

	return glyphCount;
}

#if defined (ANSI_DEFS)
long MapString4_16(uint16* map, uint16* charCodes, int16* glyphs, long glyphCount)
#else
long 
MapString4_16(map, charCodes, glyphs, glyphCount)
   uint16* map;
   uint16* charCodes;
   int16*  glyphs;
   long    glyphCount;
#endif 
{
	register short count = (short)glyphCount;

	for (--count; count >= 0; --count)
		*glyphs++ = ComputeIndex4( map, *charCodes++ );

	return glyphCount << 1;
}

/*
 * Trimmed Table Mapping - 8 bit character codes
 */
#if defined (ANSI_DEFS)
long MapString6_8(register uint16* map, uint8* charCodes, register int16* glyphs, long glyphCount)
#else
long 
MapString6_8(map, charCodes, glyphs, glyphCount)
   register uint16* map;
   uint8*           charCodes;
   register int16*  glyphs;
   long             glyphCount;
#endif 
{
	register short count = (short)glyphCount - 1;
	uint16 firstCode = *map++;
	uint16 entryCount = *map++;
	int16* origGlyphs = glyphs;

	for (; count >= 0; --count)
	{
		uint16 charCode = *charCodes++ - firstCode;
		if ( charCode < entryCount )
			*glyphs++ = map[ charCode ];
		else
			*glyphs++ = 0; /* missing char */
	}
	return glyphs - origGlyphs;
}

/*
 * Trimmed Table Mapping - 16 bit character codes
 */
#if defined (ANSI_DEFS)
long MapString6_16(register uint16* map, uint16* charCodes, register int16* glyphs, long glyphCount)
#else
long 
MapString6_16(map, charCodes, glyphs, glyphCount)
   register uint16* map;
   uint16*          charCodes;
   register int16*  glyphs;
   long             glyphCount;
#endif 
{
	register short count = (short)glyphCount - 1;
	uint16 firstCode = *map++;
	uint16 entryCount = *map++;
	int16* origGlyphs = glyphs;

	for (; count >= 0; --count)
	{
		uint16 charCode = *charCodes++ - firstCode;
		if ( charCode < entryCount )
			*glyphs++ = map[ charCode ];
		else
			*glyphs++ = 0; /* missing char */
	}
	return glyphs - origGlyphs;
}

#endif   /* TT_RDR  conditionalize entire file */

