/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/tt/sfnt.c,v 1.2 2003/05/07 23:02:30 Lann Exp $ */
/* $Log: sfnt.c,v $
/* Revision 1.2  2003/05/07 23:02:30  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.15   07 Apr 1995 11:18:12   LISA
 * Changed copyright information from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.14   13 Jan 1995 18:09:56   ROB
 * Add support for > 64KB memory blocks in DOS. Remove code which is contained
 * in ifmem.h.
 * 
 *    Rev 1.13   28 Jul 1994 10:20:46   JOE
 * In sfnt_ReadSFNT(), added additional check for the case where a PCLETTO's
 * glyph length is 0 and it is not present.
 * 
 *    Rev 1.12   02 Jun 1994 16:17:32   JOE
 * In sfnt_Classify(), defined "tag" for the case where 
 * COMPILER_CASE_INT_ONLY is undefined.
 * 
 *    Rev 1.11   22 Apr 1994 14:55:26   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.10   12 Apr 1994 15:24:08   MIKE
 * Removed test of maxSizeOfInstructions in sfnt_UnfoldCurve().
 * 
 *    Rev 1.9   15 Jun 1993 10:58:08   JOE
 * Moved "vxWorks.h" before "port.h" to resolve conflicts.
 * 
 *    Rev 1.8   28 May 1993 16:03:56   ROB
 * Changes to better support various mapping types of Kanji fonts.
 * 
 *    Rev 1.7   24 May 1993 13:40:50   JOE
 * In sfnt_UnfoldCurve(), changed key->glyphLength to keep trrack of sum of all
 * compound char pieces. This is so that if the last piece is a space (Bitstream
 * Flintstones), the Scaler doesn't treat the whole character as a space.
 * 
 *    Rev 1.6   29 Mar 1993 12:16:22   ROB
 * Include 'string.h' instead of 'mem.h' because it did not exist in Watcom.
 * 
 *    Rev 1.5   26 Mar 1993 14:48:14   MIKE
 * Initialized key->offsetTableMap array in sfnt_DoOffsetTableMap().
 * Removed else { ... } at the end of sfnt_Classify().
 * 
 *    Rev 1.4   12 Feb 1993 16:09:10   JOE
 * VXWorks support.
 * 
 *    Rev 1.3   08 Feb 1993 11:41:52   MIKE
 * sfnt_GetGlyphLocation is no longer static; changed arguments to unsigned;
 * function declaration moved to privsfnt.h; simplified MAC conditional code.
 * 
 *    Rev 1.2   15 Jan 1993 09:50:40   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.1   14 Dec 1992 09:52:42   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:16:56   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:30 $ */
/*
   File:    sfnt.c

 
 

   Copyright:  1989-1990 by Apple Computer, Inc., all rights reserved.



    Change History (most recent first):

    AGFA changes:

    27-Jul-94  mby/jfd
                    In sfnt_ReadSFNT(), added additional check for the
                    case where a PCLETTO's glyph length is 0 and it is 
                    not present.

    02-Jun-94  jfd  In sfnt_Classify(), defined "tag" for the case where
                    COMPILER_CASE_INT_ONLY is undefined.

    12-Apr-94  mby  In sfnt_UnfoldCurve(), removed test that checked
                    maxSizeOfInstructions. Some Monotype fonts have this word
                    stored as 0!

    15-Jun-93  jfd  Moved "vxWorks.h" before "port.h" to resolve conflicts.

    27-May-93  rs   Modify sfnt_ComputeIndex2() so that other SJIS mappings
                    will be easier to support.
    21-May-93  mby  In sfnt_UnfoldCurve() changed key->glyphLength to keep
                    track of sum of all compound char pieces. This is so that
                    if the last piece is a space (Bitstream Flintstones),
                    the Scaler doesn't treat the whole character as a space.

    22-Apr-93  mby  In sfnt_ReadSFNT(), if length=0, don't return MISSING_SFNT_TABLE.

    29-Mar-93  rs   Fix for Watcom - don't include 'mem.h' use 'string.h'.

    19-Mar-93  mby  Initialized key->offsetTableMap array in sfnt_DoOffsetTableMap().
                    Removed else { ... } at the end of sfnt_Classify().
                    Cleaned up some (Borland) compiler warnings.

    08-Feb-93  jfd  VXWorks support.

    18-Jan-93  mby  sfnt_GetGlyphLocation() is no longer static; changed
                    arguments to unsigned; declaration moved to privsfnt.h

    30-Nov-92  jfd  Port to Mac - ANSI function definitions using ANSI_DEFS

    22-Nov-92  rs   Port to Macintosh - sfnt_ReadSFNT().

    27-Oct-92  mby  Finished off mirrored c.c. bug. sfnt_UnfoldCurve() uses
                    the ctm and key->mirror_char to decide whether the
                    contour polarity is different from the base character.

    20-Oct-92  mby  Partly fixed mirrored compound character bug (Wingdings)
                    in sfnt_ReadSFNT(). A negative scale or matrix element
                    could be treated as positive after using SWAPW().
                    Cleaned up SWAPx() macro calls by adding casts.

    23-Sep-92  mby  Make sfnt_Classify() store offsetTableMap.Length
                    as 32-bit instead of casting to "unsigned".

    17-Jul-92  mby  Made changes to implement fonts in ROM:
                    in ReadSFNT(), replace calls to key->GetxSfntGlyphPtr()
                    with a single function call, key->GetSfntGlyphPtr();
                    change name of key->RelHSfntGlyphPtr() to RelSfntGlyphPtr().
                    GetGlyphLocation() is called if !key->extern_font

    25-Jun-92  jfd  Moved "cgconfig.h" ahead of "port.h".

    04-Jun-92  mby  Changes made to reflect that key->GetSfntFragmentPtr()
                    returns a pointer, not a handle -- DoOffsetTableMap(),
                    GetDataPtr(), GetTablePtr(), ComputeMappingX(),
                    ReadSFNTMetrics(), and GetGlyphLocation().
                    ReadSFNT() checks key->extern_font, calls different
                    routines to get glyph data if disk-based or PCLEO (ROM?)
                    Added OS_2, post, PCLT to sfnt_Classify()

    06-May-92  ss   Conditionally compile function protos for LINT_ARGS.

    22-Apr-92  mby  Changed sfnt_Classify(), sfnt_DoOffsetTable(),
                    sfnt_GetOffsetAndLength(). The application does not
                    have to pass in the sfntDirectory structure.

    07-Feb-92  rs   Conditionalize entire file.

         1991  mby  Cut #include file names to eight characters.
                    Added byte swapping macros SWAPW, SWAPL, SWAPINC
                      in various places.
                    Cast things to eliminate warnings.
                    In sfnt_Classify() replaced case with if-else
                    because MSC can only handle int expressions.
                    see #ifndef  COMPILER_CASE_INT_ONLY

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		 <6>	12/18/90	RB		Move ReleaseSfntFrag into block where table is defined. [mr]
		 <5>	12/11/90	MR		Support platform==-1 meaning this sfnt has no cmap, don't return
									an error.
		 <4>	11/16/90	MR		Fix bogus call to SWAPW in ReadSfnt for instructions after
									components. [rb]
		 <3>	 11/5/90	MR		Remove mac include files, clean up release macros. [rb]
		 <2>	10/20/90	MR		Removed unused tables in sfnt_Classify. [rb]
		<17>	 8/10/90	MR		Pass nil for textLength parameter to MapString2, checked in
									other files to their precious little system will BUILD.  Talk
									about touchy!
		<16>	 8/10/90	gbm		rolling out Mikes textLength change, because he hasnt checked
									in all the relevant files, and the build is BROKEN!
		<15>	 8/10/90	MR		Add textLength arg to MapString2
		<14>	 7/26/90	MR		dont include ToolUtils.h
		<13>	 7/23/90	MR		Change computeindex routines to call functins in MapString.c
		<12>	 7/18/90	MR		Add SWAPW macro for INTEL
		<11>	 7/13/90	MR		Lots of Ansi-C stuff, change behavior of ComputeMapping to take
									platform and script
		 <9>	 6/27/90	MR		Changes for modified format 4: range is now times two, loose pad
									word between first two arrays.  Eric Mader
		 <8>	 6/21/90	MR		Add calls to ReleaseSfntFrag
		 <7>	  6/5/90	MR		remove vector mapping functions
		 <6>	  6/4/90	MR		Remove MVT
		 <5>	  5/3/90	RB		simplified decryption.
		 <4>	 4/10/90	CL		Fixed mapping table routines for double byte codes.
		 <3>	 3/20/90	CL		Joe found bug in mappingtable format 6 Added vector mapping
									functions use pointer-loops in sfnt_UnfoldCurve, changed z from
									int32 to int16
		 <2>	 2/27/90	CL		New error code for missing but needed table. (0x1409 )	New
									CharToIndexMap Table format.
									Assume subtablenumber zero for old sfnt format.  Fixed
									transformed component bug.
	   <3.2>	11/14/89	CEL		Left Side Bearing should work right for any transformation. The
									phantom points are in, even for components in a composite glyph.
									They should also work for transformations. Device metric are
									passed out in the output data structure. This should also work
									with transformations. Another leftsidebearing along the advance
									width vector is also passed out. whatever the metrics are for
									the component at it's level. Instructions are legal in
									components. Instructions are legal in components. Glyph-length 0
									bug in sfnt.c is fixed. Now it is legal to pass in zero as the
									address of memory when a piece of the sfnt is requested by the
									scaler. If this happens the scaler will simply exit with an
									error code ! Fixed bug with instructions in components.
	   <3.1>	 9/27/89	CEL		Removed phantom points.
	   <3.0>	 8/28/89	sjk		Cleanup and one transformation bugfix
	   <2.2>	 8/14/89	sjk		1 point contours now OK
	   <2.1>	  8/8/89	sjk		Improved encryption handling
	   <2.0>	  8/2/89	sjk		Just fixed EASE comment
	   <1.5>	  8/1/89	sjk		Added composites and encryption. Plus some enhancements
	   <1.4>	 6/13/89	SJK		Comment
	   <1.3>	  6/2/89	CEL		16.16 scaling of metrics, minimum recommended ppem, point size 0
									bug, correct transformed integralized ppem behavior, pretty much
									so
	   <1.2>	 5/26/89	CEL		EASE messed up on c comments
	  <1.1>	 5/26/89	CEL		Integrated the new Font Scaler 1.0 into Spline Fonts
	   <1.0>	 5/25/89	CEL		Integrated 1.0 Font scaler into Bass code for the first time

	To Do:
		<3+>	 3/20/90	mrr		Fixed mapping table routines for double byte codes.
									Added support for font program.
									Changed count from uint16 to int16 in vector char2index routines.
*/

#define BYTEREAD
 
#include "cgconfig.h"

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include "port.h"
#include "mixmodel.h"
#include "string.h" /* memset () */

#if   TT_RDR   /* conditionalize entire file */

#include "setjmp.h"

/** FontScalers Includes **/
#include "ifmem.h"   /* add 1/11/95 - rob */
#include "fscdefs.h"
#include "fserror.h"
#include "fontmath.h"
#include "sfnt.h"
#include "fnt.h"
#include "sc.h"
#include "fontscal.h"
#include "fsglue.h"
#include "privsfnt.h"
#include "mapstrin.h"


/** SFNT Packed format **/
typedef struct {
	int16 		numberContours;
	int16 		*endPoints;				/** vector: indexes into x[], y[] **/
	int16 		numberInstructions;
	uint8  		*instructions;			/** vector **/
	uint8 		*flags;					/** vector **/
	sfnt_BBox	bbox;
} sfnt_PackedSplineFormat;


#define GetUnsignedByte( p ) ((uint8)(*p++))

/** <4> **/
#define fsg_MxCopy(a, b)	(*(b) = *(a))


/* PRIVATE PROTOTYES */

#ifdef LINT_ARGS
static int sfnt_UnfoldCurve(fsg_SplineKey *key, sfnt_PackedSplineFormat *charData,
                     int32 *sizeOfInstructions, uint8 **instructionPtr, int32 length, int16 *elementCount);
static void sfnt_Classify(register fsg_SplineKey *key, register sfnt_DirectoryEntry *dir);
static void*  sfnt_GetDataPtr(fsg_SplineKey *key, int32 offset, int32 length, sfnt_tableIndex n, boolean mustHaveTable);
static uint16 sfnt_ComputeUnkownIndex(fsg_SplineKey *key, uint16 charCode);
static uint16 sfnt_ComputeIndex0(fsg_SplineKey *key, uint16 charCode);
uint16 sfnt_ComputeIndex2(fsg_SplineKey *key, uint16 charCode);
uint16 sfnt_ComputeIndex4(fsg_SplineKey *key, uint16 charCode);
uint16 sfnt_ComputeIndex6(fsg_SplineKey *key, uint16 charCode);
#else
static int sfnt_UnfoldCurve();
static void sfnt_Classify();
static void*  sfnt_GetDataPtr();
static uint16 sfnt_ComputeUnkownIndex();
static uint16 sfnt_ComputeIndex0();
uint16 sfnt_ComputeIndex2();
uint16 sfnt_ComputeIndex4();
uint16 sfnt_ComputeIndex6();
#endif /* !LINT_ARGS */


#ifdef SEGMENT_LINK
#pragma segment SFNT_C
#endif

/*								
 *
 *	sfnt_UnfoldCurve			<3>
 *
 *	ERROR:	return NON-ZERO
 *
 */
#if defined (ANSI_DEFS)
static int sfnt_UnfoldCurve(	fsg_SplineKey* key,
								sfnt_PackedSplineFormat* charData,
								int32* sizeOfInstructions,
								uint8** instructionPtr,
								int32 length,
								int16* elementCount )
#else
static int 
sfnt_UnfoldCurve(	key, charData, sizeOfInstructions,
								instructionPtr, length, elementCount )
   fsg_SplineKey*           key;
   sfnt_PackedSplineFormat* charData;
   int32*                   sizeOfInstructions;
   uint8**                  instructionPtr;
   int32                    length;
   int16*                   elementCount;
#endif 
{
	register int z;
	register uint8 flag, *byteP, *byteP2;
	register int8 *p;
	register F26Dot6* longP;
	fnt_ElementType *elementPtr;
	int numPoints;

	elementPtr	= &(key->elementInfoRec.interpreterElements[GLYPHELEMENT]);

	if ( *elementCount == GLYPHELEMENT ) {
		key->totalContours = 0;
 		fsg_SetUpElement( key, (int32)GLYPHELEMENT );
	} else {
		/*                                       # of points in previous component */
		fsg_IncrementElement( key, (int32)GLYPHELEMENT,
           (int32)key->numberOfRealPointsInComponent, (int32)elementPtr->nc );
	}
	key->totalComponents = *elementCount;
	
	(*elementCount)++;
    key->glyphLength += length;     /* for a composite, is the sum of all pieces - mby, 5/21/93 */
	if ( length <= 0 ) {
		elementPtr->nc = 1;
		key->totalContours += 1;
		
		elementPtr->sp[0] = 0;
		elementPtr->ep[0] = 0;
		
		*sizeOfInstructions = 0;
		
		elementPtr->onCurve[0] = ONCURVE;
		elementPtr->oox[0] = 0;
		elementPtr->ooy[0] = 0;
		return NO_ERR; /***************** exit here ! *****************/
	}

	elementPtr->nc = charData->numberContours;
	z = (int)elementPtr->nc;
	key->totalContours += z;
	if ( z < 0 || z > (int16)SWAPW(key->maxProfile.maxContours) )
		return CONTOUR_DATA_ERR;

	{	/* <4> */
		register int16* sp = elementPtr->sp;
		register int16* ep = elementPtr->ep;
		int16* charDataEP = charData->endPoints;
		register LoopCount i;
		*sp++ = 0;
		*ep = (int16)SWAPWINC(charDataEP);
		for (i = z - 2; i >= 0; --i)
		{
			*sp++ = *ep++ + 1;
			*ep = (int16)SWAPWINC(charDataEP);
		}
	}
    {
        Fixed *ctm = (Fixed *)key->currentTMatrix.transform;
        int det = ( FixMul(ctm[0],ctm[4]) - FixMul(ctm[1],ctm[3]) < 0 );
        uint8* cPolR = elementPtr->cPolR;
        LoopCount i;
        for (i = z - 1; i >= 0; --i)
            *cPolR++ = (det != key->mirror_char);
    }

	numPoints = elementPtr->ep[elementPtr->nc - 1] + 1;
	key->numberOfRealPointsInComponent = (uint16)numPoints;
	if (numPoints < 0 || numPoints > (int16)SWAPW(key->maxProfile.maxPoints) )
		return POINTS_DATA_ERR;

	*sizeOfInstructions = charData->numberInstructions;
	z = (int)charData->numberInstructions;

/* Nothing ever uses maxSizeOf... There are some fonts out there that have
 * this word set to 0 in maxp, so it's better to skip the test. mby 4/12/94
 *
 *	if ( z < 0 || z > (int16)SWAPW(key->maxProfile.maxSizeOfInstructions) )
 *		return INSTRUCTION_SIZE_ERR;
 */

	*instructionPtr	= charData->instructions;
	
	/* Do flags */
	p = (int8*)charData->flags;
	byteP = elementPtr->onCurve;
	byteP2 = byteP + numPoints;			/* only need to set this guy once */
	while (byteP < byteP2)
	{
		*byteP++ = flag = GetUnsignedByte( p );
		if ( flag & REPEAT_FLAGS ) {
			LoopCount count = GetUnsignedByte( p );
			for (--count; count >= 0; --count)
				*byteP++ = flag;
		}
	}

	/* Do X first */
	z = 0;
	byteP = elementPtr->onCurve;
	longP = elementPtr->oox;
	while (byteP < byteP2)
	{
		if ( (flag = *byteP++) & XSHORT ) {
			if ( flag & SHORT_X_IS_POS )
				z += GetUnsignedByte( p );
			else
				z -= GetUnsignedByte( p );
		} else if ( !(flag & NEXT_X_IS_ZERO) ) { /* This means we have a int32 (2 byte) vector */
#ifdef BYTEREAD
			z += (int16)(*p++) << 8;
			z += (uint8)*p++;
#else
			z += *((int16 *)p); p += sizeof( int16 );
#endif
		}
		*longP++ = (F26Dot6)z;
	}

	/* Now Do Y */
	z = 0;
	byteP = elementPtr->onCurve;
	longP = elementPtr->ooy;
	while (byteP < byteP2)
	{
		if ( (flag = *byteP) & YSHORT ) {
			if ( flag & SHORT_Y_IS_POS )
				z += GetUnsignedByte( p );
			else
				z -= GetUnsignedByte( p );
		} else if ( !(flag & NEXT_Y_IS_ZERO) ) { /* This means we have a int32 (2 byte) vector */
#ifdef BYTEREAD
			z += (int16)(*p++) << 8;
			z += (uint8)*p++;
#else
			z += *((int16 *)p); p += sizeof( int16 );
#endif
		}
		*longP++ = (F26Dot6)z;
	
		*byteP++ = flag & (uint8)ONCURVE; /* Filter out unwanted stuff */
	}
	if ( !(key->compFlags & NON_OVERLAPPING) ) {
		elementPtr->onCurve[0] |= OVERLAP;
	}
	
	return NO_ERR;
}

/*
 * Internal routine		(make this an array and do a look up?)
 */
#if defined (ANSI_DEFS)
static void sfnt_Classify( register fsg_SplineKey *key,
                           register sfnt_DirectoryEntry *dir)
#else
static void 
sfnt_Classify( key, dir)
   register fsg_SplineKey       *key;
   register sfnt_DirectoryEntry *dir;
#endif 
{
    int index = -1;
#ifndef  COMPILER_CASE_INT_ONLY
    sfnt_TableTag tag = dir->tag;
    switch ( tag ) {

	case tag_FontHeader:
		index = sfnt_fontHeader;
		break;
	case tag_HoriHeader:
		index = sfnt_horiHeader;
		break;
	case tag_IndexToLoc:
		index = sfnt_indexToLoc;
		break;
	case tag_MaxProfile:
		index = sfnt_maxProfile;
		break;
	case tag_ControlValue:
		index = sfnt_controlValue;
		break;
	case tag_PreProgram:
		index = sfnt_preProgram;
		break;
	case tag_GlyphData:
		index = sfnt_glyphData;
		break;
	case tag_HorizontalMetrics:
		index = sfnt_horizontalMetrics;
		break;
	case tag_CharToIndexMap:
		index = sfnt_charToIndexMap;
		break;
	case tag_FontProgram:
		index = sfnt_fontProgram;		/* <4> */
		break;
    case tag_OS_2:
        index = sfnt_OS_2;
        break;
    case tag_Postscript:
        index = sfnt_PostScript;
        break;
    case tag_PCLT:
        index = sfnt_PCLT;
        break;
	}
#else
    sfnt_TableTag tag = dir->tag;
    if (tag == tag_FontHeader)
        index = sfnt_fontHeader;
    else if (tag == tag_HoriHeader)
        index = sfnt_horiHeader;
    else if (tag == tag_IndexToLoc)
        index = sfnt_indexToLoc;
    else if (tag == tag_MaxProfile)
        index = sfnt_maxProfile;
    else if (tag == tag_ControlValue)
        index = sfnt_controlValue;
    else if (tag == tag_PreProgram)
        index = sfnt_preProgram;
    else if (tag == tag_GlyphData)
        index = sfnt_glyphData;
    else if (tag == tag_HorizontalMetrics)
        index = sfnt_horizontalMetrics;
    else if (tag == tag_CharToIndexMap)
        index = sfnt_charToIndexMap;
    else if (tag == tag_FontProgram)
        index = sfnt_fontProgram;
    else if (tag == tag_OS_2)
        index = sfnt_OS_2;
    else if (tag == tag_Postscript)
        index = sfnt_PostScript;
    else if (tag == tag_PCLT)
        index = sfnt_PCLT;
#endif

    if (index >= 0)
        {
        key->offsetTableMap[index].Offset = (uint32)SWAPL (dir->offset);
        key->offsetTableMap[index].Length = (uint32)SWAPL (dir->length);
        }
}

/*
 * Creates mapping for finding offset table   <4>
 * Called at font initialization time (fs_NewSfnt)
 */
#if defined (ANSI_DEFS)
void sfnt_DoOffsetTableMap( register fsg_SplineKey *key )
#else
void 
sfnt_DoOffsetTableMap( key )
   register fsg_SplineKey *key;
#endif 
{
    register LoopCount i;
    sfnt_OffsetTable  *sfntDirectory;
    LoopCount cTables;

    sfntDirectory = (sfnt_OffsetTable *)
            key->GetSfntFragmentPtr( key->clientID, 0L, (long)sizeof(sfnt_OffsetTable) );
    if (sfntDirectory)
    {
        cTables = (LoopCount)((uint16)SWAPW (sfntDirectory->numOffsets) - 1);
        RELEASESFNTFRAG(key, sfntDirectory);
        sfntDirectory = (sfnt_OffsetTable *) key->GetSfntFragmentPtr( key->clientID, 0L,
               (long)(sizeof(sfnt_OffsetTable) + sizeof(sfnt_DirectoryEntry) * cTables));
    }

    if (sfntDirectory == 0)
        longjmp (key->env, NULL_SFNT_DIR_ERR);  /* Do a graceful recovery */

    /* Initialize */
    {
    register sfnt_DirectoryEntry *dir = &sfntDirectory->table[cTables];
    MEMSET (key->offsetTableMap, 0, sizeof (key->offsetTableMap));
    for (i = cTables; i >= 0; --i, --dir)
        sfnt_Classify (key, dir);
    }

    RELEASESFNTFRAG(key, sfntDirectory);
}


/*
 * Returns offset and length for table n	<4>
 */
#if defined (ANSI_DEFS)
void sfnt_GetOffsetAndLength(	register fsg_SplineKey *key,
								int32 *offsetT,
								unsigned *lengthT,
								register sfnt_tableIndex n)
#else
void 
sfnt_GetOffsetAndLength( key, offsetT, lengthT, n)
   register fsg_SplineKey  *key;
   int32                   *offsetT;
   unsigned                *lengthT;
   register sfnt_tableIndex n;
#endif 
{
    if (n >= 0) {
        *lengthT = (unsigned)key->offsetTableMap[n].Length;
        *offsetT = key->offsetTableMap[n].Offset;
	} else {
		*lengthT = 0; *offsetT = 0; /* missing table */
	}
}

/*
 * Use this when the whole table is neded.
 *
 * n is the table number.
 */
#if defined (ANSI_DEFS)
void* sfnt_GetTablePtr(	register fsg_SplineKey *key,
						register sfnt_tableIndex n,
						register boolean mustHaveTable )
#else
void*
sfnt_GetTablePtr( key, n, mustHaveTable )
   register fsg_SplineKey  *key;
   register sfnt_tableIndex n;
   register boolean         mustHaveTable;
#endif 
{
	return sfnt_GetDataPtr( key, 0L, -1L, n, mustHaveTable );	/* 0, -1 means give me whole table */
}

/*
 * Use this function when only part of the table is needed.
 *
 * n is the table number.
 * offset is within table.
 * length is length of data needed.
 * To get an entire table, pass length = -1		<4>
 */
#if defined (ANSI_DEFS)
static  void* sfnt_GetDataPtr (register fsg_SplineKey *key,
						register int32 offset,
						register int32 length,
						register sfnt_tableIndex n,
						register boolean mustHaveTable )
#else
static void*
sfnt_GetDataPtr (key, offset, length, n, mustHaveTable )
   register fsg_SplineKey  *key;
   register int32           offset;
   register int32           length;
   register sfnt_tableIndex n;
   register boolean mustHaveTable;
#endif 
{
	int32 offsetT;
	unsigned lengthT;
	void* fragment;

	sfnt_GetOffsetAndLength( key, &offsetT, &lengthT, n );

	if ( lengthT ) {
		if (length == -1) length = lengthT;
		offset += offsetT;
		fragment = key->GetSfntFragmentPtr( key->clientID, offset, length );
		if ( fragment ) {
			return fragment;
		} else {
            longjmp( key->env, CLIENT_RETURNED_NULL ); /* Do a graceful recovery */
		}
	} else {
		if ( mustHaveTable )
            longjmp( key->env, MISSING_SFNT_TABLE );   /* Do a graceful recovery */
		else
			return 0;
	}
}


/*
 * This, is when we don't know what is going on
 */
#if defined (ANSI_DEFS)
static uint16 sfnt_ComputeUnkownIndex( fsg_SplineKey* key, uint16 charCode)
#else
static uint16 
sfnt_ComputeUnkownIndex( key, charCode)
   fsg_SplineKey *key;
   uint16         charCode;
#endif 
{
#ifndef NOT_ON_THE_MAC
#pragma unused(key, charCode)
#endif
	return 0;
}

/*
 * Byte Table Mapping 256->256			<4>
 */
#if defined (ANSI_DEFS)
static uint16 sfnt_ComputeIndex0(	register fsg_SplineKey *key,
									register uint16 charCode)
#else
static uint16 
sfnt_ComputeIndex0(	key, charCode)
   register fsg_SplineKey *key;
   register uint16         charCode;
#endif 
{
	if (charCode & ~(255)) return 0;

	{
		uint8 *mappingPtr = (uint8*)sfnt_GetTablePtr( key, sfnt_charToIndexMap, true );
		register uint8* mapping = mappingPtr;
		uint16 glyph;
			
		mapping += key->mappOffset; /* Point at mapping table offset. */
		glyph = mapping[charCode];
		RELEASESFNTFRAG(key, mappingPtr);

		return glyph;
	}
}

/*
 * High byte mapping through table
 *
 * Useful for the national standards for Japanese, Chinese, and Korean characters.
 *
 * Dedicated in spirit and logic to Mark Davis and the International group.
 *
 *	Algorithm: (I think)
 *		First byte indexes into KeyOffset table.  If the offset is 0, keep going, else use second byte.
 *		That offset is from beginning of data into subHeader, which has 4 words per entry.
 *			entry, extent, delta, range
 *
 */
#if defined (ANSI_DEFS)
uint16 sfnt_ComputeIndex2( fsg_SplineKey* key, uint16 charCode )
#else
uint16 
sfnt_ComputeIndex2( key, charCode )
   fsg_SplineKey* key;
   uint16         charCode;
#endif 
{
	void* mappingPtr = sfnt_GetTablePtr( key, sfnt_charToIndexMap, true );
	uint8* mapping = (uint8*)mappingPtr;
    uint16 index;
   uint16 lcharCode = SWAPW (charCode); /* local copy of character code 5/27/93 - rs */

	mapping += key->mappOffset; /* Point at mapping table offset. */

    key->numberOfBytesTaken = (uint16)MapString2((uint16*)mapping, (uint8*)&lcharCode,
                                         (int16*)&index, 1L, 0);
    RELEASESFNTFRAG(key, mappingPtr);

	return index;
}

/*
 * Segment mapping to delta values, Yack.. !
 *
 * In memory of Peter Edberg. Initial code taken from code example supplied by Peter.
 */
#if defined (ANSI_DEFS)
uint16 sfnt_ComputeIndex4( fsg_SplineKey *key, uint16 charCode)
#else
uint16 
sfnt_ComputeIndex4( key, charCode)
   fsg_SplineKey *key;
   uint16         charCode;
#endif 
{
	void* mappingPtr = sfnt_GetTablePtr( key, sfnt_charToIndexMap, true );
	uint8* mapping = (uint8*)mappingPtr;
	uint16 index;
	
	mapping += key->mappOffset; /* Point at mapping table offset. */

	MapString4_16((uint16*)mapping, &charCode, (int16*)&index, 1L);
	RELEASESFNTFRAG(key, mappingPtr);

	return index;
}


/*
 * Trimmed Table Mapping
 */
#if defined (ANSI_DEFS)
uint16 sfnt_ComputeIndex6( fsg_SplineKey* key, uint16 charCode )
#else
uint16 
sfnt_ComputeIndex6( key, charCode )
   fsg_SplineKey* key;
   uint16         charCode;
#endif 
{
	void* mappingPtr = sfnt_GetTablePtr( key, sfnt_charToIndexMap, true );
	uint8* mapping = (uint8*)mappingPtr;
	uint16 index;
		
	mapping += key->mappOffset; /* Point at mapping table offset. */

	MapString6_16((uint16*)mapping, &charCode, (int16*)&index, 1L);	
	RELEASESFNTFRAG(key, mappingPtr);

	return index;
}

/*
 * Sets up our mapping function pointer.
 */
#if defined (ANSI_DEFS)
int sfnt_ComputeMapping(register fsg_SplineKey *key, uint16 platformID, uint16 specificID)
#else
int 
sfnt_ComputeMapping(key, platformID, specificID)
   register fsg_SplineKey *key;
   uint16                  platformID;
   uint16                  specificID;
#endif 
{
	int result = NO_ERR;

	key->numberOfBytesTaken = 2; /* Initialize */

	if (platformID == 0xffff)
		key->mappingF = sfnt_ComputeUnkownIndex;
	else
	{
		sfnt_char2IndexDirectory *table= (sfnt_char2IndexDirectory*)sfnt_GetTablePtr( key, sfnt_charToIndexMap, true );
		uint8 *mapping = (uint8 *)table;
		register uint16 format = 0xFFFF;
        register int    i;
		int16 found = false;
		
		if ( table->version == 0 )
		{
			register sfnt_platformEntry* plat = table->platform;	/* <4> */
            for ( i = (int16)SWAPW(table->numTables)-1; i >= 0; --i )
			{
                if ( (uint16)SWAPW(plat->platformID) == platformID  &&
                     (uint16)SWAPW(plat->specificID) == specificID)
				{
					found = true;
                    key->mappOffset = (int32)SWAPL(plat->offset) + 6L;  /* skip header */
					break;
				}
				++plat;
			}
		}
		
		if ( !found )
		{
			key->mappingF = sfnt_ComputeUnkownIndex;
			result = OUT_OF_RANGE_SUBTABLE;
		}
		else
		{
			mapping += key->mappOffset - 6;		/* back up for header */
            format = (uint16)SWAPW( (*(uint16 *)mapping) );

			switch ( format ) {
			case 0:
				key->mappingF = sfnt_ComputeIndex0;
				break;
			case 2:
				key->mappingF = sfnt_ComputeIndex2;
				break;
			case 4:
				key->mappingF = sfnt_ComputeIndex4;
				break;
			case 6:
				key->mappingF = sfnt_ComputeIndex6;
				break;
			default:
				key->mappingF = sfnt_ComputeUnkownIndex;
				result = UNKNOWN_CMAP_FORMAT;
				break;
			}
		}
		RELEASESFNTFRAG(key, table);
	}
	return result;
}



/*
 *
 */
#if defined (ANSI_DEFS)
void sfnt_ReadSFNTMetrics(	fsg_SplineKey* key,
							register uint16 glyphIndex )
#else
void 
sfnt_ReadSFNTMetrics( key, glyphIndex )
   fsg_SplineKey*  key;
   register uint16 glyphIndex;
#endif 
{
	register sfnt_HorizontalMetrics		*horizMetricPtr;
	register uint16						numberOf_LongHorMetrics = key->numberOf_LongHorMetrics;

	horizMetricPtr 	= (sfnt_HorizontalMetrics *)sfnt_GetTablePtr( key, sfnt_horizontalMetrics, true );

	if ( glyphIndex < numberOf_LongHorMetrics ) {
        key->nonScaledAW  = (uint16)SWAPW(horizMetricPtr[glyphIndex].advanceWidth);
        key->nonScaledLSB = (int16)SWAPW(horizMetricPtr[glyphIndex].leftSideBearing);
	} else {
		int16 *lsb = (int16 *)&horizMetricPtr[numberOf_LongHorMetrics]; /* first entry after[AW,LSB] array */
        key->nonScaledAW  = (uint16)SWAPW(horizMetricPtr[numberOf_LongHorMetrics-1].advanceWidth);
        key->nonScaledLSB = (int16)SWAPW(lsb[glyphIndex - numberOf_LongHorMetrics]);
	}
	RELEASESFNTFRAG(key, horizMetricPtr);
}

#if defined (ANSI_DEFS)
void  sfnt_GetGlyphLocation( fsg_SplineKey* key, uint16 gIndex,
                             uint32* offset, uint32* length )
#else
void  sfnt_GetGlyphLocation( key, gIndex, offset, length )
   fsg_SplineKey* key;
   uint16         gIndex;
   uint32*        offset;
   uint32*        length;
#endif 
{
	void* indexPtr = sfnt_GetTablePtr( key, sfnt_indexToLoc, true );

    if ( key->indexToLocFormat == SHORT_INDEX_TO_LOC_FORMAT )
	{
		register uint16* shortIndexToLoc = (uint16*)indexPtr;
		shortIndexToLoc += gIndex;
        *offset = (uint32)SWAPW((*shortIndexToLoc)) << 1;
        shortIndexToLoc++;
        *length = ((uint32)SWAPW((*shortIndexToLoc)) << 1) - *offset;
	}
	else
	{
		register uint32* longIndexToLoc = (uint32*)indexPtr;
		longIndexToLoc += gIndex;
        *offset = (uint32)SWAPL((*longIndexToLoc));
        longIndexToLoc++;
        *length = (uint32)SWAPL((*longIndexToLoc)) - *offset;
	}
	RELEASESFNTFRAG(key, indexPtr);
}

/*
 *	<4>
 */
#if defined (ANSI_DEFS)
int sfnt_ReadSFNT(	register fsg_SplineKey *key,
						int16 *elementCount,
						register uint16 gIndex,
						boolean useHints,
						voidFunc traceFunc)
#else
int 
sfnt_ReadSFNT(	key, elementCount, gIndex, useHints, traceFunc)
   register fsg_SplineKey *key;
   int16                  *elementCount;
   register uint16         gIndex;
   boolean                 useHints;
   voidFunc                traceFunc;
#endif 
{
	int32 		sizeOfInstructions;
	uint8* 		instructionPtr; 
    uint32      length=0, offset=0;
	int			result	= NO_ERR;
	int16*		shortP;
    MEM_HANDLE  gHndl;
    void FARPTR *glyphDataPtr = 0;
	sfnt_PackedSplineFormat charData;

	sfnt_ReadSFNTMetrics( key, gIndex );
    if (!key->extern_font)
        sfnt_GetGlyphLocation( key, gIndex, &offset, &length );
    glyphDataPtr = key->GetSfntGlyphPtr( key->clientID, gIndex, offset,
                                         &length, &gHndl );

    if (!glyphDataPtr && length > 0)
        longjmp( key->env, MISSING_SFNT_TABLE );  /* Do a graceful recovery */
	if ( length > 0 )
	{
		shortP = (int16*)glyphDataPtr;
		charData.numberContours = (int16)SWAPWINC(shortP);
		charData.bbox.xMin = (int16)SWAPWINC(shortP);
		charData.bbox.yMin = (int16)SWAPWINC(shortP);
		charData.bbox.xMax = (int16)SWAPWINC(shortP);
		charData.bbox.yMax = (int16)SWAPWINC(shortP);
	}
	else
		charData.numberContours = 1;

	if ( charData.numberContours >= 0 )	/* Not a component glyph */
	{
/*      key->lastGlyph = !( key->weGotComponents && (key->compFlags & MORE_COMPONENTS) ); */

		if ( length > 0 )
		{
			charData.endPoints = shortP;
			shortP += charData.numberContours;
			charData.numberInstructions = (int16)SWAPWINC(shortP);
			charData.instructions = (uint8*)shortP;
			charData.flags = (uint8*)shortP + charData.numberInstructions;
		}
        if (key->extern_font && !glyphDataPtr && length == 0)   /* 7-27-94 */
            goto EXIT;
        result = sfnt_UnfoldCurve( key, &charData, &sizeOfInstructions, &instructionPtr, length, elementCount );
        if (result)
            goto EXIT;

        result = fsg_InnerGridFit( key, useHints, traceFunc, &charData.bbox, sizeOfInstructions, instructionPtr, false );
        if (result)
            goto EXIT;
	}
    else if ( (key->weGotComponents = (charData.numberContours == -1)) != 0 )
	{
		uint16 flags;
		boolean transformTrashed = false;

		do {
			transMatrix ctmSaveT, localSaveT;
			uint16 glyphIndex;
			int16 arg1, arg2;

			flags 		= (uint16)SWAPWINC(shortP);
			glyphIndex	= (uint16)SWAPWINC(shortP);
		
			if ( flags & ARG_1_AND_2_ARE_WORDS ) {
				arg1 	= (int16)SWAPWINC(shortP);
				arg2 	= (int16)SWAPWINC(shortP);
			} else {
				int8* byteP = (int8*)shortP;
				if ( flags & ARGS_ARE_XY_VALUES ) {
					/* offsets are signed */
					arg1 = *byteP++;
					arg2 = *byteP;
				} else {
					/* anchor points are unsigned */
					arg1 = (uint8)*byteP++;
					arg2 = (uint8)*byteP;
				}
				++shortP;
			}
			
			if ( flags & (WE_HAVE_A_SCALE | WE_HAVE_AN_X_AND_Y_SCALE | WE_HAVE_A_TWO_BY_TWO) )
			{
				fsg_MxCopy( &key->currentTMatrix, &ctmSaveT );
				fsg_MxCopy( &key->localTMatrix, &localSaveT );
				transformTrashed = true;
				if ( flags & WE_HAVE_A_TWO_BY_TWO )
				{
					register Fixed multiplier;
					transMatrix mulT;
		
					multiplier 	= (int16)SWAPWINC(shortP); /* read 2.14 */
					mulT.transform[0][0] = (multiplier << 2); /* turn into 16.16 */
					
					multiplier 	= (int16)SWAPWINC(shortP); /* read 2.14 */
					mulT.transform[0][1] = (multiplier << 2); /* turn into 16.16 */
					
					multiplier 	= (int16)SWAPWINC(shortP); /* read 2.14 */
					mulT.transform[1][0] = (multiplier << 2); /* turn into 16.16 */
					
					multiplier 	= (int16)SWAPWINC(shortP); /* read 2.14 */
					mulT.transform[1][1] = (multiplier << 2); /* turn into 16.16 */
					
					fsg_MxConcat2x2( &mulT, &key->currentTMatrix );
					fsg_MxConcat2x2( &mulT, &key->localTMatrix );
				}
				else
				{
					Fixed xScale, yScale;
					if ( flags & WE_HAVE_AN_X_AND_Y_SCALE )
					{
						xScale 	= (int16)SWAPWINC(shortP); /* read 2.14 */
						xScale <<= 2; /* turn into 16.16 */
						yScale 	= (int16)SWAPWINC(shortP); /* read 2.14 */
						yScale <<= 2; /* turn into 16.16 */
                    }
                    else        /* xScale = yScale */
                    {
						xScale 	= (int16)SWAPWINC(shortP); /* read 2.14 */
						xScale <<= 2; /* turn into 16.16 */
						yScale = xScale;
					}
					fsg_MxScaleAB( xScale, yScale, &key->currentTMatrix );
					fsg_MxScaleAB( xScale, yScale, &key->localTMatrix );
				}
				fsg_InitInterpreterTrans( key ); /*** Compute global stretch etc. ***/
				key->localTIsIdentity = false;
			}
			key->compFlags = flags;
			key->arg1 = arg1;
			key->arg2 = arg2;

			result = sfnt_ReadSFNT( key, elementCount, glyphIndex, useHints, traceFunc );

			if ( transformTrashed )
			{
				fsg_MxCopy( &ctmSaveT, &key->currentTMatrix );
				fsg_InitInterpreterTrans( key  ); /** Restore global stretch etc. ***/	
				fsg_MxCopy( &localSaveT, &key->localTMatrix );
				transformTrashed = false;
			}
		} while ( flags & MORE_COMPONENTS && result == NO_ERR );

		/* Do Final Composite Pass */
		sfnt_ReadSFNTMetrics( key, gIndex ); /* read metrics again */
		if ( flags & WE_HAVE_INSTRUCTIONS ) {
            sizeOfInstructions = (uint16)SWAPW(*shortP);
			instructionPtr = (uint8*)(++shortP);
		} else {
			sizeOfInstructions = 0;
			instructionPtr = 0;
		}
        result = fsg_InnerGridFit( key, useHints, traceFunc, &charData.bbox, sizeOfInstructions, instructionPtr, true );
        if (result)
            goto EXIT;
    }   /* else if ( key->weGotComponents ) */

	else
		result = UNKNOWN_COMPOSITE_VERSION;

EXIT:
#ifdef RELEASE_MEM_FRAG
    if (gHndl != NIL_MH)
        key->RelSfntGlyphPtr(gHndl);
#endif

	return result;
}

#endif   /* TT_RDR  conditionalize entire file */

