/* Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 1995 all rights reserved, 
 * by Bayer Corp., Agfa Division, Wilmington, MA.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and with
 * the inclusion of the above copyright notice. This software or any
 * other copies thereof may not be provided or otherwise made
 * available to any other person except as allowed under license. No
 * title to and ownership of the software is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
/* $Header: /usr/local/cvsrep/Genesis/ufst/rts/tt/ttpcleo.c,v 1.2 2003/05/07 23:02:30 Lann Exp $ */
/* $Log: ttpcleo.c,v $
/* Revision 1.2  2003/05/07 23:02:30  Lann
/* Revision V1_00_02
/*
/* Revision 1.1.1.1  2003/05/07 21:50:39  Lann
/* First Release Version
/* 
 * 
 *    Rev 1.5   07 Apr 1995 11:18:14   LISA
 * Changed copyright information from Miles Inc. to Bayer Corp.
 * 
 *    Rev 1.4   22 Apr 1994 14:55:40   LISA
 * 
 * Made modifications to copyright/disclaimer notice.
 * 
 *    Rev 1.3   26 Mar 1993 14:39:58   MIKE
 * Read stroke weight from header as a signed byte.
 * 
 *    Rev 1.2   15 Jan 1993 09:50:22   LISA
 * Removed CtrlZ character from end of file
 * 
 *    Rev 1.1   14 Dec 1992 09:48:58   LISA
 * Made change to Log keyword
 * 
 *    Rev 1.0   10 Dec 1992 09:17:02   LISA
 * Initial revision.
*/
/* $Date: 2003/05/07 23:02:30 $ */
/*
 *                      T T P C L E O . C
 *
 *
 *  Revision history:
 *
 *  23-Mar-93  mby  Read stroke weight byte as a signed value.
 *
 *  30-Nov-92  jfd  Port to Mac - ANSI function definitions using ANSI_DEFS
 *
 *  13-Jul-92  mby  Fix bug in get_TTFHDR -- read font number with
 *                  GET_xLONG_OFF, not GET_xWORD_OFF.
 *
 *  08-Jul-92  rs   Change 'GET_$...' to 'GET_x...', could not compile with
 *                  BCC.
 *
 *  07-Jul-92  mby  ttload_PCLEO() copies TrueType global data into dynamic
 *                  memory and returns a *handle*.
 *                  Font header data is read with new macros in fscdefs.h
 *                  We shouldn't just copy memory into structures or set
 *                  structure pointers to memory blocks -- this can cause
 *                  alignment problems with some compilers or CPUs.
 *
 *  19-Jun-92  mby  Added get_TTFHDR(). Called by installPcleoHdr(); loads
 *                  HPFH->lastCode and HPFH->fontNumber.
 *
 *  29-May-92  mby  Created.
 *
 *
 *  Module to load TrueType global data from TT PCLEO.
 */



#include "cgconfig.h"
#include <string.h>

#if TT_RDR      /* TrueType Input       */
#if TT_PCLEOI   /* TrueType PCLEO Input */

#include "port.h"
#include "ifmem.h"
#include "mixmodel.h"
#include "cgif.h"
#include "ix.h"
#include "pcleo.h"
#include "cgmacros.h"

#include "ttpcleo.h"            /* PCLEO data structures; function protos */
/* #include "fscdefs.h" */


#ifdef LINT_ARGS
EXTERN MEM_HANDLE  tt_allocmem( UL32 );
#else /* !LINT_ARGS */
EXTERN MEM_HANDLE  tt_allocmem( );
#endif


#define GTTDS         0x4754    /* global TT data segment 'GT' */
#define NULL_SEGID    0xFFFF    /* null segment ID (end of segmented data) */

#define BUCKBITBOLD     0x01
#define BUCKBITNOSERIF  0x02
#define BUCKBITITALIC   0x04
#define BUCKBITNOFP     0x08
#define BUCKBITCOUR     0x10
#define BUCKBITLGOT     0x20




/****************************************************
 *********                                  *********
 **                  TTLOAD_PCLEO                  **
 *********                                  *********
 ****************************************************
 *
 *  Purpose: Loads data from TrueType PCLEO font descriptor.
 *           Saves pointer to TrueType global data following the font descr.
 *
 *  Arguments: pointer to BUCKET structure for TT font.
 *             pointer to PCLEO font header.
 *
 *  Returns: SUCCESS/FAILURE
 *           Handle to TT global data in pBucket->p.tt.GlobalTTHndl
 *           Size of TT global data in pBucket->p.tt.GlobalTTSize
 *           Universal font selection bits in pBucket->bucket_num
 */

#if defined (ANSI_DEFS)
GLOBAL int ttload_PCLEO (LPUB8 fontHeader, PBUCKET pBucket)

#else 
GLOBAL int ttload_PCLEO (fontHeader, pBucket)
   LPUB8    fontHeader;
   PBUCKET  pBucket;
#endif
{
int    found = TRUE;
LPUB8  pGTTBase;
LPUB8  pSegfd;
int    val;
UW16   word16;

pSegfd = fontHeader + GET_xWORD_OFF(fontHeader, PTO_FONTDECSCRIPTORSIZE);
                                /* ptr to global TrueType data */

/* Start at the Segmented Font Data section of the header.
 * Loop until we find the global TrueType data segment.     */
word16 = GET_xWORD(pSegfd);
while (word16 != GTTDS)
    {
    if (word16 == NULL_SEGID)
        {
        found = FALSE;
        break;
        }
    word16 = GET_xWORD_OFF(pSegfd, 2);  /* size in bytes */
    pSegfd += word16 + 4;
    word16 = GET_xWORD(pSegfd);
    }
if (!found)
    return ERR_TT_PCLEONOGT;

word16 = GET_xWORD_OFF(pSegfd, 2);
pBucket->p.tt.GlobalTTSize = (UL32)word16;
pBucket->p.tt.GlobalTTHndl = tt_allocmem((UL32)word16);
if (!pBucket->p.tt.GlobalTTHndl)
    return ERR_TT_PCLEONOMEM;
pGTTBase = (LPUB8)MEMptr(pBucket->p.tt.GlobalTTHndl);
MEMCPY(pGTTBase, pSegfd + 4, (int)word16);

/* Set bits in bucket_num for serif, posture, weight, and fixed pitch plugins. */
pBucket->bucket_num = 0;
val = (int)GET_xBYTE_OFF(fontHeader, PTO_SERIFSTYLE) & 0x3F;
if ((unsigned)val <= 1 || val == 9)
    pBucket->bucket_num |= BUCKBITNOSERIF;
if ( (SB8)(GET_xBYTE_OFF(fontHeader, PTO_STROKEWEIGHT)) > 0)
    pBucket->bucket_num |= BUCKBITBOLD;
if (GET_xBYTE_OFF(fontHeader, PTO_STYLE_LSB) & 0x03)
    pBucket->bucket_num |= BUCKBITITALIC;
if (!GET_xBYTE_OFF(fontHeader, PTO_SPACING))  /* are we fixed pitch? */
    {                               /* calculate the pitch */
    word16 = GET_xWORD_OFF(fontHeader, PTO_PITCH);
    val = (UW16)((6 * (UL32)GET_xWORD_OFF(fontHeader, PTO_SCALEFACTOR) +
                    (word16 >> 1)) / word16);
    if (val == 10)
        pBucket->bucket_num |= BUCKBITCOUR;   /* Courier plugins, pitch=10 */
    else if (val == 12)
        pBucket->bucket_num |= BUCKBITLGOT;   /* LettGoth plugins, pitch=12 */
    else
        pBucket->bucket_num |= BUCKBITNOFP;
    }

return SUCCESS;
}




#if defined (ANSI_DEFS)
GLOBAL void get_TTFHDR (PHPFH pfh, LPUB8 fontHeader)

#else 
GLOBAL void get_TTFHDR (pfh, fontHeader)
   PHPFH pfh;
   LPUB8 fontHeader;
#endif
{
pfh->lastCode = GET_xWORD_OFF(fontHeader, PTO_LASTCODE);
pfh->fontNumber = GET_xLONG_OFF(fontHeader, PTO_FONTNUMBER);
}

#endif          /* TT_PCLEOI */
#endif          /* TT_RDR    */


